/* C/bricks/API/AL0/common/AL0GatesInternal.c
 * 
 * Functions for the most basic IPC communication between Minmex and Bricks
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> (initial design, user space implementation )
 *	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> (adaption for kernel space)
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */


#include "anaLib0.h"
#include "AL0GatesInternal.h"

#include <arpa/inet.h>
#include <netinet/in.h>
#include <errno.h>

analock_t bigLock;

 
#ifdef USER
	#define NODE_CONTROL_DESC spec->minmexControlDescriptor
 	#define NODE_DATA_DESC spec->minmexDataDescriptor
	#define MY_DATA_DESC spec->myDataDescriptor
	#define MY_CONTROL_DESC spec->myControlDescriptor
#endif
#ifdef KERNEL
	#define NODE_CONTROL_DESC spec->minmexControlSocket
 	#define NODE_DATA_DESC spec->minmexDataSocket
	#define MY_DATA_DESC spec->myDataSocket
	#define MY_CONTROL_DESC spec->myControlSocket
    
        struct rw_semaphore testsem;
#endif



 
int setSockOptWrapper(void *sock, int level,int op, char * optval, int optlen){
        int ret;
#ifdef USER
	int descriptor;
	if(sock==NULL)
		return -1;
	descriptor = *(int*)sock;
      	ret = setsockopt(descriptor, level, op, optval, optlen);
       
#endif
#ifdef KERNEL
	struct socket *s =  NULL;
	mm_segment_t oldfs;
	if(sock == NULL)
		return -1;
	s = *((struct socket **)sock);
        
        oldfs = get_fs();
        set_fs( KERNEL_DS );
		ret = sock_setsockopt(s, level, op, optval, optlen);
        set_fs( oldfs );
#endif
        return ret;
} 
 
int readWrapper(void *sock, char * data, int len){
        int ret = -1;
#ifdef USER
	int descriptor;
	if(sock == NULL)
		return -1;
	descriptor = *(int*)sock;
	ret = read(descriptor, data, len);
#endif
#ifdef KERNEL
       	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
        struct sockaddr_in brick;
	struct socket *s = NULL;
	
	if(sock == NULL)
		return -1;
	
	s = *((struct socket **)sock);
        msg.msg_name = &brick;
	msg.msg_namelen = sizeof( struct sockaddr_in );
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov    = &iov;
	msg.msg_iovlen = 1;
        iov.iov_base = data;
        iov.iov_len  = len;
        oldfs = get_fs();
        set_fs( KERNEL_DS );
       
	ret = sock_recvmsg(s, &msg, len, 0 );
        set_fs( oldfs );
#endif
        return ret;
} 
 
 
/* 	Sends a control message to the Control Gate of the minmex. Will use the adequate transfer mode (UDP,PIPES...) according to   
	the mode indicated by the brick application at attach call  */
int sendControl(struct anaMinmexSpecs_s * spec, void *msg, int len) {
	int nbSent=-1;
	if(spec==NULL || msg ==NULL || len<=0){
		anaPrint(ANA_ERR, "sendControl : minmex or message is NULL\n");
		return -1;
	}

	switch(spec->minmexControlMode) {
	case UDP:{
		nbSent = sendSocketWrapper(&(NODE_CONTROL_DESC), AF_INET, spec->minmexControlAux , msg, len);
		if(nbSent < 0) {
                       	anaPrint(ANA_ERR,  "sendControl : Unable to send control" 
                                              "message through UDP socket\n");
                       	return -1;
               	}
#ifdef KERNEL
		schedule();	/* release the CPU, this is important for Kernel 2.4 which is nonpreemptive
				 * otherwise the MINMEX has no chance to execute its task_queue
 				 */	 
#endif 
        	break;
        }
#ifdef USER
	case UNIX:{
		nbSent = sendSocketWrapper(&(NODE_CONTROL_DESC), AF_LOCAL,  spec->minmexControlAux , msg, len);
		if(nbSent < 0) {
                       	anaPrint(ANA_ERR,  "sendControl : Unable to send control" 
                                              "message through UNIX socket\n");
                       	return -1;
               	}	 
        	break;
        }
        case PIPES:{
                nbSent = write(spec->minmexControlDescriptor, msg, len);
                if(nbSent < 0) {
                        anaPrint(ANA_ERR,  "sendControl : Unable to send "
                                              "control message through named pipe\n");
                        return -1;
                	}
	         break;
        }
#ifdef USERGEN
        case GENETLINK:{
                struct {
                        struct nlmsghdr n;
                        struct genlmsghdr g;
                        char buf[256];
                } req;
                struct nlattr *na;
		memset(&req, 0, sizeof(req));              
 
                req.n.nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
                req.n.nlmsg_type = *(int *)spec->minmexControlAux;
                req.n.nlmsg_flags = NLM_F_REQUEST;
                req.n.nlmsg_seq = 60;
                req.n.nlmsg_pid = *(int *)spec->myControlAux;
                req.g.cmd = COMMAND_CONTROL;  /*commands and type are defined in anaCommon.h*/
                na = (struct nlattr *) GENLMSG_DATA(&req);
                na->nla_type = ATTR_ID;
                char * message = (char *)msg; /*message*/
                int mlength = len;
                na->nla_len = mlength+NLA_HDRLEN; /*message length*/
                memcpy(NLA_DATA(na), message, mlength);
                req.n.nlmsg_len += NLMSG_ALIGN(na->nla_len);
                
                /*send message*/
                if ((nbSent = sendto_fd(spec->minmexControlDescriptor, (char *) &req, req.n.nlmsg_len)) < 0){
                        anaPrint(ANA_ERR, "error sending message via Netlink\n");
                        return -1;
                }
                break;
        }
#endif
#endif
	
	}

	return nbSent;
}


/*called after a control request to listen for a response from the minmex defined in the anaMinmexSpecs, filles the buffer argument or returns -1*/
int recvControl(struct anaMinmexSpecs_s * spec, void *buffer) {
	int nbRead=-1;
	
	if(spec==NULL || buffer==NULL)
		return -1;
	
	switch(spec->myControlMode){
        case UDP:{	/* in case it's a UDP seocket, the timeout for reading has been set with setsockopt() */
              	nbRead = readWrapper(&(MY_CONTROL_DESC), buffer, CTRLMSG_MAXSIZE);
                break;
        }
	
#ifdef USER
 	case UNIX:{	/* in case it's a UDP seocket, the timeout for reading has been set with setsockopt() */
              	nbRead = readWrapper(&(MY_CONTROL_DESC), buffer, CTRLMSG_MAXSIZE);
                break;
        }
	
        case PIPES:{ /* since the file descriptors must be set with O_NONBLOCK, we need to do a select() to timeout reception */
                int ret;
                struct timeval SelectTime;
                fd_set readFd;
                FD_ZERO(&readFd);
                FD_SET(spec->myControlDescriptor,&readFd);
                SelectTime.tv_sec = ((int*)spec->myControlAux)[1]; /* we set the timeout value indicated at the attachment*/
                SelectTime.tv_usec = 0;
                ret = select(255,&readFd,NULL,NULL,&SelectTime);
                if(ret>0)
                        nbRead = read(spec->myControlDescriptor, buffer, CTRLMSG_MAXSIZE);
		break;
        }
#ifdef USERGEN
        case GENETLINK:{
                struct {
                        struct nlmsghdr n;
                        struct genlmsghdr g;
                        char buf[CTRLMSG_MAXSIZE];
                } ans;
                struct nlattr *na;
                memset(ans.buf,0, CTRLMSG_MAXSIZE);
                nbRead = recv(spec->myControlDescriptor, &ans, sizeof(ans), 0);
               
                /* Validate response message */
                if (ans.n.nlmsg_type == NLMSG_ERROR) { /* error */
                        anaPrint(ANA_ERR, "rcvCtrl: GN: received NACK\n");
			nbRead = -1;
                        break;
                }
                if (nbRead < 0) {
                        anaPrint(ANA_ERR,  "GN len < 0\n");
			nbRead = -1;
                        break;
                }
                if (!NLMSG_OK((&ans.n), nbRead)){
                        anaPrint(ANA_ERR, "rcvCtrl: GN: invalid reply message\n");
			nbRead = -1;
                        break;
                }
                nbRead = GENLMSG_PAYLOAD(&ans.n);
                
                /*parse reply message*/
                na = (struct nlattr *) GENLMSG_DATA(&ans);
                memcpy(buffer, NLA_DATA(na), nbRead);
                break;
        }
#endif
#endif
	}

	return nbRead;
}


/* 	Sends a data message to the data Gate of the minmex. Will use the adequate transfer mode (UDP,PIPES...) according to   
	the mode indicated by the brick application at attach call   */
int sendData(struct anaMinmexSpecs_s * spec, void *msg, int len) {
	int nbSent = -1;
       	if(spec==NULL || msg ==NULL || len<=0){
		anaPrint(ANA_ERR, "sendData : minmex or message is NULL\n");
		return -1;
	}	
	

	switch(spec->minmexDataMode) {
        case UDP:{
		nbSent = sendSocketWrapper(&(NODE_DATA_DESC), AF_INET, spec->minmexDataAux ,msg, len);	
                if(nbSent < 0) {
                        anaPrint(ANA_ERR,  "sendData : Unable to send data through UDP socket\n");
                        return -1;
		}
#ifdef KERNEL
		schedule();	 /* release the CPU, this is important for Kernel 2.4 which is nonpreemptive
                                  * otherwise the MINMEX has no chance to execute its task_queue
                                  */
#endif	 
                break;
        }
#ifdef USER
	case UNIX:{
		nbSent = sendSocketWrapper(&(NODE_DATA_DESC), AF_LOCAL, spec->minmexDataAux ,msg, len);	
                if(nbSent < 0) {
                        anaPrint(ANA_ERR,  "sendData : Unable to send data through UNIX socket\n");
                        return -1;
		}	 
                break;
        }
        case PIPES:{
                nbSent = write(spec->minmexDataDescriptor, msg, len);
                if(nbSent < 0) {
                        anaPrint(ANA_ERR,  "sendData : Unable to send data message through named pipe\n");
                        return -1;
                }	 
                break;
                
        }
#ifdef USERGEN
        case GENETLINK:{
		struct {
                        struct nlmsghdr n;
                        struct genlmsghdr g;
                        char buf[1024];
                } req;
                struct nlattr *na;
                memset(&req, 0, sizeof(req));
                req.n.nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
                req.n.nlmsg_type =  *(int *)spec->minmexDataAux;
                req.n.nlmsg_flags = NLM_F_REQUEST;
                req.n.nlmsg_seq = 1; /*sequenc numbering not used it, so every msg has sequence numer = 1*/
                req.n.nlmsg_pid = *(int*)spec->minmexDataAux; /*identifier obtained by generate_local_port() at attachment */;
                req.g.cmd = COMMAND_DATA;               
                /*compose message*/
                na = (struct nlattr *) GENLMSG_DATA(&req);

                na->nla_type = 1; 
                na->nla_len = len+NLA_HDRLEN;
                memcpy(NLA_DATA(na), (char *)msg, len);
   		req.n.nlmsg_len += NLMSG_ALIGN(na->nla_len);     
                
                /*send message*/
                if ((nbSent = sendto_fd(spec->minmexDataDescriptor, (char *) &req, req.n.nlmsg_len)) < 0){
                        anaPrint(ANA_ERR, "error sending message via Netlink\n");
                        free(req.buf);
			return -1;
                }
                break;
                
        } 
#endif  
#endif
	}

	return nbSent;
}

int bindWrapper (void *sock, struct sockaddr * address, int size){
       
#ifdef USER
	int descriptor;
	if(sock == NULL || address == NULL || size <=0)
		return -1;
		
	descriptor = *(int*)sock;
        return (bind(descriptor, address, size));
#endif
#ifdef KERNEL 
	struct socket *s = NULL;
	if(sock == NULL || address == NULL || size <=0)
		return -1;
	 s = *((struct socket **)sock);
	/*in case of kernel mode sock is a struct socket ** (2 Stars) */
    	return s->ops->bind(s, address, size);
/*	return (kernel_bind(s, address, size));*/
#endif
}


static int threadFunction(void * data){
	struct callbackArgs * args = (struct callbackArgs *)data;

	args->function(args->data, args->len, args->input, args->aux); 

	free(args->data);
	free(args);

	return 1;


}

int threadNotification(struct notifThreadArg *notifArg){
	anaPrint(ANA_DEBUG, "****** Threading the notification \n ");
	notifArg->handler(notifArg->notif);
	
	if(notifArg->notif->name)
		free(notifArg->notif->name);

	free(notifArg->notif);
	free(notifArg);

	return 1;


}

/* this function sets the age of the minmex to zero when triggered */
void handleHbeat(char *data, int len, anaLabel_t input, void *aux) {
	myMinmex->age = 0;
}




/* the default function executed, if a callback should be executed as a separate thread */
void anaL0_launchThread(char *data, int len, anaLabel_t input, void *aux) {
	
	char * message = NULL;
	struct threadAux * thrd = (struct threadAux *)aux;
	struct callbackArgs * cbargs = malloc(sizeof(struct callbackArgs));
	if(data==NULL || aux ==NULL)
		return;

	message = malloc(len);
	memcpy(message, data, len);
	cbargs->function = thrd->fct;

	cbargs->data = message;
	cbargs->len = len;
	cbargs->input = input;
	if(thrd->cpy != NULL)
		cbargs->aux = thrd->cpy(thrd->aux);
	_anaThread(0, (void *) threadFunction, cbargs, "callbackThread");
}

 void * anaL0_copyAux(void * aux, copyAuxFunction copyAux, freeAuxFunction freeAux, anaCallback_t fct){
	struct threadAux * ret = malloc(sizeof(struct threadAux));
	ret->fct = fct;

	ret->cpy = copyAux;
	ret->free = freeAux;
	if(copyAux != NULL){
		ret->aux = copyAux(aux);
		anaPrint(ANA_DEBUG, "anaL0_copyAux: %p\n", ret->aux);
	}
	return ret;
}
 void anaL0_freeAux(void * aux){
	struct threadAux * ret = (struct threadAux *)aux;
	if(ret->aux != NULL && ret->free != NULL)
		ret->free(ret->aux);
	free(ret);
}

/* initializes information about minmex by brick and builds the attach message to be sent to minmex,  Parameters :
	 minmexControlURI : communication mode (UDP,PIPES,..) to send control messages (attach, registerCallback,..) to minmex expressed as URI
	 myControlURI   : communication mode chosen by brick to receive control messages replies on 
	 myDataURI   : communication mode chosen by brick to receive data messages (encapsulated with labels) 
	 myDataURI   : same as minmexCtrlLink
	 desiredDataMode : desired mode for data sending to minmex
	 timeoutVal   : timeout value for read() calls on brick's control gate                                                           */

struct anaMinmexSpecs_s * anaL0_initMinmexSpecs(char *myDataURI, char *myControlURI, 
                                   char *minmexControlURI, int desiredDataMode, int timeoutVal){
	/* expected URIs:  [mode]://[Aux] 
		in case of UDP mode Aux would be: [IP]:[port]
	*/
	
	/* The attach message to build must be formatted this way: 
           [a] [mycontrolMode] [mycontrolLink] [mydataMode] [mydataLink][nameLen][name] [datadesiredMode] */
	   								/* 1 byte */
	int len;
	char *myDataAux;
	char *myControlAux;
	char *minmexControlAux;

	struct timeval  timeo;

	struct anaMinmexSpecs_s * spec;
        int messageIndex=0;
	int minmexCtlMode=-1;
	int myDataMode=-1; 
	int myControlMode=-1;
	#ifdef USER
		int sock;
	#endif
	#ifdef KERNEL
                struct anaMinmexFunctions * minmexfct = NULL;
		struct socket *sock;
         #endif            
	if(myDataURI==NULL || myControlURI==NULL || minmexControlURI==NULL)
		return NULL; 
		
	analock_init(&bigLock);
	
	spec =  (struct anaMinmexSpecs_s *) malloc(sizeof(struct anaMinmexSpecs_s));
	if(spec == NULL){
		anaPrint(ANA_ERR, "anaL0_initMinmexSpecs: unable to allocate memory\n");
		return NULL;
	}
	
	spec->attachMessage = (unsigned char *) malloc(CTRLMSG_MAXSIZE * sizeof(unsigned char));
	if(spec->attachMessage==NULL){
		anaPrint(ANA_ERR, "anaL0_initMinmexSpecs: unable to allocate memory (for attach message)\n");
		free(spec);
		return NULL;	
	}
	
	if(!strncmp(myDataURI,"udp",3)|| !strncmp(myDataURI,"UDP",3))
		myDataMode = UDP;
	/*add  tests for other modes (PIPES) here */		
        if(!strncmp(myDataURI,"pipe",4)|| !strncmp(myDataURI,"PIPE",4))
                myDataMode = PIPES;
        if(!strncmp(myDataURI,"genetlink",9)|| !strncmp(myDataURI,"GENETLINK",9))
                myDataMode = GENETLINK;
	if(!strncmp(myDataURI,"unix",4)|| !strncmp(myDataURI,"UNIX",4))
                myDataMode = UNIX;
		
		
	if(!strncmp(myControlURI,"udp",3)|| !strncmp(myControlURI,"UDP",3))
		myControlMode = UDP;
        if(!strncmp(myControlURI,"pipe",4)|| !strncmp(myControlURI,"PIPE",4))
                myControlMode = PIPES;
        if(!strncmp(myControlURI,"genetlink",9)|| !strncmp(myControlURI,"GENETLINK",9))
                myControlMode = GENETLINK;
        if(!strncmp(myControlURI,"unix",4)|| !strncmp(myControlURI,"UNIX",4))
                myControlMode = UNIX;
      
	
	if(!strncmp(minmexControlURI,"udp",3)|| !strncmp(minmexControlURI,"UDP",3))
		minmexCtlMode = UDP;
        if(!strncmp(minmexControlURI,"pipe",4)|| !strncmp(minmexControlURI,"PIPE",4))
                minmexCtlMode = PIPES;
        if(!strncmp(minmexControlURI,"genetlink",9)|| !strncmp(minmexControlURI,"GENETLINK",9))
                minmexCtlMode = GENETLINK;
	if(!strncmp(minmexControlURI,"unix",4)|| !strncmp(minmexControlURI,"UNIX",4)){
                minmexCtlMode = UNIX;
        }

	switch(minmexCtlMode) {

        case UDP:{ /* if the mode is UDP, [AUX] would be [IP]:[port] */
                char *IP=NULL;
                char *portChar=NULL;
                int port=-1;
                struct sockaddr_in *tmp=NULL;
	
				
                minmexControlAux = (char *)minmexControlURI + 6; /* + 6 is for "udp://" */
                
                IP = strsep(&minmexControlAux, ":");
                portChar = strsep(&minmexControlAux, ":");
                
                if (portChar !=NULL)
                        port = atoi(portChar);
                
                if(port <=0) {
                        anaPrint(ANA_ERR,  "initMinmexSpecs : Invalid UDP port \n");
                        free(spec->attachMessage);
                        free(spec);
                        return NULL;
                }
                
                spec->minmexControlMode = UDP;

		if(sockCreateWrapper(&sock, PF_INET, SOCK_DGRAM, 0)<0){
                        anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create minmex Control gate descriptor \n");
                        free(spec->attachMessage);
                        free(spec);
                        return NULL;
                }
               	
                tmp = (struct sockaddr_in*) malloc(sizeof(struct sockaddr_in));
                memset(tmp, 0, sizeof(struct sockaddr_in));
                tmp->sin_addr.s_addr = inet_addr(IP);
                tmp->sin_family = AF_INET;
                tmp->sin_port = htons(port);
                spec->minmexControlAux = tmp;
	
		NODE_CONTROL_DESC = sock;
			
                anaPrint(ANA_DEBUG, "initMinmexSpecs : Sending control messages to minmex will be done"
                            " with UDP to IP %s on port %d\n", IP, port);	
                break;
        }

#ifdef USER
	case UNIX:{ /* if the mode is UNIX, URL would be:  unix://socket_name */
                struct sockaddr_un *tmp;
		char *sockPath = (char *)minmexControlURI + 7; /* + 7 is for "pipe://" */
				
		spec->minmexControlMode = UNIX;

		if(sockCreateWrapper(&sock, AF_LOCAL, SOCK_DGRAM, 0)<0){
                        anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create minmex Control gate descriptor \n");
                        free(spec->attachMessage);
                        free(spec);
                        return NULL;
                }
               	
                tmp = (struct sockaddr_un*) malloc(sizeof(struct sockaddr_un));
                memset(tmp, 0, sizeof(struct sockaddr_un));
		
		tmp->sun_family = AF_LOCAL;
        	strcpy(tmp->sun_path, sockPath);
        		       
                spec->minmexControlAux = tmp;
	
		NODE_CONTROL_DESC = sock;
			
                anaPrint(ANA_DEBUG, "initMinmexSpecs : Sending control messages to minmex will be done"
                            " with UNIX on socket %s\n", sockPath);	
                break;
        }

        case PIPES:{	/*in this case URL would be:  pipe://pipes_path */
                char *pipePath = (char *)minmexControlURI + 7; /* + 7 is for "pipe://" */
	
                spec->minmexControlMode = PIPES;
                spec->minmexControlDescriptor = open(pipePath, O_WRONLY|O_NONBLOCK);
                
                if(spec->minmexControlDescriptor < 0){
                        anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create "
                                              "minmex Control gate descriptor \n");
                        free(spec->attachMessage);
                        free(spec);
                        return NULL;	
                }
                break;
        }
#ifdef USERGEN
        case GENETLINK:{
		minmexControlAux = (char *)minmexControlURI +12; /*for genetlink*/
                int pid = generate_local_port();  /* generates a local id to which messages can be sent */
                spec->minmexControlMode = GENETLINK;
                spec->minmexControlDescriptor = create_nl_socket(NETLINK_GENERIC, 0, pid);
                if(spec->minmexControlDescriptor <0) {
                        anaPrint(ANA_ERR,  "initMinmexSpecs :"
                                              " Unable to create minmex Control gate descriptor \n");
                        free(spec->attachMessage);
                        free(spec);
                        return NULL;
                }
                int * id = (int *) malloc(sizeof (int));
                *id = get_family_id(spec->minmexControlDescriptor, minmexControlAux);
                anaPrint(ANA_DEBUG,  "minmex control with genetlink family: %i \n", *id);

                spec->minmexControlAux = id;
                spec->myControlAux = malloc(sizeof(int));
                *(int *)spec->myControlAux = pid;
                break;
        }
#endif
#endif

        default: {
                anaPrint(ANA_ERR,  "initMinmexSpecs : mode %d for sending "
                                      "control data to minmex is unhandled\n",minmexCtlMode);
		
                free(spec->attachMessage);
                free(spec);
                return NULL;
        }
	}
	
	switch(myControlMode) {

		case UDP:{ 	/* expected [AUX] : [IP]:[port] */
			char *IP;
			char *portChar;
			struct sockaddr_in my_addr;
		
			int port=-1;
			myControlAux = myControlURI + 6; /* + 6 is for "udp://" */
			/* <message formatting>*/
			spec->attachMessage[0]= 'a';
			spec->attachMessage[1]= (unsigned char) UDP;
			len = strlen(myControlAux) ;
			spec->attachMessage[2]= (unsigned char) len;
			memcpy(spec->attachMessage+3,  myControlAux, len);
			
			messageIndex = len + 3;
			
			/* <End of message formatting> */
			
			IP = strsep(&myControlAux, ":");
			portChar = strsep(&myControlAux, ":");
			
			if (portChar!= NULL)
				port = atoi(portChar);
	
			spec->myControlMode = UDP;
                        
			if(sockCreateWrapper( &sock ,PF_INET, SOCK_DGRAM, 0)<0){
                                anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick Control gate descriptor \n");
			
				closeWrapper(&(NODE_CONTROL_DESC));
						
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			MY_CONTROL_DESC = sock;
			
			/* configuring local address for listening socket */
			memset(&my_addr, 0, sizeof(struct sockaddr_in));
			my_addr.sin_family = AF_INET;	
			my_addr.sin_addr.s_addr = inet_addr(IP) ; /*htonl(INADDR_ANY); */
			my_addr.sin_port = htons(port); 
		
                        if(bindWrapper(&(MY_CONTROL_DESC), (struct sockaddr *) &my_addr, sizeof(struct sockaddr))<0) {
				anaPrint(ANA_ERR, "initMinmexSpecs : Unable to bind brick UDP control socket to port %d\n", port);
                                closeWrapper(&(NODE_CONTROL_DESC));
				closeWrapper(&(MY_CONTROL_DESC));
				free(spec->minmexControlAux);
				free(spec->attachMessage);
				free(spec);
			    return NULL;
			}
					
			/* Setting the timeout value for the socket */
			timeo.tv_sec  = timeoutVal;
			timeo.tv_usec = 0;
			if (setSockOptWrapper(&(MY_CONTROL_DESC), SOL_SOCKET, SO_RCVTIMEO, (char *)&timeo, sizeof(timeo)) < 0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : unable to set timeout on control socket\n");
				closeWrapper(&(NODE_CONTROL_DESC));
				closeWrapper(&(MY_CONTROL_DESC));
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
			 	return NULL;
			}
						
			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of control messages from minmex will be done with UDP on port %d\n", port);	
			break;
		}

#ifdef USER
		case UNIX:{ 	
			
			char *sockPath = myControlURI + 7; /* +7 for unix:// */
			
			struct sockaddr_un  *my_addr =NULL;
			
			/* <message formatting>*/
			spec->attachMessage[0]= 'a';
			spec->attachMessage[1]= (unsigned char) UNIX;
			len = strlen(sockPath) ;
			spec->attachMessage[2]= (unsigned char) len;
			memcpy(spec->attachMessage+3,  sockPath, len);
			
			messageIndex = len + 3;
			
			/* <End of message formatting> */
			
			spec->myControlMode = UNIX;
                        
			if(sockCreateWrapper( &sock ,AF_UNIX, SOCK_DGRAM, 0)<0){
                                anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick Control gate descriptor \n");
			
				closeWrapper(&(NODE_CONTROL_DESC));
						
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			MY_CONTROL_DESC = sock;
			
			/* configuring local address for listening socket */
			my_addr = (struct sockaddr_un *) malloc(sizeof(struct sockaddr_un ));
			
			my_addr->sun_family = AF_LOCAL;
        		strcpy(my_addr->sun_path, sockPath);
        		unlink(sockPath);
	
       		        if(bindWrapper(&(MY_CONTROL_DESC), (struct sockaddr *) my_addr, SUN_LEN(my_addr))<0) {
				anaPrint(ANA_ERR, "initMinmexSpecs : Unable to bind brick UNIX control socket to %s\n", sockPath);
                                closeWrapper(&(NODE_CONTROL_DESC));
				closeWrapper(&(MY_CONTROL_DESC));
				free(spec->minmexControlAux);
				free(spec->attachMessage);
				free(my_addr);
				free(spec);
			    return NULL;
			}
			
			/* set permissions to rwxrwxrwx (because vlink is ran as root) */
			
			if (chmod(sockPath, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH) == -1) {
				anaPrint(ANA_ERR, "initMinmexSpecs : unable to set permissions on UNIX socket %s", sockPath);
			}
			
			/* Setting the timeout value for the socket */
			timeo.tv_sec  = timeoutVal;
			timeo.tv_usec = 0;
			if (setSockOptWrapper(&(MY_CONTROL_DESC), SOL_SOCKET, SO_RCVTIMEO, (char *)&timeo, sizeof(timeo)) < 0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : unable to set timeout on control socket\n");
				closeWrapper(&(NODE_CONTROL_DESC));
				closeWrapper(&(MY_CONTROL_DESC));
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(my_addr);
				free(spec);
			 	return NULL;
			}
			
			spec->myControlAux = my_addr;		
			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of control messages from minmex will be done with UNIX on %s\n", sockPath);	
			break;
		}
		case PIPES:{
			int dummyDesc;
			char *pipePath = myControlURI + 7; /* +7 for pipe:// */ 
			spec->myControlMode = PIPES;
			/* <message formatting>*/
			spec->attachMessage[0]= 'a';
			spec->attachMessage[1]= (unsigned char) PIPES;
			len = strlen(pipePath) ;
			spec->attachMessage[2]= (unsigned char) len;
			memcpy(spec->attachMessage+3,  pipePath, len);
			messageIndex = len + 3;
			
			/* <End of message formatting> */
			/* we create the FIFO pipe */
			/* the pipe is created with access rights to other users */
			if ((mkfifo(pipePath, 0666) == -1) && (errno != EEXIST)) {
    				anaPrint(ANA_ERR, "InitMinmexSPecs :Error creating control named pipe errno=%d\n",errno);
        			closeWrapper(&(NODE_CONTROL_DESC));
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
    			}
			
			/* we open the the pipe in read mode */
			spec->myControlDescriptor = open(pipePath, O_NONBLOCK|O_RDONLY);
    					
			if(spec->myControlDescriptor <0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick Control gate descriptor \n");
                                closeWrapper(&(NODE_CONTROL_DESC));
                                free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			/* we open the pipe in write mode : [WARNING] we open the write mode to always have a writing descriptor pointing at the fifo
				otherwise, on all bricks detachment, the pipes would be no more useable (even worse with select causing tight loop),
				in which case we would have to close the pipe and reopen it => therefore we open this dummy descriptor, DONT USE IT TO WRITE
			 */
			dummyDesc = open(pipePath, O_NONBLOCK|O_WRONLY);
			if(dummyDesc <0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to open dummy Descriptor on Control gate FIFO\n");
				closeWrapper(&(NODE_CONTROL_DESC));
                                closeWrapper(&(MY_CONTROL_DESC));
                                free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			/* in the myControlAux we store a 2 integers array: containing the dummy descriptor(index 0) and the timeout value(index 1) */
			spec->myControlAux = (int *) malloc(2*sizeof(int)); 
			((int *)spec->myControlAux)[0] = dummyDesc;
			((int*) spec->myControlAux)[1] = timeoutVal;
			
						
			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of control messages from minmex will be done with NAMED PIPES on %s\n", pipePath);	
			break;
			
		}
#ifdef USERGEN
                case GENETLINK:{
                        spec->myControlMode = GENETLINK;
			/* <message formatting>*/
			spec->attachMessage[0]= 'a';
			spec->attachMessage[1]= (unsigned char) GENETLINK;
                        len = sizeof(int);
                        spec->attachMessage[2]= (unsigned char) len;
			/* if there is already a socket created we can use this socket / id */
			if(spec->minmexControlMode == GENETLINK){
                                spec->myControlDescriptor = spec->minmexControlDescriptor;
                        }
                        else{
                                int pid = generate_local_port();
                                spec->myControlDescriptor = create_nl_socket(NETLINK_GENERIC, 0, pid);
                                spec->myControlAux = malloc(sizeof(int));
                                *(int *)spec->myControlAux = pid;
                        }
                        *(int*)(spec->attachMessage+3) = *(int *)spec->myControlAux;
                        messageIndex = len + 3;
                        break;
                }
#endif
#endif
		default:{
			anaPrint(ANA_NOTICE, "initMinmexSpecs : mode %d for receiving control data from minmex is unhandled\n", myControlMode);
      
           		closeWrapper(&(NODE_CONTROL_DESC));

			free(spec->attachMessage);
			free(spec->minmexControlAux);
			free(spec);
			return NULL;
		}
	}	
	
	switch(myDataMode) {

		case UDP:{  /*expected [AUX] : [IP]:[port]*/
			struct sockaddr_in my_addr;
			char *IP; 
			char *portChar;
			int port=-1;
			
			myDataAux = myDataURI + 6; /* + 6 is for "udp://" */
			
			/* <message formatting>*/
			
			spec->attachMessage[messageIndex]= (unsigned char) UDP;
			len = strlen(myDataAux) ;
			spec->attachMessage[messageIndex+1]= (unsigned char) len;
			memcpy(spec->attachMessage+messageIndex+2,  myDataAux, len);
			
			messageIndex = messageIndex + len + 2;
			
			/* <End of message formatting> */  
				
			IP = strsep(&myDataAux, ":");
			portChar = strsep(&myDataAux, ":");
			
			if (portChar !=NULL)
				port = atoi(portChar);
				
			spec->myDataMode = UDP;
			if(sockCreateWrapper( &sock , PF_INET, SOCK_DGRAM, 0 )<0){
                                anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick Data gate descriptor (UDP) \n");
				closeWrapper(&(NODE_CONTROL_DESC));
                               closeWrapper(&(MY_CONTROL_DESC));
                                if(spec->myControlMode == PIPES)
#ifdef USER
					close(((int *)spec->myControlAux)[0]);
#endif
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			MY_DATA_DESC = sock;

			/* configuring local address for listening socket */
			memset(&my_addr, 0, sizeof(struct sockaddr_in));
			my_addr.sin_family = AF_INET;	
			my_addr.sin_addr.s_addr = inet_addr(IP) ;
			my_addr.sin_port = htons(port); 
		
			if(bindWrapper(&(MY_DATA_DESC), (struct sockaddr *) &my_addr, 
                                       sizeof(struct sockaddr))<0) {
				anaPrint(ANA_ERR, "initMinmexSpecs : Unable to bind brick "
                                                     "UDP Data socket to port %d\n", port);
                               closeWrapper(&(NODE_CONTROL_DESC));
                               closeWrapper(&(MY_CONTROL_DESC));
#ifdef USER
                                if(spec->myControlMode == PIPES)
                                        close(((int *)spec->myControlAux)[0]);
#endif
                                free(spec->attachMessage);
                                closeWrapper(&(MY_DATA_DESC));
                                free(spec->minmexControlAux);
                                free(spec);
                                return NULL;
			}
					
			/* Setting the timeout value for the socket */
			timeo.tv_sec  = timeoutVal;
			timeo.tv_usec = 0;
			if (setSockOptWrapper(&(MY_DATA_DESC), SOL_SOCKET, SO_RCVTIMEO, (char *)&timeo, sizeof(timeo)) < 0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : unable to set timeout on Data socket\n");
                                closeWrapper(&(NODE_CONTROL_DESC));
                              	closeWrapper(&(MY_CONTROL_DESC));
#ifdef USER
				if(spec->myControlMode == PIPES)
					close(((int *)spec->myControlAux)[0]);
#endif
				free(spec->attachMessage);
                                closeWrapper(&(MY_DATA_DESC));
				free(spec->minmexControlAux);
				free(spec);
			 	return NULL;
			}

#ifdef KERNEL
                        MY_DATA_DESC->sk->sk_data_ready = cb_data;
#endif

			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of data messages from minmex will be done with UDP on port %d\n", port);	
			break;
		}

#ifdef USER
		case UNIX:{ 
			struct sockaddr_un *my_addr=NULL;
			char *sockPath = myDataURI + 7; /* + 7 is for "unix://" */
			
			/* <message formatting>*/
			
			spec->attachMessage[messageIndex]= (unsigned char) UNIX;
			len = strlen(sockPath) ;
			spec->attachMessage[messageIndex+1]= (unsigned char) len;
			memcpy(spec->attachMessage+messageIndex+2,  sockPath, len);
			
			messageIndex = messageIndex + len + 2;
			
			/* <End of message formatting> */  
			
			spec->myDataMode = UNIX;
	
			if(sockCreateWrapper( &sock , AF_UNIX, SOCK_DGRAM, 0)<0){
                                anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick Data gate descriptor (UNIX) \n");
				closeWrapper(&(NODE_CONTROL_DESC));
                                closeWrapper(&(MY_CONTROL_DESC));
                                if(spec->myControlMode == PIPES)

					close(((int *)spec->myControlAux)[0]);

				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			MY_DATA_DESC = sock;

			/* configuring local address for listening socket */
			my_addr = (struct sockaddr_un  *) malloc(sizeof(struct sockaddr_un));
			
			my_addr->sun_family = AF_LOCAL;
        		strcpy(my_addr->sun_path, sockPath);
		
			unlink(sockPath);
			
			if(bindWrapper(&(MY_DATA_DESC), (struct sockaddr *) my_addr, 
                                       SUN_LEN(my_addr))<0) {
				anaPrint(ANA_ERR, "initMinmexSpecs : Unable to bind brick "
                                                     "Unix Data socket %s\n", sockPath);
                               closeWrapper(&(NODE_CONTROL_DESC));
                               closeWrapper(&(MY_CONTROL_DESC));

                                if(spec->myControlMode == PIPES)
                                        close(((int *)spec->myControlAux)[0]);

                                free(spec->attachMessage);
                                closeWrapper(&(MY_DATA_DESC));
                                free(spec->minmexControlAux);
				free(my_addr);
                                free(spec);
                                return NULL;
			}
			
			/* set permissions to rwxrwxrwx (because vlink is ran as root) */
			
			if (chmod(sockPath, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IWOTH|S_IXOTH) == -1) {
				anaPrint(ANA_ERR, "initMinmexSpecs : unable to set permissions on UNIX socket %s", sockPath);
			}
			
			/* Setting the timeout value for the socket */
			timeo.tv_sec  = timeoutVal;
			timeo.tv_usec = 0;
			if (setSockOptWrapper(&(MY_DATA_DESC), SOL_SOCKET, SO_RCVTIMEO, (char *)&timeo, sizeof(timeo)) < 0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : unable to set timeout on Data socket\n");
                                closeWrapper(&(NODE_CONTROL_DESC));
                              	closeWrapper(&(MY_CONTROL_DESC));

				if(spec->myControlMode == PIPES)
					close(((int *)spec->myControlAux)[0]);

				free(spec->attachMessage);
                                closeWrapper(&(MY_DATA_DESC));
				free(spec->minmexControlAux);
				free(my_addr);
				free(spec);
			 	return NULL;
			}
			
			spec->myDataAux = my_addr;
			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of data messages from minmex will be done with UNIX on %s\n", sockPath);	
			break;
		}
		case PIPES:{
			int dummyDesc;
			char *pipePath = myDataURI + 7; /* +7 for pipe:// */ 
			spec->myDataMode = PIPES;

			/* <message formatting>*/
			
			spec->attachMessage[messageIndex]= (unsigned char) PIPES;
			len = strlen(pipePath) ;
			spec->attachMessage[messageIndex+1]= (unsigned char) len;
			memcpy(spec->attachMessage+messageIndex+2,  pipePath, len);
			
			messageIndex = messageIndex + len + 2;
			
			/* <End of message formatting> */ 
			
			
			/* we create the FIFO pipe */
			/* the pipe is created with access rights to other users */
			if ((mkfifo(pipePath, 0666) == -1) && (errno != EEXIST)) {
    			anaPrint(ANA_ERR, "InitMinmexSPecs :Error creating data named pipe errno=%d\n",errno);
        		return NULL;
    		}
			
			/* we open the the pipe in read mode */
			spec->myDataDescriptor = open(pipePath, O_NONBLOCK|O_RDONLY);
    					
			if(spec->myDataDescriptor <0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to create brick data gate descriptor \n");
                                closeWrapper(&(NODE_CONTROL_DESC));
                               	closeWrapper(&(MY_CONTROL_DESC));

				if(spec->myControlMode == PIPES)
					close(((int *)spec->myControlAux)[0]);

				free(spec->myControlAux);
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			/* we open the pipe in write mode : [WARNING] we open the write mode to always have a writing descriptor pointing at the fifo
				otherwise, on all bricks detachment, the pipes would be no more useable (even worse with select causing tight loop),
				in which case we would have to close the pipe and reopen it => therefore we open this dummy descriptor, DONT USE IT TO WRITE
			 */
			 dummyDesc = open(pipePath, O_NONBLOCK|O_WRONLY);
			 if(dummyDesc <0) {
				anaPrint(ANA_ERR,  "initMinmexSpecs : Unable to open dummy Descriptor on Control gate FIFO\n");
                                closeWrapper(&(NODE_CONTROL_DESC));
                               	closeWrapper(&(MY_CONTROL_DESC));

                                if(spec->myControlMode == PIPES )
					close(((int*)spec->myControlAux)[0]);

				free(spec->myControlAux);
				free(spec->attachMessage);
				free(spec->minmexControlAux);
				free(spec);
				return NULL;
			}
			
			/* in the myDataAux we store a 2 integers array: containing the dummy descriptor(index 0) and the timeout value(index 1) */
			spec->myDataAux = (int *) malloc(2*sizeof(int)); 
			((int*)spec->myDataAux)[0] = dummyDesc;
			((int*)spec->myDataAux)[1] = timeoutVal;
		
			anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of data messages from minmex will be done with NAMED PIPES on %s\n", pipePath);	
			break;
						
		}
#ifdef USERGEN
                case GENETLINK:
                {
                        spec->attachMessage[messageIndex]= (unsigned char) GENETLINK;
			len = sizeof(int);
			spec->attachMessage[messageIndex+1]= (unsigned char) len;
                        int pid = generate_local_port(); /* generate a local identifier to which data can be sent */
                        memcpy(spec->attachMessage+messageIndex+2, &pid, len);
			
			messageIndex = messageIndex + len + 2;
			spec->myDataMode = GENETLINK;
                        spec->myDataAux = malloc(sizeof(int));
                        *((int *)(spec->myDataAux)) = pid;

                        spec->myDataDescriptor = create_nl_socket(NETLINK_GENERIC, 0, pid);
                        if(spec->myDataDescriptor <0) {
                                anaPrint(ANA_ERR,  "initMinmexSpecs :"
                                                       " Unable to create minmex Data gate descriptor \n");
                                free(spec->attachMessage);
                                free(spec);
                                return NULL;
                        }
               		anaPrint(ANA_DEBUG, "initMinmexSpecs : Reception of data messages from minmex will be done with GENERIC NETLINK\n ");	
                        break;
                }
#endif
#endif

		default: {
			anaPrint(ANA_NOTICE, "initMinmexSpecs : mode %d for receiving data from minmex is unhandled\n", myDataMode);
                        
	    	        closeWrapper(&(NODE_CONTROL_DESC));

			free(spec->attachMessage);
			free(spec->minmexControlAux);
			free(spec);
			return NULL;
		}
	}
	/* we add the name of the Brick before the optional desired mode*/
	
	len = strlen(mymodename);
	if(len>(CTRLMSG_MAXSIZE- messageIndex))  /* if the module name length is bigger than space left in the attach message buffer*/
		len = CTRLMSG_MAXSIZE- messageIndex;
		
	/*if the size is bigger than 255 set it to 255 since the name size is coded on 1 Byte*/
	if(len>255)  /* if the module name length is bigger than space left in the attach message buffer*/
		len = 255;
	

	spec->attachMessage[messageIndex]= (unsigned char) len;
	memcpy(spec->attachMessage+messageIndex+1, mymodename, len);
	
	messageIndex = messageIndex + len + 1;
	
	
	switch(desiredDataMode){

        case UDP: {
                spec->attachMessage[messageIndex]= (unsigned char) UDP;
                messageIndex = messageIndex+1;
                break;
        }

#ifdef USER 
	case UNIX: {
                spec->attachMessage[messageIndex]= (unsigned char) UNIX;
                messageIndex = messageIndex+1;
                break;
        }
        case PIPES: {
                spec->attachMessage[messageIndex]= (unsigned char) PIPES;
                messageIndex = messageIndex+1;
                break;
        }
#ifdef USERGEN
        case GENETLINK: {
                spec->attachMessage[messageIndex]= (unsigned char) GENETLINK;
                messageIndex = messageIndex+1;
                break;
        }  
#endif
#endif

        default: {
                anaPrint(ANA_NOTICE, "initMinmexSpecs : [Warning] desired mode %d "
                                     "for data sending to minmex is unhandled \n", desiredDataMode);
        }
	}
	
		
	/*  we store the size of the attach message in the age field of the minmexSpec 
            so that attach can know the number of bytes to indicate to sendto() or write() 
            without doing strlen() that causes bugs (if the UDP mode is on for control, 
            strlen returns 1 since the attach mesage starts as 'a0...' , 
            and we end up sending only one byte 'a' to the minmex)
            The age field will be replaced with the correct value if the attach succeeds 
		
	*/
	spec->sizeAttachMsg = messageIndex;
	
	return spec;
}

/* Closes the descriptors opened for a minmex and frees the allocated memory and shadow IDT*/
void freeMinmexSpecs(struct anaMinmexSpecs_s * spec){
	if(spec==NULL){
		anaPrint(ANA_ERR, "freeMinmexSpecs : minmexSpec is already NULL\n");
		return ;
	}

        closeWrapper(&(NODE_CONTROL_DESC));
        closeWrapper(&(NODE_DATA_DESC));
        closeWrapper(&(MY_DATA_DESC));
        closeWrapper(&(MY_CONTROL_DESC));

	
#ifdef USER	
	if(spec->myDataMode==UNIX) /*we need to unlink the sockets*/
		unlink(((struct sockaddr_un *)spec->myDataAux)->sun_path);
		
	if(spec->myControlMode==UNIX) /*we need to unlink the sockets*/
		unlink(((struct sockaddr_un *)spec->myControlAux)->sun_path);
	
	if(spec->myControlMode == PIPES)
		close(((int*)spec->myControlAux)[0]); /* If PIPES, we must close the dummy writing descriptor */

	if(spec->myDataMode == PIPES)
		close(((int*)spec->myDataAux)[0]); /* If PIPES, we must close the dummy writing descriptor */
#endif
	
        free(spec->myDataAux);
        free(spec->myControlAux);

        free(spec->minmexControlAux);
     	spec->minmexControlAux = NULL;

	free(spec->minmexDataAux);

        free(spec->attachMessage);
        free(spec->buffer);
        freeSDTTable(spec->shadowT);
     	free(spec);
}
