/* C/bricks/API/anaLib0_gates.c
 * 
 * implementation of API level 0
 * 
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> (inital design, user space)
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> (kernel)
 * 	
 * 	   	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */


#include "anaLib0.h" 
#include "AL0GatesInternal.h"
#include "anaThread.h"
//#include "anaError.h"

#include <arpa/inet.h>
#include <netinet/in.h>

#ifdef USERGEN
#include <linux/genetlink.h> 
#include <linux/netlink.h>
#endif

/*socketWrapper functions*/
#ifdef KERNEL
struct workqueue_struct *wq;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
DECLARE_DELAYED_WORK( work_hbc, anaL0multi_heartBeat); 
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
DECLARE_WORK(work_hbc, anaL0multi_heartBeat,NULL);
#endif

struct wq_wrapper wq_data;
#endif

#ifdef USER
int anaDebugON=1;
int attached = 0;

#ifdef USERGEN
/*functions for genetlink*/ 
/* generate_local_port is taken from the libnl library
 * it generates a unique identifier. This is needed that the kernel
 * may send generic netlink messages to the correct userspace 
 * generic netlink sockets. If only one socket would be open per process
 * the pid would suffice
 */
static uint32_t used_ports_map[32];
uint32_t generate_local_port(void)
{
        int i, n;
        uint32_t pid = getpid() & 0x3FFFFF;

        for (i = 0; i < 32; i++) {
                if (used_ports_map[i] == 0xFFFFFFFF)
                        continue;

                for (n = 0; n < 32; n++) {
                        if (1UL & (used_ports_map[i] >> n))
                                continue;

                        used_ports_map[i] |= (1UL << n);
                        n += (i * 32);

                        /* PID_MAX_LIMIT is currently at 2^22, leaving 10 bit
                         * to, i.e. 1024 unique ports per application. */
                        return pid + (n << 22);

                }
        }

        /* Out of sockets in our own PID namespace, what to do? FIXME */
        return UINT_MAX;
}	
 

/* generic netlink functions have been taken from getdelay.c file of the Linux Kernel */
/* Create a raw netlink socket and bind */
int create_nl_socket(int protocol, int groups, int pid)
{
        int fd;
        struct sockaddr_nl local;
        
        fd = socket(AF_NETLINK, SOCK_RAW, protocol);
        if (fd < 0)
                return -1;
        
        memset(&local, 0, sizeof(local));
        local.nl_family = AF_NETLINK;
        local.nl_groups = groups;
        local.nl_pid = pid;
        int i = -1;    
        if ((i = bind(fd, (struct sockaddr *) &local, sizeof(local))) < 0)
                goto error;
        return fd;
 error:
        close(fd);
        return -1;
}

/*
 * Send netlink message to kernel
 */
int sendto_fd(int s, const char *buf, int bufLen)
{
        struct sockaddr_nl nladdr;
        int r;
        
        memset(&nladdr, 0, sizeof(nladdr));
        nladdr.nl_family = AF_NETLINK;
	while ((r = sendto(s, buf, bufLen, 0, (struct sockaddr *) &nladdr,
                           sizeof(nladdr))) < bufLen) {
                if (r > 0) {
                        buf += r;
                        bufLen -= r;
                } else if (errno != EAGAIN)
                        return -1;
        }
        return 0;
}
/*
 * Probe the controller in genetlink to find the family id
 * for the family with the given name
 */
int get_family_id(int sd, char * family)
{
        struct {
                struct nlmsghdr n;
                struct genlmsghdr g;
                char buf[256];
        } family_req;
        
        struct {
                struct nlmsghdr n;
                struct genlmsghdr g;
                char buf[256];
        } ans;
	memset(&family_req, 0, sizeof(family_req));
	memset(&ans, 0, sizeof(ans));
	
        int id = -1;
        struct nlattr *na;
        int rep_len;
        /* Get family name */
        family_req.n.nlmsg_type = GENL_ID_CTRL;
        family_req.n.nlmsg_flags = NLM_F_REQUEST;
        family_req.n.nlmsg_seq = 0;
        family_req.n.nlmsg_pid = getpid();
        family_req.n.nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN);
        family_req.g.cmd = CTRL_CMD_GETFAMILY;
        family_req.g.version = 0x1;
        
        na = (struct nlattr *) GENLMSG_DATA(&family_req);
        na->nla_type = CTRL_ATTR_FAMILY_NAME;
      	na->nla_len = strlen(family) + 1 + NLA_HDRLEN;
	if(na->nla_len < 256)
                strcpy(NLA_DATA(na), family);
        else
                return -1;
	
        family_req.n.nlmsg_len += NLMSG_ALIGN(na->nla_len);
	
        if (sendto_fd(sd, (char *) &family_req, family_req.n.nlmsg_len) < 0){
		anaPrint(ANA_ERR, "error sending message via Netlink\n");
		return -1;
	}
	rep_len = recv(sd, &ans, sizeof(ans), 0);
        if (rep_len < 0){
                anaPrint(ANA_ERR, "error receiving reply message via Netlink\n");
                return -1;
        }
       
        /* Validate response message */
        if (!NLMSG_OK((&ans.n), rep_len)){
                anaPrint(ANA_ERR, "invalid reply message received via Netlink\n");
                return -1;
        }

        if (ans.n.nlmsg_type == NLMSG_ERROR) { /* error */
                anaPrint(ANA_ERR, "error received NACK - leaving\n");
                return -1;
        }

        na = (struct nlattr *) GENLMSG_DATA(&ans);
        na = (struct nlattr *) ((char *) na + NLA_ALIGN(na->nla_len));
        if (na->nla_type == CTRL_ATTR_FAMILY_ID) {
                id = *(__u16 *) NLA_DATA(na);
        }
        return id;
}
#endif //USERGN

void notifWrapper(void *data, int len, anaLabel_t input, void *aux) {

	/* received format: [eventID(4bytes)][handler pointer][threadFlag][label][viewer][destLabel][destViewer][name if any]*/
	/*struct anaEventNotif_s *notif*/
	struct anaEventNotif_s *notif = NULL;
	uint32_t *ip = NULL;
	anaLabel_t *lp = NULL;
	anaHandle_t *hp = NULL;
	char *sp = NULL;
	
	int offset = 0;
	eventHandler_t handler; 
	int threadFlag = 0;
	
	struct notifThreadArg *notifArg = NULL;
	
	anaPrint(ANA_DEBUG, "received Notification \n");
		
	notif = (struct anaEventNotif_s *) malloc(sizeof(struct anaEventNotif_s));
	
	ip = (uint32_t *) data;
	notif->eventID = ntohl(*ip);
	offset += sizeof(uint32_t);
	
		
	memcpy(&handler, data+offset, sizeof(void *)); 
	offset += sizeof(void *);
		
	sp = (char *) (data + offset);
	if(sp[0])	
		threadFlag = 1;
	offset += 1;
	
	lp = (anaLabel_t *) (data + offset);
	notif->label = ntohl(*lp); 
	offset += sizeof(anaLabel_t);

	hp = (anaHandle_t *) (data + offset);
	notif->viewer = ntohl(*hp);
	offset += sizeof(anaHandle_t);

	lp = (anaLabel_t *) (data + offset);
	notif->destLabel = ntohl(*lp); 
	offset += sizeof(anaLabel_t);
	

	hp = (anaHandle_t *) (data + offset);
	notif->destViewer = ntohl(*hp);
	offset += sizeof(anaHandle_t);


	if(len > offset){
		sp = (char *) (data + offset);
		notif->name = (char *) malloc(strlen(sp)+1);	
	}
	
	if(threadFlag){
		notifArg = (struct notifThreadArg *) malloc(sizeof(struct notifThreadArg));
		notifArg->notif = notif;
		notifArg->handler = handler;
	
		/* we start the thread for the notification */
		_anaThread(0, (void *) threadNotification, notifArg, "notificationThread");
		return;
	}

	/* otherwise we call the handler locally */
	handler(notif);
	
	if(notif->name)
		free(notif->name);
	
	free(notif);
		
}
int anaL0_eventUnsubscribe(struct anaEventSub_s* sub){
	/* message format : ['g'][MY_HANDLE][eventMask][label][viewer]([name])*/	
	char buff[CTRLMSG_MAXSIZE];	
	
	int offset = 0;
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;	
	char *sp = NULL; 
	uint32_t *ip = NULL;
	int nbRead;

	memset(buff, 0, CTRLMSG_MAXSIZE);
	buff[0] = 'g';
	offset += 1;
	
	hp = (anaHandle_t *)(buff + offset); 
	*hp = htonl(MY_HANDLE);
	offset += sizeof(anaHandle_t);
	
	ip = (uint32_t *) (buff + offset);
	*ip = htonl(sub->eventMask);
	offset += sizeof(uint32_t); 

	lp = (anaLabel_t *) (buff + offset);
	*lp = htonl(sub->label);
	offset += sizeof(anaLabel_t);

	hp = (anaHandle_t *)(buff + offset); 
	*hp = htonl(sub->viewer);
	offset += sizeof(anaHandle_t);

	if(sub->name){
		sp = (char *) (buff + offset);
		strcpy(sp, sub->name);
		offset += (strlen(sub->name) + 1);
	}
	
	/* we send the message to the gates plugin */
	if(sendControl(myMinmex, buff, offset) < 0){
		anaPrint(ANA_ERR,  "anaL0_eventUnsubscribe : Unable to send Unsubscribe message to minmex\n");
                return -1;
	}
	
	/* we wait to receive an answer confirming the subscription or announcing an error at the minmex */
	memset(buff, 0, CTRLMSG_MAXSIZE);	
	nbRead = recvControl(myMinmex, buff);	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_eventUnsubscribe : Unable to receive reply from minmex\n");
              	return -1;
	}

	/* we check if the reply is an error message */	
	if(buff[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_eventUnsubscribe : Ana minmex returned error code: %s\n", getErrorString((int)buff[1]));
		return (int)buff[1];
	}	
	
	return 0;

}


int anaL0_eventSubscribe(struct anaEventSub_s* sub, eventHandler_t handler, int threadFlag) {
	/* message format : ['b'][MY_HANDLE][handler pointer][threadFlag(1 byte)][eventMask][label][viewer]([name])*/	
	char buff[CTRLMSG_MAXSIZE];	
	int offset = 0;
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;	
	char *sp = NULL; 
	uint32_t *ip = NULL;
	int nbRead;

	memset(buff, 0, CTRLMSG_MAXSIZE);
	buff[0] = 'b';
	offset += 1;
	
	hp = (anaHandle_t *)(buff + offset); 
	*hp = htonl(MY_HANDLE);
	offset += sizeof(anaHandle_t);
	
	/* we copy the pointer to the handler in the buffer */
	memcpy(buff+offset, &handler, sizeof(void *));	
	offset += sizeof(void *);
	
	if(threadFlag){
		sp = buff + offset;
		sp[0] = 1;
	}
	offset += 1;
	
	ip = (uint32_t *) (buff + offset);
	*ip = htonl(sub->eventMask);
	offset += sizeof(uint32_t); 

	lp = (anaLabel_t *) (buff + offset);
	*lp = htonl(sub->label);
	offset += sizeof(anaLabel_t);

	hp = (anaHandle_t *)(buff + offset); 
	*hp = htonl(sub->viewer);
	offset += sizeof(anaHandle_t);

	if(sub->name){
		sp = (char *) (buff + offset);
		strcpy(sp, sub->name);
		offset += (strlen(sub->name) + 1);
	}
	
	/* we send the message to the gates plugin */
	if(sendControl(myMinmex, buff, offset) < 0){
		anaPrint(ANA_ERR,  "anaL0_eventSubscribe : Unable to send subscribe message to minmex\n");
                return -1;
	}
	
	/* we wait to receive an answer confirming the subscription or announcing an error at the minmex */
	memset(buff, 0, CTRLMSG_MAXSIZE);	
	nbRead = recvControl(myMinmex, buff);	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_eventSubscribe : Unable to receive reply from minmex\n");
              	return -1;
	}

	/* we check if the reply is an error message */	
	if(buff[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_eventSubscribe : Ana minmex returned error code: %s\n", getErrorString((int)buff[1]));
		return (int)buff[1];
	}	
	
	return 0;
}

int anaL0_eventNotify( struct anaEventNotif_s *notif){
	
	char buff[CTRLMSG_MAXSIZE];	
	int offset = 0;
	int nbRead;
	uint32_t *ip;
	anaLabel_t *lp;
	anaHandle_t *hp;
	char *tmp = NULL;

	memset(buff, 0, CTRLMSG_MAXSIZE);

	/* we serialize the notification structure */
		/* format ['j'][MY_HANDLE][eventID(4bytes)][label][viewer][destLabel][destViewer][name if any] */
	
	buff[0] = 'j';
	offset += 1 ;
	
	hp = (anaHandle_t *) (buff + offset);
	*hp = htonl(MY_HANDLE);
	offset += sizeof(anaHandle_t);
	
	ip = (uint32_t *) (buff +offset);
	*ip = htonl(notif->eventID);
	offset +=  sizeof(uint32_t);
	

	lp = (anaLabel_t *) (buff + offset);
	*lp = htonl(notif->label);
	offset +=  sizeof(anaLabel_t);
	
	hp = (anaHandle_t *) (buff + offset);
	*hp = htonl(notif->viewer);
	offset += sizeof(anaHandle_t);

	lp = (anaLabel_t *) (buff + offset);
	*lp = htonl(notif->destLabel);
	offset +=  sizeof(anaLabel_t);

	hp = (anaHandle_t *) (buff + offset);
	*hp = htonl(notif->destViewer);
	offset += sizeof(anaHandle_t);

	if(notif->name){
		tmp = (char *) (buff+offset);
		strcpy(tmp, notif->name);
		offset += strlen(notif->name) + 1 ;
	}
	
	
	/* we send the message to the gates plugin */
	if(sendControl(myMinmex, buff, offset) < 0){
		anaPrint(ANA_ERR,  "anaL0_eventNotify : Unable to send notification message to minmex\n");
                return -1;
	}
	
	/* we wait to receive an answer confirming the subscription or announcing an error at the minmex */
	memset(buff, 0, CTRLMSG_MAXSIZE);	
	nbRead = recvControl(myMinmex, buff);	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_eventNotify : Unable to receive reply from minmex\n");
              	return -1;
	}

	/* we check if the reply is an error message */	
	if(buff[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_eventNotify : Ana minmex returned error code: %s\n", getErrorString((int)buff[1]));
		return (int)buff[1];
	}	
	
	return 0;
}
	


#endif //USER

/*	minmexSpec : contains the specifics of the minmex to attach to 
 *	(minmex control gate, brick data/control gates dedicated to this minmex)
 * 	This structure is also filled(complemented) according to the attach reply.	 
 *	tableBitSize : size of the Mask used by the hash function => size of the IDT table must be 2^tableBitSize
 */
int anaL0_attach(struct anaMinmexSpecs_s * minmexSpec, int tableBitSize) {
	char buff[CTRLMSG_MAXSIZE];
	int sizeAttachMsg = 0, ret = -1, dataAuxLength;

	int minmexPort;
	char *minmexIP=NULL;
	char *minmexUDPport=NULL;
	struct sockaddr_in *tmp=NULL;


	struct SDTentry_s *entry=NULL;
	struct SDTentry_s *notifEntry=NULL;
	anaLabel_t myNotifLabel;
	
	anaLabel_t *lp;
	anaHandle_t *hp;
#ifdef USER
	int sock;
	#define NODE_DATA_DESC minmexSpec->minmexDataDescriptor
#endif
#ifdef KERNEL
	struct socket *sock=NULL;
	#define NODE_DATA_DESC minmexSpec->minmexDataSocket
#endif
	
	int minmexDataMode = -1;
	char *minmexDataGateChar = NULL;
        uint32_t *tmpInt = NULL;
      
        if(minmexSpec == NULL ||tableBitSize<=0 ){
                anaPrint(ANA_ERR, "anaL0 attach: invalid minmexSpec\n");
                return -1;
        }

	if(attached){
		anaPrint(ANA_ERR, "anaL0 attach: This brick is already attached to a minmex \n");
                return -1;
	}
	
	myHBLabel = generateLabel(); /* there is a slight risk of collisions here, should be fixed in future */
	myNotifLabel = 	generateLabel();
	
#ifdef USER
	srandom(time(0));
#endif
/*the kernel uses get_random_bytes, which does not need to be initialized*/

	sizeAttachMsg = minmexSpec->sizeAttachMsg; 
	/* we try to add the heartBeat label to the attach message */
	
	sizeAttachMsg+=sizeof(anaLabel_t);
	
	if(sendControl(minmexSpec, minmexSpec->attachMessage, sizeAttachMsg) < 0) {
        	anaPrint(ANA_ERR, "Attach : Unable to send attach message\n");
                return -1;
        }
        /* < Reception attach reply> */
        memset(buff, 0, CTRLMSG_MAXSIZE);
        ret = recvControl(minmexSpec, buff); 
        if(ret <= 0) {
                anaPrint(ANA_ERR,  "Unable to receive attach reply from minmex\n");
                return -1;
        }
        if(buff == NULL) {
                anaPrint(ANA_ERR,  "buff == NULL\n");
                return -1;
        } 
              
	/* <Reply message parsing >*/
	/*firt we check if the response is an error code */
	if (buff[0]=='e') { 	/* minmex sent an error code*/
		anaPrint(ANA_ERR,   "Attach : Ana minmex returned error code : %s \n", buff);
	    return -1;
	}
			
	/* <Reply message format >
	['r'][handle][controlLabel][minmexDataMode][dataAuxLength][dataAux]  */

		
	if(ret < sizeof(anaHandle_t) +  sizeof(anaLabel_t)) { 
		anaPrint(ANA_ERR,  "Attach : Bad attach reply from minmex\n");
		return -1;
	}
	
	minmexSpec->buffer = malloc(DEFAULT_UDP_MTU*sizeof(char));
	
	tmpInt = (uint32_t*) (buff+1); /* careful, this should be changed in case we move to 64 bits handles*/
	minmexSpec->handle = ntohl(*tmpInt);
	myHandle = minmexSpec->handle;
	
	
	tmpInt = (uint32_t*) (buff+1+sizeof(anaLabel_t)); /* careful, this should be changed in case we move to 64 bits handles*/
	minmexSpec->minmexCompLabel = ntohl(*tmpInt);
	myNodeCompLabel = minmexSpec->minmexCompLabel;
	
	
	minmexDataMode = (int)buff[1 + sizeof(anaLabel_t) + sizeof(anaHandle_t)];
	anaPrint(ANA_DEBUG,   "Attach : minmex Data Mode : %d \n", minmexDataMode);

	dataAuxLength = (int)buff[2 + sizeof(anaLabel_t) + sizeof(anaHandle_t)];
	
	minmexDataGateChar = (char*) (buff + 3 + sizeof(anaLabel_t) + sizeof(anaHandle_t));
	
	/* <End of reply message parsing >*/
		
	/* <complementing minmexSpec according to the Response> */
	/* < adding minmexs data gate to the minmexSpecs > */	
	switch(minmexDataMode) {

	case UDP:{ 
		minmexSpec->minmexDataMode = UDP;
                if(sockCreateWrapper(&sock, PF_INET, SOCK_DGRAM, 0)<0){
                        anaPrint(ANA_ERR,  "Attach :Unable to create minmex Data gate descriptor \n");
			free(minmexSpec->buffer);
			minmexSpec->buffer = NULL;
                        return -1;
		}
		
		NODE_DATA_DESC = sock;
			
		/*in this case the minmexDataGateChar will be fromatted as : [IP]:[PORT]*/
                minmexIP = strsep(&minmexDataGateChar, ":");
		minmexUDPport = strsep(&minmexDataGateChar, ":");
              
		if(minmexUDPport == NULL || minmexIP == NULL) {
			anaPrint(ANA_ERR,  "Attach : Bad attach reply from minmex \n");
			free(minmexSpec->buffer);
			minmexSpec->buffer = NULL;
			closeWrapper(&(NODE_DATA_DESC)) ;
			return -1;
		}
		
		minmexPort = atoi(minmexUDPport);
		tmp = (struct sockaddr_in*) malloc(sizeof(struct sockaddr_in));
		if(tmp==NULL){
			anaPrint(ANA_ERR,  "Attach : Bad attach reply from minmex \n");
			free(minmexSpec->buffer);
			minmexSpec->buffer = NULL;
			closeWrapper(&(NODE_DATA_DESC)) ;
			return -1;
		}
		memset(tmp, 0, sizeof(struct sockaddr_in));
		tmp->sin_addr.s_addr = inet_addr(minmexIP);
		tmp->sin_family = AF_INET;
		tmp->sin_port = htons(minmexPort);
		
		minmexSpec->minmexDataAux = tmp;
		anaPrint(ANA_DEBUG, "Sending data messages to minmex will be done with UDP at IP %s on port %d \n",minmexIP, minmexPort);	
		break;
	}
#ifdef USER
	case UNIX:{ 
		struct sockaddr_un *tmpu;
		char *sockPath = minmexDataGateChar;
		minmexSpec->minmexDataMode = UNIX;
                if(sockCreateWrapper(&sock, AF_UNIX, SOCK_DGRAM, 0)<0){
                        anaPrint(ANA_ERR,  "Attach :Unable to create minmex Data gate descriptor \n");
			free(minmexSpec->buffer);
                        return -1;
		}
		
		NODE_DATA_DESC = sock;
			
		tmpu = (struct sockaddr_un*) malloc(sizeof(struct sockaddr_un));
		if(tmpu==NULL){
			anaPrint(ANA_ERR,  "Attach :Unable to create minmex Data gate descriptor \n");
			free(minmexSpec->buffer);
                        return -1;
		}
		tmpu->sun_family = AF_UNIX;
		strcpy(tmpu->sun_path, sockPath);
		
		minmexSpec->minmexDataAux = tmpu;
		anaPrint(ANA_DEBUG, "Sending data messages to minmex will be done with UNIX at %s \n", sockPath);	
		break;
	}
	case PIPES:{  /*minmexDAtaGateURL : pipe://pipe_path*/
		char *pipePath = minmexDataGateChar ;
		minmexSpec->minmexDataMode = PIPES;
		minmexSpec->minmexDataDescriptor = open(pipePath, O_NONBLOCK|O_WRONLY);
		if(minmexSpec->minmexDataDescriptor<0) {
			anaPrint(ANA_ERR,  "Attach :Unable to create minmex Data gate descriptor (PIPES)\n");
			free(minmexSpec->buffer);
			return -1;
		}
		
		anaPrint(ANA_DEBUG, "Sending data messages to minmex will be done with NAMED PIPES at %s \n",pipePath);	
		break;
	
	}
#ifdef USERGEN
        case GENETLINK:{ 
                char * temp;
                int * id;
		minmexSpec->minmexDataMode = GENETLINK;
                if(minmexSpec->myDataMode != GENETLINK){
			/* generates a unique identifier to which generic netlink messages are sent */
                        int pid = generate_local_port();
                        minmexSpec->minmexDataDescriptor = create_nl_socket(NETLINK_GENERIC, 0, pid);
                        if(minmexSpec->minmexDataDescriptor <0) {
                                anaPrint(ANA_ERR, "Attach :Unable to create minmex Data gate descriptor (GENETLINK)\n");
                                free(minmexSpec->buffer);
				minmexSpec->buffer = NULL;
                                return -1;
                        };
                      
                }
                else{
                        minmexSpec->minmexDataDescriptor = minmexSpec->myDataDescriptor;
                }
                temp = (char *)minmexDataGateChar + 12; /*for genetlink*/
		anaPrint(ANA_DEBUG, "looking for family %s\n", temp);
                id = (int *) malloc(sizeof (int));
		/* converts the generic netlink name (e.g. "ANA") to the id to be used to send data too*/ 
		*id = get_family_id(minmexSpec->minmexDataDescriptor,temp);
		minmexSpec->minmexDataAux = id;
		anaPrint(ANA_DEBUG, "Sending data messages to minmex will be done with GENETLINK at %i \n",*id);
                break;
	
	}  
#endif
#endif

	default:{
		anaPrint(ANA_NOTICE, "Attach : mode %d for sending data messages to minmex is unhandled\n", minmexDataMode);
		return -1;	
	}
	}	
	
	/* <Shadow IDT initialization>*/
	minmexSpec->shadowT = (struct shadowIDT_s *) malloc(sizeof(struct shadowIDT_s));
		
	if (initializeSDT (minmexSpec->shadowT, tableBitSize)<0) {
		anaPrint(ANA_ERR,  "Attach : Unable to allocate memory for Shadow IDT \n");
		free(minmexSpec->shadowT);	
		free(minmexSpec->buffer);
		free(minmexSpec->minmexDataAux);
		closeWrapper(&(NODE_DATA_DESC));
		return -1;
	}
	/* <End of Shadow IDT initialization>*/	
	
	/* <End of complementing minmexSpec according to the Response> */
		
	/* < launching/notifying thread listening on myDataGate > */
	
		attached = 1;
		/* we start the listenLoop thread */
#ifdef USER	
	
	/* we add our heartbeat entry to the SDT */

	
	entry = (struct SDTentry_s *) malloc(sizeof(struct SDTentry_s));
	entry->label = myHBLabel;
	entry->callBackFct = (anaCallback_t)handleHbeat;

	addSDTEntry(myMinmex->shadowT,entry);
	
	notifEntry = (struct SDTentry_s *) malloc(sizeof(struct SDTentry_s));
	notifEntry->label = myNotifLabel;
	notifEntry->callBackFct = (anaCallback_t)notifWrapper;

	addSDTEntry(myMinmex->shadowT, notifEntry);
	/* we send a message to the gates plugin to declare this notification label */
		/* format ['n'][MY_HANDLE][notifLabel] */
	buff[0] = 'n';
	hp = (anaHandle_t *) (buff+1);
	*hp = htonl(myHandle);	
		
	lp = (anaLabel_t *) (buff + 1 + sizeof(anaHandle_t));
	*lp = htonl(myNotifLabel);	
	
	if(sendControl(minmexSpec, buff, 1+sizeof(anaHandle_t)+sizeof(anaLabel_t)) < 0) {
        	anaPrint(ANA_ERR, "Attach : Unable to send notification label\n");
               if(freeSDTTable(minmexSpec->shadowT)<0)
				anaPrint(ANA_ERR, "Attach : After failure, Unable to free IDT table\n");
			
		free(minmexSpec->buffer);
		free(minmexSpec->minmexDataAux);
		closeWrapper(&(NODE_DATA_DESC));
	       
	        return -1;
        }	
	
	
		anatimer_init();
		
		if ( anatimer_add(HEARTBEAT_INTERVAL*1000, 1, 1, (fct_t)anaL0_heartBeat, NULL, 0) == -1 ){
			anaPrint(ANA_ERR, "Attach: problem adding heartBeat timer\n");
			if(freeSDTTable(minmexSpec->shadowT)<0)
				anaPrint(ANA_ERR, "Attach : After failure, Unable to free IDT table\n");
			
			free(minmexSpec->buffer);
			free(minmexSpec->minmexDataAux);
			closeWrapper(&(NODE_DATA_DESC));
			return -1;
		}
                
		/*ret = pthread_create(&loopThread, NULL, (void *) listenLoop, NULL);
                pthread_detach(loopThread); */
		
		ret = _anaThread(0, (void *)listenLoop, NULL, "dataLoop");
		
		sleep(1); /*hack to avoid that listen threads starts to late...*/
                
                if (ret == 0) {
                        anaPrint(ANA_ERR,  "Attach : problem creating listenLoop thread \n");
			if(freeSDTTable(minmexSpec->shadowT)<0)
				anaPrint(ANA_ERR, "Attach : After failure, Unable to free IDT table\n");
			
			free(minmexSpec->buffer);
			free(minmexSpec->minmexDataAux);
			closeWrapper(&(NODE_DATA_DESC));
			return -1;
		}

		
#endif
#ifdef KERNEL
		/* initialize the work queue for the UDP mode */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                if(minmexSpec->myDataMode == UDP){
                        INIT_WORK(&wq_data.worker, parseDataTemp);
                }
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                if(minmexSpec->myDataMode == UDP){
                        INIT_WORK(&wq_data.worker, parseDataTemp, &wq_data);
                }
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		tq_udp.routine = parseDataTemp;
		tq_udp.data = (void *)&wq_data;
#endif
#endif
		
		
	
	
	/* < End of launching/notifying thread listening on myDataGate > */	
	

#ifdef USER
	signal(SIGINT, anaL0_stopANA);
#endif

#ifdef KERNEL
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	wq = create_singlethread_workqueue( "hbcWq" ); 
                if (!wq)
                        return -ENOMEM;
                if( !queue_delayed_work( wq, &work_hbc, HZ ) ) {                 	
                        anaPrint(ANA_ERR,  "queue_work not successful, no heart beat check\n");
                }
#else
		tq_hbc.routine = anaL0multi_heartBeat;
		tq_hbc.data = NULL;
		/* for the heart beat check, setup a timer.
 		 * upon firing the sleepfunction is executed 
 		 * which inserts the anaL0_heartBeat in the task queue
 		 */		
		init_timer(&timerA);
		timerA.function = sleepfunction;
		timerA.data = 0;
		timerA.expires = jiffies + HZ * HEARTBEAT_INTERVAL;
		add_timer(&timerA);
#endif
        /* register the data receive function of the brick with the MINMEX */
	if(minmexSpec->myDataMode == KERN){
                struct anaMinmexFunctions * minmexfct = (struct anaMinmexFunctions *)minmexSpec->minmexControlAux;
                minmexfct->brickDataChannel(minmexSpec->myDataFunctions, minmexSpec->handle);
        }
     	schedule();
 
#endif /*kernel*/
        anaPrint(ANA_DEBUG, "attached successfully\n");
	return 0;
} 



/* the detachment function sends a detach message (containinfg the handle) to the minmex in order to inform it of the
	  brick departure, so that it can free all labels associated with this brick
	  the function sets also the global variable Attached to 0 in order to stop the listenLoop thread 
	detach() also removes this minmex from the minmexs list in order for listenLoop to stop expecting messages from it	  
*/
int anaL0_detach() {
	char msg[CTRLMSG_MAXSIZE];
	int ret=0;
	int attempts=3;
	
	anaHandle_t *handle = NULL;
	
	analock_lock(&bigLock);  
	msg[0] = 'd';
	handle = (anaHandle_t *) (msg+1);
	*handle = htonl(myMinmex->handle);
	
	/* in some case, example UNIX sockets, when the minmex crashes, the brick can not send the
	   detach messages and thus if we retrn -1. the brick cannot free the minmexSpec and would be
	   trying to send the detach message in an endless loop. Therefore, we try 3 times to send
	   the message, if it fails we free the minmexSPecs anyway .
	 */
	while(attempts>0 && sendControl(myMinmex, msg, sizeof(anaHandle_t)+1) < 0){
		anaPrint(ANA_ERR,  "Detach : Unable to send detach message %d attempts remaining \n",attempts-1);
		attempts--;
	}
		
	if(attempts==0)
		ret = -2; /* -2 return error code means detached without being able to send detach message*/
	


	attached = 0;
	
	/* we free all data and control gates*/
	freeMinmexSpecs(myMinmex);

	analock_unlock(&bigLock);  
	return ret;
}

int anaL0_setIDPinfo( struct IDPinfo_s *info) {
	anaHandle_t callee;
	
	char msg[CTRLMSG_MAXSIZE];
	anaHandle_t *hp = NULL;
	anaHandle_t *lp = NULL;
	int offset, nbRead;
	char *t;
	uint16_t *ip;
	
	if (info ==NULL)
		return -1;
	
	callee = (info->binding).brick.handle; 
	if(callee == 0) {
		info->IDPtype = BRICK;
		(info->binding).brick.handle = MY_HANDLE;
	}
	
	/* message format : ['s'][MY_HANDLE][label][callee][outputIDP][IC flag]([ctxtLen (2bytes )][ctxt][srvLen (2 bytes)][srv][MTU][chanType]) */
	msg[0] = 's';
	offset = 1;
	hp = (anaHandle_t *) (msg+offset);
	*hp = htonl(MY_HANDLE);
	offset += sizeof(anaHandle_t);
	
	lp = (anaLabel_t *) (msg+offset);
	*lp = htonl(info->label);
	offset += sizeof(anaLabel_t);
	
	hp = (anaHandle_t *) (msg+offset);
	*hp = htonl(info->binding.brick.handle);
	offset += sizeof(anaHandle_t);
	
	lp = (anaLabel_t *) (msg+offset);
	*lp = htonl(info->binding.brick.outputIDP);
	offset += sizeof(anaLabel_t);
	
	t = (char *) (msg+offset);
	t[0] = info->ICflag;
	offset += 1;

	if(info->ICflag){
		if(info->destContext.valueLen > 0 && info->destContext.value != NULL){
			ip = (uint16_t *)  (msg+offset);
			*ip = htons(info->destContext.valueLen);
			offset += sizeof(uint16_t);
			
			memcpy(msg+offset, info->destContext.value, info->destContext.valueLen);
			offset += info->destContext.valueLen;
						
		}
		else{
			ip = (uint16_t *)  (msg+offset);
			*ip = 0;
			offset += sizeof(uint16_t);
		}
		if(info->destService.valueLen > 0 && info->destService.value != NULL){
			ip = (uint16_t *)  (msg+offset);
			*ip = htons(info->destService.valueLen);
			offset += sizeof(uint16_t);
			
			memcpy(msg+offset, info->destService.value, info->destService.valueLen);
			offset += info->destService.valueLen;
						
		}
		else{
			ip = (uint16_t *)  (msg+offset);
			*ip = 0;
			offset += sizeof(uint16_t);
		}
		ip = (uint16_t *)  (msg+offset);
		*ip = htons(info->MTU);
		offset += sizeof(uint16_t);
		
		t = (char *) (msg+offset);
		t[0] = info->chanType;
		offset += 1;
	}
	
	if(sendControl(myMinmex, msg, offset ) < 0) {
                anaPrint(ANA_ERR,  "Unable to send setinfo request to minmex\n");
              	return -1;
        }
	
	/* we wait for the response */
	memset(msg, 0, CTRLMSG_MAXSIZE);
        nbRead = recvControl(myMinmex, msg);
	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_setIDPinfo : Unable to receive reply from minmex\n");
                return -1;
	}
	
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "setIDPinfo : Ana minmex returned error: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1];
	}
	
	return 0;	
}



int anaL0_getIDPinfo(anaLabel_t label,  anaHandle_t viewer, struct IDPinfo_s *result){

	/* the rquest message format : [i][myHandle][label][viewer] */

	char msg[CTRLMSG_MAXSIZE];
	anaHandle_t *hp = NULL;
	anaHandle_t *lp = NULL;
	int offset, nbRead;
	char *t;
	uint16_t *ip;
	memset(msg, 0, CTRLMSG_MAXSIZE);
	
	msg[0] = 'i';
	hp = (anaHandle_t *)(msg+1);
	*hp = htonl(MY_HANDLE);

	lp = (anaLabel_t *)(msg + 1 + sizeof(anaHandle_t));
	*lp = htonl(label);
	hp = (anaHandle_t *)(msg + 1 + sizeof(anaHandle_t) + sizeof(anaLabel_t));
	*hp = htonl(viewer);
	
	if(sendControl(myMinmex, msg, 1+sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t) ) < 0) {
                anaPrint(ANA_ERR,  "Unable to send getinfo request to minmex\n");
              	return -1;
        }

	/* we wait for the response */
	memset(msg, 0, CTRLMSG_MAXSIZE);
        nbRead = recvControl(myMinmex, msg);
	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_getIDPinfo : Unable to receive reply from minmex\n");
                return -1;
	}
	
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "getIDPinfo : Ana minmex returned error: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1];
	}
	
	/* response format : [i][label][viewer][owner][permFlag 1 byte][ICFlag]([contextLen (2 bytes)][context][serviceLen (2 bytes)][service][MTU (2 bytes)][chanType])[IDPtype(1 byte)][destination] */
		/* depending on the IDP types destination is :*/
		 	/* Brick : [callee handle][outpuIDP] */
			/* Redirect : [dest label][destViewer] */
			/* FORK : [nbForks 1 byte][dest Labe1][dest Viewer1]...*/
	
	lp = (anaLabel_t *)(msg + 1);
	result->label = ntohl(*lp);
	offset = 1+ sizeof(anaLabel_t);
	
	hp = (anaHandle_t *) (msg + offset);
	result->viewer = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	hp = (anaHandle_t *) (msg + offset);
	result->owner = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	t = (char *) (msg + offset);
	if(t[0])
		result->permFlag = 1;
	else
		result->permFlag = 0;
	
	offset += 1;
	
	t = (char *) (msg + offset);
	if(t[0]){
		int ctxLen, srvLen;
		
		result->ICflag = 1;
		offset += 1;
		ip = (uint16_t *) (msg + offset);
		
		ctxLen = ntohs(*ip);
		offset+= sizeof(uint16_t);
		if(ctxLen>0){
			result->destContext.valueLen = ctxLen;
			result->destContext.value = (void *) malloc(ctxLen);
			memcpy(result->destContext.value, msg + offset, ctxLen);
			offset += ctxLen;
		}
		
		ip = (uint16_t *) (msg + offset);
		srvLen = ntohs(*ip);
		offset+= sizeof(uint16_t);
		if(srvLen>0){
			result->destService.valueLen = srvLen;
			result->destService.value = (void *) malloc(srvLen);
			memcpy(result->destService.value, msg + offset, srvLen);
			offset += srvLen;
		}	
		
		ip = (uint16_t *) (msg + offset);
		result->MTU = ntohs(*ip);
		offset += sizeof(uint16_t);
		
		t = (char *)(msg + offset);
		result->chanType = t[0];
		offset += 1;
	}
	else{
		result->ICflag = 0;
		offset += 1;
	}

	t = (char *)(msg + offset);
	offset += 1; 
	
		if(t[0] == BRICK){
			
			result->IDPtype = BRICK;
			hp = (anaHandle_t *) (msg + offset);
			offset += sizeof(anaHandle_t);
			result->binding.brick.handle = ntohl(*hp);
			
			lp = (anaLabel_t *) (msg + offset);
			offset += sizeof(anaLabel_t);
			result->binding.brick.outputIDP = ntohl(*lp);
	
		}
		if(t[0] == REDIRECT){
			result->IDPtype = REDIRECT;
			lp = (anaLabel_t *) (msg + offset);
			offset += sizeof(anaLabel_t);
			result->binding.IDPs = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));  
			result->binding.IDPs->label = ntohl(*lp);
		
			hp = (anaHandle_t *) (msg + offset);
			offset += sizeof(anaHandle_t);
			result->binding.IDPs->viewer = ntohl(*hp);
			
		}
		if(t[0] == FORK){
			int nbForks,i;
			struct nextIDP_s *tmp;
			result->IDPtype = FORK;
			t = (char *) (msg + offset);
			offset += 1;
			nbForks = t[0];
			
			result->binding.IDPs = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));
			tmp = result->binding.IDPs;
			for(i=0;i<nbForks;i++){
				lp = (anaLabel_t *) (msg + offset);
				offset += sizeof(anaLabel_t);
				tmp->label = ntohl(*lp);
				
				hp = (anaHandle_t *) (msg + offset);
				offset += sizeof(anaHandle_t);
				tmp->viewer = ntohl(*hp);
			
				if(i!= nbForks-1){
					tmp->next = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));	
					tmp = tmp->next;	
				}
			}
		}
	
	
	return 0;
}



anaLabel_t anaL0_registerCallback( anaCallback_t fct, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux, 
				const anaHandle_t viewer, int perm, int thread) {
	
	/* format : ['r'][MY_HANDLE][permFlag][viewer] */
	char msg[CTRLMSG_MAXSIZE];
	
	struct SDTentry_s *entry=NULL;
	
	anaHandle_t *handle= NULL; 
	
	anaLabel_t *labelP = NULL;
	anaLabel_t result;
	
	int nbRead = -1;
	int len = 0;

	if( fct ==NULL){
		anaPrint(ANA_ERR, "anaL0 registerCallback: NULL arguments\n");
		set_ana_errno(-1);
		return 0;
	}
	   
	memset(msg, 0, CTRLMSG_MAXSIZE);
	msg[0]='r';
	
	handle = (anaHandle_t *) (msg+1); 
	*handle = (anaHandle_t) htonl(MY_HANDLE);
	
	/* we add the permanent option */
	if(perm)
		msg[1+sizeof(anaHandle_t)] = 'p';  /* p for permanent */
	else
		msg[1+sizeof(anaHandle_t)] = 'v';  /* v for volatile  */
	
	/* in case the viewer argument is not NULL we add it to the message */
	len = 2+sizeof(anaHandle_t);
	
	if(viewer != 0){
		msg[len]=1;
		handle = (anaHandle_t *) (msg+len+1);
		*handle = (anaHandle_t) htonl(viewer);
		len = len + 1 + sizeof(anaHandle_t);
	}
	else{
		msg[len]=0;
		len++;
	}
	
			
        /* sending request to register callback IDP */
      
        if(sendControl(myMinmex, msg, len) < 0) {
                anaPrint(ANA_ERR,  "Unable to send register callback message to minmex\n");
		set_ana_errno(-1);
              	return 0;
        }

	/* we wait to receive the answer containing the label of the created IDP */
        memset(msg, 0, CTRLMSG_MAXSIZE);
        nbRead = recvControl(myMinmex, msg);

	
	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_registerCallback : Unable to receive reply from minmex\n");
                return 0;
	}
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "registerCallback : Ana minmex returned error: %s\n", getErrorString((int)msg[1]));
		set_ana_errno((int)msg[1]);
		
		return 0;
	}	
	if(nbRead < sizeof(anaLabel_t) + 1 ) {
		anaPrint(ANA_ERR,  "registerCallback : Bad register Callback response\n");
		set_ana_errno(-1);
                return 0;
	}   
	
	
	/* expected format of response : [r][label] */
	labelP = (anaLabel_t *) (msg+1);
	result = (anaLabel_t) ntohl(*labelP);	
				
				
	/* we then contruct and add the SDT entry */
	entry = (struct SDTentry_s *) malloc(sizeof(struct SDTentry_s));
	if(entry==NULL){
		anaPrint(ANA_ERR,  "registerCallback : Unable to allocate memory for SDT entry\n");
		set_ana_errno(-1);
		return 0;
	}
		
	entry->label = result;

	if(thread == 1){ /* launch in separate thread */
		entry->callBackFct = (anaCallback_t)anaL0_launchThread;
		entry->aux = anaL0_copyAux(aux, copyAux, freeAux, fct);	
		entry->copyAux = (copyAuxFunction)anaL0_copyAux;
		entry->freeAux = anaL0_freeAux;
	}
	else{
		entry->callBackFct = fct;
		if(aux!=NULL && copyAux !=NULL)
			entry->aux = copyAux(aux);

		entry->freeAux = freeAux;
		entry->copyAux = copyAux;
	}

	entry->next = NULL;
	
	addSDTEntry(myMinmex->shadowT,entry);

	return result;
}


int anaL0_send(anaLabel_t label, void *msg, int len) {
	
	int totalSize;
	char *fmtMsg;
        
       	anaLabel_t *labelP = NULL;
	anaHandle_t *handleP = NULL;
       
       
     	/* <formatMsg> */
	/* we format the message as: [label][handle][ack flag (1byte)][message] */	
	
	totalSize = len + sizeof(anaLabel_t) + sizeof(anaHandle_t) +1; /* +1 for the ack flag byte*/ 
	if(totalSize > DEFAULT_UDP_MTU) {   /* fragmentation in future ?? */
		anaPrint(ANA_NOTICE,  "Send : message is too big\n");
	}


	fmtMsg = (char*) malloc(totalSize); 
	if(fmtMsg == NULL){
		anaPrint(ANA_ERR,  "Send : unable to allocate memory for formatted message\n");
		return -1;	
	}
	
	labelP = (anaLabel_t *) fmtMsg;
	*labelP = htonl(label);
	
	handleP = (anaHandle_t *) (fmtMsg+sizeof(anaLabel_t));
	*handleP = htonl(MY_HANDLE);
	
/*	fmtMsg[sizeof(anaLabel_t) + sizeof(anaHandle_t)] = 0; */ /*this instruction is implicit since buffer was initialized to 0*/
	
	memcpy(fmtMsg + sizeof(anaLabel_t) + sizeof(anaHandle_t) +1, msg, len);

	
        if(sendData( myMinmex, fmtMsg, totalSize)<0) {
               anaPrint(ANA_ERR,   "Send : Unable to send data to minmex\n");
               return -1;
        }
	  
       	free(fmtMsg);
 	return 0;

}

int anaL0_send_ack(anaLabel_t label, void *msg, int len) {
	
	int totalSize, nbRead;
	char *fmtMsg;
        char ackBuf[2] = {0,0};
       
       	anaLabel_t *labelP = NULL;
	anaHandle_t *handleP = NULL;
       
       
     	/* <formatMsg> */
	/* we format the message as: [label][handle][ack flag (1byte)][message] */	
	
	totalSize = len + sizeof(anaLabel_t) + sizeof(anaHandle_t)+1 ; /* +1 for the ack flag byte*/ 
	if(totalSize > DEFAULT_UDP_MTU) {   /* fragmentation in future ?? */
		anaPrint(ANA_NOTICE,  "Send : message is too big\n");
	}


	fmtMsg = (char*) malloc(totalSize); 
	if(fmtMsg == NULL){
		anaPrint(ANA_ERR,  "Send : unable to allocate memory for formatted message\n");
		return -1;	
	}
	
	labelP = (anaLabel_t *) fmtMsg;
	*labelP = htonl(label);
	
	handleP = (anaHandle_t *) (fmtMsg+sizeof(anaLabel_t));
	*handleP = htonl(MY_HANDLE);
	
	fmtMsg[sizeof(anaLabel_t) + sizeof(anaHandle_t)] = 1;
	
	memcpy(fmtMsg + sizeof(anaLabel_t) + sizeof(anaHandle_t) +1, msg, len);
	
        if(sendData( myMinmex, fmtMsg, totalSize)<0) {
               anaPrint(ANA_ERR,   "Send : Unable to send data to minmex\n");
               return -1;
        }
	
	free(fmtMsg); 
	
	/* we wait for the response */
	nbRead = recvControl(myMinmex, ackBuf);
	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_send_ack : Unable to receive reply from minmex\n");
                return -1;
	}
	
	/* we check if the reply is an error message */	
	if(ackBuf[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_send_ack : Ana minmex returned error: %s\n", getErrorString((int)ackBuf[1]));
		return (int)ackBuf[1];
	}	
	
 	return 0;
}

int anaL0_redirect(const anaLabel_t originLabel, const anaHandle_t originViewer, const anaLabel_t destLabel, const anaHandle_t destViewer){
	
	/* format of message : [c][handle][label origin][viewer origin][label dest][viewer dest]*/
	char msg[CTRLMSG_MAXSIZE];
	int nbRead=-1;
     	anaHandle_t *handle= NULL;
	anaLabel_t *label = NULL;
	
	msg[0] = 'c';
	handle = (anaHandle_t *) (msg+1);
	*handle = htonl(MY_HANDLE);
	
	label = (anaLabel_t *) (msg+1+sizeof(anaHandle_t));
	*label = htonl(originLabel);
	
	handle = (anaHandle_t *) (msg+1+sizeof(anaHandle_t)+sizeof(anaLabel_t));
	*handle = htonl(originViewer);
	
	label = (anaLabel_t *) (msg+1+2*sizeof(anaHandle_t)+sizeof(anaLabel_t));
	*label = htonl(destLabel);
	
	handle = (anaHandle_t *) (msg+1+2*sizeof(anaHandle_t)+2*sizeof(anaLabel_t));
	*handle = htonl(destViewer);
	
		
       	if(sendControl(myMinmex, msg, 1+3*sizeof(anaHandle_t)+2*sizeof(anaLabel_t)) < 0) {
		anaPrint(ANA_ERR,  "Unable to send redirect message to minmex\n");
                return -1;
       	}
        
	
	/* we wait to receive the answer: confirmation or error*/
       	memset(msg, 0, CTRLMSG_MAXSIZE);	
       	nbRead = recvControl(myMinmex, msg);	
      
      
       	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_redirect : Unable to receive reply from minmex\n");
        	return -1;
	}
	
	
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "redirect : Ana minmex returned error code: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1] ;
	}	
		
	return 0;
}


 
int  anaL0_unregisterCallback(const  anaLabel_t label, const  anaHandle_t viewer) {	
	/*format of message to send : [u][label][viewer][myhandle] */
	char msg[CTRLMSG_MAXSIZE];
	int nbRead = -1;
	
	anaLabel_t *lp=NULL;
	anaLabel_t *hp=NULL;
	
	memset(msg, 0, CTRLMSG_MAXSIZE);
	msg[0]='u';
	lp = (anaLabel_t *)(msg+1);
	*lp = htonl(label);
	
	hp = (anaHandle_t *) (msg+1+sizeof(anaLabel_t));
	*hp = htonl(viewer);
	
	hp = (anaHandle_t *) (msg+1+sizeof(anaLabel_t)+sizeof(anaHandle_t));
	*hp = htonl(MY_HANDLE);
	
				
	if(sendControl(myMinmex, msg, 1+sizeof(anaLabel_t)+2*sizeof(anaHandle_t)) < 0){
		anaPrint(ANA_ERR,  "anaL0_unregisterCallback : Unable to send unregister message to minmex\n");
                return -1;
	}
	
	/* we wait to receive an answer confirming the deregistration or announcing an error at the minmex */
	memset(msg, 0, CTRLMSG_MAXSIZE);	
	nbRead = recvControl(myMinmex, msg);	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_unregisterCallback : Unable to receive reply from minmex\n");
              	return -1;
	}

	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "unregisterCallback : Ana minmex returned error code: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1];
	}	

	/* we then remove the label from the dispatch table */
	if (freeSDTEntry(myMinmex->shadowT,label)<0) {
		anaPrint(ANA_ERR,  "unregisterCallback : Unable to remove entry from SDT \n");
		return -1;
	}

	return 0;
}


int anaL0_changeLabelAux(const anaLabel_t label, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux){
	
	struct SDTentry_s *entry;

	entry = getSDTEntry(myMinmex->shadowT, label);
	if(entry == NULL)
		return -1;
	if(entry->aux && entry->freeAux)
		entry->freeAux(entry->aux);
	

	entry->freeAux = freeAux;
	entry->copyAux = copyAux;
		
	if(aux !=NULL ) 
		entry->aux = copyAux(aux); 
	else
		entry->aux = NULL;
		 	
	return 0;
}

int anaL0_changeLabelView(anaLabel_t label, anaHandle_t oldViewer, anaHandle_t newViewer){
	
	/* format of message to send : [v][label][myhandle][oldViewerHandle][newViewerHandle] */
	char msg[CTRLMSG_MAXSIZE];
	int nbRead = -1;

	
	anaHandle_t *hp;
	anaLabel_t *lp;
	
	memset(msg, 0, CTRLMSG_MAXSIZE);
	msg[0]='v';
	
	
	lp = (anaLabel_t *) (msg+1);
	*lp = htonl(label);
	
	hp = (anaHandle_t *) (msg+1+sizeof(anaLabel_t));
	*hp = htonl(MY_HANDLE);
	
	hp = (anaHandle_t *)(msg+1+sizeof(anaLabel_t) + sizeof(anaHandle_t));
	*hp = htonl(oldViewer);
	
	hp = (anaHandle_t *)(msg+1+sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t));
	*hp = htonl(newViewer);
	
	if(sendControl(myMinmex, msg, 1+sizeof(anaLabel_t) + 3 * sizeof(anaHandle_t)) < 0) {
		anaPrint(ANA_ERR,   "anaL0_changeLabelView: Unable to send  message to minmex\n");
               	return -1;
	}
	
	/* we wait to receive an answer confirming the change or announcing an error at the minmex */
	memset(msg, 0, CTRLMSG_MAXSIZE);	
	nbRead = recvControl(myMinmex, msg);	
	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0multi_changeLabelView : Unable to receive reply from minmex\n");
                return -1;
	}
	
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_changeLabelView : Ana minmex returned error code: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1];
	}
	
	return 0;		
}


int anaL0_forkData(anaLabel_t originLabel,  anaHandle_t originViewer,  anaLabel_t destLabel,  anaHandle_t destViewer){
	/* format of message : [f][handle][originViewer][originLabel][destViewer][destLabel]*/
	char msg[CTRLMSG_MAXSIZE];
	int nbRead=-1;

	
	anaHandle_t *hp = NULL;
	anaLabel_t *lp = NULL;
	
	memset(msg, 0, CTRLMSG_MAXSIZE);

	msg[0] = 'f';
	
	hp = (anaHandle_t *) (msg+1);
	*hp = htonl(MY_HANDLE);
	
	hp = (anaHandle_t *) (msg+1 +sizeof(anaHandle_t));
	*hp = htonl(originViewer);
	
	
	lp = (anaLabel_t *) (msg+1+2*sizeof(anaHandle_t));
	*lp = htonl(originLabel);
	
	hp = (anaHandle_t *) (msg+1 + 2*sizeof(anaHandle_t)+sizeof(anaLabel_t));
	*hp = htonl(destViewer);
	
	
	lp = (anaLabel_t *) (msg+1 + 3*sizeof(anaHandle_t)+sizeof(anaLabel_t));
	*lp = htonl(destLabel);

        if(sendControl(myMinmex, msg, 1 + 3*sizeof(anaHandle_t)+2*sizeof(anaLabel_t)) < 0) {
        	anaPrint(ANA_ERR,  "Unable to send Fork message to minmex\n");
                return -1;
        }
     
     
       /* we wait to receive the answer: confirmation or error*/
      	memset(msg, 0, CTRLMSG_MAXSIZE);	
        nbRead = recvControl(myMinmex, msg);	
        
       	if(nbRead <= 0) {
		anaPrint(ANA_ERR,  "anaL0_forkData : Unable to receive reply from minmex\n");
       		return -1;
	}
	
	
	/* we check if the reply is an error message */	
	if(msg[0]== 'e') {
		anaPrint(ANA_ERR,  "anaL0_forkData: minmex returned error code: %s\n", getErrorString((int)msg[1]));
		return (int)msg[1];
	}	

	return 0;
}


/* this function stops ANA activity correctly*/
void  anaL0_stopANA(int exitVal){
	
	/*First thing call brick_exit */
	if(attached) {
		brick_exit();
		validate_exit();	
		attached = 0;
		anaL0_detach(myMinmex);  

		/* we then kill all remaining threads */
		anaThreadExit();
	}
#ifdef USER
        exit(0); /*kill our process, just in case there is some infinite loop etc */
#endif

}


void anaL0_heartBeat(int sig) {
	char msg[CTRLMSG_MAXSIZE];
	
	myMinmex->age++;
	
	if(myMinmex->age > DEAD_NODE_THRESHOLD){
		anaPrint(ANA_ERR, "Minmex not responding to heartbeat for a while => anaL0_stopANA()\n");
		anaL0_stopANA(1);
	}
	else {	
		/* we send a heartBeat request,  Format : [h][handle][notificationLabel]*/
		anaHandle_t *tmp;
		anaLabel_t *lp;
		msg[0]='h';
		
		tmp = (anaHandle_t *) (msg+1);
		*tmp = htonl(MY_HANDLE);
		
		lp = (anaLabel_t *) (msg+1+sizeof(anaHandle_t));
		*lp = htonl(myHBLabel); 
		
		sendControl(myMinmex, msg, sizeof(anaHandle_t) + 1 + sizeof(anaLabel_t));

	}
}

#ifdef USER

int listenLoop() {
	int ret,nbRead = 0;
	fd_set readFd;
	anaLabel_t *idp = NULL;
	anaLabel_t idpVal;
	void *data=NULL;
	struct SDTentry_s *entry=NULL;
#ifdef USERGEN
        struct {
                struct nlmsghdr n;
                struct genlmsghdr g;
                char buf[CTRLMSG_MAXSIZE];
        } ans;
        struct nlattr *na;
#endif
	anaPrint(ANA_DEBUG, "data listenloop started\n");
        
	
	while(attached) {
                
		FD_ZERO(&readFd);
              
             	FD_SET(myMinmex->myDataDescriptor,&readFd);
                	
		ret = select(255,&readFd,NULL,NULL,NULL /*, &our_sigset*/);
		if(ret<=0) 
                        continue;
              
	        if (FD_ISSET(myMinmex->myDataDescriptor, &readFd)) {
                	unsigned char * buf = NULL;
			memset(myMinmex->buffer, 0, DEFAULT_UDP_MTU);  /*we  clean the buffer*/
                        if(myMinmex->myDataMode == GENETLINK){
#ifdef USERGEN
                                        memset(ans.buf,0, CTRLMSG_MAXSIZE);
                                        nbRead = recv(myMinmex->myDataDescriptor, &ans, sizeof(ans), 0);
                                      
                                        /* Validate response message */
                                        if (ans.n.nlmsg_type == NLMSG_ERROR) { /* error */
                                                anaPrint(ANA_ERR, "GN: received NACK\n");
                                                continue;
                                        }
                                        if (nbRead < 0) {
                                                anaPrint(ANA_ERR,  "GN len < 0\n");
                                                continue;
                                        }
                                        if (!NLMSG_OK((&ans.n), nbRead)){
                                                anaPrint(ANA_ERR, "GN: invalid reply message\n");
                                                continue;
                                        }
                                                                     
                                        /*parse reply message*/
                                        na = (struct nlattr *) GENLMSG_DATA(&ans);
					nbRead = na->nla_len - NLA_HDRLEN;

				        buf = (unsigned char *)NLA_DATA(na);
#endif
                         }
                         else{
                         	nbRead = read(myMinmex->myDataDescriptor, myMinmex->buffer, DEFAULT_UDP_MTU);
                                if(nbRead>0) {
                                	buf = myMinmex->buffer;
                                }
                                else{
                                	if(nbRead==0 && myMinmex->myDataMode == PIPES) { /* when read returns 0 on FIFOs, means wrtiting descriptor was closed  ==> detach from minmex */
                                        	anaPrint(ANA_ERR, "listenLoop: Named pipe for data gate has been closed on minmex's side ==> detach \n");
                                                anaL0_stopANA(1);
                                        }
                                        continue;
                                }
                         }
                         idp = (anaLabel_t *)buf;
			 idpVal = (anaLabel_t) ntohl(*idp);
                        /*  anaPrint(ANA_DEBUG, "listenLoop : identified IDP %s\n", idpChar));*/
                         data = buf + sizeof(anaLabel_t);
                                
                        /* </ Message IDPHeader/Payload split> */
                                
                        /* we look for the entry corresponding to this IDP in the shadow IDT associated with this dataGate*/
                        entry = getSDTEntry(myMinmex->shadowT , idpVal);
                        if(entry==NULL)
                                anaPrint(ANA_ERR, "listenLoop : No entry mapped to Label %d in shadowIDT\n", idpVal);
                        else {
                                        if((entry->callBackFct) != 0)
					/*	if((void*)(entry->callBackFct) == (void*)handleNotification)  
							handleNotification(data, nbRead - tmp->labelSize, idp, tmp, entry->aux);
						
						else
					*/		/* if we are using the simple AL0, we call the callback 
							function without indicating the origin minmex*/
							
						entry->callBackFct(data, nbRead - sizeof(anaLabel_t), idpVal, entry->aux);
						
					
					else
                                                anaPrint(ANA_ERR, "listenLoop: callBack function mapped to IDP %d is NULL \n", idpVal);
                                }
                                
                        }
                        
                }
        
	anaPrint(ANA_NOTICE, "listenLoop : Brick detached  -> Ending listenLoop \n");

	return 0;
}
#endif
