/* C/bricks/API/API0/anaLib0_so.c
 * 
 * implementation of API calls for the so plugin functioning mode
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */
/** @defgroup anaL0 ANA API level 0
 * Definition of the ANA API level 0
 */

#include "anaLib0.h"


/** \ingroup anaL0
 * \brief sets the information on a certain IDP
 * \param info pointer to the structure that contains the IDP's details 
 * \return 0 on success, negative error value otherwise
 */
int anaL0_setIDPinfo( struct IDPinfo_s *info){
	anaHandle_t callee;
	if (info ==NULL)
		return -1;
	
	callee = (info->binding).brick.handle; 
	if(callee == 0) {
		info->IDPtype = BRICK;
		(info->binding).brick.handle = MY_HANDLE;
	}
		
	return  minmex_setIDPinfo(MY_HANDLE, info);
}

/** \ingroup anaL0
 * \brief returns information about a given IDP 
 * \param label the label of the IDP 
 * \param viewer the handle of the veiwer of this IDP
 * \param result pointer to the structure that will be filled with the IDP's details (must be allocated) 
  * \return 0 on success, negative error value otherwise
 */
int anaL0_getIDPinfo(anaLabel_t label,  anaHandle_t viewer, struct IDPinfo_s *result){
	if (result == NULL)
		return -1;
	
	return minmex_getIDPinfo(MY_HANDLE, viewer, label, result);	
}

/** \ingroup anaL0
 * \brief redirects label originLabel to label destLabel
 * everything sent to originLabel will be forwarded to destLabel
 * \param originLabel the label to be redirected
 * \param originViewer the viewer of the origin label
 * \param destLabel the destination label
 * \param destViewer the viewer of the destination label
 * \return 0 on success, negative error value otherwise
 */
int anaL0_redirect(const anaLabel_t originLabel, const anaHandle_t originViewer, const anaLabel_t destLabel, const anaHandle_t destViewer){
	return handleRedirect(myHandle, originViewer, originLabel,  destViewer, destLabel);
}

/** \ingroup anaL0
 * \brief everything sent to originLabel will be received by
 * originLabel and destLabel
 * \param originLabel the original label
 * \param originViewer the viewer of the origin label
 * \param destLabel the label that will receive a copy of the data
 * \param destViewer the viewer of the destination label
 * \return 0 on success, negative error value otherwise
 */
int anaL0_forkData(const anaLabel_t originLabel, const anaHandle_t originViewer, const anaLabel_t destLabel, const anaHandle_t destViewer){
	return handleForkData(myHandle , originViewer, originLabel, destViewer, destLabel);
}

/** \ingroup anaL0
 * \brief sends the message msg to the idp label
 * \param label the destination idp
 * \param msg the message to be sent
 * \param len the length of the message
 * \return number of bytes sent on success, negative error value otherwise
 */
int anaL0_send(const anaLabel_t label, void *msg, int len) {
	return dispatchData(myHandle, label, myHandle, msg, len);
}

/** \ingroup anaL0
 * \brief registers a new callback function
 * \param fct a pointer to the callback function
 * \param aux auxiliary argument that the callback function will receive 
 * \param copyAux function that copies the aux argument
 * \param freeAux function that frees the aux argument
 * \param viewer the viewer of the newly generated idp
 * \param perm IDP_PERM: The IDP is permanent (lifetime of the brick), IDP_VOL: the IDP gets deleted when unused
 * \param thread NOTHREAD the callback function is executed directly (attention: no other messages can be received
 * during the execution of the callback function). THREAD the callback function is executed in a dedicated thread.
 * \return new IDP on success, 0 on failure
 */
anaLabel_t anaL0_registerCallback( anaCallback_t fct, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux, 
				const anaHandle_t viewer, int perm, int thread) {
	
	anaLabel_t result;
	int ret = handleRegister(myHandle, fct, (void *)aux, copyAux,  freeAux, viewer, perm, thread, &result);
	
	if (!ret)
		return result;

	else {
		set_ana_errno(ret);
		return 0;
	}
}

/** \ingroup anaL0
 * \brief change the auxiliary argument of a callback function
 * \param label the idp of the callback function
 * \param aux the new auxiliary argument
 * \param copyAux the function to copy the new aux
 * \param freeAux the function to free the new aux
 * \return 0 on success, negative error value else
 */
int anaL0_changeLabelAux(const anaLabel_t label, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux){
	return 	handleChangeLabelAux( myHandle, label, aux,  copyAux,  freeAux);
}

/** \ingroup anaL0
 * \brief unregister a given callback function
 * \param label the idp to unregister
 * \param viewer the viewer of the idp
 * \return 0 on success, negative error value else
 */
int anaL0_unregisterCallback(const anaLabel_t label, anaHandle_t viewer) {	
	return handleUnregister(myHandle, viewer, label);
}

/** \ingroup anaL0
 * \brief change the viewer of an IDP
 * \param label the idp for which the viewer needs to be changed
 * \param oldViewer the original viewer
 * \param newViewer the new viewer
 * \return 0 on success, negative error value otherwise
 */

int anaL0_changeLabelView(const anaLabel_t label, const anaHandle_t oldViewer,  const anaHandle_t newViewer){
	return handleChangeView(myHandle,  oldViewer, label, newViewer);
}




/** \ingroup anaL0
 * \brief Correctly terminates th calling brick  
 * \param exitVal the exit status number returned to the minmex or to the running shell
 * \return void
 */
void  anaL0_stopANA(int exitVal){
	
	/* we stop the timer of the heartBeat */
	anatimer_del(heartBeatTimer);
	
	/*First thing call brick_exit */
	brick_exit();
	validate_exit();
	/* we stop the timers */
/*	anatimer_exit();  commented for now until anThreadExit does not kill the minmex anymore */ 
	
	/* Detach from the minmex  */
	/* in .so mode, the minmex will take care of killing 
	             all our threads                    */
	
	attached = 0;
	handleDetach(MY_HANDLE,exitVal);  
}

void  whenKicked(int exitVal){
	
	anaPrint(ANA_EMERG, "Brick is being kicked by the minmex\n");
	/* we stop the timer of the heartBeat */
	anatimer_del(heartBeatTimer);
	
	brick_exit();

}

void wrapInThread(struct dataMessageQueue_s *msg){

	if (msg == NULL)
		return;
		
	/* we launch the callBack function */
	msg->callBackFct(msg->data, msg->dataLen, msg->inputLabel, msg->aux);
	
	/* and we free the message  IMPORTANT : WE DO NOT TOUCH THE AUX*/
	free(msg->data);
	free(msg);
}


void heartBeat(){

	handleHBRequest(MY_HANDLE);
		
}
/* this function is the one that receives all notifications for the brick and then calls the proper event handler */
void notificationWrapper(struct anaEventNotif_s *notif, int len, anaLabel_t input,  void *aux){
	/* aux contains the pointer to the event handler */
	eventHandler_t *Phandler = (eventHandler_t *) aux;
	eventHandler_t handler = *Phandler;

	handler(notif);
	
	if(BRICK_ATTACHED_EVENT&notif->eventID || BRICK_DETACHED_EVENT&notif->eventID || IDP_UNPUBLISHED_EVENT&notif->eventID)
		free(notif->name);
	
	free(aux);
}
/** \ingroup anaL0
 * \brief subscribes the brick to an ANA event 
 * \param sub  structure containing the details of IDP or Brick the event is about and the event bitMask
 * \param handler  pointer to the function to be triggered when the event is received
 * \param threadFlag indicates wether the handler function should be executed in a separate thread 
 * \return 0 on success, negative error value otherwise
 */
int anaL0_eventSubscribe(struct anaEventSub_s* sub, eventHandler_t handler, int threadFlag){
	return minmex_registerEventHandler(MY_HANDLE, MY_HANDLE, sub, (anaCallback_t) notificationWrapper, &handler,  sizeof(void*), threadFlag);
}


/** \ingroup anaL0
 * \brief unsubscribes the brick from an ANA event 
 * \param sub  structure containing the details of IDP or Brick the event is about and the event bitMask
 * \return 0 on success, negative error value otherwise
 */
int anaL0_eventUnsubscribe(struct anaEventSub_s* sub){
	return minmex_unregisterEventHandler(MY_HANDLE, sub );
}

/** \ingroup anaL0
 * \brief sends a notification of the occurence of an event 
 * \param notif  structure containing the details of IDP or Brick the event is about and the event value
 * \return 0 on success, negative error value otherwise
 */
int anaL0_eventNotify( struct anaEventNotif_s *notif){
	return  minmex_notifyEvent(MY_HANDLE, notif );
}


void dataLoop(struct condLock0_s *lock){
	struct dataMessageQueue_s *msg = NULL;
	int ret;
	
	/* we unlock the init(main) thread */	
	analock_lock(&(lock->condLock));
	lock->flag = 1;
	ret = analock_cond_signal(&(lock->cond));
	analock_unlock(&(lock->condLock));


	anaPrint(ANA_DEBUG, "Data loop thread started (%s).\n", mymodename);

	
	/* we wait for messages from the minmex */	
	while (attached){
	
		analock_lock(&queueLock);

		while ((messageQueue) == NULL) {
			analock_cond_wait( &waitLock , &queueLock);	
		}

		/* we unqueue the head message */	
		msg = messageQueue;
		messageQueue = msg->next;
	
		analock_unlock(&queueLock);
		if (msg->thread)
			_anaThread(myHandle, (void *) wrapInThread, msg, mymodename);
		else {
			/* we execute the callback function*/
			msg->callBackFct(msg->data, msg->dataLen, msg->inputLabel, msg->aux);
			/* and we free the data  IMPORTANT : WE DO NOT TOUCH THE AUX*/
			free(msg->data);
			free(msg);
		}
	}
}

