#include "shadowIDT.h"
#include "hashFunction.h"

/***************************************************************

Hash functions by Bob Jenkins, December 1996, Public Domain.

2005-11-09: tested by CJ with 128-bit keys to 10 bits values 
            --> hash function has a uniform distribution

***************************************************************/

ub4 hash( k, length, initval, maskLength)
register ub1 *k;        /* the key */
register ub4  length;   /* the length of the key in bytes */
register ub4  initval;    /* the previous hash, or an arbitrary value */
register int  maskLength;    /* the previous hash, or an arbitrary value */
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */

	return (c & hashmask(maskLength));	 
}


/* hashes an anaLabel into an index */
int label2hash(struct shadowIDT_s *IDT, anaLabel_t l) {
	int h;
	if(IDT==NULL)
		return -1;
	h = hash((ub1*)&l, sizeof(anaLabel_t), 0, IDT->hashMaskSize);
	return h;
}


/* hashes an anaLabel into an index */
int initializeSDT (struct shadowIDT_s *IDT, int hashMaskSize) { 
	/* the allocated size of the table will be (2^hashMaskSize), the hashMaskSize will be also used in the hash  function */
	if(IDT==NULL || hashMaskSize<0)
		return -1;
	IDT->size = pow2(hashMaskSize);
	IDT->hashMaskSize =hashMaskSize;
	IDT->tab = (struct SDTentry_s **) malloc(IDT->size * sizeof(struct SDTentry_s *));
	if(IDT->tab==NULL){
		anaPrint(ANA_ERR, "initializeSDT : unable to allocate memory for SDT hash table\n");
		return -1;
	}
	anaPrint(ANA_DEBUG, "ShadowIDT table initialized with size %u\n", IDT->size);
	return 0;
}

/* returns the IDT entry corresponding to an IDP */
struct SDTentry_s* getSDTEntry(struct shadowIDT_s *IDT, anaLabel_t l) {
	int i ;
	struct SDTentry_s *entry=NULL;
	
	if(IDT==NULL)
		return NULL;
	i = label2hash(IDT, l);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "getSDTEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	entry = IDT->tab[i];

	if (entry == NULL ) { 	/* meaning that there is no value mapped to this  label in the Hash table */
		anaPrint(ANA_NOTICE, "getSDTEntry : entry in IDT at index %d  is NULL\n", i);
		return NULL;	
	}
			
	if  ( entry->label == l) { /* if the keys are equal */
		return entry; 
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
		 * the entry with the correct key */
		entry = entry->next; 
		while (entry != NULL) {
			if(l == entry->label) 
				return entry;
			else 
				entry = entry->next;
		} 
	}
	return NULL;
}


/* Frees the entire shadowIDT */
int freeSDTTable(struct shadowIDT_s *IDT) { 
	int i;
	struct SDTentry_s *entry=NULL;
	struct SDTentry_s *next=NULL;
	
	if (IDT==NULL)
		return -1;
		
	for (i=0; i< (int) IDT->size; i++) {
		entry = IDT->tab[i];
		while (entry!=NULL) {
			next = entry->next;
			if (entry->aux && entry->freeAux)
				entry->freeAux((entry->aux));
			free(entry);
			entry = next;
		}
	}
	IDT->size = 0;
	free(IDT->tab);
	free(IDT);
	return 0;
}

int freeSDTEntry(struct shadowIDT_s *IDT, anaLabel_t l) { 
	int i ;
	struct SDTentry_s *entry=NULL;
	struct SDTentry_s *tmp=NULL;
	
	if (IDT==NULL)
		return -1;
	
	i = label2hash(IDT, l);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "freeIDTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = IDT->tab[i];
	if (entry == NULL) /* meaning that there is no value mapped to this */
		return -1;	  /*  label in the Hash table */
			
	if (l == entry->label) {
		IDT->tab[i] = entry->next; 
		
		if(entry->aux!=NULL && entry->freeAux !=NULL)
			entry->freeAux(entry->aux);
		free(entry); 
		return 0;
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
			  the entry with the correct key */
		while (entry != NULL ) {
			if(entry->next != NULL) {
				tmp = entry->next;
				if(l == entry->label) {
					entry->next = tmp->next;
					if(tmp->aux !=NULL && tmp->freeAux !=NULL)
						tmp->freeAux(tmp->aux);
				
					free(tmp);
					return 0;
				}	 
			}
			entry = entry->next;
		}
	}
	return -1;
}

int addSDTEntry (struct shadowIDT_s *IDT, struct SDTentry_s *toAdd) {
	struct SDTentry_s *entry=NULL;
	int i ;
	if(IDT==NULL || toAdd==NULL)
		return -1;
	
	i= label2hash(IDT, toAdd->label);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "addIDTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	entry = IDT->tab[i];
	if(entry == NULL ){
		toAdd->next = NULL;
		IDT->tab[i] = toAdd;
	}
	else { /* hash collison */
		anaPrint(ANA_DEBUG, "addIDTEntry : Hash collison\n");
		toAdd->next = IDT->tab[i];
		IDT->tab[i]=toAdd;
	}
	return 0;
}


