/*
 * C/bricks/API/AL2/anaLib2.c
 * 
 * Implementation of the AL2 functions
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/** @defgroup anaL2 ANA API level 2
 * Definition of the ANA API level 2 
 */


#include "anaLib2.h"
//#include "anaError.h"
#include "anaThread.h"
#include "messageQueue.h"

struct MQTable_s MQTable;

struct timespec default_timeout = {1,0}; /* default timeout value for various functions */

void * identityFct(void *pointer){
	return pointer;
}

void freeCondLock(struct condLock_s *cond){

	if(cond !=NULL && cond->value !=NULL && cond->value != (void*) 1)
		free(cond->value);
	if(cond)
		free(cond);
}

void initCondLock(struct condLock_s *cond){ 
	if(cond == NULL)
		return;
	analock_cond_init(&(cond->cond));
	
	analock_init(&(cond->condLock));

	analock_init(&(cond->valueLock));
	
	cond->value = NULL;
}

/* will be called after a publish message was sent, checks whether an error occured */
void anaL2_handlePublishReply(char *data, int len, anaLabel_t input, struct condLock_s *responseCond) {
	
	char *name=NULL;
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**)&name);
	
	if(anaL1_isError(data)) {/*if ERROR command*/
		if(name!=NULL)
			anaPrint(ANA_DEBUG, "anaL2_publish : Publish of entry %s Failed\n",name);
		else
			anaPrint(ANA_DEBUG, "anaL2_publish : Publish Error\n");
				
		responseCond->valueLen = 0; 
	}
	else {
		if(name!=NULL)
			anaPrint(ANA_NOTICE, "anaL2_publish : Publish of entry %s succeeded\n",name);
		else
			anaPrint(ANA_NOTICE, "anaL2_publish : Publish Success\n");
		
		responseCond->valueLen =  1; 
		
	}
	
	responseCond->value = (void*)1;
	/* we unblock the awaiting thread */
	analock_lock(&(responseCond->condLock)); 
	analock_cond_signal(&(responseCond->cond));	
	analock_unlock(&(responseCond->condLock));
}


void anaL2_handleUnPublishReply(char *data, int len, anaLabel_t input, struct condLock_s *responseCond) {
	
	char *descrip=NULL;
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**)&descrip);
	
	if(anaL1_isError(data)) {
		if(descrip!=NULL)
			anaPrint(ANA_DEBUG, "anaL2_unpublish : Unpublish of entry %s Failed\n",descrip);
		else
			anaPrint(ANA_DEBUG, "anaL2_unpublish : Unpublish Error\n");
				
		responseCond->valueLen = 0; 
	}
	else {
		if(descrip!=NULL)
			anaPrint(ANA_NOTICE, "anaL2_unpublish : Unpublish of entry %s succeeded\n",descrip);
		else
			anaPrint(ANA_NOTICE, "anaL2_unpublish : Unublish Success\n");
		
		responseCond->valueLen =  1; 
	}
	
	responseCond->value = (void*)1;
	
	analock_lock(&(responseCond->condLock)); 
	analock_cond_signal(&(responseCond->cond));	
	analock_unlock(&(responseCond->condLock));
	
}

/* this function just sets the received message in the value field of the responseCond and awakes the
waiting threads */
void anaL2_handleLookupReply(char *data, int len, anaLabel_t input, struct condLock_s *responseCond) {
	analock_lock(&(responseCond->valueLock));
	responseCond->valueLen = len;
	responseCond->value = (void *) malloc(len);
	memcpy(responseCond->value, data, len);
	analock_unlock(&(responseCond->valueLock));
	
	analock_lock(&(responseCond->condLock));
	analock_cond_signal(&(responseCond->cond));
	analock_unlock(&(responseCond->condLock));
}


/* This function is a default Handler that queues messages in case no callback function is specified in publish*/

void anaL2_queueMessage(char *data, int len, anaLabel_t input, void *aux) {
        
	struct MQEntry_s *mqtEntry=NULL;
	struct MessageQueue_s *mq = NULL;
	
	if(data==NULL || len<=0)
		return;
			
	/* first we get the message queue mapped to the input label*/
	mqtEntry = getMQTEntry(&MQTable, input);
	if(mqtEntry==NULL)
		return;
	
	/*we lock the access to the queue*/
	analock_lock(&(mqtEntry->queueLock));
	if(mqtEntry->queue==NULL){
		mq = (struct MessageQueue_s *) malloc(sizeof(struct MessageQueue_s));
		mq->message = (void*) malloc(len);
		memcpy(mq->message, data, len);
		mq->msgLen = len;
		mqtEntry->queue = mq;
	}
	else{
		mq = mqtEntry->queue;
		while(mq->next !=NULL)
			mq=mq->next;
		mq->next = (struct MessageQueue_s *) malloc(sizeof(struct MessageQueue_s));
		mq->next->message = (void*) malloc(len);
		memcpy(mq->next->message, data, len);
		mq->next->msgLen = len;
	}
	analock_unlock(&(mqtEntry->queueLock));

	/*now we send a signal on the condition to wake up the awaiting threads, if any*/
	/*first we need to lock the access to the condition variable*/
	analock_lock(&(mqtEntry->condLock));
	analock_cond_signal(&(mqtEntry->cond));
	analock_unlock(&(mqtEntry->condLock));
}

void anaL2_threadWrapper(struct anaL2_serviceWrap *wrp){

	if(wrp && wrp->funct && wrp->msg)
		wrp->funct(wrp->msg);
	anaL2_freeMsg(wrp->msg);
	free(wrp);
}



void anaL2_launchThread(char *data, int len, anaLabel_t input, void *aux) {

	struct anaL2_message *msg = NULL;
	struct anaL2_serviceWrap *wrp = NULL;
	anaLabel_t *tmpLabel = NULL;
	int ret;
	
	if (data == NULL || aux == NULL)
		return;
		
	/* we need to copy the data (since separate thread)*/
	msg = (struct anaL2_message *) malloc(sizeof(struct anaL2_message));
	
	msg->bufferPointer = (void *) malloc(len);
	memcpy(msg->bufferPointer, data, len);
	msg->idp = input;
	
	if (equalXRPCommands((char *)msg->bufferPointer, XRP_CMD_DATA)){
		
		msg->dataLen = anaL1_decResponse((xrpMsg_t)msg->bufferPointer, XRP_CLASS_MESSAGE, 0, &msg->data);

		msg->senderContext = (struct context_s *) malloc(sizeof(struct context_s));
		msg->senderContext->valueLen = anaL1_decResponse((xrpMsg_t)msg->bufferPointer, XRP_CLASS_SRC_CXT, 0, &(msg->senderContext->value));
		
		
		msg->senderService = (struct service_s *) malloc(sizeof(struct service_s));	
 		msg->senderService->valueLen = anaL1_decResponse((xrpMsg_t)msg->bufferPointer, XRP_CLASS_SRC_SRV, 0, &(msg->senderService->value));
		
		ret = anaL1_decResponse((xrpMsg_t)msg->bufferPointer, XRP_CLASS_LABEL, 0, (void**)&tmpLabel);

		
		if(tmpLabel)
			msg->responseLabel = *tmpLabel;
	}	
	else {
		msg->dataLen = len;
		msg->data = msg->bufferPointer;
	}
	
	wrp = (struct anaL2_serviceWrap *) malloc(sizeof(struct anaL2_serviceWrap));
	wrp->funct = aux;
	wrp->msg = msg;
	/*the function to lauch in a separate thread is in aux argument*/
	_anaThread(myHandle, (void *) anaL2_threadWrapper, wrp, NULL);
}

void anaL2_serviceWrapper(char *data, int len, anaLabel_t input, AL2Callback_t aux) {

	struct anaL2_message msg;
	anaLabel_t *tmpLabel = NULL;
	struct context_s senderContext;
	struct service_s senderService;
	
	if (data == NULL || aux == NULL)
		return;
		
	/* at the difference of launchThread, here we do not copy the message */ 
	/* calling anaL2_freeMsg later is NOT dangerous*/
	
/*	msg = (struct anaL2_message *) malloc(sizeof(struct anaL2_message)); */
	memset(&msg, 0, sizeof(struct anaL2_message));
		
	msg.idp = input;
	
	if (equalXRPCommands((char *)data, XRP_CMD_DATA)){
		memset(&senderContext, 0, sizeof(struct context_s));
		memset(&senderService, 0, sizeof(struct service_s));
		
		msg.dataLen = anaL1_decResponse((xrpMsg_t)data, XRP_CLASS_MESSAGE, 0, &(msg.data));
		
		senderContext.valueLen = anaL1_decResponse((xrpMsg_t)data, XRP_CLASS_SRC_CXT, 0, &(senderContext.value));
 		senderService.valueLen = anaL1_decResponse((xrpMsg_t)data, XRP_CLASS_SRC_SRV, 0, &(senderService.value));
		
		msg.senderContext = &senderContext;
		msg.senderService = &senderService;
		
		anaL1_decResponse((xrpMsg_t)data, XRP_CLASS_LABEL, 0, (void**) &tmpLabel);
		
		if(tmpLabel)
			msg.responseLabel = *tmpLabel;
	}
	else {
		msg.dataLen = len;
		msg.data = data;
	}
	
	/*the function to lauch is in aux argument*/
	aux(&msg);
}


int anaL2_initDefault(void){
	
	if(initializeMQT(&MQTable, DEFAULT_MQ_BITSIZE) < 0)
		return -1;
	return 0;
}


/*! \ingroup anaL2
 * \brief frees the L2 message \a msg. This function is to be used only if the message was popped via the anaL2_receive function. DO NOT use it at the end of the AL2 callback functions.  
 * \param msg The message to free
 */
void anaL2_freeMsg(struct anaL2_message *msg){
	
	if (msg==NULL)
		return;
	if (msg->bufferPointer)
		free(msg->bufferPointer);
	
	if(msg->senderContext && msg->senderContext->value)
		free(msg->senderContext->value);
	
	if(msg->senderService && msg->senderService->value)
		free(msg->senderService->value);
	
	free(msg);
}

/*! \ingroup anaL2
 * \brief frees the lookup resopnse \a resp
 * \param resp The response to free
 */
void anaL2_freeLkpResponse(struct anaL2_lkpResponse *resp){
	
	struct anaL2_lkpResponse *next;
	
	while (resp!=NULL){
		free(resp->description);
		next = resp->next;
		free(resp);
		resp=next;
	}
}


 /** \ingroup anaL2
 * \brief function allowing a compartment provider brick to pass data to a "higher layer" brick while indicating the sender's context and service and IDP label to respond
 * \param destIDP IDP of the "higher layer" brick
 * \param data received data to be forwarded to the "higher layer" brick
 * \param lenData length of the data
 * \param senderContext sender Context
 * \param senderTarget sender Target
 * \param responseLabel IDP label to respond to the sender, in case the compartment provider wishes to open a channel in reverse direction
 * \return number of bytes sent on success, -1 otherwise
 */
int anaL2_forwardData(anaLabel_t destIDP, void *data , int lenData, struct context_s *senderContext, struct service_s *senderTarget, anaLabel_t responseLabel){
	
	int msgLen;
	int ret;
	xrpMsg_t al1msg=NULL;
	
	if(data ==NULL || lenData < 0)
		return -1;
	
	al1msg = anaL1_allocateMessage();
	
	msgLen=anaL1_encData(al1msg, data, lenData , senderContext, senderTarget, responseLabel);
	if(msgLen<0){
		anaPrint(ANA_NOTICE, "anaL2: unable to encode forward data message\n");
		free(al1msg);
		return -1;
	}
	
	ret = anaL0_send(destIDP, al1msg, msgLen);
	if( ret <0 ){
		anaPrint(ANA_NOTICE, "Unable to forward data message (with al0_send)\n");
		free(al1msg);
		return ret;
	}
		
	free(al1msg);
	return msgLen;
	
}

/** \ingroup anaL2
 * \brief This function pops out the oldest message from the IDP's message queue.
 * if no message is received during a given time period the function returns
 * \param input the label on which the message should be received
 * \param timeout delay after a timeout occurs (if NULL, call is blocking)
 * \return the message received or NULL otherwise */
struct anaL2_message * anaL2_receive(anaLabel_t input, struct timespec *timeout){
	
	struct MQEntry_s *mqtEntry=NULL;
	struct MessageQueue_s *mq = NULL;
	int ret=0;	
	struct anaL2_message *res=NULL;
	anaLabel_t *tmpLabel = NULL;

	mqtEntry = getMQTEntry(&MQTable, input);
	if(mqtEntry==NULL)	/* this means there was no message Queue set for this IDP */
		return NULL;
	
	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	ret = 0;
	
	analock_lock(&(mqtEntry->condLock));
	
	if (mqtEntry->queue == NULL)
	{
		if (timeout)
			ret = analock_cond_timedwait(&(mqtEntry->cond), &(mqtEntry->condLock), 
						     timeout->tv_sec, timeout->tv_nsec);
		else
			ret = analock_cond_wait(&(mqtEntry->cond), &(mqtEntry->condLock));
	}
	
	analock_unlock(&(mqtEntry->condLock));
	
	/* once we're here means that there is a message in the queue or we timed out */
	if (ret!=0 || mqtEntry->queue == NULL)
		return NULL;
	
	/* we lock the access to the queue */	
	analock_lock(&(mqtEntry->queueLock));

	/* we pop out the first message in the queue (the oldest) */
	res = (struct anaL2_message *) malloc(sizeof(struct anaL2_message));
	res->idp = input;
	
	if(equalXRPCommands((char *)mqtEntry->queue->message, XRP_CMD_DATA)){
		res->dataLen = anaL1_decResponse((xrpMsg_t)mqtEntry->queue->message, XRP_CLASS_MESSAGE, 0, &res->data);
		
		res->senderContext = (struct context_s *) malloc(sizeof(struct context_s));
		
		res->senderContext->valueLen = anaL1_decResponse((xrpMsg_t)mqtEntry->queue->message, XRP_CLASS_SRC_CXT, 0, &(res->senderContext->value));
 		
		res->senderService = (struct service_s *) malloc(sizeof(struct service_s));
		res->senderService->valueLen = anaL1_decResponse((xrpMsg_t)mqtEntry->queue->message, XRP_CLASS_SRC_SRV, 0, &(res->senderService->value));
		
		anaL1_decResponse((xrpMsg_t)mqtEntry->queue->message, XRP_CLASS_LABEL, 0, (void **)&tmpLabel);
		res->responseLabel = *tmpLabel;
		res->bufferPointer = mqtEntry->queue->message;
	}
	else {
		res->dataLen = mqtEntry->queue->msgLen;
		res->data = mqtEntry->queue->message;
		res->bufferPointer = mqtEntry->queue->message;
	}
	
	/* we pop out this message from the queue */
	mq = mqtEntry->queue->next;
	free(mqtEntry->queue);
	mqtEntry->queue = mq;
	analock_unlock(&(mqtEntry->queueLock));
	
	return res;
}


/** \ingroup anaL2
 * \brief Implements the lookup directive 
 *
 * \param compIDP IDP of the compartment in which the lookup will be
 * \param ctx the context (e.g. "*", ".", "255.255.255.0")
 * \param srv the service (what is looked up) 
 * \param result place to store the result 
 * \param description self description
 * \param timeout delay after which the request times out
 * \return success: how many elements where found, failure: negative error value
*/
int anaL2_lookup(anaLabel_t compIDP, struct context_s *ctx, struct service_s *srv, struct anaL2_lkpResponse **result, struct service_s *description, struct timespec *timeout){
	
	anaLabel_t handleReplyIDP;
	anaLabel_t *tmpLabel = NULL;
	char *tmpDescr=NULL;
  	struct condLock_s  *responseCondition=NULL;
	xrpMsg_t al1msg;
	int al1msgLen, arglen, ret=0;
	int counter,tmpdescLen;
	struct anaL2_lkpResponse *tmpRes = NULL;
	struct anaL2_lkpResponse *predecessor = NULL;
	struct timespec *t;
	
	if (!ctx || !srv || !result)
	{
		set_ana_errno(ANA_FCT_ARG_NULL);
		return ANA_FCT_ARG_NULL;
	}
	
	if (timeout)
		t = timeout;
	else
		t = &default_timeout;
	
	/* first we register the lookup reply handler*/
	responseCondition = (struct condLock_s *) malloc(sizeof(struct condLock_s));
	initCondLock(responseCondition);
		
	handleReplyIDP = anaL0_registerCallback((anaCallback_t) & anaL2_handleLookupReply, responseCondition, (copyAuxFunction) identityFct, (freeAuxFunction) freeCondLock, 0, 0, NOTHREAD );
	if(handleReplyIDP == 0) { 
		free(responseCondition);
		anaPrint(ANA_DEBUG, "anaL2_lookup: unable to register a new IDP to handle the lookup response\n");
		set_ana_errno(ANA_UNABLE_TOSET_RESPONSELABEL_ERROR);
		return ANA_UNABLE_TOSET_RESPONSELABEL_ERROR;
	}
    		
	al1msg = anaL1_allocateMessage();
	if((al1msgLen=anaL1_encLookup(al1msg, ctx, srv, description)) < 0 ){
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0); /*freeing of response condition is handled by the unregisterCallback*/
		anaPrint(ANA_DEBUG, "anaL2_lookup: unable to encode AL1 lookup message\n");
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return ANA_UNABLE_TOENC_QUERY_ERROR;
	}
  
 	 /*now we add the handleReplyIDP to the message as a replyTo label*/

	arglen = addXRPArg((xrpMsg_t) al1msg, XRP_CLASS_LABEL, &handleReplyIDP, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_DEBUG, " anaL2_resolve: Unable to add response label to message\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return ANA_UNABLE_TOENC_QUERY_ERROR;
	}
	al1msgLen += arglen;
	
	/* we send the message to the CPE*/
  	if(anaL0_send(compIDP , al1msg , al1msgLen)<0){
		anaPrint(ANA_DEBUG, " anaL2_lookup: Unable to send the al1 message\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_UNABLE_TOSEND_QUERY_ERROR);
		return ANA_UNABLE_TOSEND_QUERY_ERROR;
	}  
	
	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	ret = 0;
		
	/* now we need to block until the lookup response gets handled */
	analock_lock(&(responseCondition->condLock));
	
	if(responseCondition->value==NULL)
		ret = analock_cond_timedwait(&(responseCondition->cond), &(responseCondition->condLock), 
					     t->tv_sec, t->tv_nsec);	
	
	analock_unlock(&(responseCondition->condLock));
	
	if(ret!=0 || responseCondition->value == NULL ){
		anaPrint(ANA_DEBUG, " anaL2_lookup: timedout before lookup response, ret = %d \n", ret);
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_TIMEOUT_ERROR);
		return ANA_TIMEOUT_ERROR;
	}
	
	/* now we need to extract the label from the reply and return it */	
	
	if(anaL1_isError((xrpMsg_t)responseCondition->value)) {/*if ERROR command*/
		anaPrint(ANA_ERR, "anaL2_lookup: Lookup Failed error msg received\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_REMOTE_ERROR);
		return ANA_REMOTE_ERROR;
	}
		
	/* now we build the chained list of anaL2_lkpResponse */
	counter = 0;
		
	while(anaL1_decResponse((xrpMsg_t)responseCondition->value, XRP_CLASS_LABEL, counter, (void**)&tmpLabel)>0){
		tmpRes = (struct anaL2_lkpResponse *) malloc(sizeof(struct anaL2_lkpResponse));

		if(predecessor == NULL)
			*result = tmpRes;
		else
			predecessor->next = tmpRes;
		
		tmpRes->label = *tmpLabel;
		tmpdescLen = anaL1_decResponse((xrpMsg_t)responseCondition->value, XRP_CLASS_LKP_DESCR, counter, (void**)&tmpDescr);
		if(tmpdescLen>0){
			tmpRes->description = (void *) malloc(tmpdescLen);
			memcpy(tmpRes->description, tmpDescr,tmpdescLen);
			tmpRes->descLen = tmpdescLen;
		}
		else{
			tmpRes->description = NULL;
			tmpRes->descLen = 0;
		}
		
		predecessor = tmpRes;
		counter++;
	}
	
	if(counter==0){ /* means no results */
		anaPrint(ANA_DEBUG, "anaL2_resolve: Failed : Empty results\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_EMPTY_RESULT_ERROR);
		return ANA_EMPTY_RESULT_ERROR;
	}
	
	free(al1msg);
	anaL0_unregisterCallback(handleReplyIDP, 0);
	return counter;
}

/** \ingroup anaL2
 * \brief Implements the resolve directive 
 *
 * \param compIDP IDP of the compartment in which the lookup will be
 * \param ctx the context (e.g. "*", ".", "255.255.255.0")
 * \param srv the service (what is resolved) 
 * \param chanType 'u' unicast 'm' multicast 'a' anycast 'b' broadcast 
 * \param description self description
 * \param timeout delay after which the request times out
 * \return success: the idp for the resolved target, failure: returns 0
*/
anaLabel_t anaL2_resolve(anaLabel_t compIDP, struct context_s *ctx, struct service_s *srv, char chanType, struct service_s *description, struct timespec *timeout){
  	
	anaLabel_t handleReplyIDP;	
	anaLabel_t *tmpLabel = NULL;
  	struct condLock_s  *responseCondition=NULL;
	xrpMsg_t al1msg;
	int al1msgLen, arglen, ret=0;
	struct timespec *t;
	
	if (!ctx || !srv)
	{
		set_ana_errno(ANA_FCT_ARG_NULL);
		return 0;
	}
	
	if (timeout)
		t = timeout;

	else
		t = &default_timeout;
	
	/* first we register the lookup reply handler*/
	responseCondition = (struct condLock_s *) malloc(sizeof(struct condLock_s));
	initCondLock(responseCondition); 
	
	handleReplyIDP = anaL0_registerCallback((anaCallback_t) & anaL2_handleLookupReply, responseCondition, (copyAuxFunction)identityFct, (freeAuxFunction)freeCondLock, 0, 0, NOTHREAD );
	if( handleReplyIDP == 0) { 
		free(responseCondition);
		anaPrint(ANA_DEBUG, "anaL2_resolve: unable to register a new IDP to handle the lookup response\n");
		set_ana_errno(ANA_UNABLE_TOSET_RESPONSELABEL_ERROR);
		return 0;
	}
  	/* we then encode and send the AL1 lookup message */
	al1msg = anaL1_allocateMessage();
	if((al1msgLen=anaL1_encResolve(al1msg, ctx, srv, chanType, description)) < 0 ){
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0); /*freeing of response condition is handled by the unregisterCallback*/
		anaPrint(ANA_DEBUG, "anaL2_resolve: unable to encode AL1 lookup message\n");
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return 0;
	}
  
 	 /*now we add the handleReplyIDP to the message as a replyTo label*/
	arglen = addXRPArg((xrpMsg_t) al1msg, XRP_CLASS_LABEL, &handleReplyIDP, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_DEBUG, " anaL2_resolve: Unable to add response label to message\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return 0;
	}
	al1msgLen += arglen;
		
	/* we send the message to the CPE*/
  	if(anaL0_send(compIDP , al1msg , al1msgLen)<0){
		anaPrint(ANA_DEBUG, " anaL2_resolve: Unable to send the al1 message to %i\n", compIDP);
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_UNABLE_TOSEND_QUERY_ERROR);
		return 0;
	}  
	
	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	
	ret = 0;
	/* now we need to block until the lookup response gets handled */
	analock_lock(&(responseCondition->condLock));
	if(responseCondition->value == NULL)
		ret = analock_cond_timedwait(&(responseCondition->cond), &(responseCondition->condLock), t->tv_sec, t->tv_nsec);	
	
	analock_unlock(&(responseCondition->condLock));
	
	if(ret!=0 || responseCondition->value == NULL ){
		anaPrint(ANA_DEBUG, " anaL2_resolve: timedout before resolve response, ret = %d , condition %s, lock: %p\n", ret, (char *)responseCondition->value, &responseCondition->cond );
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_TIMEOUT_ERROR);
		return 0;
	}
	/* now we need to extract the label from the reply and return it */	
	
	if(anaL1_isError((xrpMsg_t)responseCondition->value)) {/*if ERROR command*/
		anaPrint(ANA_ERR, "anaL2_resolve: Lookup Failed error msg received\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_REMOTE_ERROR);
		return 0;
	}

	if(anaL1_decResponse((xrpMsg_t)responseCondition->value, XRP_CLASS_LABEL, 0, (void**)&tmpLabel)>0){
				
		anaLabel_t res = *tmpLabel;
	
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		return res; /* idp */
	}
	else{
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_EMPTY_RESULT_ERROR);
		return 0;
	}
}

/** \ingroup anaL2
 * \brief sends a request and waits until exact one message is arrived (or a timer expired)
 * \param dest destination IDP
 * \param msg the xrp message containing the request (note, it has to be an xrp message allocated with anaL1_allocateMessage
 * \param reqlen the length of msg 
 * \param respLen upon returning this will contain the length of the answer
 * \param timeout after which the request times out
 * \return success: the idp for the resolved target, failure: negative error value
*/
char * anaL2_requestReply(anaLabel_t dest, xrpMsg_t msg, int reqlen, int *respLen, struct timespec *timeout){
	
	/* we register first a callback for the service*/
	anaLabel_t replyTo;
	int ret = 0, arglen;
	char * result;
	struct condLock_s  *responseCondition=NULL;
	struct timespec *t;
	
	if (timeout)
		t = timeout;
	else
		t = &default_timeout;
	
	/* will be freed upon unregistering the Callback function */
	responseCondition = (struct condLock_s *) malloc(sizeof(struct condLock_s));
	initCondLock(responseCondition);
			
	/* all messages to be received are sent to this function. */
	replyTo = anaL0_registerCallback((anaCallback_t) & anaL2_handleLookupReply, responseCondition, (copyAuxFunction)identityFct, (freeAuxFunction)freeCondLock, 0, 0, NOTHREAD);
	if (replyTo == 0) {
		anaPrint(ANA_DEBUG, "anaL2_requestReply: unable to register a new IDP for the service\n");
		return NULL; /* ana_errno is already set by AL0 */
	}

	/*now we add the replyTo IDP to the message as a replyTo label*/
	arglen = addXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &replyTo, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_DEBUG, " anaL2_requestReply: Unable to add response label to message\n");
		anaL0_unregisterCallback(replyTo, 0);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return NULL;
	}

	reqlen += arglen;

	/* now we send the message */
	if(anaL0_send(dest, msg, reqlen)<0){
		anaPrint(ANA_DEBUG, " anaL2_requestReply: Unable to send the message\n");
		anaL0_unregisterCallback(replyTo, 0);
		set_ana_errno(ANA_UNABLE_TOSEND_QUERY_ERROR);
		return NULL;
	}

	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	ret = 0; 
	
	analock_lock(&(responseCondition->condLock));
	if(responseCondition->value==NULL)
		ret = analock_cond_timedwait(&(responseCondition->cond) ,&(responseCondition->condLock), 
					     t->tv_sec, t->tv_nsec);
	analock_unlock(&(responseCondition->condLock));

	if(ret!=0){
		anaPrint(ANA_DEBUG, " anaL2_requestReply: timedout before resolve response \n");
		anaL0_unregisterCallback(replyTo, 0);
		set_ana_errno(ANA_TIMEOUT_ERROR);
		return NULL;
	}

	/* when we arrive here, means that a reponse was received */
	if(responseCondition->value == NULL){
		anaPrint(ANA_DEBUG, " anaL2_requestReply: condition unlocked, even though no lookup reply\n");
		anaL0_unregisterCallback(replyTo, 0);
		set_ana_errno(ANA_EMPTY_RESULT_ERROR);
		return NULL;
	}

	/* the responseCondition will be freed with unregistering the callback */
	result = (char *)malloc(responseCondition->valueLen);
	memcpy(result, responseCondition->value, responseCondition->valueLen);
	
	if(respLen != NULL)
		*respLen = responseCondition->valueLen;
	anaL0_unregisterCallback(replyTo, 0);
	return result;
}


/** \ingroup anaL2
 * \brief Implements the publish directive 
 *
 * \param compIDP IDP of the compartment in which the lookup will be
 * \param ctx the context (e.g. "*", ".", "255.255.255.0")
 * \param srv what is published
 * \param callback the callback function associated with the description
 * \param separateThread NOTHREAD: service will be directly executed (note: it is not possible to receive any other messages when
 * 	the service function will be executed). THREAD: service will be executed in a dedicated thread  
 * \param timeout timespec struct that specifies when the operation times out.
 * \return success: the idp for the resolved target, failure: returns 0
*/
anaLabel_t anaL2_publish(anaLabel_t compIDP, struct context_s *ctx, struct service_s *srv, AL2Callback_t callback, int separateThread, struct timespec *timeout){

	/* we register first a callback for the service*/
	anaLabel_t result;
	anaLabel_t handleReplyIDP;	
	xrpMsg_t al1msg;
	int al1msgLen, arglen, ret = 0;
	struct condLock_s  *responseCondition=NULL;
	int nullService = 0;
	void *aux=NULL;
	copyAuxFunction cp = NULL;
	freeAuxFunction fr = NULL;
	anaCallback_t funct;
	struct timespec *t;
	
	if ( !ctx || !srv )
	{
		set_ana_errno(ANA_FCT_ARG_NULL);
		return 0;
	}
	
	if (timeout)
		t = timeout;
	else
		t = &default_timeout;
	
	if (!callback){
		funct = (anaCallback_t) & anaL2_queueMessage;	
		nullService = 1 ;
	}
	else{

		if(separateThread){
			aux = callback;
			funct = (anaCallback_t) & anaL2_launchThread;
			cp = identityFct;
		}
		else{   /* we register the anaL2_serviceWrapper function as an AL0 callback function, this function will parse the incoming message and build an
			anaL2_message structure to pass to the published service */
			aux = callback;
			funct = (anaCallback_t) & anaL2_serviceWrapper;
			cp = identityFct;
		}
	} 
	/* First we register an IDP to handle the publish reply */
	
	responseCondition = (struct condLock_s *) malloc(sizeof(struct condLock_s));
	initCondLock(responseCondition);
			
	handleReplyIDP = anaL0_registerCallback((anaCallback_t) & anaL2_handlePublishReply, responseCondition, (copyAuxFunction) identityFct, (freeAuxFunction) freeCondLock, 0, 0, NOTHREAD);
	if( handleReplyIDP == 0 ) { 
		free(responseCondition);
		anaPerror(ANA_ERR, "anaL2_publish: unable to register a new IDP to handle the publish response");
		set_ana_errno(ANA_UNABLE_TOSET_RESPONSELABEL_ERROR);
		return 0;
	}
	
	result = anaL0_registerCallback((anaCallback_t)funct, aux, cp, fr, 0, IDP_PERM, NOTHREAD );
	if( result == 0 ) { /*TOCHANGE :for now we asunme we don't pass auxillary args in L2, and no viewer spec*/
		anaL0_unregisterCallback(handleReplyIDP, 0);
		anaPrint(ANA_DEBUG, "anaL2_publish: unable to register a new IDP for the service\n");
		return 0; /* AL0 has already set ana_errno */
	}
	
	/*we then publish it to the destination compartment*/
	
	/*encoding */
	al1msg = anaL1_allocateMessage();
	if((al1msgLen=anaL1_encCompartmentPublish(al1msg, result, ctx, srv, 1)) < 0 ){
		free(al1msg);
		anaL0_unregisterCallback(result, 0);
		anaL0_unregisterCallback(handleReplyIDP, 0);
			
		anaPrint(ANA_DEBUG, "anaL2_publish: unable to encode publish message for the new service's IDP\n");
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return 0;
	}
	
	/*now we add the handleReplyIDP to the message as a replyTo label*/
	
	arglen = addXRPArg((xrpMsg_t) al1msg, XRP_CLASS_LABEL, &handleReplyIDP, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_DEBUG, "anaL2_publish: Unable to add response label to message\n");
		free(al1msg);
		anaL0_unregisterCallback(result, 0);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return 0;
	}
	al1msgLen += arglen;
	
	/*Block here waiting for a confirmation or error from the default callback function handling the publication */	
	/*depending on the result we return the rgistered IDP (success) or NULL (failure)*/
	
	
	/* we send the publish message now */
	if(anaL0_send(compIDP, al1msg, al1msgLen)<0){
		anaPrint(ANA_DEBUG, "anaL2_publish: Unable to send the publish message\n");
		free(al1msg);
		anaL0_unregisterCallback(result, 0);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		
		set_ana_errno(ANA_UNABLE_TOSEND_QUERY_ERROR);
		return 0;
	}  
	
	/* now we need to block until the publish response gets handled */
	
	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	ret = 0; 
		
	analock_lock(&(responseCondition->condLock));
	if(responseCondition->value==NULL)
		ret = analock_cond_timedwait(&(responseCondition->cond), &(responseCondition->condLock),
					     t->tv_sec, t->tv_nsec);
	if(ret!=0){
		anaPrint(ANA_DEBUG, "anaL2_publish: timedout before publish response ret = %d\n", ret);
		free(al1msg);
		anaL0_unregisterCallback(result, 0);
		anaL0_unregisterCallback(handleReplyIDP, 0);

		set_ana_errno(ANA_TIMEOUT_ERROR);
		return 0;
	}  
	
	analock_unlock(&(responseCondition->condLock));
	/*when we arrive here, means that a reponse was received */

	if( responseCondition->valueLen == 1 ){ /* success of publish*/
		struct MQEntry_s *mqEntry;
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		
		if(nullService){
		
			/* before returning the resulting IDP we add an entry in the Message QueueTable */
			mqEntry = (struct MQEntry_s *) malloc(sizeof(struct MQEntry_s));
			mqEntry->label = result;		
			if(addMQTEntry(&MQTable, mqEntry)<0){
				free(al1msg);
				free(mqEntry);
				anaL0_unregisterCallback(handleReplyIDP, 0);
				anaL0_unregisterCallback(result, 0);
				
				set_ana_errno(ANA_UNKNOWN_ERROR);		
				return 0;
			}
		}
		return result; /* return idp value */
	}
	else{
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		anaL0_unregisterCallback(result, 0);
		
		set_ana_errno(ANA_REMOTE_ERROR);
		return 0;
	}
	return 0;
}  

/** \ingroup anaL2
 * \brief unpublishes a keyword in a compartment. 
 * This function does not destroy the IDP (meaning no unregisterCallback), it just unpublishes 
 * it from the target compartment (since the same IDP can be used in multiple compartments) 
 * \param compIDP IDP of the compartment in which to unpublish
 * \param label the label to unpublish
 * \param ctxt context in which service was published
 * \param srv service that was published
 * \param timeout delay after which the request times out
 * \return success: 0, failure: negative error value
*/
int anaL2_unpublish(anaLabel_t compIDP, anaLabel_t label, struct context_s *ctxt, struct service_s *srv, struct timespec *timeout){
	
	xrpMsg_t al1msg=NULL;
	anaLabel_t handleReplyIDP;
	int msgLen, arglen, ret;
	struct condLock_s  *responseCondition=NULL;
	struct timespec *t;
	
	if ( !srv || !ctxt )
	{
		set_ana_errno(ANA_FCT_ARG_NULL);
		return ANA_FCT_ARG_NULL;
	}
	
	if (timeout)
		t = timeout;
	else
		t = &default_timeout;
	
	responseCondition = (struct condLock_s *) malloc(sizeof(struct condLock_s));
	initCondLock(responseCondition);
	
	/* here we need to take care to not double free the condlock in gates mode
         * we free it after each unregisterCallback call */
	handleReplyIDP = anaL0_registerCallback((anaCallback_t)& anaL2_handleUnPublishReply, responseCondition, (copyAuxFunction) identityFct, NULL,  0, 0, NOTHREAD);
	if (handleReplyIDP == 0) { 
		free(responseCondition);
		anaPrint(ANA_DEBUG, "anaL2_publish: unable to register a new IDP to handle the publish response\n");
		set_ana_errno(ANA_UNABLE_TOSET_RESPONSELABEL_ERROR);
		return ANA_UNABLE_TOSET_RESPONSELABEL_ERROR;
	}
		
	al1msg = anaL1_allocateMessage();
	if (al1msg == NULL)
		return ANA_UNABLE_TOENC_QUERY_ERROR;
	/* encode an unpublish request and sends is to the compartment */
	msgLen = anaL1_encCompartmentUnpublish(al1msg, label, ctxt, srv);
	
	if (msgLen < 0)
	{
		anaPrint(ANA_ERR, "anaL2_unpublish : unable to encode unpublish command\n");
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		free(al1msg);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return ANA_UNABLE_TOENC_QUERY_ERROR;
	}
	

	/*now we add the handleReplyIDP to the message as a replyTo label*/
	arglen = addXRPArg((xrpMsg_t) al1msg, XRP_CLASS_LABEL, &handleReplyIDP, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_DEBUG, "anaL2_publish: Unable to add response label to message\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		set_ana_errno(ANA_UNABLE_TOENC_QUERY_ERROR);
		return ANA_UNABLE_TOENC_QUERY_ERROR;
	}
	msgLen += arglen;
	
	/*we send the publish message now */
	if (anaL0_send(compIDP , al1msg , msgLen)<0){
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		free(al1msg);
		set_ana_errno(ANA_UNABLE_TOSET_RESPONSELABEL_ERROR);
		return ANA_UNABLE_TOSET_RESPONSELABEL_ERROR;
	}
	
	/* setting ret=0 this is very important with .so plugins ! */
	/* sometimes the reply callback is triggered before we do the lock => */
	/* responseCondition->value is != NULL and ret!=0 => bugs */
	ret = 0;
	
	analock_lock(&(responseCondition->condLock));
	if (responseCondition->value==NULL)
		ret = analock_cond_timedwait(&(responseCondition->cond), &(responseCondition->condLock), 
					     t->tv_sec, t->tv_nsec);
	
	if (ret!=0){
		anaPrint(ANA_DEBUG, "anaL2_unpublish: timedout before unpublish response\n");
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		set_ana_errno(ANA_TIMEOUT_ERROR);
		return ANA_TIMEOUT_ERROR;
	}  
	
	analock_unlock(&(responseCondition->condLock));
	/*when we arrive here, means that a reponse was received */

	if (responseCondition->valueLen == 1){ /* success of unpublish*/
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		free(al1msg);
		return 0;
	}
	else{
		free(al1msg);
		anaL0_unregisterCallback(handleReplyIDP, 0);
		free(responseCondition);
		set_ana_errno(ANA_REMOTE_ERROR);
		return ANA_REMOTE_ERROR;
	}
}
