
#include "messageQueue.h"
#include "hashFunction.h"

/***************************************************************

Hash functions by Bob Jenkins, December 1996, Public Domain.

2005-11-09: tested by CJ with 128-bit keys to 10 bits values 
            --> hash function has a uniform distribution

***************************************************************/

ub4 hashMQT( k, length, initval, maskLength)
register ub1 *k;        /* the key */
register ub4  length;   /* the length of the key in bytes */
register ub4  initval;    /* the previous hash, or an arbitrary value */
register int  maskLength;    /* the previous hash, or an arbitrary value */
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */

	return (c & hashmask(maskLength));	 
}

/* hashes an anaLabel into an index */
int label2hashMQT(struct MQTable_s *MQT, anaLabel_t l) {
	int h;
	if(MQT==NULL)
		return -1;
	h = hashMQT((ub1 *)&l, sizeof(anaLabel_t), 0, MQT->hashMaskSize);
	return h;
}


/* hashes an anaLabel into an index */
int initializeMQT (struct MQTable_s *MQT, int hashMaskSize) { 
	/* the allocated size of the table will be (2^hashMaskSize), the hashMaskSize will be also used in the hash  function */
	if(MQT==NULL || hashMaskSize<0)
		return -1;
	MQT->size = pow2(hashMaskSize);
	MQT->hashMaskSize =hashMaskSize;
	MQT->tab = (struct  MQEntry_s**) malloc(MQT->size * sizeof(struct  MQEntry_s*));
	if(MQT->tab==NULL){
		anaPrint(ANA_ERR, "initializeMQT : unable to allocate memory for Message Queue hash table\n");
		return -1;
	}
	anaPrint(ANA_DEBUG, "Message Queue table initialized with size %u\n", MQT->size);
	return 0;
}

/* returns the MQT entry corresponding to an IDP */
struct MQEntry_s* getMQTEntry(struct MQTable_s *MQT, anaLabel_t l) {
	int i ;
	struct MQEntry_s *entry=NULL;
	
	if(MQT==NULL)
		return NULL;
	i = label2hashMQT(MQT, l);
	if (i<0 || i>(int)(MQT->size)-1) {
		anaPrint(ANA_ERR, "getMQTEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	entry = MQT->tab[i];

	if (entry == NULL ) { 	/* meaning that there is no value mapped to this  label in the Hash table */
		anaPrint(ANA_NOTICE, "getMQTEntry : entry in MQT at index %d  is NULL\n", i);
		return NULL;	
	}
			
	if  (l== entry->label) { /* if the keys are equal */
		return entry; 
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
		 * the entry with the correct key */
		entry = entry->next; 
		while (entry != NULL) {
			if( l==entry->label ) 
				return entry;
			else 
				entry = entry->next;
		} 
	}
	return NULL;
}


/* Frees the entire Message Queue Table */
int freeMQTable(struct MQTable_s *MQT) { 
	int i;
	struct MQEntry_s *entry=NULL;
	struct MQEntry_s *next=NULL;
	struct MessageQueue_s *mqNext=NULL;
	
	if(MQT==NULL)
		return -1;
		
	for (i=0; i< (int)(MQT->size); i++) {
		entry = MQT->tab[i];
		while  (entry!=NULL) {
			next = entry->next;
			while(entry->queue != NULL){	
				mqNext = entry->queue->next;
				free(entry->queue->message);
				free(entry->queue);
				entry->queue = mqNext;
			}
		
			free(entry);
			entry = next;
		}
	}
	free(MQT->tab);
	free(MQT);
	return 0;
}

int freeMQTEntry(struct MQTable_s *MQT, anaLabel_t l) { 
	int i ;
	struct MQEntry_s *entry=NULL;
	struct MQEntry_s *tmp=NULL;
	struct MessageQueue_s *mqNext=NULL;
	if(MQT==NULL)
		return -1;
	
	i = label2hashMQT(MQT, l);
	if (i<0 || i>(int)(MQT->size)-1) {
		anaPrint(ANA_ERR, "freeMQTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = MQT->tab[i];
	if (entry == NULL) /* meaning that there is no value mapped to this */
		return -1;	  /*  label in the Hash table */
			
	if (l==entry->label) {
		MQT->tab[i] = entry->next; 
		while(entry->queue != NULL){	
			mqNext = entry->queue->next;
			free(entry->queue->message);
			free(entry->queue);
			entry->queue = mqNext;
		}
				
		free(entry); 
		return 0;
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
			  the entry with the correct key */
		while (entry != NULL ) {
			if(entry->next != NULL) {
				tmp = entry->next;
				if(l==entry->label) {
					entry->next = tmp->next;
					while(tmp->queue != NULL){	
						mqNext = tmp->queue->next;
						free(tmp->queue->message);
						free(tmp->queue);
						tmp->queue = mqNext;
					}
								
					free(tmp);
					return 0;
				}	 
			}
			entry = entry->next;
		}
	}
	return -1;
}

int addMQTEntry (struct MQTable_s *MQT, struct MQEntry_s *toAdd) {
	struct MQEntry_s *entry=NULL;
	int i ;
	if(MQT==NULL || toAdd==NULL)
		return -1;
	
	i= label2hashMQT(MQT, toAdd->label);
	if (i<0 || i>(int)(MQT->size)-1) {
		anaPrint(ANA_ERR, "addMQTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	entry = MQT->tab[i];
	if(entry == NULL ){
		toAdd->next = NULL;
		MQT->tab[i] = toAdd;
	}
	else { /* hash collison */
		anaPrint(ANA_DEBUG, "addMQTEntry : Hash collison\n");
		toAdd->next = MQT->tab[i];
		MQT->tab[i]=toAdd;
	}
	return 0;
}
