/* C/bricks/cfinder/cfinder.c
 *  
 * Compartment finder system
 * 
 * Creates KVR entry: cfinder
 * 
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

#include "brick_template.h"

#ifdef KERNEL
#include <linux/module.h>
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Christophe Jelger");
//static char *cd = "kern://cfinder_data"; //"udp://127.0.0.1:5559"; /* brick data url*/
//static char *cc = "kern://cfinder_ctrl"; //"udp://127.0.0.1:8889"; /* brick control url*/
//static char *nc = "kern://anaMinmex"; //"udp://127.0.0.1:6666"; /* minmex control url*/
//static char * debug  = "ANA_DEBUG";
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

#define MODULE_NAME cfinder
char mymodename[]="cfinder";

/*-------------------------------------------------------------------------------------*/

#include "cfinder.h"

anaLabel_t kvr_idp = 0;
struct context_s cfinder_ctx;
struct service_s cfinder_srv;

/* for future use */
// struct cpmt_entry *cpmt_list = NULL; /* the list of compartments we use (to talk to peers) */

void brick_exit()
{
	anaPrint(ANA_DEBUG, "brick_exit\n");
	
	if (kvr_idp)
	{
		anaL2_unpublish(NODE_LABEL, kvr_idp, &cfinder_ctx, &cfinder_srv, NULL);
		kvr_idp = 0;
	}
}

int brick_start()
{
	anaL2_initDefault();
	
	cfinder_ctx.value = ".";
	cfinder_ctx.valueLen = 2;
	cfinder_srv.value = "cfinder";
	cfinder_srv.valueLen = strlen(cfinder_srv.value)+1;
	
	/* publish cfinder in KVR */
	
	kvr_idp = anaL2_publish(NODE_LABEL, &cfinder_ctx, &cfinder_srv, recv_msg, THREAD, NULL);
	if (!kvr_idp)
	{
		/* publish in KVR failed */
	
		anaPrint(ANA_ERR, "brick_start --> cannot publish cfinder IDP in KVR\n");
		anaL0_stopANA(-1);
	}
	
	return 1;
}

void recv_msg(struct anaL2_message *msg)
{
	xrpCmd_t cmd;
	
	if (!msg || !(msg->data))
	{
		anaPrint(ANA_ERR, "recv_msg --> received NULL message! \n");
		return;
	}
	
	cmd = msg->data;
	
	if (equalXRPCommands(cmd, XRP_CMD_RESOLVE) ||
	    equalXRPCommands(cmd, XRP_CMD_LOOKUP))
	{
		/* resolve or lookup  message */
	
		anaPrint(ANA_NOTICE, "recv_msg --> received resolve/lookup request (%d bytes)\n", msg->dataLen);
		handle_resolve(msg);
	}
		
	else if (equalXRPCommands(cmd, XRP_CMD_PUBLISH ))
	{
		/* publish message */
		
		anaPrint(ANA_NOTICE, "recv_msg --> received publish request (%d bytes)\n", msg->dataLen);
		handle_publish(msg);
	}
		
	else if (equalXRPCommands(cmd, XRP_CMD_UNPUBLISH))
	{
		/* unpublish message */
		
		anaPrint(ANA_NOTICE, "recv_msg --> received unpublish request (%d bytes)\n", msg->dataLen);
		handle_unpublish(msg);
	}
	
	else
		anaPrint(ANA_ERR, "recv_msg --> unknown XRP request (%d bytes)\n", msg->dataLen);
}

void handle_resolve(struct anaL2_message *msg)
{
	anaLabel_t *replyto;
	int len, size = 0, labelsize;
	char *chan_type;
	xrpMsg_t reply_msg;
	struct regexp_entry *reg;
	struct context_s ctx;
	struct service_s srv;
	
	if (!msg || !(msg->data))
		return;
	
	replyto = (anaLabel_t *) getXRPArg(msg->data, XRP_CLASS_LABEL, &labelsize, 0);
	srv.value = getXRPArg(msg->data, XRP_CLASS_DST_SRV, &srv.valueLen, 0);
	ctx.value = getXRPArg(msg->data, XRP_CLASS_DST_CXT, &ctx.valueLen, 0);
	chan_type = (char *) getXRPArg(msg->data, XRP_CLASS_CHANTYPE, &len, 0);
	
	if (allocateXRPMsg(&reply_msg) == -1)
	{
		anaPrint(ANA_ERR, "handle_resolve --> cannot allocate XRP message!\n");
		return;
	}
	
	if (!srv.value || !ctx.value || !replyto)
	{
               	anaPrint(ANA_ERR, "handle_resolve --> cannot extract some information!\n");
		
		if (replyto)
		{
			size += fillXRPCommand(reply_msg, XRP_CMD_ERROR);
			anaL0_send(*replyto, reply_msg, size);
		}
		
		free(reply_msg);
               	return;
        }

	anaPrint(ANA_NOTICE, "handle_resolve --> %s | %s\n", 
		 (char *)ctx.value, (char *)srv.value);

	reg = lookup_entry(srv.value);

	if (!reg)
	{
		size += fillXRPCommand(reply_msg, XRP_CMD_ERROR);
		anaL0_send(*replyto, reply_msg, size);
		return;
	}

	size += fillXRPCommand(reply_msg, XRP_CMD_RESOLVE);
	size += addXRPArg(reply_msg, XRP_CLASS_LABEL, &reg->idp, sizeof(anaLabel_t)); 
	anaL0_send(*replyto, reply_msg, size);
	free(reply_msg);
}

void handle_publish(struct anaL2_message *msg)
{
	anaLabel_t *idp, *replyto;
	int len;
	char flags = 0;
	struct context_s ctx;
	struct service_s srv;
	xrpMsg_t reply_msg;
	
	if (!msg || !(msg->data))
		return;
	
	reply_msg = msg->data;
	
	srv.value = getXRPArg(msg->data, XRP_CLASS_SRC_SRV, &srv.valueLen, 0);
	ctx.value = getXRPArg(msg->data, XRP_CLASS_SRC_CXT, &ctx.valueLen, 0);
	idp = (anaLabel_t *) getXRPArg(msg->data, XRP_CLASS_LABEL, &len, 0);
	replyto = (anaLabel_t *) getXRPArg(msg->data, XRP_CLASS_LABEL, &len, 1);
	
	if (!idp || !srv.value || !replyto || !ctx.value)
	{
               	anaPrint(ANA_ERR, "handle_publish --> cannot extract some information!\n");
               	
		if (replyto)
		{
			fillXRPCommand(reply_msg, XRP_CMD_ERROR);
			anaL0_send(*replyto, reply_msg, msg->dataLen);
		}
		return;
        }
	
	if (strcmp(ctx.value, "."))
	{
		/* for the moment we only accept publish in the "." context */
		
		anaPrint(ANA_DEBUG, "handle_publish --> publish context different than \".\"\n");
		
		if (replyto)
		{
			fillXRPCommand(reply_msg, XRP_CMD_ERROR);
			anaL0_send(*replyto, reply_msg, msg->dataLen);
		}
		return;
	}
	
	anaPrint(ANA_NOTICE, "handle_publish --> %s\n", (char *)srv.value);
	
	flags |= REGEXP_SCOPE_LOCAL; /* we only handle the "." context for now */
	
	if (add_entry(srv.value, *idp, flags, 0) == -1)
		fillXRPCommand(reply_msg, XRP_CMD_ERROR);
	
	/* send back reply */
	anaL0_send(*replyto, reply_msg, msg->dataLen);
}

void handle_unpublish(struct anaL2_message *msg)
{
	int len;
	anaLabel_t *idp, *replyto; 
	xrpMsg_t reply_msg;
	struct service_s srv;
	struct context_s ctx;

	if (!msg || !(msg->data))
		return;

	reply_msg = msg->data;

	srv.value = getXRPArg(msg->data, XRP_CLASS_SRC_SRV, &srv.valueLen, 0);
	ctx.value = getXRPArg(msg->data, XRP_CLASS_SRC_CXT, &ctx.valueLen, 0);
	idp = (anaLabel_t *) getXRPArg(msg->data, XRP_CLASS_LABEL, &len, 0);
	replyto = (anaLabel_t *) getXRPArg(msg->data, XRP_CLASS_LABEL, &len, 1);
	
	if (!srv.value || !ctx.value || !idp || !replyto)
	{
               	anaPrint(ANA_ERR, "handle_unpublish --> cannot extract some information!\n");
               	
		if (replyto)
		{
			fillXRPCommand(reply_msg, XRP_CMD_ERROR);
			anaL0_send(*replyto, reply_msg, msg->dataLen);
		}	
		return;
	}
	
	anaPrint(ANA_NOTICE, "handle_unpublish --> %s\n", (char *)srv.value);
	
	if (del_entry(srv.value, *idp) == -1)
		fillXRPCommand(reply_msg, XRP_CMD_ERROR);
	
	anaL0_send(*replyto, reply_msg, msg->dataLen);	
}
