/* C/bricks/cfinder/table.c
 *    
 * internal stuff of the compartment finder
 *  
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

#include "cfinder.h"

#define ERRBUF_SIZE 256

/* the link list of table entries */
struct regexp_entry *regexp_table = NULL;

int add_entry(char *regexp_str, anaLabel_t idp, char flags, uint16_t hops)
{
	struct regexp_entry *prev = NULL, *reg = regexp_table;
	int i = 0, errno;
	regex_t tmp_regex;
	char errbuf[ERRBUF_SIZE];
	
	/* check first if string is a valid regular expression */
	
	if (!regexp_str || !idp)
	{
		anaPrint(ANA_ERR, "add_entry --> NULL argument!\n");
		return -1;
	}
	
	errno = regcomp(&tmp_regex, regexp_str, REG_EXTENDED);
	
	if (errno)
	{
		regerror(errno, &tmp_regex, errbuf, ERRBUF_SIZE);
		anaPrint(ANA_ERR, "add_entry --> error when compiling regexp = %s\n", errbuf);
		return -1;
	}
	
	while (reg)
	{
		prev = reg;
		reg = reg->next;
		i++;
	}	
	
	reg = malloc(sizeof(struct regexp_entry));
	
	if (!reg)
	{
		anaPrint(ANA_ERR, "add_entry --> cannot allocate memory for new entry!\n");
		return -1;
	}

	reg->idp = idp;
	reg->regexp_str = malloc(strlen(regexp_str)+1);
	if (!reg->regexp_str)
	{
		anaPrint(ANA_ERR, "add_entry --> cannot allocate memory for regexp string!\n");
		free_entry(reg);
		return -1;
	}
	strcpy(reg->regexp_str, regexp_str);
	reg->regex = tmp_regex;
	reg->hops = hops;
	reg->flags = flags;
	reg->next = NULL;

	if (prev)
		prev->next = reg;
	else
		regexp_table = reg;
	
	anaPrint(ANA_NOTICE, "add_entry --> added new entry (idx %d, idp %x)\n", i, idp);
	
	return i;
}

int del_entry(char *regexp_str, anaLabel_t idp)
{
	struct regexp_entry *prev = NULL, *reg = regexp_table;
	int i = 0;

	if (!regexp_table)
	{
		anaPrint(ANA_ERR, "del_entry --> table is empty!\n");
		return -1;	
	}
	
	if (!regexp_str || !idp)
	{
		anaPrint(ANA_ERR, "del_entry --> regexp or IDP NULL!\n");
		return -1;	
	}
	
	while (reg)
	{
		if (!strcmp(reg->regexp_str, regexp_str) && (reg->idp == idp) )
		{
			if (prev)
				prev->next = reg->next;
			else
				regexp_table = reg->next;
			
			anaPrint(ANA_NOTICE, "del_entry --> removing entry (idx %d, idp %x)\n", i, idp);
			free_entry(reg);
			return 1;		
		}
	
		prev = reg;
		reg = reg->next;
		i++;
	}
	
	anaPrint(ANA_NOTICE, "del_entry --> entry (%s idp %x) not found\n", regexp_str, idp);
	return -1; /* not found */
}

struct regexp_entry* lookup_entry(char *str)
{
	struct regexp_entry *reg = regexp_table;
	int result;
	regmatch_t match;
	
	if (!str)
	{
		anaPrint(ANA_ERR, "lookup_entry --> string is NULL!\n");
		return NULL;	
	}

	if (!regexp_table)
	{
		anaPrint(ANA_ERR, "lookup_entry --> table is empty!\n");
		return NULL;	
	}
	
	while (reg)
	{
		result = regexec(&reg->regex, str, 1, &match, 0);
	
		if (!result)
		{
			anaPrint(ANA_NOTICE, "lookup_entry --> match found for %s (%d %d)\n", 
				 str, match.rm_so, match.rm_eo);
			return reg;
		}
		
		reg = reg->next;
	}
	
	anaPrint(ANA_NOTICE, "lookup_entry --> no match found for %s\n", str);
	return NULL; 
}

void free_entry(struct regexp_entry *reg)
{
	if (!reg)
		return;
		
	if (reg->regexp_str)
		free(reg->regexp_str);
		
	free(reg);
	
	return;
}
