/* C/bricks/eth-vl/dataGramvEthBrick.c
 *  
 * implementation of the ethernet brick
 * depends on: vlink 
 * KVR entries: eth01+interface+wired+ethernet (on default) 
 * 
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*----------------------- Default brick header code -----------------------*/
/* brick_template.h contains the functions that initialize the bricks */
#include "brick_template.h"

/* Update the fields below with the name of your brick.
 * This name in combination with the process ID 
 * will be used for the log files */
char mymodename[]="ethBrick";

/* Standard ANA code uses the ISC licence.
 * To be able to load our brick to the kernel we use
 * the compatible "Dual BSD/GPL" license
 */
#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif
/*----------------------- End default brick header code -------------------*/
/*----------------------- Brick specific code starts here------------------*/

#include "dataGramEthBrick.h"
#include "ana_vlinkAPI.h"


struct QREP dataRep;
anaLabel_t vlinkLabel;
anaLabel_t highLayIDP;
anaLabel_t bindreply, datarecv;
anaLabel_t unbind;
anaLabel_t cfinder, regexp; /* for publication of Ethernet regexp in compartment finder */

int vlinkMTU; /* hold the value of the vlink MTU */

void recvFromLL(char *data, int len, anaLabel_t input, void *aux);
void recvFromHL(char *data, int len, anaLabel_t input, void *aux);

/* wrapper for L2 messages: added by Christophe for compartment finder code */
void anaL2_msg(struct anaL2_message *msg); /* simply calls recvFromHL() */

struct resolveRequestList *requestList;

unsigned char  myAddress[ETH_ALEN] ;
unsigned char  broadcastAddress[ETH_ALEN] ;


/* Global declarations */
static void sendToVlink(char *data, int len, anaLabel_t input, struct fullEthHeader *fullHeader);
void freeFullHeader(struct fullEthHeader *fh);
struct fullEthHeader *cloneFullHeader(struct fullEthHeader *fh);
anaLabel_t *cloneLabel2(anaLabel_t *aux);
static void freeValue(void *value);


uint16_t  generateNextHeader(void){
	uint16_t res;
	int exists = 1;
	
	while(exists){
		res = (uint16_t) random();
		/*we check that the header is different from our constatnts*/

		if(res!=0 && res != ETH_CMD  &&  getQREPEntryByName(&dataRep, (char*)&res, sizeof(uint16_t))==NULL)

			exists = 0;
	}
	return res;
}

void  handleUnpublishReply(char *data, int len, anaLabel_t input, void *aux) {
	if(anaL1_isError(data)) 
		anaPrint(ANA_ERR, "Unpublish Error\n");
		
	else
		anaPrint(ANA_ERR, "Unpublish Success\n");
}

anaLabel_t *cloneLabel2(anaLabel_t *aux){
	anaLabel_t *aux2 = (anaLabel_t *) malloc(sizeof(anaLabel_t));
	*aux2 = *aux;
	return aux2;
}


struct fullEthHeader * cloneFullHeader(struct fullEthHeader *fh){
	struct fullEthHeader *clone = NULL;
	if(fh==NULL)
		return NULL;

	clone = (struct fullEthHeader *)malloc(sizeof(struct fullEthHeader));
	clone->ethhdr = (struct ether_header *) malloc(sizeof(struct ether_header));
	memcpy(clone->ethhdr, fh->ethhdr, sizeof(struct ether_header));
	if(fh->destDescrip!=NULL){
		clone->destDescrip = (char*) malloc(fh->destDescLen);
		memcpy(clone->destDescrip,fh->destDescrip, fh->destDescLen);	
		clone->destDescLen = fh->destDescLen;
		
	}
	if(fh->srcDescrip!=NULL){
		clone->srcDescrip = (void*) malloc(fh->srcDescLen);
		memcpy(clone->srcDescrip,fh->srcDescrip, fh->srcDescLen);	   
		clone->srcDescLen =fh->srcDescLen;	
	}
	clone->opCode = fh->opCode;
	return clone;
}

void freeFullHeader(struct fullEthHeader *fh){
	free(fh->ethhdr);
	free(fh->srcDescrip);
	free(fh->destDescrip);
}


/* cleanup function which is called upon leaving the ehtBrick */

void  brick_exit() {

      	xrpMsg_t msg;

	struct service_s service;
	struct context_s context;
	
	int ret = 0, len = 0;
	anaPrint(ANA_NOTICE, "ETH Brick EXIT\n");

	msg = anaL1_allocateMessage();
	

	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	
	
	context.value = ".";
	context.valueLen = 2;
	
	service.value = "eth01+interface+wired+ethernet";
	service.valueLen = strlen("eth01+interface+wired+ethernet")+1;
	
	/* unpublish the the entries in the QREP of the ethBrick */
	len = anaL1_encCompartmentUnpublish(msg, highLayIDP, &context, &service);
	ret = anaL1_requestReply(NODE_LABEL, msg, len, (anaCallback_t)&  handleUnpublishReply, NULL, NULL, NULL, 0);
	if(ret < 0){
		anaPrint(ANA_ERR, "problem Sending unpublish request for \"eth0\"\n");
	}
	free(msg);
	
	
	/* unbind from vlink */
	anaL0_send(unbind, &datarecv, sizeof(anaLabel_t));
	
	anaL0_unregisterCallback(highLayIDP, 0);	
	anaL0_unregisterCallback(datarecv, 0);
	anaL0_unregisterCallback(bindreply, 0);
	
	/* remove compartment finder entry (if any) */
	if (cfinder && regexp)
	{
		struct service_s srv;
		struct context_s ctx;
		
		ctx.value = ".";
		ctx.valueLen = 2;
		srv.value = ETH_REGEXP;
		srv.valueLen = strlen(srv.value)+1;
		
		anaL2_unpublish(cfinder, regexp, &ctx, &srv, NULL);
	}
	
	freeQRepository(&dataRep, freeValue);
}

/* checks whether "eth" was successfully published in the KVR */
 void  handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name=NULL;
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**)&name);

	if(anaL1_isError(data)) {/*if ERROR command*/
		if(name!=NULL)
			anaPrint(ANA_ERR, "Publish of entry %s Failed\n",name);
		else
			anaPrint(ANA_ERR, "Publish Error\n");
		return;
	}
	else
		if(name!=NULL)
			anaPrint(ANA_NOTICE, "Publish of entry %s succeeded\n",name);
		else
			anaPrint(ANA_NOTICE, "Publish Success\n");
}

 void  handle_bindreply(char *data, int len, anaLabel_t input, anaLabel_t *dataRecv) {
	
	struct IDPinfo_s vlinkinfo;
	int ret;

	vlinkLabel = vlinkBindResponse(data, len, *dataRecv, &unbind, myAddress);
	if (vlinkLabel == 0){
		anaPrint(ANA_EMERG, "no vlink Label received -> cannot work without vlink!\n");
		return;
	}
	
	memset(&vlinkinfo, 0, sizeof(struct IDPinfo_s));
	/* we ask for IDP info */
	ret = anaL0_getIDPinfo(vlinkLabel, MY_HANDLE, &vlinkinfo);
	
	if(ret< 0){
		anaPrint(ANA_DEBUG, " handle_bindreply failed to fetch vlink MTU : %s\n",getErrorString(ret));
		vlinkMTU = 0;
	}
	else	
		vlinkMTU = vlinkinfo.MTU;
	
		
	anaPrint(ANA_DEBUG, "Bound vlink to this ethBrick (MAC is " ETH_FMT "), vlink MTU is %d\n", ETH_HEX(myAddress), vlinkMTU);

}
/* receives the answer from the lookup request in the KVR for "vlink"
 * registers the function which receives data from the vlink with the vlinkBrick
 */
 void  handleLookupReply(char *data, int len, anaLabel_t input, void *aux) {

	char *name, *publ_name;
	anaLabel_t *label = NULL;
	int labelSize=0,i, ret, lenMsg;
	xrpMsg_t msg;
	
	struct service_s service;
	struct context_s context;
	
	anaPrint(ANA_NOTICE, "received Lookup reply for \"vlink\" \n");
	
	if(anaL1_isError(data)) {/*if ERROR command*/
		anaPrint(ANA_EMERG, "Lookup for vlink failed .-> can not work without vlink!\n");
		return;
	}

	i=0;
	if(anaL1_decResponse(data, XRP_CLASS_NAME, 0, (void**)&name)>0){
		labelSize =  anaL1_decResponse(data, XRP_CLASS_LABEL, 0, (void**)&label);
		if(name!=NULL && label !=NULL){
			anaPrint(ANA_NOTICE, "\tname:%s  IDP:%d\n", name, *label);
			
			/* register data callback function (function which receives data from vlink) with MINMEX*/
			datarecv = anaL0_registerCallback((anaCallback_t)& recvFromLL, NULL, NULL, NULL, 0, IDP_PERM, 0);
			if(datarecv == 0){
				anaPerror(ANA_EMERG, "Unable register datarecv IDP, -> cannot receive messages from vlink\n");	
				return;
			}

			/* register function which handles bind reply with MINMEX */
			bindreply = anaL0_registerCallback((anaCallback_t)&  handle_bindreply, &datarecv, (copyAuxFunction) cloneLabel2, (freeAuxFunction) free, 0, 0, 0);
			if (bindreply == 0) {
				anaPerror(ANA_EMERG, "Unable to register bind response IDP: -> cannot bind to vlink!");
				return;
			}
			/* bind this ethBrick to this vlinkBrick */
			vlinkBind(*label, bindreply, datarecv);
		}
	}
	/* if a vlink was found, the bind will work -> other Bricks can send data over it
	 * -> publish ethBrick in the KVR
	 */
	
	msg = anaL1_allocateMessage();
	/* register callback function for higher layer bricks*/
	highLayIDP = anaL0_registerCallback((anaCallback_t)&  recvFromHL , NULL, NULL, NULL, 0, IDP_PERM, 0);
	if (highLayIDP == 0) {
		anaPerror(ANA_ERR, "Brick :Unable to register Callback IDP to ANA minmex: -> cannot receive messages from higher level bricks");
		free(msg);
		return;
	}
	/* publish callback function for higher layer bricks*/
	if (getAuxArg(0)==NULL)
		publ_name ="eth01+interface+wired+ethernet";
	else {
		publ_name = (char *) malloc(50);
		strcat(publ_name, getAuxArg(0));
		strcat(publ_name, "+interface+wired+ethernet");
	}
	memset(&service, 0, sizeof(struct service_s));
	memset(&context , 0, sizeof(struct context_s));
	
	context.value = ".";
	context.valueLen = 2;
	
	service.value = publ_name;
	service.valueLen = strlen(publ_name)+1;
	
	lenMsg = anaL1_encCompartmentPublish(msg, highLayIDP, (struct context_s *)&context, (struct service_s *) &service,  1);
	if(lenMsg < 0){
		anaPrint(ANA_ERR, "problem encoding publish request \n");
		free(msg);
		return;
	}
	
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t)&  handlePublishReply ,NULL, NULL, NULL, 0); 

	if(ret < 0){
		anaPrint(ANA_ERR, "problem Sending publish request: cannot publish \"eth0\" in KVR \n");
		free(msg);
		return;

	}
	free(msg);
}
 void  checkResolveReply(char *ID) {

	struct resolveRequestList *request = NULL;
	struct resolveRequestList *predecessor = NULL;
	int found=0, totalSize=0;
	xrpMsg_t replyMsg=NULL;
	/* we check if the request ID is still in the queue */
	request = requestList;
	while(!found && request != NULL){
		if(!memcmp(ID, request->resId, RES_ID_SIZE))
			found=1;
		else {
			predecessor = request;
			request = request->next;
		}
	}

	/*if not found means that the request was answered*/
	if(!found)
		return;
	
	replyMsg = anaL1_allocateMessage();
	
	/* we send an error message to the querier */
	totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
     	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
     	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
    					
	anaL0_send(request->replyTo, replyMsg, totalSize);
	free(replyMsg);
	
	/* Now we free the request and unqueue it */
	if(predecessor==NULL)
		requestList = request;
	else
		predecessor->next = request->next;

	free(request->context);
	free(request->target);
	free(request->description);
	free(request);
}


 void  updateResolveReply(char *ID, struct ether_header *ethHdr , void *response, int responseLen){

	struct resolveRequestList *request = NULL;
	struct resolveRequestList *predecessor = NULL;
	int found=0;
	int  i;
	struct ether_header tmphed;
	struct fullEthHeader fh;
	anaLabel_t dataIDP;
	uint16_t *header;
	xrpMsg_t replyMsg=NULL;
	struct IDPinfo_s info;
	
	/* we check if the request ID is still in the queue */
	request = requestList;
	while(!found && request != NULL){
		if(!memcmp(ID, request->resId, RES_ID_SIZE))
			found=1;
		else {
			predecessor = request;
			request = request->next;
		}
	}
	
	/*if not found means that the response comes too late (timer expired)*/
	if(!found)
		return;
	
	/* Now we stop the timer */
	anatimer_del(request->timerId);
	
	replyMsg =anaL1_allocateMessage();
	
	for(i=0;i<6;i++)
			tmphed.ether_dhost[i] = ethHdr->ether_shost[i];
	
	for(i=0;i<6;i++)
			tmphed.ether_shost[i]= myAddress[i];

	header = (uint16_t *)response;
	if(*header != 0 ){

		tmphed.ether_type = ntohs(*header);
	
		fh.destDescrip= NULL;
		if(request->description != NULL){
			fh.srcDescrip= (void *) malloc(request->descLen) ;
			memcpy(fh.srcDescrip, request->description, request->descLen );
			fh.srcDescLen = request->descLen;
		}
		else
			fh.srcDescrip=NULL;
			
		fh.ethhdr = &tmphed;	

		/* we create an IDP mapped to sendtoVlink with the right aux argument*/
		dataIDP = anaL0_registerCallback((anaCallback_t) &  sendToVlink, &fh, (copyAuxFunction)cloneFullHeader, (freeAuxFunction) freeFullHeader, 0, 0, 0);
		
		if(dataIDP == 0){
			int totalSize;
			
          		anaPerror(ANA_ERR, "updateResolveReply: unable to register and IDP for resolve response");	
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
            		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
           						
			anaL0_send(request->replyTo, replyMsg, totalSize);
            		free(replyMsg);
		}
		else{ /* we send a successfull response */
			int totalSize;
			int ret = 0;	
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_RESOLVE);
            		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_LABEL, &dataIDP, sizeof(anaLabel_t));
           		
			if((ret = anaL0_send(request->replyTo, replyMsg, totalSize)) < 0 )
				anaPrint(ANA_ERR, "updateResolveReply: failed to send response to querier brick: label: %i, error: %i\n", request->replyTo, ret);	
            		
			free(replyMsg);
		
			/* we set the information about the IDP */
			memset(&info, 0, sizeof(struct IDPinfo_s));
	
			info.label = dataIDP;
			info.ICflag = 1;
			info.destContext.value = ethHdr->ether_shost;
			info.destContext.valueLen = 6;
	
			info.destService.value = request->target;
			info.destService.valueLen = request->targetLen;
	
			info.MTU = vlinkMTU - 300; /* 300 is just an estimation(hack) for context and description lengths */	
			info.chanType = 'u'; 
			info.binding.brick.outputIDP = vlinkLabel;
	
			ret =  anaL0_setIDPinfo( &info);
			if(ret < 0)
				anaPrint(ANA_DEBUG, " handleResolve : setIDPinfo failed \n");
			}
	}
	else{
		int totalSize;
		anaPrint(ANA_ERR, "updateResolveReply: No next header field in response\n");
      		
		totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
        	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
        	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
        					
		anaL0_send(request->replyTo, replyMsg, totalSize);
        	free(replyMsg);
	}
	
	if(predecessor==NULL)
		requestList = request->next;
	else
		predecessor->next = request->next;

	free(request->context);
	free(request->target);
	free(request->description);
	free(request);
}

int  sendResolveRequest(unsigned char *context, void *target, int targetLen, void *desc, int descLen,
char chanType, anaLabel_t replyTo){

	struct fullEthHeader fullHeader;
	struct ether_header tmphed;

	struct resolveRequestList *resRequest = NULL;	
	int timerId=-1;
	int i;

	for(i=0;i<6;i++)
		tmphed.ether_dhost[i] = context[i];	
	
	for(i=0;i<6;i++)
		tmphed.ether_shost[i] = myAddress[i];

	tmphed.ether_type = ETH_CMD;
	fullHeader.opCode =ETH_RES;
	fullHeader.ethhdr = &tmphed;
	fullHeader.destDescrip = target;
	fullHeader.srcDescrip = desc;
	fullHeader.srcDescLen = descLen;
	fullHeader.destDescLen = targetLen; 

	resRequest = (struct resolveRequestList *) malloc(sizeof(struct resolveRequestList));
	for(i=0;i<RES_ID_SIZE;i++)
		resRequest->resId[i] = (int)random()%256;


	resRequest->replyTo= replyTo;
	resRequest->context = (char *) malloc(6);
	memcpy(resRequest->context, context, 6);
	resRequest->target = (char *) malloc(targetLen);
	memcpy(resRequest->target, target, targetLen);
	
	resRequest->targetLen = targetLen;
	resRequest->contextLen = 6;
	
	if(desc!=NULL){
		resRequest->description = (void*) malloc(descLen);
		memcpy(resRequest->description, desc, descLen);
		resRequest->descLen = descLen;
	}

	resRequest->chanType = chanType;
	resRequest->next = requestList;
	requestList = resRequest;	
	
	timerId = anatimer_add(RESOLVE_TIMEOUT, 0, 0, (fct_t) &  checkResolveReply, resRequest->resId, RES_ID_SIZE);

	if (timerId == -1)
                anaPrint(ANA_ERR, "Problem adding timer\n");
	
	resRequest->timerId = timerId;	
	
	sendToVlink(resRequest->resId, RES_ID_SIZE, 0, &fullHeader); 
	return 0;
}

 void handleResolve(void *msg, int len) {

	int targetLen, contextLen,labelSize, descLen, ret, dummy,  i = 0;
        char *target;
	unsigned char *context;
	char *description = NULL;
        anaLabel_t replyTo;
	anaLabel_t result;
	
	anaLabel_t *lp = NULL;

	struct fullEthHeader *fh;
	struct ether_header *tmphed;
	char *chanType=NULL;	
	struct QREPKeyObjectList *res=NULL; 
	struct IDPinfo_s info;
	
	lp =  (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &labelSize,0);
        if(lp == NULL ) { 
                anaPrint(ANA_ERR, "handleResolve : No label to reply to \n");
                return;
        }
	
	replyTo = *lp;

   	target = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_SRV, &targetLen,0);
	context = (unsigned char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_CXT, &contextLen,0);
	description = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_SRC_SRV, &descLen,0);
	chanType = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_CHANTYPE, &dummy,0);
	
	
	if(target==NULL || context ==NULL )
		return; 

	if(context[0]=='.'){ 
		ret = QREPfindByQuery(&dataRep, target, &res);
		if(ret < 0){
			fillXRPCommand((xrpMsg_t)msg, XRP_CMD_ERROR);
                        anaL0_send(replyTo, msg, len);
			return ;		
		}
		else{
			/* IN CASE MULTIPLE RESPONSES ADD MULTICAST FUNCTION HERE */
			if(res == NULL) {
				anaPrint(ANA_DEBUG, "couldnt find entry for %s\n", target);
				return;
			}
			anaPrint(ANA_DEBUG, "hello ");
			anaPrint(ANA_DEBUG, "--- %i ", *(anaLabel_t *)res->entry->value);
			anaPrint(ANA_DEBUG, "world\n");			
			*lp = *(anaLabel_t *)res->entry->value; 
			anaL0_send(replyTo, msg, len);					
			
			QREPfreeQueryRes(res);	
			return;	
		}
	}		
	tmphed = (struct ether_header *)malloc(sizeof(struct ether_header));
	for(i=0;i<6;i++)
		tmphed->ether_shost[i]= myAddress[i];
	
	if(chanType[0]=='m' || chanType[0]=='b'){ /* in this case we use deliver to send the message*/

	/* we prepare the header*/
		if(context[0]=='*')
			for(i=0;i<6;i++)
				tmphed->ether_dhost[i] = broadcastAddress[i];
		else
			for(i=0;i<6;i++)
				tmphed->ether_dhost[i] = context[i];
			
		tmphed->ether_type = ETH_CMD;
		fh = (struct fullEthHeader *)malloc(sizeof(struct fullEthHeader));
		fh->opCode = ETH_DLV;
		fh->destDescrip= target;
		fh->destDescLen =targetLen;
		fh->srcDescrip=description;
		fh->srcDescLen=descLen;
		fh->ethhdr = tmphed;
	
		/* we register callback mapped to the sendToVlink function */ 
		result = anaL0_registerCallback((anaCallback_t) &  sendToVlink,
		fh, (copyAuxFunction) cloneFullHeader, (freeAuxFunction)
		freeFullHeader, 0, 0, 0);
		
		if(result == 0){
			anaPerror(ANA_ERR, "handleResolve: unable to create an IDP");
			fillXRPCommand((xrpMsg_t)msg, XRP_CMD_ERROR);
                       	anaL0_send(replyTo, msg, len);
			return;
		}

		/* otherwise we send a successfull response */			
		*lp = result;
		/* Now we send the response */
		anaL0_send(replyTo, msg, len);
		
		/* we set the information about the IDP */
		memset(&info, 0, sizeof(struct IDPinfo_s));
	
		info.label = result;
		info.ICflag = 1;
		info.destContext.value = "*";
		info.destContext.valueLen = 2;
	
		info.destService.value = target;
		info.destService.valueLen = targetLen;
	
		info.MTU = vlinkMTU - 300; /* 300 is just an estimation(hack) for context and description lengths */	
		info.chanType = chanType[0];
		info.binding.brick.outputIDP = vlinkLabel;
	
		ret =  anaL0_setIDPinfo( &info);
		if(ret < 0)
			anaPrint(ANA_DEBUG, " handleResolve : setIDPinfo failed \n");
			
		return;				
	}
			
	else{ /* the chanType is either unicast or anycast =>  we send a request */

		if(context[0]=='*')
			ret = sendResolveRequest(broadcastAddress, target, targetLen,description, descLen, chanType[0], replyTo);
		else
			ret = sendResolveRequest(context, target, targetLen, description, descLen, chanType[0], replyTo);
		if (ret<0) {
			anaPrint(ANA_ERR, "handleResolve:unable to send Resolve request\n");
			fillXRPCommand((xrpMsg_t)msg, XRP_CMD_ERROR);
                       	anaL0_send(replyTo, msg, len);
		}	
	 
		return;
	}
	
	return;
}



/* this function is called from the higher level Brick (e.g. dummyBrick) to 
 * publish its description and its IDP in the QREP of this Brick (e.g. "IP" 123)
 * a confirmation might be sent back if there is an answerIDP supplied
 * the aux field holds the repository to which the IDP should be added: dataRep or unsolRep
 */
 void  handlePublish(void *msg, int len) {
	anaLabel_t IDP = 0;
        anaLabel_t replyTo = 0;
	anaLabel_t *lp = NULL;
		
        int labelLen, dummy;
     	char *description = NULL;
        int descLen;
        struct QREPEntry *entry;
        char *keyWord;
	struct QREP * rep = (struct QREP *)&dataRep;
	uint16_t nextHeader;
	anaPrint(ANA_ERR, "handlePublish\n");
        lp =  (anaLabel_t *) getXRPArg( (xrpMsg_t) msg, XRP_CLASS_LABEL, &labelLen, 0);
 	
	if(lp!=NULL)
		IDP = *lp;
	
	description =  (char*) getXRPArg( (xrpMsg_t) msg, XRP_CLASS_SRC_SRV, &descLen, 0);
        	
	if(description ==NULL || IDP ==0){
                anaPrint(ANA_ERR, "handlePublish: unable to identify name or IDP\n");
                return;
        }

	anaPrint(ANA_DEBUG, "publish service %s\n", description);

        /* the replyTo label is optional, we look for an argument with class label
 	 * at the second occurence => the first label is the one associated to the entry
 	 */
	 
	lp = (anaLabel_t *) getXRPArg( (xrpMsg_t) msg, XRP_CLASS_LABEL, &dummy,1);
  	
	if(lp != NULL)
		replyTo = *lp;
	anaPrint(ANA_ERR, "replyTo = %i\n", replyTo);
	/* we generate a unique next header field for this service (that will be stored as its name in QREP)*/
	nextHeader = generateNextHeader();
	
	keyWord = strsep(&description, "+");

	entry = allocateQREPEntry((char *)&nextHeader, sizeof(uint16_t) ,&IDP, sizeof(anaLabel_t));

	if((keyWord==NULL) || entry==NULL){
	        /*we send an error message, if a replyTo label was specified*/
                if(lp!=NULL){
    			fillXRPCommand((xrpMsg_t)msg, XRP_CMD_ERROR);
                        anaL0_send(replyTo, msg, len);
                }
                return;
        }
	
	anaPrint(ANA_ERR, "handlePublish: IDP %i, keyWord %s\n", IDP, keyWord);
	
	while(keyWord != NULL){
		QREPaddKeywordToEntry(entry, keyWord);
                keyWord = strsep(&description, "+");
        }
	
        if(addQREPEntry(rep, entry)<0){
                if(lp!=NULL){
                        fillXRPCommand((xrpMsg_t)msg, XRP_CMD_ERROR);
			anaL0_send(replyTo, msg, len);
		}
 		freeQREPEntry(rep, entry, freeValue);

        }
 	else{ /*everything went well*/
		if(lp!=NULL){ /*as a confirmation we send the same received message*/
                        anaL0_send(replyTo, msg, len);
                }
        }

}

/* this function is called upon an unpublish request, either for the dataRep or for the 
 * unsolicited bootstrapping rep. Looks for an entry with the name given in data and frees it.
 */
void  handleUnpublish(char * data, int len){


	char * query;
	anaLabel_t label=0;
	anaLabel_t replyTo=0;
	
	anaLabel_t *lp;
	
	int lenQuery, lenLabel, dummy;
	struct QREPKeyObjectList * res =NULL;
	struct QREPKeyObjectList * tmp =NULL;
	
	query =  (char *) getXRPArg(data, XRP_CLASS_SRC_SRV, &lenQuery,0);
	
	lp = (anaLabel_t *) getXRPArg(data, XRP_CLASS_LABEL, &lenLabel,0);
	if(lp != NULL)
		label = *lp;
	
	lp =  (anaLabel_t *) getXRPArg(data, XRP_CLASS_LABEL, &dummy,1);
	if(lp != NULL)
		replyTo = *lp;
	
	anaPrint(ANA_DEBUG, "Unpublish description %s\n", query);
	
	if(label==0)
		anaPrint(ANA_DEBUG, "No label\n");
	
	if(replyTo==0)
		anaPrint(ANA_DEBUG, "No replyTo\n");
	
	if(QREPfindByQuery(&dataRep, query, &res) < 0){
		anaPrint(ANA_ERR, "Unpublish : Could not find an entry matching the target description\n");
		if(replyTo!=0){
			/* we send an error message */
			fillXRPCommand (data, XRP_CMD_ERROR);
			anaL0_send(replyTo, data, len);
		}		
		return;
	}
	
	tmp = res;
	while(tmp!=NULL){
		if(label == *((anaLabel_t *)(tmp->entry->value))  ){
			if(freeQREPEntry(&dataRep, tmp->entry, freeValue) == 0 ){
				if(replyTo !=0){	
					/* we send a confirmation message : same one received*/
					anaL0_send(replyTo, data, len);
				}
			}
			else{
				if(replyTo != 0){	
					/* we send an error message : same one received but with XRP_CMD_ERROR*/
					fillXRPCommand (data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
				}
			}
			
			QREPfreeQueryRes(res);
			return;
			
		}
		tmp = tmp->next;
	}
	
	/*if we arrive here, means no entry with same IDP + descript*/	
	anaPrint(ANA_ERR, "Unpublish : Could not find an entry matching the target description and label\n");
	QREPfreeQueryRes(res);
	
	if(replyTo != 0){	
		/* we send an error message : same one received but with XRP_CMD_ERROR*/
		fillXRPCommand (data, XRP_CMD_ERROR);
		anaL0_send(replyTo, data, len);
	}
	
	return;
}


/* callback which will be used from higher layer protocols
 * forwards message to handler function corresponding to the XRP_CMD specified */

void anaL2_msg(struct anaL2_message *msg) {

	if (!msg || !msg->data)
		return;

	recvFromHL(msg->data, msg->dataLen, msg->idp, NULL);
}
 
void recvFromHL(char *data, int len, anaLabel_t input, void *aux) {

	/*parse the received xrp message */
	xrpCmd_t command = data;
	
	/* resolve request */
        if(equalXRPCommands(command, XRP_CMD_RESOLVE)) {
		handleResolve(data, len);
	}
	
	/* publish request */
        if(equalXRPCommands(command, XRP_CMD_PUBLISH )){
		handlePublish(data, len);
	}
		
	/* unpublish  request */
	if(equalXRPCommands(command, XRP_CMD_UNPUBLISH)){
		anaPrint(ANA_NOTICE, "received unpublish request\n");
		handleUnpublish(data, len);
	}

}


/* callback function which will be used upon receiving a messag from the lower layer (vlink)
 * for a BP_REQUEST it checks the QREP and sends an answer to the vlink
 * answer consists of data IDP, along with all key words
 * handles unsolicited information
 * forwards ANA_IDP messages to the corresponding IDP
 */
void recvFromLL(char *data, int len, anaLabel_t input, void *aux){


	struct ether_header *ethhdr = NULL;
	int offset;
	uint16_t destDescLen, srcDescLen;
	int dataLen, i;
	struct QREPKeyObjectList * res =NULL;
	xrpMsg_t msg = NULL;
	int msgLen; 
	void * srcDesc=NULL;
	
	struct service_s serviceStr;
	struct context_s contextStr;
	anaLabel_t *tmpL = NULL;
		
	if(data==NULL ||  len < (int) sizeof(struct ether_header))
		return;

	ethhdr = (struct ether_header *)data;
	anaPrint(ANA_NOTICE, "\n\n-->received from Vlink %i bytes\n\n", len);
	
	/* First we check if we accept this datagram (i.e myAddress in dstAddr or a broadcast one) */
	if( (memcmp(ethhdr->ether_dhost, myAddress, ETH_ALEN))  && memcmp(ethhdr->ether_dhost, broadcastAddress, ETH_ALEN)){
		anaPrint(ANA_DEBUG, "received message not for us\n");
		return; /*not destined to us*/ 
	}

	switch(ntohs(ethhdr->ether_type)) {
	case ETH_CMD:{
		uint16_t *opCodeP = (uint16_t *) (data + sizeof(struct ether_header)); 
		uint16_t opCode;
		anaPrint(ANA_NOTICE, "-->received Command\n");
		offset = sizeof(struct ether_header) + sizeof(uint16_t); /* +2 is for the opCode */
		opCode = ntohs(*opCodeP);
		if(opCode == ETH_DLV){
			uint16_t *tmpLen;  
			void *destDesc=NULL;
			anaPrint(ANA_NOTICE, "-->received DELIVER\n");
			
			tmpLen = (uint16_t *) (data+offset);
			destDescLen = ntohs(*tmpLen); /* *tmp;*/
			destDesc = data +offset + sizeof(uint16_t);
			
			anaPrint(ANA_DEBUG, "--> deliver destDEscLen =%d value in buffer %d , offset %d\n",
			destDescLen, *tmpLen, offset);	
			anaPrint(ANA_DEBUG, "--> deliver destDesc =%s \n", (char *) destDesc);	
			offset += sizeof(uint16_t); /* this is for the dest Description Length*/ 
			
			/*destDescLen = strlen( data+offset)+1;*/
			
			/*lookup the target description in the QREP*/
			if(QREPfindByQuery(&dataRep, data+offset, &res) < 0)
				anaPrint(ANA_ERR, "Could not find an entry matching the target description\n");
		
			if(res==NULL)
				return; /*no matching target*/
	
			
			offset+=destDescLen;
	
			tmpLen =(uint16_t *)(data+offset);
			srcDescLen = ntohs(*tmpLen); /* *tmp; */
			
			offset += sizeof(uint16_t); /* this is for the dest Description Length*/ 
			
			/*srcDescLen = strlen(data+offset)+1;*/

			srcDesc = data+offset;
			offset+=srcDescLen;
	
			anaPrint(ANA_NOTICE, "-->srcDescLen = %d  srcDesc=%s\n", srcDescLen, (char*)srcDesc);
			
			/* Now we need to allocate a buffer that wil contain the data received in the message + the cooked ID at the end of it 
			The format id the following : [dataLen][data][cookedID]	
			*/
	
			dataLen = len - offset; /* length of the data contained in the received message */
	
			msg = anaL1_allocateMessage();
			
			memset(&contextStr, 0, sizeof(struct context_s));
			memset(&serviceStr, 0, sizeof(struct service_s));
			
			contextStr.value = ethhdr->ether_shost;
			contextStr.valueLen = ETH_ALEN;
			
			serviceStr.value = srcDesc;
			serviceStr.valueLen = srcDescLen;
			
			msgLen=anaL1_encData(msg, data+offset, dataLen, &contextStr, &serviceStr, 0);
			if(msgLen<0){
				anaPrint(ANA_ERR, "recvFromLL: unable to encode data message (deliver case)\n");
				free(msg);
				return;
			}
		
	
			/* now we send the message to the target bricks : IMPORTANT / TO CHANGE ? : if many bricks fit this description, they would receive a copy as well !! */	
			if(res == NULL)
				anaPrint(ANA_DEBUG, "cannot forward\n");
			while(res != NULL){
				tmpL = (anaLabel_t *)res->entry->value;
			//	anaPrint(ANA_DEBUG, "forwarding data to %i\n" *tmpL);
				anaL0_send(*tmpL, msg, msgLen);
				res = res->next;
			}
			
			QREPfreeQueryRes(res);
			free(msg);

		}
		if(opCode == ETH_RES){ /*commande resolve*/
			char *resId=NULL;
			uint16_t *tmp;
			
			tmp =(uint16_t *) (data+offset);
			destDescLen = ntohs(*tmp);
			
			anaPrint(ANA_DEBUG, "---> RESOLVE REQUEST : size of target descr %d \n",destDescLen );	
			offset += sizeof(uint16_t); /* this is for the dest Description Length*/ 
			
			/*destDescLen = strlen(data+offset)+1;*/
			
			anaPrint(ANA_DEBUG, "-->received remote RESOLVE request for %s\n", data+offset); 
			/*lookup the target description in the QREP*/
			if(QREPfindByQuery(&dataRep, data+offset, &res) < 0)
				anaPrint(ANA_ERR, "Could not find an entry matching the target description \n");
			
			offset+=destDescLen;
		
			tmp =(uint16_t *) (data+offset);
			srcDescLen = ntohs(*tmp);
			

			anaPrint(ANA_DEBUG, "---> RESOLVE REQUEST : size of source descr %d \n",srcDescLen );
			offset += sizeof(uint16_t); /* this is for the src Description Length*/ 

			
			/*srcDescLen = strlen(data+offset)+1;*/
			if(srcDescLen)
				srcDesc = data+offset;
			else
				srcDesc = NULL;
			offset+=srcDescLen;


			/* we need to find the resolve queryidentifier */
			resId  = data + offset;
		
			if(res==NULL){	
				/* we don't respond otherwise it would be easy to flood network :)*/
				anaPrint(ANA_DEBUG, "res == NULL\n");
				return;
			}
			else{
				struct fullEthHeader fh;
				struct ether_header tmphed;
				struct QREPKeyObjectList * tmp =NULL;
				int nbRes=0;
				void *response = NULL;
				int offset = 0;
				uint16_t *tmpP = NULL;
				uint16_t tmpI;
				/* first we check how many matches we have */
				tmp = res;
				while(tmp!=NULL){
					nbRes++;
					tmp=tmp->next;
				}
				
				/* then we allocate a data buffer to hold the request ID + nbRes*headers */
				response = (void *) malloc((sizeof(uint16_t)*nbRes)+RES_ID_SIZE);
				memcpy(response, resId, RES_ID_SIZE);
				offset += RES_ID_SIZE;
			
				while(res!=NULL){
					anaPrint(ANA_DEBUG, "res != NULL\n");
					tmpP = (uint16_t *) (res->entry->name);
					tmpI = htons(*tmpP);
					//anaPrint(ANA_DEBUG, "%i %i\n", *tmpP, tmpI);	
					tmpP = (uint16_t *)((char *)response + offset);
					*tmpP = tmpI;
				
				//	anaPrint(ANA_DEBUG, "%i\n", tmpI
					memcpy((char *)response+offset, res->entry->name, sizeof(uint16_t));
					offset += sizeof(uint16_t);
					res = res->next; 
				}
			
				QREPfreeQueryRes(res);
			
				for(i=0;i<6;i++)
					tmphed.ether_dhost[i] = ethhdr->ether_shost[i];
	
				for(i=0;i<6;i++)
					tmphed.ether_shost[i]= myAddress[i];
						
				tmphed.ether_type = ETH_CMD;
				
				fh.opCode = ETH_RSP;
				fh.destDescrip= NULL;
				fh.srcDescrip=NULL;
				fh.ethhdr = &tmphed;


				sendToVlink((char *)response, offset, 0, &fh);

				anaPrint(ANA_DEBUG, "sent reply for resolve\n");
				free(response);				
			}
		
		}
		if(opCode  == ETH_RSP){ /*resolve response*/
			char *resId;
			int responseLen;
			anaPrint(ANA_NOTICE, "-->received RESOLVE response \n");
			offset+=2*sizeof(u_int16_t); /* 4 is for the 2 description length fileds*/
			
			/*lookup the target description in the QREP*/
			resId  = data + offset;
			offset += RES_ID_SIZE;
			responseLen = len - offset;
		
			updateResolveReply(resId, ethhdr, data+offset, responseLen); 
		
		}
		break;
	}	
	default:{ /* here we use the next header fields*/
		struct QREPEntry *entry=NULL;
		uint16_t *tmp;
		uint16_t tmpI;
		int tmpI2;
		anaPrint(ANA_ERR, "forward message to next header\n");	
		offset = sizeof(struct ether_header);
		tmp =(uint16_t *)(data+offset);
		
		srcDescLen = ntohs(*tmp);
		offset+= sizeof(uint16_t);
		
		srcDesc =data+offset;
		/*srcDescLen = strlen(srcDesc)+1;*/
		
		offset += srcDescLen;
		tmpI = ntohs(ethhdr->ether_type);
		tmpI2 = ethhdr->ether_type;

		entry = getQREPEntryByName(&dataRep, (char*)&tmpI2 ,sizeof(uint16_t));	

	
		if(entry==NULL){
			anaPrint(ANA_DEBUG, "entry == NULL\n");
			return; /*no matching target*/
		}
		dataLen = len - offset; /* length of the data contained in the received message */
	
		msg = anaL1_allocateMessage();
		msgLen = 0;
		
		memset(&contextStr, 0, sizeof(struct context_s));
		memset(&serviceStr, 0, sizeof(struct service_s));
			
		contextStr.value = ethhdr->ether_shost;
		contextStr.valueLen = ETH_ALEN;
			
		if(srcDescLen>0) {
			serviceStr.value = srcDesc;
			serviceStr.valueLen = srcDescLen;
		}
				
		
		if(srcDescLen>0)
			msgLen=anaL1_encData(msg,  data+offset, dataLen, &contextStr, &serviceStr, 0);
		else
			msgLen=anaL1_encData(msg, data+offset, dataLen, &contextStr, NULL, 0);
	
		if(msgLen<0){
			anaPrint(ANA_ERR, "recvFromLL: unable to encode data message (next header case)\n");
			free(msg);
			return;
		}
		
		tmpL = (anaLabel_t *) entry->value;
		anaPrint(ANA_ERR, "forward message to next header IDP = %i, msg %s, msgLen %i\n", *tmpL, msg, msgLen);	
		
		if(anaL0_send(*tmpL, msg, msgLen) < 0 )
			anaPrint(ANA_ERR, "recvFromLL: unable to forward received data message to IDP:%d\n", *tmpL);

		free(msg);
	}
	
	}
	
}

/* send a data message to the vlink */

 void  sendToVlink(char *data, int len, anaLabel_t input, struct fullEthHeader *fullHeader){

	char * msg=NULL;
	int  offset = 0;
	int newLen=0;

	uint16_t *tmp=NULL;
	if(data==NULL || fullHeader ==NULL || fullHeader->ethhdr==NULL)
		return;

	/* first we calculate the length of the message to send */
	newLen = len + sizeof(struct ether_header);
	
	if(fullHeader->ethhdr->ether_type == ETH_CMD){
		newLen += sizeof(uint16_t); /* 2 is the size of the opCode */
		newLen += 2*sizeof(uint16_t); /* 4 is for 2 times the length fileds (2 bytes each) of target and description */
		if(fullHeader->destDescrip!=NULL)
			newLen += fullHeader->destDescLen; 
			
		if(fullHeader->srcDescrip!=NULL)
			newLen += fullHeader->srcDescLen;
			
	}
	else{
		newLen += sizeof(uint16_t); /* 2 is the size of the length field for the source description  */
		if(fullHeader->srcDescrip!=NULL)
			newLen += fullHeader->srcDescLen;
	}
	
	
	if(newLen > ETH_MTU){
		anaPrint(ANA_ERR, "Message cannot be sent to Vlink, length bigger than MTU");
		return;
	}
		
	msg = (char *) malloc(newLen);

	
	memcpy(msg+offset, fullHeader->ethhdr, sizeof(struct ether_header));
	tmp = (uint16_t *) (msg + 2 * ETH_ALEN);
	*tmp = htons(fullHeader->ethhdr->ether_type);
	offset += sizeof(struct ether_header);
	
	if(fullHeader->ethhdr->ether_type == ETH_CMD){
		uint16_t *opCodeP = (uint16_t *)(msg+offset);
		*opCodeP = htons(fullHeader->opCode);
		offset += sizeof(uint16_t);	
		
		if(fullHeader->destDescrip!=NULL){
			/*we store the length of the description field*/
			tmp = (uint16_t *)(msg+offset);
			*tmp = htons(fullHeader->destDescLen);/* (u_int16_t) fullHeader->destDescLen;*/
			anaPrint(ANA_DEBUG, "sendToVlink cmd put into buffer %d as destDescLen at offset %d\n", *tmp, offset);
			offset += sizeof(uint16_t); /* 2 is the length of the description field */		
			memcpy(msg+offset, fullHeader->destDescrip, fullHeader->destDescLen);
			offset += fullHeader->destDescLen;
		}
		else{	/*we store the length of the description field*/
			tmp = (uint16_t *) (msg+offset);
			*tmp= htons(0);
			anaPrint(ANA_DEBUG,"sendToVlink cmd put into buffer %d as destDescLen \n", *tmp);	
			offset += sizeof(uint16_t); /* 2 is the length of the description field */	
		}
		
		if(fullHeader->srcDescrip!=NULL){
			/*we store the length of the description field*/
			tmp = (uint16_t *)(msg+offset);
			*tmp=htons(fullHeader->srcDescLen);/* (u_int16_t) fullHeader->srcDescLen;*/
			anaPrint(ANA_DEBUG,"sendToVlink cmd put into buffer %d as srcDescLen  at offset %d\n", *tmp, offset);	
			offset += sizeof(uint16_t); /* 2 is the length of the description field */		
			memcpy(msg+offset, fullHeader->srcDescrip, fullHeader->srcDescLen);
			offset += fullHeader->srcDescLen;
		}
		else{	/*we store the length of the description field*/
			tmp = (uint16_t *) (msg+offset);
			*tmp= htons(0);
			anaPrint(ANA_DEBUG,"sendToVlink cmd put into buffer %d as srcDescLen \n", *tmp);	
			offset += sizeof(uint16_t); /* 2 is the length of the description field */	
		}
	}
	else{
		if(fullHeader->srcDescrip!=NULL){
			/*we store the length of the description field*/
			tmp = (uint16_t *) (msg+offset);
			*tmp= htons(fullHeader->srcDescLen);/* (u_int16_t) fullHeader->srcDescLen;*/
			offset += sizeof(uint16_t); /* 2 is the length of the description field */		
			anaPrint(ANA_DEBUG,"sendToVlink NOT cmd put into buffer %d as srcDescLen  at offset %d\n", *tmp, offset);	
			memcpy(msg+offset, fullHeader->srcDescrip, fullHeader->srcDescLen);
			offset += fullHeader->srcDescLen;
		}
		else{	/*we store the length of the description field*/
			tmp = (uint16_t *) (msg+offset);
			*tmp= htons(0);
			anaPrint(ANA_DEBUG,"sendToVlink NOT cmd put into buffer %d as srcDescLen \n", *tmp);	
			offset += sizeof(u_int16_t); /* 2 is the length of the description field */	
		}
	}
	
	memcpy(msg+offset, data, len);
	offset +=len;	

	anaL0_send(vlinkLabel, msg, newLen);
	anaPrint(ANA_NOTICE, "data message sent to vlink\n");
	free(msg);
}


int  brick_start() {
	int ret = 0;
	int lenMsg;
	int i;
	xrpMsg_t msg;
	
	struct service_s service;
	struct context_s context;

	requestList =NULL;
	anaPrint(ANA_NOTICE, "aux arg 1 = %s 2:%s 3:%s\n", getAuxArg(0), getAuxArg(1), getAuxArg(2));
	
	for(i=0;i<6;i++)
		broadcastAddress[i] = 0xFF;
	
	/*dummy intialisation until we bind to a vlink*/
	myAddress[0]=0x04;
	myAddress[1]=0x05;
	myAddress[2]=0x06;
	myAddress[3]=0x07;
	myAddress[4]=0x08;
	myAddress[5]=0x09;
	
	/* create the repository for lookup requests */
	initQRepository(&dataRep, QREP_ENTRYTAB_DEFAULTMASK, QREP_KEYTAB_DEFAULTMASK);

	msg = anaL1_allocateMessage();

	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	service.value = "vlink";
	service.valueLen = 6;
	
	context.value =  ".";
	context.valueLen = 2;

	/* lookup the vlink brick */
	lenMsg = anaL1_encResolve(msg, &context, &service, 'u', NULL);
        if (lenMsg<0) {
                anaPrint(ANA_EMERG, "Brick : Unable to format Lookup request \n");
		free(msg);
                return -1;
        }

        ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t)&  handleLookupReply ,NULL, NULL, NULL, 0);
		
        if (ret<0) {
                anaPrint(ANA_EMERG, "Brick : Lookup requestReply  Failed \n");
		free(msg);
                return -1;
        }

       	free(msg);

	/* added by Christophe for publication in compartment finder */

	anaL2_initDefault();
	service.value = "cfinder";
	service.valueLen = strlen(service.value)+1;	
	cfinder = anaL2_resolve(NODE_LABEL, &context, &service, 'u', NULL, NULL);
	if (!cfinder)
		anaPrint(ANA_ERR, "cannot find cfinder brick in KVR\n");
	else
	{
		service.value = ETH_REGEXP;
		service.valueLen = strlen(service.value)+1;
		regexp = anaL2_publish(cfinder, &context, &service, anaL2_msg, NOTHREAD, NULL);
		
		if (!regexp)
			anaPrint(ANA_ERR, "cannot publish ETH regexp in cfinder\n");
		else
			anaPrint(ANA_DEBUG, "published ETH regexp in cfinder\n");
	}
		
	return 1;
}


/* free function which has to be provided for the QREP
 * to free all the entries
 */
static void freeValue(void *value){
	free(value);
}
