/** \file
 * FBSD header file with necessary data structures, defines (documented in the report) and functions.
 */

#include "quickRepository.h"
#include "anaCommon.h"

#define XRP_CMD_FIELD "fld"
#define XRP_CMD_PUBLISHUP "pup"
#define XRP_CMD_ROUTEUP "rup"
#define XRP_CMD_SUBSCRIBE "sub"
#define XRP_CMD_RESPONSE "rsp"
#define XRP_CMD_CHANGE "chn"
#define XRP_CMD_DISCOVER "dis"

#define XRP_CLASS_DESCRIP "dsc"
#define XRP_CLASS_SERVICETYPE "sty"
#define XRP_CLASS_SERVICEID "sid"
#define XRP_CLASS_MESSAGEID "mid"
#define XRP_CLASS_SEQNR "snr"
#define XRP_CLASS_FIELDTYPE "fty"
#define XRP_CLASS_PARAM1 "pm1"
#define XRP_CLASS_PARAM2 "pm2"
#define XRP_CLASS_TTL "ttl"
#define XRP_CLASS_POTENTIAL "pot"
#define XRP_CLASS_REQUESTER "rqr"

#define PUBLISH_TTL 15				/* time to life for publish messages */
#define SUBSCRIBE_TTL 15			/* time to life for subscribe messages */
#define SERVICE_UPDATE_TIME 90		/* seconds between service updates */
#define ROUTE_UPDATE_TIME 90		/* seconds between route updates */
#define GARBAGE_UPDATE_TIME 90		/* seconds between garbage updates */
#define GARBAGE_LIFES 2			/* survival count */

#define MAX_INTERFACES 10

//! struct services
/*! A data structure to store the properties of node local service instances. */
struct services {
	char *type;			/*!< Service type */
	char *id;				/*!< Service ID */
	int seq;				/*!< Ongoing sequence number for advertisements */
	char *field;			/*!< Field type */
	int param1;				/*!< Parameter 1, e.g. capacity */
	int param2;				/*!< Parameter 2, e.g. hop count */
	int ttl;				/*!< Time to live for advertisements */
	int age;				/*!< Age for garbage collection issues */
	struct services *next;
};

//! struct ids
/*! A data structure to store service IDs and according sequence numbers of all service instances. */
struct ids {
	char *type;			/*!< Service type */
	char *id;				/*!< Service ID */
	int seq;				/*!< Ongoing sequence number */
	int param1;				/*!< Parameter 1, e.g. capacity */
	int param2;				/*!< Parameter 2, e.g. hop count */
	int age;				/*!< Age for garbage collection issues */
	struct ids *next;
};

//! struct pots
/*! A data structure to store field potentials of all service instances. */
struct pots {
	char *type;			/*!< Service type */
	char *field;			/*!< Field type */
	int potential;			/*!< Field potential */
	int age;				/*!< Age for garbage collection issues */
	struct pots *next;
};

//! struct tab
/*! A data structure to store the routing table. */
struct tab {
	char *type;			/*!< Service type */
	char *potential;			/*!< Field potential */
	int own;				/*!< Indicates type of next hop (local or remote node) */
	anaLabel_t label;			/*!< Next hop */
	int age;				/*!< Age for garbage collection issues */
	struct tab *next;
};

//! struct neighbour
/*! A data structure to store all neighbour nodes. */
struct fbr_neighbour {
	char *id;				/*!< Unique ID for neighbour, e.g. MAC address */
	anaLabel_t label;			/*!< Outgoing interface */
	int age;				/*!< Age for garbage collection issues */
	struct fbr_neighbour *next;
};

//! struct mids
/*! A data structure to store the message IDs of all messages. */
struct mids {
	char *type;			/*!< Service type */
	char *id;				/*!< Message ID */
	int cown;				/*!< Client application is node local*/
	int sown;				/*!< Service instance is node local */
	anaLabel_t clabel;		/*!< Next IDP towards client application*/
	anaLabel_t slabel;		/*!< Next IDP towards service instance*/
	char *requester;			/*!< Origin requester */
	struct mids *next;
};

//! A function to generate a random number between a lower and upper bound.
    /*!
      \param a Lower bound
      \param e Upper bound
      \return Random number
    */
int randomsnr(int a, int e) {
	int rand = random();
	return a + rand % (e - a);
}

/* default values for the repository */
#define QREP_KEYTAB_DEFAULTMASK 10
#define QREP_ENTRYTAB_DEFAULTMASK 10

