/* C/bricks/fbr/fbr_diss.c
 *  
 * dissemination for field-based publish / subscribe
 * depends on: dgrmEthBrick
 * runs in user space
 * KVR entries: fbr_diss
 * 
 * IMPORTANT: use minmex with -b option to extend default message size
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"
#include "anaLib2.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_diss";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvNodeIDP;
anaLabel_t recvEthIDP;
anaLabel_t potfLabel;
anaLabel_t rtabLabel;
anaLabel_t ethLabel;

void recvNodeData(char *data, int len, anaLabel_t input, void *aux);
void recvEthData(char *data, int len, anaLabel_t input, void *aux);

struct ids *sid;
struct ids *sid_test;
struct ids *sid_start;
struct fbr_neighbour *nb;
struct fbr_neighbour *nb_test;
struct fbr_neighbour *nb_start;


int garbage = GARBAGE_LIFES;

/* called upon leaving fbr_diss */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_DISS HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex! = NULL\n");
	
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_diss";
	service.valueLen = strlen("fbr_diss")+1;

	len = anaL1_encCompartmentUnpublish(msg, recvNodeIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in KVR!\n");
	}
	free(msg);
	msg = anaL1_allocateMessage();
	len = anaL1_encCompartmentUnpublish(msg, recvEthIDP, &context, &service);
	ret = anaL0_send(ethLabel, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in ethernet brick!\n");
	}
	free(msg);
	
	anatimer_exit();
	
	anaL0_unregisterCallback(recvNodeIDP, 0);
	anaL0_unregisterCallback(recvEthIDP, 0);
	
}

void sendToNbh(void *data, int len) {
	int j, sent = 0;
	char *eth, *jch;
	anaLabel_t tmpLabel, broadcastLabel;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 
	anaPrint(ANA_DEBUG, "sendTo Nbh\n");

	for (j=1; j<MAX_INTERFACES; j++) {
		eth = (char *) malloc(20);
		jch = (char *) malloc(10);
		strcat(eth, "eth0");
		sprintf(jch, "%i", j);
		strcat(eth, jch);
	
		context.value = ".";
		context.valueLen = 2;
		service.value = eth;
		service.valueLen = strlen(eth)+1;

		tmpLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if(tmpLabel == 0)
			anaPrint(ANA_DEBUG, "service %s not found\n", eth);
		if (tmpLabel != 0) {
			context.value = "*";
			context.valueLen = 2;
			service.value = "fbr_diss";
			service.valueLen = strlen("fbr_diss")+1;

			broadcastLabel = anaL2_resolve(tmpLabel, &context, &service, 'b', &description, &timeout);
			if (broadcastLabel != 0) {
				anaL0_send(broadcastLabel, data, len);
				sent = 1;
			}
			else
				anaPrint(ANA_DEBUG, "no neighbours\n");
		}
		free(eth);
		free(jch);
	}
	
	if (sent==0)
		anaPrint(ANA_EMERG, "no neighbours found!\n");
	else
		anaPrint(ANA_NOTICE, "sent to neighbours\n");
}

void localField(void *data, int len) {
	int typeLen, idLen, seqLen, seq, param2, param2Len, id_known = 0;
	char *type, *id, *seq_m, *param2_m;
	struct ids *sid_stored = NULL;
	anaPrint(ANA_DEBUG, "localField\n");	
	type =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	id =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICEID, &idLen, 0);
	seq_m =  (char*) getXRPArg((xrpMsg_t)data, XRP_CLASS_SEQNR, &seqLen, 0);
	param2_m =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_PARAM2, &param2Len, 0);
	sscanf(seq_m, "%i", &seq);
	sscanf(param2_m, "%i", &param2);
	
	while (sid_test->type!=NULL) {
		if (strcmp(sid_test->id, id)==0) {
			id_known = 1;
			sid_stored = sid_test;
		}
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	if (id_known) {
		memcpy(&sid_stored->seq, &seq, 4);
		memcpy(&sid_stored->param2, &param2, 4);
		memcpy(&sid_stored->age, &garbage, 4);
	} else {
		sid->next = (struct ids *) malloc(sizeof(struct ids));
		sid->type = (char*) malloc(strlen(type) + 1);
		memcpy(sid->type, type, strlen(type) + 1);
		sid->id = (char*) malloc(strlen(id) + 1);
		memcpy(sid->id, id, strlen(id) + 1);
		sid->seq = seq;
		sid->param1 = 0;
		sid->param2 = param2;
		sid->age = GARBAGE_LIFES;
		sid = sid->next;
		sid->type = NULL;
		sid->id = NULL;
		sid->seq = 0;
		sid->param1 = 0;
		sid->param2 = 0;
		sid->age = 0;
		sid->next = NULL;
	}
	
	sendToNbh(data, len);
}

void localRoute(void *data, int len) {
	sendToNbh(data, len);
}

void ethRoute(void *data, int len, char *sender) {
	anaLabel_t label = 0, tmpLabel;
	int arglen, neighbour_known = 0, j;
	char *eth, *jch;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	while (nb_test->id!=NULL) {
		if (strcmp(nb_test->id, sender)==0) {
			neighbour_known = 1;
			label = nb_test->label;
			memcpy(&nb_test->age, &garbage, 4);
		}
		nb_test = nb_test->next;
	}
	nb_test = nb_start;
	
	if (!neighbour_known) {
		for (j=1; j<MAX_INTERFACES; j++) {
			eth = (char *) malloc(20);
			jch = (char *) malloc(10);
			strcat(eth, "eth0");
			sprintf(jch, "%i", j);
			strcat(eth, jch);
			context.value = ".";
			context.valueLen = 2;
			service.value = eth;
			service.valueLen = strlen(eth) + 1;
			/* we are looking for our own Ethernet devices to ask them to resolve the fbr_forw of the sender */
			tmpLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
			if (tmpLabel != 0) {
				context.value = sender;
				context.valueLen = strlen(sender)+1;
				service.value = "fbr_forw";
				service.valueLen = strlen("fbr_forw") + 1;
				label = anaL2_resolve(tmpLabel,  &context, &service, 'u', &description, &timeout);
				if (label != 0) {
					break;
				}
			}
			free(eth);
			free(jch);
		}
		
		if (label == 0) {
			/* hm, we have received an updat from that sender, but we cannot resolve him
 			 * can happen, if we are in a wireless setting
 			 */
			anaPrint(ANA_EMERG, "fbr_diss: sender not found! \n");
			return;
		}
		
		nb->next = (struct fbr_neighbour *) malloc(sizeof(struct fbr_neighbour));
		nb->id = (char*) malloc(strlen(sender) + 1);
		memcpy(nb->id, sender, strlen(sender) + 1);
		nb->label = label;
		nb->age = GARBAGE_LIFES;
		nb = nb->next;
		nb->id = NULL;
		nb->label = 0;
		nb->age = 0;
		nb->next = NULL;
	}
	
	if (label != 0) {
		arglen = addXRPArg(data, XRP_CLASS_LABEL, &label, sizeof(anaLabel_t));
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "ethRoute: unable to add label to message!\n");
			return;
		}
		len += arglen;
	} else {
		anaPrint(ANA_EMERG, "could not create label to sender!\n");
		return;
	}
	
	if (rtabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_rtab";
		service.valueLen = strlen("fbr_rtab")+1;
		rtabLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (rtabLabel == 0) {
			anaPrint(ANA_EMERG, "routing table brick not found!\n");
			return;
		}
	}
	anaL0_send(rtabLabel, data, len);
	anaPrint(ANA_NOTICE, "route update sent to routing table brick\n");
}

void ethField(void *data, int len, char *sender) {
	int typeLen, idLen, seqLen, ttlLen, ttl, seq, newttl, fieldLen, param1Len, param2Len, id_known = 0, param2;
	char *type, *id, *seq_m, *ttl_m, *param1, *param2_m, *field, *newttl_m;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	struct ids *sid_stored = NULL;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 
	
	type =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	id =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICEID, &idLen, 0);
	seq_m =  (char*) getXRPArg((xrpMsg_t)data, XRP_CLASS_SEQNR, &seqLen, 0);
	field =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_FIELDTYPE, &fieldLen, 0);
	param1 =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_PARAM1, &param1Len, 0);
	param2_m =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_PARAM2, &param2Len, 0);
	ttl_m =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_TTL, &ttlLen, 0);
	sscanf(seq_m, "%i", &seq);
	sscanf(ttl_m, "%i", &ttl);
	sscanf(param2_m, "%i", &param2);
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "service-id: %s\n", id);
	anaPrint(ANA_NOTICE, "seq-nr: %s\n", seq_m);
	anaPrint(ANA_NOTICE, "param1: %s\n", param1);
	anaPrint(ANA_NOTICE, "param2: %s\n", param2_m);
	
	/* decrease TTL */
	if (ttl<=1) {
		anaPrint(ANA_EMERG, "TTL is over!\n");
		return; /* drop */
	}
	else {
		newttl = ttl - 1;
		newttl_m = (char *) malloc(5);
		sprintf(newttl_m, "%i", newttl);
		memcpy(data + xrpSpec->commandSize + sizeof(uint16_t) + ((xrpSpec->classSize + sizeof(uint16_t)) * 7) + strlen(type) + strlen(id) + strlen(seq_m) + strlen(field) + strlen(param1) + strlen(param2_m) + 6, newttl_m, strlen(newttl_m) + 1);
		anaPrint(ANA_NOTICE, "new TTL: %i\n", newttl);
	}
	
	param2 = param2 + 1;
	param2_m = (char *) malloc(5);
	sprintf(param2_m, "%i", param2);
	memcpy(data + xrpSpec->commandSize + sizeof(uint16_t) + ((xrpSpec->classSize + sizeof(uint16_t)) * 6) + strlen(type) + strlen(id) + strlen(seq_m) + strlen(field) + strlen(param1) + 5, param2_m, strlen(param2_m) + 1);
	anaPrint(ANA_NOTICE, "new param2: %i\n", param2);
	
	while (sid_test->type!=NULL) {
		if (strcmp(sid_test->id, id)==0) {
			id_known = 1;
			sid_stored = sid_test;
		}
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	if (id_known) {
		if (seq<sid_stored->seq || (seq==sid_stored->seq && param2>sid_stored->param2)) {
			anaPrint(ANA_NOTICE, "message not interesting\n");
		} else {
			memcpy(&sid_stored->seq, &seq, 4);
			memcpy(&sid_stored->param2, &param2, 4);
			memcpy(&sid_stored->age, &garbage, 4);
			
			if (potfLabel == 0) {
				context.value = ".";
				context.valueLen = 2;
				service.value = "fbr_potf";
				service.valueLen = strlen("fbr_potf")+1;
				potfLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
				if (potfLabel == 0) {
					anaPrint(ANA_EMERG, "field computing brick not found!\n");
					return;
				}
			}
			anaL0_send(potfLabel, data, len);
			anaPrint(ANA_NOTICE, "field message sent to field computing brick\n");
		
			sendToNbh(data, len);
		}
	} else {
		sid->next = (struct ids *) malloc(sizeof(struct ids));
		sid->type = (char*) malloc(strlen(type) + 1);
		memcpy(sid->type, type, strlen(type) + 1);
		sid->id = (char*) malloc(strlen(id) + 1);
		memcpy(sid->id, id, strlen(id) + 1);
		sid->seq = seq;
		sid->param1 = 0;
		sid->param2 = param2;
		sid->age = GARBAGE_LIFES;
		sid = sid->next;
		sid->type = NULL;
		sid->id = NULL;
		sid->seq = 0;
		sid->param1 = 0;
		sid->param2 = 0;
		sid->age = 0;
		sid->next = NULL;
		
		if (potfLabel == 0) {
			context.value = ".";
			context.valueLen = 2;
			service.value = "fbr_potf";
			service.valueLen = strlen("fbr_potf")+1;
			potfLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
			if (potfLabel == 0) {
				anaPrint(ANA_EMERG, "field computing brick not found!\n");
				return;
			}
		}
		anaL0_send(potfLabel, data, len);
		anaPrint(ANA_NOTICE, "field message sent to field computing brick\n");
		
		sendToNbh(data, len);
	}
}

/* receive data from node function */
void recvNodeData(char *data, int len, anaLabel_t input, void *aux) {
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* field message received*/
        if (equalXRPCommands(command, XRP_CMD_FIELD)) {
		anaPrint(ANA_NOTICE, "\n\n--> received field message from local node\n\n");
		localField(data, len);
	}
	
	/* route update received */
        if (equalXRPCommands(command, XRP_CMD_ROUTEUP)) {
		anaPrint(ANA_NOTICE, "\n\n--> received route update from local node\n\n");
		localRoute(data, len);
	}
}

/* receive data form ethernet function */
void recvEthData(char *data, int len, anaLabel_t input, void *aux) {
	int lenMsg, senderContextLen, senderTargetLen;
	void *message, *senderContext, *senderTarget;
	xrpCmd_t command; 
	lenMsg = anaL1_decResponse(data, XRP_CLASS_MESSAGE, 0, (void**) &message);
	senderContextLen = anaL1_decResponse(data, XRP_CLASS_SRC_CXT, 0, &senderContext);
       	senderTargetLen = anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, &senderTarget);
	
	/* parse the received xrp message */
	command = message;
	
	/* field message received*/
        if (equalXRPCommands(command, XRP_CMD_FIELD)) {
		anaPrint(ANA_NOTICE, "\n\n--> received field message from ethernet\n\n");
		ethField(message, lenMsg, (char *) senderContext);
	}
	
	/* route update received*/
        if (equalXRPCommands(command, XRP_CMD_ROUTEUP)) {
		anaPrint(ANA_NOTICE, "\n\n--> received route update from ethernet\n\n");
		ethRoute(message, lenMsg, (char *) senderContext);
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

void garbageUpdate(void *unused) {
	int garbage_new;
	struct ids *sid_back;
	struct fbr_neighbour *nb_back;
	
	sid_back = sid_start;
	while (sid_test->type!=NULL) {
		if (sid_test->age<=1) {
			if (sid_test==sid_start) {
				free(sid_test->type);
				free(sid_test->id);
				sid_start = sid_test->next;
			} else {
				free(sid_test->type);
				free(sid_test->id);
				if (sid_test->next==NULL) {
					sid_back->next = NULL;
				} else {
					sid_back->next = sid_test->next;
				}
			}
		} else {
			garbage_new = sid_test->age - 1;
			memcpy(&sid_test->age, &garbage_new, 4);
		}
		if (sid_test!=sid_start)
			sid_back = sid_back->next;
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	nb_back = nb_start;
	while (nb_test->id!=NULL) {
		if (nb_test->age<=1) {
			if (nb_test==nb_start) {
				free(nb_test->id);
				nb_start = nb_test->next;
			} else {
				free(nb_test->id);
				if (nb_test->next==NULL) {
					nb_back->next = NULL;
				} else {
					nb_back->next = nb_test->next;
				}
			}
		} else {
			garbage_new = nb_test->age - 1;
			memcpy(&nb_test->age, &garbage_new, 4);
		}
		if (nb_test!=nb_start)
			nb_back = nb_back->next;
		nb_test = nb_test->next;
	}
	nb_test = nb_start;
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 
	
	anaL2_initDefault();
	anatimer_init();
	
	potfLabel = 0;
	rtabLabel = 0;
	
	/* initializes ids list */
	sid = (struct ids *) malloc(sizeof(struct ids));
	sid_start = sid;
	sid_test = sid;
	sid->type = NULL;
	sid->id = NULL;
	sid->seq = 0;
	sid->param1 = 0;
	sid->param2 = 0;
	sid->age = 0;
	sid->next = NULL;
	
	/* initializes neighbour list */
	nb = (struct fbr_neighbour *) malloc(sizeof(struct fbr_neighbour));
	nb_start = nb;
	nb_test = nb;
	nb->id = NULL;
	nb->label = 0;
	nb->age = 0;
	nb->next = NULL;
	
	/* register data receive from ethernet callback function */
	recvEthIDP = anaL0_registerCallback((anaCallback_t) & recvEthData, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (recvEthIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register recvEthIDP!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: recvEthIDP registered\n");
	
	context.value = ".";
	context.valueLen = 2;
	service.value = "wired";
	service.valueLen = strlen("wired")+1;
	ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	anaPrint(ANA_ERR, "received label of eth01 %i\n", ethLabel); 

	/*context.value = ".";
	context.valueLen = 2;
	service.value = "eth01"; //"wired";
	service.valueLen = strlen("eth01");// strlen("wired")+1;
	ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	*/
	if (ethLabel == 0) {
		anaPrint(ANA_EMERG, "ethernet brick not found!\n");
		return -1;
	}
	
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_diss";
	service.valueLen = strlen("fbr_diss")+1;
	lenMsg = anaL1_encCompartmentPublish(msg, recvEthIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for recvEthData!\n");
		return -1;
	}
	ret =  anaL1_requestReply(ethLabel, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for recvEthData!\n");
		return -1;
	}
	free(msg);

	/* register data receive from node callback function */
	recvNodeIDP = anaL0_registerCallback((anaCallback_t) & recvNodeData, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (recvNodeIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register recvNodeIDP!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: recvNodeIDP registered\n");
	
	/* publish ip along with the data IDP in the KVR */
	msg = anaL1_allocateMessage();
	lenMsg = anaL1_encCompartmentPublish(msg, recvNodeIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for recvNodeIDP!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for recvNodeIDP!\n");
		return -1;
	}
	free(msg);
	
	anatimer_add((1000 * GARBAGE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)garbageUpdate, NULL, 0);
	
	return 1;
}


