/* C/bricks/fbr/fbr_potf.c
 *  
 * computes potential field for field-based publish / subscribe
 * runs in user space
 * KVR entries: fbr_potf
 * 
 * IMPORTANT: use minmex with -b option to extend default message size
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */
char mymodename[] = "fbr_potf";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvIDP;
anaLabel_t dissLabel;
anaLabel_t rtabLabel;

struct services *serv;
struct services *serv_test;
struct services *serv_start;
struct ids *sid;
struct ids *sid_test;
struct ids *sid_start;
struct pots *pot;
struct pots *pot_test;
struct pots *pot_start;
	
void recvData_potf(char *data, int len, anaLabel_t input, void *aux);

int garbage = GARBAGE_LIFES;

/* called upon leaving fbr_potf */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_POTF HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex! = NULL\n");
	
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_potf";
	service.valueLen = strlen("fbr_potf")+1;

	len = anaL1_encCompartmentUnpublish(msg, recvIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in KVR!\n");
	}
	free(msg);
	
	anatimer_exit();
	
	anaL0_unregisterCallback(recvIDP, 0);
}

void publishUpdate(void *data, int len) {
	char *descriptionStr, *type, *id = NULL, *capacity_m, *hopcount_m, *ttl_m, *seq_m;
	int totalLength, arglen, capacity, descLen, new_seq = 0;
	xrpMsg_t msg;
	struct services *serv_stored = NULL;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	descriptionStr =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_DESCRIP, &descLen, 0);
	
	type = strsep(&descriptionStr, "+");
	sscanf(descriptionStr, "%i", &capacity);
	
	while (serv_test->type!=NULL) {
		if (strcmp(serv_test->type, type)==0) {
			id = serv_test->id;
			serv_stored = serv_test;
			new_seq = serv_test->seq + 1;
			memcpy(&serv_test->seq, &new_seq, 4);
			memcpy(&serv_test->param1, &capacity, 4);
			memcpy(&serv_test->age, &garbage, 4);
		}
		serv_test = serv_test->next;
	}
	serv_test = serv_start;
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "service-id: %s\n", id);
	anaPrint(ANA_NOTICE, "capacity: %i\n", capacity);
	
	while (sid_test->type!=NULL) {
		if (strcmp(sid_test->id, id)==0) {
			memcpy(&sid_test->seq, &new_seq, 4);
			memcpy(&sid_test->age, &garbage, 4);
			memcpy(&sid_test->param1, &capacity, 4);
		}
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	while (pot_test->type!=NULL) {
		if (strcmp(pot_test->type, type)==0) {
			memcpy(&pot_test->age, &garbage, 4);
		}
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_FIELD);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_SERVICEID, id, strlen(id) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add service-id to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	seq_m = (char *) malloc(30);
	sprintf(seq_m, "%d", serv_stored->seq);
	arglen = addXRPArg(msg, XRP_CLASS_SEQNR, seq_m, strlen(seq_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add seq-nr to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_FIELDTYPE, serv_stored->field, strlen(serv_stored->field) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add field-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	capacity_m = (char *) malloc(20);
	sprintf(capacity_m, "%i", serv_stored->param1);
	arglen = addXRPArg(msg, XRP_CLASS_PARAM1, capacity_m, strlen(capacity_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add capacity to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	hopcount_m = (char *) malloc(5);
	sprintf(hopcount_m, "%i", serv_stored->param2);
	arglen = addXRPArg(msg, XRP_CLASS_PARAM2, hopcount_m, strlen(hopcount_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add hopcount to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	ttl_m = (char *) malloc(5);
	sprintf(ttl_m, "%i", serv_stored->ttl);
	arglen = addXRPArg(msg, XRP_CLASS_TTL, ttl_m, strlen(ttl_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add ttl to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (dissLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_diss";
		service.valueLen = strlen("fbr_diss")+1;
		dissLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
		if (dissLabel == 0) {
			anaPrint(ANA_EMERG, "dissemination brick not found!\n");
			return;
		}
	}
	anaL0_send(dissLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "publish update sent to dissemination brick\n");
	free(msg);
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, "infinite", strlen("infinite") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add potential to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_LABEL, "KVR", 4);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add label to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (rtabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_rtab";
		service.valueLen = strlen("fbr_rtab")+1;
		rtabLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
		if (rtabLabel == 0) {
			anaPrint(ANA_EMERG, "routing table brick not found!\n");
			return;
		}
	}
	anaL0_send(rtabLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "publish update sent to routing table brick\n");
	free(msg);
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, "infinite", strlen("infinite") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "publishUpdate: unable to add potential to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (dissLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_diss";
		service.valueLen = strlen("fbr_diss")+1;
		dissLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
		if (dissLabel == 0) {
			anaPrint(ANA_EMERG, "dissemination brick not found!\n");
			return;
		}
	}
	anaL0_send(dissLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "route update sent to dissemination brick\n");
	free(msg);
}

void fieldAssembly(void *data, int len) {
	char *descriptionStr, *type, *field, *id, *capacity_m, *hopcount_m, *ttl_m, *seq_m;
	int totalLength, arglen, capacity, hopcount, ttl, descLen, seq, service_known = 0, dummy;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	descriptionStr =  (char*) getXRPArg((xrpMsg_t) data, XRP_CLASS_SRC_SRV, &descLen, 0);
	anaPrint(ANA_DEBUG, "descriptionStr = %s\n", descriptionStr); 
	
	type = strsep(&descriptionStr, "+");
	sscanf(descriptionStr, "%i", &capacity);
	hopcount = 0;
	field = "electrostatic";
	id = (char *) malloc(7);
	sprintf(id, "%d", randomsnr(100000, 999999));
	seq = 100;
	ttl = PUBLISH_TTL;
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	
	while (serv_test->type!=NULL) {
		if (strcmp(serv_test->type, type)==0) {
			anaPrint(ANA_EMERG, "this service is already published!\n");
			return;
		}
		serv_test = serv_test->next;
	}
	serv_test = serv_start;
	
	anaPrint(ANA_NOTICE, "service-id: %s\n", id);
	anaPrint(ANA_NOTICE, "capacity: %i\n", capacity);
	anaPrint(ANA_NOTICE, "field for routing: %s\n", field);
	
	serv->next = (struct services *) malloc(sizeof(struct services));
	serv->type = (char*) malloc(strlen(type) + 1);
	memcpy(serv->type, type, strlen(type) + 1);
	serv->id = (char*) malloc(strlen(id) + 1);
	memcpy(serv->id, id, strlen(id) + 1);
	serv->seq = seq;
	serv->field = (char*) malloc(strlen(field) + 1);
	memcpy(serv->field, field, strlen(field) + 1);
	serv->param1 = capacity;
	serv->param2 = hopcount;
	serv->ttl = ttl;
	serv->age = GARBAGE_LIFES;
	serv = serv->next;
	serv->type = NULL;
	serv->id = NULL;
	serv->seq = 0;
	serv->field = NULL;
	serv->param1 = 0;
	serv->param2 = 0;
	serv->ttl = 0;
	serv->age = 0;
	serv->next = NULL;
	
	sid->next = (struct ids *) malloc(sizeof(struct ids));
	sid->type = (char*) malloc(strlen(type) + 1);
	memcpy(sid->type, type, strlen(type) + 1);
	sid->id = (char*) malloc(strlen(id) + 1);
	memcpy(sid->id, id, strlen(id) + 1);
	sid->seq = seq;
	sid->param1 = capacity;
	sid->param2 = hopcount;
	sid->age = GARBAGE_LIFES;
	sid = sid->next;
	sid->type = NULL;
	sid->id = NULL;
	sid->seq = 0;
	sid->age = 0;
	sid->next = NULL;
	
	while (pot_test->type!=NULL) {
		if (strcmp(pot_test->type, type)==0) {
			service_known = 1;
		//	pot_test->field = realloc(pot_test->field, strlen(field) + 1);
			free(pot_test->field);
			pot_test->field = malloc(strlen(field) + 1);
			memcpy(pot_test->field, field, strlen(field) + 1);
			dummy = 0;
			memcpy(&pot_test->potential, &dummy, 4);
			memcpy(&pot_test->age, &garbage, 4);
		}
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
	
	if (!service_known) {
		pot->next = (struct pots *) malloc(sizeof(struct pots));
		pot->type = (char*) malloc(strlen(type) + 1);
		memcpy(pot->type, type, strlen(type) + 1);
		pot->field = (char*) malloc(strlen(field) + 1);
		memcpy(pot->field, field, strlen(field) + 1);
		pot->potential = 0;
		pot->age = GARBAGE_LIFES;
		pot = pot->next;
		pot->type = NULL;
		pot->field = NULL;
		pot->potential = 0;
		pot->age = 0;
		pot->next = NULL;
	}
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_FIELD);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_SERVICEID, id, strlen(id) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add service-id to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	seq_m = (char *) malloc(30);
	sprintf(seq_m, "%d", seq);
	arglen = addXRPArg(msg, XRP_CLASS_SEQNR, seq_m, strlen(seq_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add seq-nr to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_FIELDTYPE, field, strlen(field) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add field-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	capacity_m = (char *) malloc(20);
	sprintf(capacity_m, "%i", capacity);
	arglen = addXRPArg(msg, XRP_CLASS_PARAM1, capacity_m, strlen(capacity_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add capacity to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	hopcount_m = (char *) malloc(5);
	sprintf(hopcount_m, "%i", hopcount);
	arglen = addXRPArg(msg, XRP_CLASS_PARAM2, hopcount_m, strlen(hopcount_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add hopcount to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	ttl_m = (char *) malloc(5);
	sprintf(ttl_m, "%i", ttl);
	arglen = addXRPArg(msg, XRP_CLASS_TTL, ttl_m, strlen(ttl_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add ttl to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (dissLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_diss";
		service.valueLen = strlen("fbr_diss")+1;
		dissLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (dissLabel == 0) {
			anaPrint(ANA_EMERG, "dissemination brick not found!\n");
			return;
		}
	}
	anaL0_send(dissLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "publish sent to dissemination brick\n");
	free(msg);
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, "infinite", strlen("infinite") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add potential to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_LABEL, "KVR", 4);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add label to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (rtabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_rtab";
		service.valueLen = strlen("fbr_rtab")+1;
		rtabLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (rtabLabel == 0) {
			anaPrint(ANA_EMERG, "routing table brick not found!\n");
			return;
		}
	}
	anaL0_send(rtabLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "publish sent to routing table brick\n");
	free(msg);
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, "infinite", strlen("infinite") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "fieldAssembly: unable to add potential to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (dissLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_diss";
		service.valueLen = strlen("fbr_diss")+1;
		dissLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (dissLabel == 0) {
			anaPrint(ANA_EMERG, "dissemination brick not found!\n");
			return;
		}
	}
	anaL0_send(dissLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "route update sent to dissemination brick\n");
	free(msg);
}

void handleUnpublish(char *type) {
	int totalLength, arglen;
	xrpMsg_t msg;
	struct services *serv_back;
	struct ids *sid_back;
	struct pots *pot_back;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	serv_back = serv_start;
	while (serv_test->type!=NULL) {
		if (strcmp(serv_test->type, type)==0) {
			if (serv_test==serv_start) {
				free(serv_test->type);
				free(serv_test->id);
				free(serv_test->field);
				serv_start = serv_test->next;
			} else {
				free(serv_test->type);
				free(serv_test->id);
				free(serv_test->field);
				if (serv_test->next==NULL) {
					serv_back->next = NULL;
				} else {
					serv_back->next = serv_test->next;
				}
			}
		}
		if (serv_test!=serv_start)
			serv_back = serv_back->next;
		serv_test = serv_test->next;
	}
	serv_test = serv_start;
	
	sid_back = sid_start;
	while (sid_test->type!=NULL) {
		if (strcmp(sid_test->type, type)==0) {
			if (sid_test==sid_start) {
				free(sid_test->type);
				free(sid_test->id);
				sid_start = sid_test->next;
			} else {
				free(sid_test->type);
				free(sid_test->id);
				if (sid_test->next==NULL) {
					sid_back->next = NULL;
				} else {
					sid_back->next = sid_test->next;
				}
			}
		}
		if (sid_test!=sid_start)
			sid_back = sid_back->next;
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	pot_back = pot_start;
	while (pot_test->type!=NULL) {
		if (strcmp(pot_test->type, type)==0) {
			if (pot_test==pot_start) {
				free(pot_test->type);
				free(pot_test->field);
				pot_start = pot_test->next;
			} else {
				free(pot_test->type);
				free(pot_test->field);
				if (pot_test->next==NULL) {
					pot_back->next = NULL;
				} else {
					pot_back->next = pot_test->next;
				}
			}
		}
		if (pot_test!=pot_start)
			pot_back = pot_back->next;
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "handleUnpublish: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "handleUnpublish: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, "zero", strlen("zero") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "handleUnpublish: unable to add potential to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_LABEL, "KVR", 4);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "handleUnpublish: unable to add label to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	if (rtabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_rtab";
		service.valueLen = strlen("fbr_rtab")+1;
		rtabLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (rtabLabel == 0) {
			anaPrint(ANA_EMERG, "routing table brick not found!\n");
			return;
		}
	}
	anaL0_send(rtabLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "unpublish sent to routing table brick\n");
	free(msg);
}

void fieldUpdate(void *data, int len) {
	char *type, *field, *id, *capacity_m, *hopcount_m, *ttl_m, *seq_m, *potential_m;
	int capacity, hopcount, ttl, seq, typeLen, idLen, seqLen, fieldLen, param1Len, param2Len, ttlLen, id_known = 0, seq_stored = 0, totalLength, arglen, routeup = 0, serv_known = 0;
	int potential = 0, pot_old, pot_sub, pot_add;
	struct pots *pot_stored = NULL, *pot_back;
	struct ids *sid_stored = NULL;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 

	
	type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICEID, &idLen, 0);
	seq_m =  (char *) getXRPArg((xrpMsg_t)data, XRP_CLASS_SEQNR, &seqLen, 0);
	field =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_FIELDTYPE, &fieldLen, 0);
	capacity_m =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_PARAM1, &param1Len, 0);
	hopcount_m =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_PARAM2, &param2Len, 0);
	ttl_m =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_TTL, &ttlLen, 0);
	sscanf(seq_m, "%i", &seq);
	sscanf(capacity_m, "%i", &capacity);
	sscanf(hopcount_m, "%i", &hopcount);
	sscanf(ttl_m, "%i", &ttl);
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "field for routing: %s\n", field);
	
	while (sid_test->type!=NULL) {
		if (strcmp(sid_test->id, id)==0) {
			id_known = 1;
			sid_stored = sid_test;
			seq_stored = sid_test->seq;
			memcpy(&sid_test->age, &garbage, 4);
		}
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	while (pot_test->type!=NULL) {
		if (strcmp(pot_test->type, type)==0) {
			pot_stored = pot_test;
			memcpy(&pot_test->age, &garbage, 4);
		}
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
	
	while (serv_test->type!=NULL) {
		if (strcmp(serv_test->type, type)==0) {
			serv_known = 1;
		}
		serv_test = serv_test->next;
	}
	serv_test = serv_start;
	
	if (pot_stored!=NULL && pot_stored->potential==0 && serv_known==0) {
		pot_back = pot_start;
		while (pot_test->type!=NULL) {
			if (strcmp(pot_test->type, type)==0) {
				if (pot_test==pot_start) {
					free(pot_test->type);
					free(pot_test->field);
					pot_start = pot_test->next;
				} else {
					free(pot_test->type);
					free(pot_test->field);
					if (pot_test->next==NULL) {
						pot_back->next = NULL;
					} else {
						pot_back->next = pot_test->next;
					}
				}
			}
			if (pot_test!=pot_start)
				pot_back = pot_back->next;
			pot_test = pot_test->next;
		}
		pot_test = pot_start;
	}
	
	if (id_known &&  !(seq<seq_stored)) {
		if (pot_stored->potential!=0) {
			if (strcmp(field, "electrostatic")==0) {
				pot_old =  pot_stored->potential;
				pot_sub =  (sid_stored->param1 / sid_stored->param2);
				pot_add =  (capacity / hopcount);
				potential = pot_old - pot_sub + pot_add;
				memcpy(&pot_stored->potential, &potential, 4);
				routeup = 1;
			}
		} else {
			potential = 0;
		}
		memcpy(&sid_stored->seq, &seq, 4);
		memcpy(&sid_stored->param1, &capacity, 4);
		memcpy(&sid_stored->param2, &hopcount, 4);
	} else {
		if (pot_stored!=NULL) {
			/* overwrite */
			if (pot_stored->potential!=0) {
				if (strcmp(field, "electrostatic")==0) {
					pot_old =  pot_stored->potential;
					pot_add =  (capacity / hopcount);
					potential = pot_old + pot_add;
					memcpy(&pot_stored->potential, &potential, 4);
					routeup = 1;
				}
			} else {
				potential = 0;
			}
		} else {
			/* new */
			pot->next = (struct pots *) malloc(sizeof(struct pots));
			pot->type = (char*) malloc(strlen(type) + 1);
			memcpy(pot->type, type, strlen(type) + 1);
			pot->field = (char*) malloc(strlen(field) + 1);
			memcpy(pot->field, field, strlen(field) + 1);
			if (strcmp(field, "electrostatic")==0) {
				potential =  (capacity / hopcount);
			}
			pot->potential = potential;
			pot->age = GARBAGE_LIFES;
			pot = pot->next;
			pot->type = NULL;
			pot->field = NULL;
			pot->potential = 0;
			pot->age = 0;
			pot->next = NULL;
			routeup = 1;
		}
		sid->next = (struct ids *) malloc(sizeof(struct ids));
		sid->type = (char*) malloc(strlen(type) + 1);
		memcpy(sid->type, type, strlen(type) + 1);
		sid->id = (char*) malloc(strlen(id) + 1);
		memcpy(sid->id, id, strlen(id) + 1);
		sid->seq = seq;
		sid->param1 = capacity;
		sid->param2 = hopcount;
		sid->age = GARBAGE_LIFES;
		sid = sid->next;
		sid->type = NULL;
		sid->id = NULL;
		sid->param1 = 0;
		sid->param2 = 0;
		sid->seq = 0;
		sid->age = 0;
		sid->next = NULL;
	}
	
	if (potential==0)
		anaPrint(ANA_NOTICE, "potential is already infinite\n");
	else
		anaPrint(ANA_NOTICE, "new potential: %i\n", potential);
	
	if (routeup) {
		msg = anaL1_allocateMessage();
		totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "fieldUpdate: unable to fill command in message!\n");
			free(msg);
			return;
		}  
		arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "fieldUpdate: unable to add service-type to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		if (potential==0)
			potential_m = "infinite";
		else {
			potential_m = (char *) malloc(20);
			sprintf(potential_m, "%i", potential);
		}
		arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, potential_m, strlen(potential_m) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "fieldUpdate: unable to add potential to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		if (dissLabel == 0) {
			context.value = ".";
			context.valueLen = 2;
			service.value = "fbr_diss";
			service.valueLen = strlen("fbr_diss")+1;
			dissLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
			if (dissLabel == 0) {
				anaPrint(ANA_EMERG, "dissemination brick not found!\n");
				return;
			}
		}
		anaL0_send(dissLabel, msg, totalLength);
		anaPrint(ANA_NOTICE, "potential sent to dissemination brick\n");
		free(msg);
	}
}

void updateRoute(void *unused) {
	xrpMsg_t msg;
	int totalLength, arglen;
	char *potential_m;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 

	while (pot_test->type!=NULL) {
		msg = anaL1_allocateMessage();
		totalLength = fillXRPCommand(msg, XRP_CMD_ROUTEUP);
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "updateRoute: unable to fill command in message!\n");
			free(msg);
			return;
		}  
		arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, pot_test->type, strlen(pot_test->type) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "updateRoute: unable to add service-type to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		if (pot_test->potential==0)
			potential_m = "infinite";
		else {
			potential_m = (char *) malloc(20);
			sprintf(potential_m, "%i", pot_test->potential);
		}
		arglen = addXRPArg(msg, XRP_CLASS_POTENTIAL, potential_m, strlen(potential_m) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "updateRoute: unable to add potential to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		if (dissLabel == 0) { 
			context.value = ".";
			context.valueLen = 2;
			service.value = "fbr_diss";
			service.valueLen = strlen("fbr_diss")+1;
			dissLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
			if (dissLabel == 0) {
				anaPrint(ANA_EMERG, "dissemination brick not found!\n");
				return;
			}
		}
		anaL0_send(dissLabel, msg, totalLength);
		anaPrint(ANA_NOTICE, "route update sent to dissemination brick\n");
		free(msg);
		
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
}

void garbageUpdate(void *unused) {
	int garbage_new;
	int pot_old, pot_sub, potential;
	struct ids *sid_back;
	struct pots *pot_back;
	struct services *serv_back;
	
	sid_back = sid_start;
	while (sid_test->type!=NULL) {
		if (sid_test->age<=1) {
			while (pot_test->type!=NULL) {
				if (strcmp(pot_test->type, sid_test->type)==0) {
					if (sid_test->param2!=0) {
						if (pot_test->potential!=0) {
							if (strcmp(pot_test->field, "electrostatic")==0) {
								pot_old =  pot_test->potential;
								pot_sub =  (sid_test->param1 / sid_test->param2);
								potential = pot_old - pot_sub;
								memcpy(&pot_test->potential, &potential, 4);
							}
						}
					}
				}
				pot_test = pot_test->next;
			}
			pot_test = pot_start;
			if (sid_test==sid_start) {
				free(sid_test->type);
				free(sid_test->id);
				sid_start = sid_test->next;
			} else {
				free(sid_test->type);
				free(sid_test->id);
				if (sid_test->next==NULL) {
					sid_back->next = NULL;
				} else {
					sid_back->next = sid_test->next;
				}
			}
		} else {
			garbage_new = sid_test->age - 1;
			memcpy(&sid_test->age, &garbage_new, 4);
		}
		if (sid_test!=sid_start)
			sid_back = sid_back->next;
		sid_test = sid_test->next;
	}
	sid_test = sid_start;
	
	serv_back = serv_start;
	while (serv_test->type!=NULL) {
		if (serv_test->age<=1) {
			if (serv_test==serv_start) {
				free(serv_test->type);
				free(serv_test->id);
				free(serv_test->field);
				serv_start = serv_test->next;
			} else {
				free(serv_test->type);
				free(serv_test->id);
				free(serv_test->field);
				if (serv_test->next==NULL) {
					serv_back->next = NULL;
				} else {
					serv_back->next = serv_test->next;
				}
			}
		} else {
			garbage_new = serv_test->age - 1;
			memcpy(&serv_test->age, &garbage_new, 4);
		}
		serv_test = serv_test->next;
	}
	serv_test = serv_start;
	
	pot_back = pot_start;
	while (pot_test->type!=NULL) {
		if (pot_test->age<=1) {
			anaPrint(ANA_NOTICE, "\n\n*** %s died ***\n\n", pot_test->type);
			if (pot_test==pot_start) {
				free(pot_test->type);
				free(pot_test->field);
				pot_start = pot_test->next;
			} else {
				free(pot_test->type);
				free(pot_test->field);
				if (pot_test->next==NULL) {
					pot_back->next = NULL;
				} else {
					pot_back->next = pot_test->next;
				}
			}
		} else {
			garbage_new = pot_test->age - 1;
			memcpy(&pot_test->age, &garbage_new, 4);
		}
		if (pot_test!=pot_start)
			pot_back = pot_back->next;
		pot_test = pot_test->next;
	}
	pot_test = pot_start;
}

/* receive data function */
void recvData_potf(char *data, int len, anaLabel_t input, void *aux) {
	char *type, *description;
	int descLen;
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* publish received */
        if (equalXRPCommands(command, XRP_CMD_PUBLISH)) {
		anaPrint(ANA_NOTICE, "\n\n--> received publish request\n\n");
		fieldAssembly(data, len);
	}
	
	/* publish update received */
        if (equalXRPCommands(command, XRP_CMD_PUBLISHUP)) {
		anaPrint(ANA_NOTICE, "\n\n--> received publish update\n\n");
		publishUpdate(data, len);
	}
	
	/* unpublish received */
	if (equalXRPCommands(command, XRP_CMD_UNPUBLISH)) {
		anaPrint(ANA_NOTICE, "\n\n--> received unpublish request\n\n");
		description =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SRC_SRV, &descLen, 0);
		type = strsep(&description, "+");
		anaPrint(ANA_NOTICE, "service-type: %s\n", type);
		handleUnpublish(type);
	}
	
	/* field message received */
	if (equalXRPCommands(command, XRP_CMD_FIELD)) {
		anaPrint(ANA_NOTICE, "\n\n--> received field message\n\n");
		fieldUpdate(data, len);
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	anaL2_initDefault();
	anatimer_init();
	
	dissLabel =  0;
	rtabLabel = 0;
	
	/* initializes services list */
	serv = (struct services *) malloc(sizeof(struct services));
	serv_start = serv;
	serv_test = serv;
	serv->type = NULL;
	serv->id = NULL;
	serv->seq = 0;
	serv->field = NULL;
	serv->param1 = 0;
	serv->param2 = 0;
	serv->ttl = 0;
	serv->age = 0;
	serv->next = NULL;
	
	/* initializes ids list */
	sid = (struct ids *) malloc(sizeof(struct ids));
	sid_start = sid;
	sid_test = sid;
	sid->type = NULL;
	sid->id = NULL;
	sid->seq = 0;
	sid->param1 = 0;
	sid->param2 = 0;
	sid->age = 0;
	sid->next = NULL;
	
	/* initializes pots list */
	pot = (struct pots *) malloc(sizeof(struct pots));
	pot_start = pot;
	pot_test = pot;
	pot->type = NULL;
	pot->field = NULL;
	pot->potential = 0;
	pot->age = 0;
	pot->next = NULL;
	
	/* register data receive callback function */
	recvIDP = anaL0_registerCallback((anaCallback_t) & recvData_potf, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (recvIDP<0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register recvIDP!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: recvIDP registered\n");
	
	/* publish ip along with the data IDP in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_potf";
	service.valueLen = strlen("fbr_potf")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for recvIDP!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for recvIDP!\n");
		return -1;
	}
	free(msg);

	anatimer_add((1000 * ROUTE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)updateRoute, NULL, 0);
	anatimer_add((1000 * GARBAGE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)garbageUpdate, NULL, 0);
	
	return 1;
}


