
#include "attachedTable.h"
#include "hashFunction.h"

/***************************************************************

Hash functions by Bob Jenkins, December 1996, Public Domain.

2005-11-09: tested by CJ with 128-bit keys to 10 bits values 
            --> hash function has a uniform distribution

***************************************************************/

ub4 hashATTT( k, length, initval, maskLength)
register ub1 *k;        /* the key */
register ub4  length;   /* the length of the key in bytes */
register ub4  initval;    /* the previous hash, or an arbitrary value */
register int  maskLength;    /* the previous hash, or an arbitrary value */
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */

	return (c & hashmask(maskLength));	 
}



int handle2hashATTT(struct attachedTable_s *table, anaHandle_t handle) {
	int h;
	if(table==NULL)
		return -1;
	h = hashATTT((ub1 *)&handle, sizeof(anaHandle_t), 0, table->hashMaskSize);
	return h;
}



int initializeATTT (struct attachedTable_s *table, int hashMaskSize) { 
	/* the allocated size of the table will be (2^hashMaskSize), the hashMaskSize will be also used in the hash  function */
	if(table==NULL || hashMaskSize<0)
		return -1;
	table->size = pow2(hashMaskSize);
	table->hashMaskSize =hashMaskSize;
	table->tab = (struct  attachedBrick_s**) malloc(table->size * sizeof(struct  attachedBrick_s*));
	if(table->tab==NULL){
		anaPrint(ANA_ERR, "initializetable : unable to allocate memory for attached bricks hash table\n");
		return -1;
	}
	anaPrint(ANA_DEBUG, "Attached brick's table initialized with size %u\n", table->size);
	return 0;
}


struct attachedBrick_s* gettableEntry(struct attachedTable_s *table, anaHandle_t h) {
	int i ;
	struct attachedBrick_s *entry=NULL;
	
	if(table==NULL)
		return NULL;
	i = handle2hashATTT(table, h);
	if (i<0 || i>(int)(table->size)-1) {
		anaPrint(ANA_ERR, "gettableEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	entry = table->tab[i];

	if (entry == NULL ) { 	/* meaning that there is no value mapped to this  label in the Hash table */
		anaPrint(ANA_NOTICE, "gettableEntry : entry in table at index %d  is NULL\n", i);
		return NULL;	
	}
			
	if  (h== entry->handle) { /* if the keys are equal */
		return entry; 
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
		 * the entry with the correct key */
		entry = entry->next; 
		while (entry != NULL) {
			if( h==entry->handle ) 
				return entry;
			else 
				entry = entry->next;
		} 
	}
	return NULL;
}



int freetable(struct attachedTable_s *table) { 
	int i;
	struct attachedBrick_s *entry=NULL;
	struct attachedBrick_s *next=NULL;
		
	if(table==NULL)
		return -1;
		
	for (i=0; i< (int)(table->size); i++) {
		entry = table->tab[i];
		while  (entry!=NULL) {
			next = entry->next;
			free(entry->dataSocket);
			free(entry->dataAux);
			free(entry->controlSocket);
			free(entry->controlAux);
			
			free(entry);
			entry = next;
		}
	}

	
	free(table->tab);
	free(table);
	return 0;
}

int freetableEntry(struct attachedTable_s *table, anaHandle_t h) { 
	int i ;
	struct attachedBrick_s *entry=NULL;
	struct attachedBrick_s *tmp=NULL;

	if(table==NULL)
		return -1;
	
	i = handle2hashATTT(table, h);
	if (i<0 || i>(int)(table->size)-1) {
		anaPrint(ANA_ERR, "freetableEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = table->tab[i];
	if (entry == NULL) /* meaning that there is no value mapped to this */
		return -1;	  /*  label in the Hash table */
			
	if (h==entry->handle) {
		table->tab[i] = entry->next; 
		
		free(entry->dataSocket);
		free(entry->dataAux);
		free(entry->controlSocket);
		free(entry->controlAux);
				
		free(entry);
		table->nbAttached--;
		return 0;
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
			  the entry with the correct key */
		while (entry != NULL ) {
			if(entry->next != NULL) {
				tmp = entry->next;
				if(h==entry->handle) {
					entry->next = tmp->next;
					
					free(entry->dataSocket);
					free(entry->dataAux);
					free(entry->controlSocket);
					free(entry->controlAux);
			
					free(tmp);
					table->nbAttached--;
					return 0;
				}	 
			}
			entry = entry->next;
		}
	}
	return -1;
}

int addtableEntry (struct attachedTable_s *table, struct attachedBrick_s *toAdd) {
	struct attachedBrick_s *entry=NULL;
	int i ;
	if(table==NULL || toAdd==NULL)
		return -1;
	
	i= handle2hashATTT(table, toAdd->handle);
	if (i<0 || i>(int)(table->size)-1) {
		anaPrint(ANA_ERR, "addtableEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	entry = table->tab[i];
	if(entry == NULL ){
		toAdd->next = NULL;
		table->tab[i] = toAdd;
	}
	else { /* hash collison */
		anaPrint(ANA_DEBUG, "addtableEntry : Hash collison\n");
		toAdd->next = table->tab[i];
		table->tab[i]=toAdd;
	}
	table->nbAttached++;
	return 0;
}
