/* C/bricks/examples/templates/brickAL0_template.c
 * 
 * This plugin is executed either as .so plugin in user space
 * or as a kernel module. It implements IPC connectivity between
 * bricks running in gates mode and the minmex.
 * For this it listens in Kernel space for generic netlink and udp messages,
 * in user space it listens for udp, pipes, and unix sockets.
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> Initial implementation
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> extension for kernel space
 * 	 
 * 	   	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */
/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"
#include "gatesPlug.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */
char mymodename[]="gatesPlug";

#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif

/*------------------------------- End default brick header code --------------------------*/

char relayDataBuff[DEFAULT_UDP_MTU];

/****************************************/
/*        helper functions              */
/****************************************/

int addDataGate(struct commGate_s *gate){
	struct commGate_s *tmp = defaultDataGate;
	if(gate==NULL)
		return -1;
		
	if (defaultDataGate == NULL) {
		defaultDataGate = gate;
		return 0;
	}	
	while(tmp->next!=NULL) 
		tmp = tmp->next;

	tmp->next = gate;
	return 0;
}


int addControlGate(struct commGate_s *gate) {
	struct commGate_s *tmp = defaultControlGate;
	if(gate==NULL)
		return -1;
	
	if (defaultControlGate == NULL) {
		defaultControlGate = gate;
		return 0;
	}
	while(tmp->next!=NULL) 
		tmp = tmp->next;

	tmp->next = gate;
	return 0;
}


void * copyData(void * data){
	char * new_data = (char *)malloc(strlen((const char *)data)+1);
	memcpy(new_data, (char *)data, strlen((const char *)data));
	new_data[strlen((const char *)data)]='\0';
	return (char *)new_data;
}

anaHandle_t* copyHandle(anaHandle_t *h){
	anaHandle_t *res = (anaHandle_t *) malloc(sizeof(anaHandle_t)); 
	*res = *h;
	return res;
}

void __attribute__((visibility("default"))) whenBrickKicked(anaHandle_t h);

/************************************/
/*       relay functions            */
/************************************/

int relayControl(anaHandle_t handle, void *msg, int len) {
	
	struct attachedBrick_s* entry = NULL ;
	int nbSent = -1;
	
	if (msg==NULL || len<0){
		return -1;
	}
			
	entry = gettableEntry(&attBricksTable, handle);
	if (entry == NULL) {
		anaPrint(ANA_ERR, "relayControl: no entry corresponding to handle \n");
		return -1;
	}
	switch(entry->controlMode) {
		/* UDP works for user and kernel space */
		case UDP:{
		
	        #ifdef USER
			int sock = entry->controlDescriptor;
		#endif
		#ifdef KERNEL
			struct socket *sock = entry->controlSocket;
		#endif
                                     
			nbSent = sendSocketWrapper(&sock, AF_INET, entry->controlAux, msg, len);
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayControl : Unable to send contol data through UDP socket\n");
            			return -1;
			}	 
			break;
            
		}
#ifdef USER
		case UNIX:{
		
			int sock = entry->controlDescriptor;
		
			nbSent = sendSocketWrapper(&sock, AF_LOCAL, entry->controlAux,msg, len);
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayControl : Unable to send contol data through UNIX socket\n");
            			return -1;
			}	 
			break;
		}
		case PIPES:{
			nbSent = write(entry->controlDescriptor, msg, len);
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayControl : Unable to send control data through Named PIPE\n");
                                return -1;
			}
                        break;
		}
#endif /*USER */
#ifdef KERNEL
                case GENETLINK:{
			nbSent = sendGenetlink(entry->controlDescriptor, msg, len, COMMAND_CONTROL);
			if (nbSent < 0) {
				anaPrint(ANA_ERR, "relayControl : Unable to send control data through Generic Netlink Sockets\n");
                                return -1;
			}
			break;
		}
#endif
	}
	return nbSent;
}



void relayData(char *data, int len, anaLabel_t input,  anaHandle_t *h) {
		
    	anaLabel_t *label;
	struct attachedBrick_s* entry = NULL ;
	int nbSent = -1;
	
	if (data==NULL || len<0){
		return ;
	}
	
	memset(relayDataBuff, 0, DEFAULT_UDP_MTU);
	label = (anaLabel_t *) relayDataBuff;
	*label = htonl(input);
	
	/* We should find a way better than copying the data just to format the message */
	memcpy(relayDataBuff+sizeof(anaLabel_t), data, len);
	
	
	entry = gettableEntry(&attBricksTable, *h);
	if (entry == NULL) {
		anaPrint(ANA_ERR, "relayData: no entry corresponding to handle \n");
		return;
	}
	switch(entry->dataMode) {

		case UDP:{
		#ifdef USER
			int sock = entry->dataDescriptor;
		#endif
		#ifdef KERNEL
			struct socket *sock = entry->dataSocket;
		#endif                                         
			nbSent = sendSocketWrapper(&sock, AF_INET,entry->dataAux, relayDataBuff, len+sizeof(anaLabel_t));
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayData : Unable to send  data through UDP socket\n");
            			return;
			}	 
			break;
            
		}
#ifdef USER
		case UNIX:{
		
			int sock = entry->dataDescriptor;
		
			nbSent = sendSocketWrapper(&sock, AF_LOCAL, entry->dataAux, relayDataBuff, len+sizeof(anaLabel_t));
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayData : Unable to send  data through UNIX socket\n");
            			return;
			}	 
			break;
		}
		case PIPES:{
			nbSent = write(entry->controlDescriptor, relayDataBuff, len+sizeof(anaLabel_t));
			if(nbSent < 0) {
				anaPrint(ANA_ERR, "relayDta: Unable to send  data through Named PIPE\n");
                                return;
			}
                        break;
		}
#endif /* USER */
#ifdef KERNEL
	  	 case GENETLINK:{
			nbSent = sendGenetlink(entry->dataDescriptor, relayDataBuff, len + sizeof(anaLabel_t), COMMAND_DATA);
			if (nbSent < 0) {
				anaPrint(ANA_ERR, "relayControl : Unable to send control data through Generic Netlink Sockets\n");
                                return;
			}
			break;
		}
#endif /*KERNEL*/
	}
	
}


void relayNotification(struct anaEventNotif_s *notif, int len, anaLabel_t input,  struct notifAuxGates *aux){
	
	/* we first get the notification label of the target brick */
	char *data = NULL;
	char *tmp = NULL;
	int dataLen;
	int offset = 0;
	
	uint32_t *ip;
	anaLabel_t *lp;
	anaHandle_t *hp;

	/* we serialize the notification structure */
		/* format [eventID(4bytes)][handler pointer][threadFlag][label][viewer][destLabel][destViewer][name if any]*/
	
	dataLen = sizeof(int)+2*sizeof(anaHandle_t)+2*sizeof(anaLabel_t)+sizeof(void *) + 1; /* +1 for the threadFlag*/
	if(notif->name)
		dataLen += strlen(notif->name)+1;
		
	data = (void *) malloc(dataLen);
	ip = (uint32_t *) data;
	*ip = htonl(notif->eventID);
	offset +=  sizeof(uint32_t);
	
	memcpy(data+offset, &(aux->handlerPointer), sizeof(void *));
	offset += sizeof(void *);
	
		
	if(aux->threadFlag){
		tmp = (char *) (data + offset);
		tmp[0] = 1; 
	}
	offset += 1;
	
	lp = (anaLabel_t *) (data + offset);
	*lp = htonl(notif->label);
	offset +=  sizeof(anaLabel_t);
	
	hp = (anaHandle_t *) (data + offset);
	*hp = htonl(notif->viewer);
	offset += sizeof(anaHandle_t);

	lp = (anaLabel_t *) (data + offset);
	*lp = htonl(notif->destLabel);
	offset +=  sizeof(anaLabel_t);

	hp = (anaHandle_t *) (data + offset);
	*hp = htonl(notif->destViewer);
	offset += sizeof(anaHandle_t);

	if(notif->name){
		tmp = (char *) (data+offset);
		strcpy(tmp, notif->name);
		offset += strlen(notif->name) + 1 ;
	}

	/* we send the structure to the brick on its notif label*/
	relayData(data, dataLen, aux->notifLabel,  &(aux->handle));
	
	/* we free the allocated data */
	free(data);
}


void whenBrickKicked(anaHandle_t h){
	
	/* we send a message to the brick */
	
	/* we free the brick's entry in our table */
	freetableEntry(&attBricksTable, h);
}

void handleSetNotifLabel(char *msg, int length){
	/*message format : ['n'][MY_HANDLE][notifLabel]*/
	
	anaLabel_t *lp;
	anaHandle_t *hp;
	struct attachedBrick_s* entry = NULL ;
	
	if(msg == NULL || length < (1+sizeof(anaHandle_t) + sizeof(anaLabel_t)))
		return;
	
	hp = (anaHandle_t *) (msg+1);
	lp = (anaLabel_t *) (msg+1+sizeof(anaHandle_t));

	entry = gettableEntry(&attBricksTable, ntohl(*hp));
	if(entry == NULL){
		anaPrint(ANA_DEBUG, "handleSetNotifLabel : no brick matching the received handle\n");
		return;
	}
	
	entry->notifLabel = ntohl(*lp);
}

void relayDetach(char *msg, int length) {
	
	anaHandle_t *handle = (anaHandle_t *)(msg+1);
	anaHandle_t val = (anaHandle_t) ntohl(*handle);
#ifdef USER	
	handleDetach(val, -255); /* the -255 values is important since it indicates to the minmex that the brick is a remote one
	 (the minmex will then not try to kill all the brick's threads ) */
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached)
		myminmex->detach(val, -255);
	analock_read_unlock(&attachLock);
#endif
	/* we remove the brick from our attached brick's table */
	
	freetableEntry(&attBricksTable, val);
	anaPrint(ANA_NOTICE, "Detached brick, current number of bricks: %d\n",attBricksTable.nbAttached);	
}

void relayRegister(char *msg, int len) {
	
	/*  request format : [r][owner handle][1 byte indication permanence][1 byte indicating if entry i public or private][viewer handle] */

	char m[CTRLMSG_MAXSIZE];
	int ret;
	
	anaLabel_t *label=NULL;
    	anaLabel_t result;
	
	
	anaHandle_t *owner=NULL;
	anaHandle_t ownerVal;
	
	anaHandle_t *viewer=NULL;
	anaHandle_t viewerVal;
		
	char *c=NULL;
	int permanent=0;
	if(msg==NULL || len<(sizeof(anaHandle_t) + 2)){
		anaPrint(ANA_ERR, "handleRegister: Wrong Register message\n");
		return;
	}	
	
	owner = (anaHandle_t *) (msg + 1);
	ownerVal = (anaHandle_t) ntohl(*owner);
	
	
	c = (char *)(msg +1+sizeof(anaHandle_t));
	if(c[0]=='p')
		permanent = 1;
	
	c = (char *)(msg + 2 + sizeof(anaHandle_t));
	if (c[0]){
		viewer = (anaHandle_t *)(msg + 3 + sizeof(anaHandle_t));
		viewerVal = ntohl(*viewer);
	}
	else
		viewerVal = 0;	

#ifdef USER	
	ret = handleRegister (ownerVal, (void *) relayData, &ownerVal, (copyAuxFunction) copyHandle, (freeAuxFunction) free, viewerVal, permanent, 0, &result);
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached) {
		ret = myminmex->regist(ownerVal, (void *) relayData, &ownerVal, (copyAuxFunction) copyHandle, (freeAuxFunction) free, viewerVal, permanent, 0, &result);
	} else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif
	if(ret <0){
		/*we return the error code :  [e][error_code]*/ 
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'e';
		m[1] = ret;	
		relayControl(ownerVal, m, 2);
	}
	else{
		/* format of response : [r][label] */
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'r';
		
		label = (anaLabel_t *) (m+1);
		*label = htonl(result);	
		relayControl(ownerVal, m, 1+sizeof(anaLabel_t));
	}
		
}

void relayRedirect(char *msg, int len) {
	
	/* format of message : [c][handle][label origin][viewer origin][label dest][viewer dest]*/
	

	char m[CTRLMSG_MAXSIZE];
	int ret;
	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	
	anaHandle_t querier;
	anaLabel_t originL;
	anaHandle_t originV;
	
	anaLabel_t destL;
	anaHandle_t destV;
		

	if(msg==NULL || len<(3*sizeof(anaHandle_t)) +(2*sizeof(anaLabel_t) +1)){
		anaPrint(ANA_ERR, "relayRedirect: Wrong message\n");
		return;
	}	
	
	hp = (anaHandle_t *) (msg + 1);
	querier = (anaHandle_t) ntohl(*hp);
	
	lp = (anaLabel_t *) (msg + 1+ sizeof(anaHandle_t));
	originL = (anaLabel_t) ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1+ sizeof(anaHandle_t) + sizeof(anaLabel_t));
	originV = (anaHandle_t) ntohl(*hp);
	
	lp = (anaLabel_t *) (msg + 1+ 2*sizeof(anaHandle_t)+ sizeof(anaLabel_t));
	destL = (anaLabel_t) ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1+ 2*sizeof(anaHandle_t) + 2*sizeof(anaLabel_t));
	destV = (anaHandle_t) ntohl(*hp);
	
	
#ifdef USER	
	ret = handleRedirect(querier, originV, originL,  destV, destL);
#endif
#ifdef KERNEL
	ret = myminmex->redirect(querier, originV, originL,  destV, destL);
#endif
	if(ret <0){
		/*we return the error code :  [e][error_code]*/ 
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'c';
		
		relayControl(querier, m, 1);
	}
		
}


void relayUnregister(char *msg, int len) {
	
	/*format of message : [u][label][viewer][myhandle] */
	

	char m[CTRLMSG_MAXSIZE];
	int ret;
	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	
	anaHandle_t viewer;
	anaHandle_t querier;
	anaLabel_t label;
		

	if(msg==NULL || len<( 1 + sizeof(anaHandle_t) + sizeof(anaLabel_t) )){
		anaPrint(ANA_ERR, "relayUnregister: Wrong message\n");
		return;
	}	
	
	lp = (anaLabel_t *) (msg + 1);
	label = (anaLabel_t) ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1+sizeof(anaLabel_t));
	viewer = (anaHandle_t) ntohl(*hp);
	
	hp = (anaHandle_t *) (msg + 1+sizeof(anaLabel_t)+sizeof(anaHandle_t));
	querier = (anaHandle_t) ntohl(*hp);
	
	
#ifdef USER		
	ret = handleUnregister(querier, viewer, label); 
#endif
#ifdef KERNEL	
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->unregist(querier, viewer, label);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
#endif

	if(ret <0){
		anaPerror(ANA_ERR, "handleUnregister failed err:");
		/*we return the error code :  [e][error_code]*/ 
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'u';
		
		relayControl(querier, m, 1);
	}
		
}


void relayEventSubscribe(char *msg, int len){
	/* message format : ['b'][MY_HANDLE][handler pointer][threadFlag(1 byte)][eventMask][label][viewer]([name])*/	
	struct notifAuxGates notifAuxG;
	anaLabel_t *lp = NULL;
	anaHandle_t *hp = NULL;
	char *sp = NULL;
	char *name = NULL;
	uint32_t *ip = NULL;
	struct attachedBrick_s *entry = NULL;
	int offset = 1; 
	uint32_t eventMask;
	int ret;
	struct anaEventSub_s sub;
	char m[2];
	
	memset(&notifAuxG, 0, sizeof(struct notifAuxGates)); 
	hp = (anaHandle_t *) (msg + offset);
	notifAuxG.handle = ntohl(*hp);
	offset += sizeof(anaHandle_t);

	/* we need to find the notification label for this brick */
	entry =  gettableEntry(&attBricksTable, notifAuxG.handle);
	if(entry == NULL){
		/* we do not answer the brick (that we do not know) */
		return;
	}
	
	notifAuxG.notifLabel = entry->notifLabel;
	
	/* we copy the value of the handler pointer */
	memcpy(&(notifAuxG.handlerPointer), msg+offset, sizeof(void *));
	offset += sizeof(void *);
	
	sp = (char *) (msg + offset);
	if(sp[0])
		notifAuxG.threadFlag = 1; 
	offset += 1;	
		
	ip = (uint32_t *) (msg + offset);
	eventMask = ntohl(*ip);
	offset += sizeof(uint32_t);
	
	lp = (anaLabel_t *) (msg + offset);
	offset += sizeof(anaLabel_t);	
	
	hp = (anaHandle_t *) (msg + offset);
	offset += sizeof(anaHandle_t);
	
	if(len > offset){
		name = msg + offset; 
		offset += strlen(name) + 1;
	}
	
	/* we fill the subscription structure */
	memset(&sub, 0, sizeof(struct anaEventSub_s));
	sub.eventMask = eventMask;
	sub.label = ntohl(*lp);
	sub.viewer = ntohl(*hp);
	if(name)
		sub.name = name;
#ifdef USER	
	ret = minmex_registerEventHandler(MY_HANDLE, notifAuxG.handle, &sub, (anaCallback_t) relayNotification, &notifAuxG,  sizeof(struct notifAuxGates), 0);
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached) {
		ret = myminmex->registerEventHandler(MY_HANDLE, notifAuxG.handle, &sub, (anaCallback_t) relayNotification, &notifAuxG,  sizeof(struct notifAuxGates), 0);
	} else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif
	if(ret <0){
		anaPrint(ANA_DEBUG, "relaySubscribeEvent failed \n");
		/*we return the error code :  [e][error_code]*/ 
		m[0] = 'e';
		m[1] = ret;	
		relayControl(notifAuxG.handle, m, 2);
	}
	else{
		/* format of response : [c] */
		m[0] = 'b';
		relayControl(notifAuxG.handle, m, 1);
	}
	
	
}	
void relayEventUnsubscribe(char *msg, int len){
	/* message format : ['g'][MY_HANDLE][eventMask][label][viewer]([name])*/	
	anaLabel_t *lp = NULL;
	anaHandle_t *hp = NULL;
	anaHandle_t querier;
	
	char *name = NULL;
	uint32_t *ip = NULL;
	int offset = 1; 
	uint32_t eventMask;
	int ret;
	struct anaEventSub_s sub;
	char m[2];
	
	hp = (anaHandle_t *) (msg + offset);
	querier = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	ip = (uint32_t *) (msg + offset);
	eventMask = ntohl(*ip);
	offset += sizeof(uint32_t);
	
	lp = (anaLabel_t *) (msg + offset);
	offset += sizeof(anaLabel_t);	
	
	hp = (anaHandle_t *) (msg + offset);
	offset += sizeof(anaHandle_t);
	
	if(len > offset){
		name = msg + offset; 
		offset += strlen(name) + 1;
	}
	
	/* we fill the subscription structure */
	memset(&sub, 0, sizeof(struct anaEventSub_s));
	sub.eventMask = eventMask;
	sub.label = ntohl(*lp);
	sub.viewer = ntohl(*hp);
	if(name)
		sub.name = name;
#ifdef USER	
	ret = minmex_unregisterEventHandler(querier,  &sub);
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached) {
		ret = myminmex->unregisterEventHandler(querier,  &sub);
	} else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif
	if(ret <0){
		anaPrint(ANA_DEBUG, "relaySubscribeEvent failed \n");
		/*we return the error code :  [e][error_code]*/ 
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		m[0] = 'b';
		relayControl(querier, m, 1);
	}
	
	
}	

void relayEventNotify(char *msg, int len){

	/* format ['j'][handle][eventID(4bytes)][label][viewer][destLabel][destViewer][name if any] */
	struct anaEventNotif_s notif;
	
	anaLabel_t *lp = NULL;
	anaHandle_t *hp = NULL;
	anaHandle_t querier;
	
	char *name = NULL;
	uint32_t *ip = NULL;
	int offset = 1; 

	int ret;
	char m[2];
	
	anaPrint(ANA_DEBUG, "********* received notification to relay \n");
	
	memset(&notif, 0, sizeof(struct anaEventNotif_s));
	
	hp = (anaHandle_t *) (msg + offset);
	querier = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	ip = (uint32_t *) (msg + offset);
	notif.eventID = ntohl(*ip);
	offset += sizeof(uint32_t);
	
	lp = (anaLabel_t *) (msg + offset);
	notif.label = ntohl(*lp);
	offset += sizeof(anaLabel_t);	
	
	hp = (anaHandle_t *) (msg + offset);
	notif.viewer = ntohl(*hp);
	offset += sizeof(anaHandle_t);

	lp = (anaLabel_t *) (msg + offset);
	notif.destLabel = ntohl(*lp);
	offset += sizeof(anaLabel_t);	

	hp = (anaHandle_t *) (msg + offset);
	notif.destViewer = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	if(len > offset){
		notif.name = msg + offset; 
		offset += strlen(name) + 1;
	}
	
	anaPrint(ANA_DEBUG, "relayEventNotify : relaying notification to the minmex \n");
#ifdef USER
	ret = minmex_notifyEvent(querier, &notif );
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached) {
		ret = myminmex->notifyEvent(querier, &notif );
	} else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif
	if(ret <0){
		anaPrint(ANA_DEBUG, "relayEventNotify failed \n");
		/*we return the error code :  [e][error_code]*/ 
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		anaPrint(ANA_DEBUG, "relayEventNotify succeeded \n");
		/* format of response : [c] */
		m[0] = 'j';
		relayControl(querier, m, 1);
	}
}

void relaySetInfo(char *msg, int len) {
	
/* message format : ['s'][MY_HANDLE][label][callee][outputIDP][IC flag]([ctxtLen (2bytes )][ctxt][srvLen (2 bytes)][srv][MTU][chanType]) */
	
	char m[CTRLMSG_MAXSIZE];
	int ret, offset;
	struct IDPinfo_s info;
	char *t;
	uint16_t *ip;

	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	
	anaHandle_t querier;

	memset(&info, 0, sizeof(struct IDPinfo_s));
	info.IDPtype = BRICK;	
	
	hp = (anaHandle_t *) (msg + 1);
	querier = ntohl(*hp);
	offset = 1 +sizeof(anaHandle_t);
	
	lp = (anaLabel_t *) (msg + offset);
	info.label = ntohl(*lp);
	offset += sizeof(anaLabel_t);
			
	hp = (anaHandle_t *) (msg + offset);
	info.binding.brick.handle = ntohl(*hp);
	offset += sizeof(anaHandle_t);
	
	lp = (anaLabel_t *) (msg + offset);
	info.binding.brick.outputIDP = ntohl(*lp);
	offset += sizeof(anaLabel_t);
	
	t = (char *) (msg + offset);
	if(t[0]){
		int ctxLen, srvLen;
		info.ICflag = 1;
		offset += 1; 
		
		ip = (uint16_t *) (msg +offset);
		ctxLen = ntohs(*ip);
		offset+= sizeof(uint16_t);
		if(ctxLen>0){
			info.destContext.valueLen = ctxLen;
			info.destContext.value = (void *) malloc(ctxLen);
			memcpy(info.destContext.value, msg+offset,ctxLen );
			offset+= ctxLen;
		}
		
		ip = (uint16_t *) (msg +offset);
		srvLen = ntohs(*ip);
		offset+= sizeof(uint16_t);
		if(srvLen>0){
			info.destService.valueLen = srvLen;
			info.destService.value = (void *) malloc(srvLen);
			memcpy(info.destService.value, msg+offset,srvLen );
			offset+= srvLen;
		}
		ip = (uint16_t *) (msg +offset);
		info.MTU = ntohs(*ip);
		offset += sizeof(uint16_t);
		
		t = (char *) (msg + offset);
		info.chanType = t[0];
		offset += 1;
			
	}
	else{
		info.ICflag = 0;
		offset += 1; 	
	}
	
	/* we call minmex_setinfo */ 
#ifdef USER		

	ret = minmex_setIDPinfo(querier, &info);
#endif
#ifdef KERNEL	
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->setIDPinfo(querier, &info);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
#endif

	if(ret <0){
		anaPrint(ANA_DEBUG, "relaySetinfo failed err");
		/*we return the error code :  [e][error_code]*/ 
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		m[0] = 's';
		relayControl(querier, m, 1);
	}
	
	
}	
	
void relayGetInfo(char *msg, int len) {
	

	/* format of message : [i][myHandle][label][viewer]*/
	char m[CTRLMSG_MAXSIZE];
	int ret, offset;
	struct IDPinfo_s info;
	char *t;
	uint16_t *ip;
	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	
	anaHandle_t querier;
	anaHandle_t viewer;
	anaLabel_t label;
		
	hp = (anaHandle_t *) (msg + 1);
	querier = ntohl(*hp);
	
	lp = (anaLabel_t *) (msg + 1 + sizeof(anaHandle_t));
	label = ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1 + sizeof(anaHandle_t) + sizeof(anaLabel_t));
	viewer = ntohl(*hp);

	/* we call the minmex_getIDPinfo function */
	memset(&info, 0, sizeof(struct IDPinfo_s));
	
	anaPrint(ANA_DEBUG, "loking for info of IDP %8x\n", label);
#ifdef USER		
	ret = minmex_getIDPinfo(querier, viewer, label, &info);
#endif
#ifdef KERNEL	
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->getIDPinfo(querier, viewer, label, &info);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
#endif
	
	if(ret < 0){
		m[0] = 'e';
		m[1] = ret;
		relayControl(querier, m, 2);
	}
	else{
		/* response format : [i][label][viewer][owner][permFlag 1 byte][ICFlag]([contextLen (2 bytes)][context][serviceLen (2 bytes)][service][MTU (2 bytes)][chanType])[IDPtype(1 byte)][destination] */
		/* depending on the IDP types destination is :*/
		 	/* Brick : [callee handle][outpuIDP] */
			/* Redirect : [dest label][destViewer] */
			/* FORK : [nbForks 1 byte][dest Labe1][dest Viewer1]...*/

		memset(m, 0, CTRLMSG_MAXSIZE);
		m[0] = 'i';
		lp = (anaLabel_t *) (m+1);
		*lp = htonl(info.label);
		
		hp = (anaHandle_t *) (m + 1 + sizeof(anaLabel_t));
		*hp = htonl(info.viewer);
		
		hp = (anaHandle_t *) (m + 1 + sizeof(anaLabel_t) + sizeof(anaHandle_t));
		*hp = htonl(info.owner);
		
		t = (char *) (m + 1 + sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t));
		if(info.permFlag)
			t[0] = 1;
		/*else
			t[0] = 0 ; this instruction is implicit since array set to 0 */
		
		t = (char *) (m + 2 + sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t));
		if(info.ICflag){
			t[0] = 1;
			offset = 3 + sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t);
			
			if((offset + info.destContext.valueLen + info.destService.valueLen) + sizeof(uint16_t) > CTRLMSG_MAXSIZE){
				anaPrint(ANA_ERR, "relayGetInfo : adding the context and service to the response message will go over the CTRLMSG_MAXSIZE limit \n");
				return;
			}
			ip = (uint16_t *) (m+offset);
			*ip = htons(info.destContext.valueLen);
			
			offset += sizeof(uint16_t);
			
			if(info.destContext.valueLen > 0 && info.destContext.value!=NULL){
				memcpy(m+offset,info.destContext.value, info.destContext.valueLen );
				offset += info.destContext.valueLen;
			}
			
			ip = (uint16_t *) (m+offset);
			*ip = htons(info.destService.valueLen);	
			offset += sizeof(uint16_t);
			
			if(info.destService.valueLen > 0 && info.destService.value!=NULL){
				memcpy(m+offset,info.destService.value, info.destService.valueLen );
				anaPrint(ANA_DEBUG, "******** destService %s \n",m+offset);
				offset += info.destService.valueLen;
			}
			
			ip = (uint16_t *) (m+offset);
			*ip = htons(info.MTU);
			anaPrint(ANA_DEBUG, " ########### MTU %d info.MTU %d\n",  ntohs(*ip), info.MTU);			
			offset += sizeof(uint16_t);
		
			t = (char *) (m+offset);
			t[0] = info.chanType;
			offset += 1;
		
		}
		else{
			/*t[0] = 0; implicit */
			offset = 3 + sizeof(anaLabel_t) + 2 * sizeof(anaHandle_t);
		}
		
		t = (char *) (m + offset);
	/*	switch(info.IDPtype) {*/
			if(info.IDPtype == BRICK){
				if( (1+sizeof(anaHandle_t) + sizeof(anaLabel_t) + offset )  > CTRLMSG_MAXSIZE){
					anaPrint(ANA_ERR, "relayGetInfo : adding the IDP type BRICK info to the response message will go over the CTRLMSG_MAXSIZE limit \n");
					return;
				}
				t[0] = BRICK;
				offset += 1;
				hp = (anaHandle_t *) (m+offset);
				*hp = htonl(info.binding.brick.handle);
				offset += sizeof(anaHandle_t);
				
				
				lp = (anaLabel_t *) (m+offset);
				*lp = htonl(info.binding.brick.outputIDP);
				offset += sizeof(anaLabel_t);
				
			}
			if( info.IDPtype == REDIRECT){
				
				if( (1+sizeof(anaHandle_t) + sizeof(anaLabel_t) + offset )  > CTRLMSG_MAXSIZE){
					anaPrint(ANA_ERR, "relayGetInfo : adding the IDP type REDIRECT info to the response message will go over the CTRLMSG_MAXSIZE limit \n");
					return;
				}
				t[0] = REDIRECT;
				offset += 1;
				lp = (anaLabel_t *) (m+offset);
				*lp = htonl(info.binding.IDPs->label);
				offset += sizeof(anaLabel_t);
				
				hp = (anaHandle_t *) (m+offset);
				*hp = htonl(info.binding.IDPs->viewer);
				offset += sizeof(anaHandle_t);
							
				
			}
			if( info.IDPtype == FORK){
				
				int nbForks = 0;
				struct nextIDP_s *tmp =   info.binding.IDPs;
				int i;
				while(tmp !=NULL){
					nbForks++;
					tmp = tmp->next;
				}
				
				if( ( 1 + nbForks * (sizeof(anaHandle_t) + sizeof(anaLabel_t)) + offset )  > CTRLMSG_MAXSIZE){ /* 1 is for the number of forks */
					anaPrint(ANA_ERR, "relayGetInfo : adding the IDP type FORK info to the response message will go over the CTRLMSG_MAXSIZE limit \n");
					return;
				}
				t[0] = FORK;
				offset += 1;
				t = (char *) (m + offset);
				t[0] = nbForks;
				offset += 1; 
				tmp = info.binding.IDPs;
				
				for(i=0;i<nbForks;i++){
					lp = (anaLabel_t *)(m +offset);
					*lp = htonl(tmp->label);
					offset+=sizeof(anaLabel_t);
					
					hp = (anaHandle_t *)(m +offset);
					*hp = htonl(tmp->viewer);
					offset+=sizeof(anaHandle_t); 
					
					tmp = tmp->next;
				}	
					
				
			}
		
		
		/* we send the response back */	
		relayControl(querier, m, offset);	
	}
	
	
}
void relayForkData(char *msg, int len) {
	

	/* format of message : [f][handle][originViewer][originLabel][destViewer][destLabel]*/

	char m[CTRLMSG_MAXSIZE];
	int ret;
	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	
	anaHandle_t originViewer;
	anaHandle_t destViewer;
	
	anaHandle_t querier;
	
	anaLabel_t originLabel;
	anaLabel_t destLabel;	
	

	if(msg==NULL || len<( 1 + 3*sizeof(anaHandle_t) + 2*sizeof(anaLabel_t) )){
		anaPrint(ANA_ERR, "relayForkData: Wrong message\n");
		return;
	}	
	
	
	hp = (anaHandle_t *) (msg + 1);
	querier = (anaHandle_t) ntohl(*hp);
	
	hp = (anaHandle_t *) (msg + 1+sizeof(anaHandle_t));
	originViewer = (anaHandle_t) ntohl(*hp);
	
	lp = (anaLabel_t *) (msg + 1+2*sizeof(anaHandle_t));
	originLabel = (anaLabel_t) ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1 + 2*sizeof(anaHandle_t)+sizeof(anaLabel_t));
	destViewer = (anaHandle_t) ntohl(*hp);
	
	lp = (anaLabel_t *) (msg + 1 + 3* sizeof(anaHandle_t)+sizeof(anaLabel_t));
	destLabel = (anaLabel_t) ntohl(*lp);
	
	
#ifdef USER	
	ret = handleForkData(querier, originViewer, originLabel, destViewer, destLabel);
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->forkData(querier , originViewer, originLabel, destViewer, destLabel);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif

	if(ret <0){
		anaPerror(ANA_ERR, "handleForkData failed err");
		/*we return the error code :  [e][error_code]*/ 
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'u';
		
		relayControl(querier, m, 1);
	}
		
}

void relayChangeView(char *msg, int len) {
	
	
	/* format of message : [v][label][myhandle][oldViewerHandle][newViewerHandle] */

	char m[CTRLMSG_MAXSIZE];
	int ret;
	
	anaHandle_t *hp= NULL;
	anaLabel_t *lp= NULL;
	

	anaLabel_t label;
	anaHandle_t querier;
	anaHandle_t oldViewer;
	anaHandle_t newViewer;	

	if(msg==NULL || len<(3*sizeof(anaHandle_t)) +sizeof(anaLabel_t) +1){
		anaPrint(ANA_ERR, "relayChangeView: Wrong Register message\n");
		return;
	}	
	
		
	lp = (anaLabel_t *) (msg + 1);
	label = (anaLabel_t) ntohl(*lp);
	
	hp = (anaHandle_t *) (msg + 1+ sizeof(anaLabel_t));
	querier = (anaHandle_t) ntohl(*hp);
	
	
	hp = (anaHandle_t *) (msg + 1+ sizeof(anaLabel_t) + sizeof(anaHandle_t));
	oldViewer = (anaHandle_t) ntohl(*hp);
	
	
	hp = (anaHandle_t *) (msg + 1+ sizeof(anaLabel_t) + 2*sizeof(anaHandle_t));
	newViewer = (anaHandle_t) ntohl(*hp);
	
	
#ifdef USER	
	ret = handleChangeView(querier, oldViewer, label, newViewer);
#endif
#ifdef KERNEL
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->changeView(querier, oldViewer, label, newViewer);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);

#endif
	if(ret <0){
		/*we return the error code :  [e][error_code]*/ 
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'e';
		m[1] = ret;	
		relayControl(querier, m, 2);
	}
	else{
		/* format of response : [c] */
		memset(m, 0,CTRLMSG_MAXSIZE );
		m[0] = 'v';
		
		relayControl(querier, m, 1);
	}
		
}



void relayHBRequest(char *msg, int length){
	anaHandle_t *hp = (anaHandle_t *)(msg+1);
	anaHandle_t h = ntohl(*hp);
	
	anaLabel_t *lp = (anaLabel_t *)(msg+1+sizeof(anaHandle_t));
	anaLabel_t l = ntohl(*lp);
#ifdef USER	
	handleHBRequest(h);
#endif
#ifdef KERNEL
	anaPrint(ANA_DEBUG, "hbrequest\n");
	analock_read_lock(&attachLock);
	if(attached)
		myminmex->heartBeat(h);
	analock_read_unlock(&attachLock);
#endif
	/* and now we send a response to the brick */
	relayData("h", 1, l,  &h);
}


int relayAttach(char *msg, int length) {
	
	int  controlMode, controlAuxLength, len, found,dataMode, replySize,dataAuxLength, desiredMode;
	char *controlLink;
	char *dataLink;
		
	struct attachedBrick_s *entry = NULL;
	
	char *brickName;

	char *IP;
	int port;
	char *portChar;
	struct sockaddr_in *tmp;

	unsigned char replyMsg[CTRLMSG_MAXSIZE];
	struct commGate_s *tmpGate;
     	
	struct attachRes_s *attachRes = NULL;
#ifdef USER
	char *pipePath;
     	struct commSpecs_s specs;
#endif
#ifdef KERNEL
	struct anaBrickFunctions *myFunctions = NULL;
#endif
	anaLabel_t *label;
	anaHandle_t *handle;

#ifdef USER	
	memset(&specs, 0,  sizeof(struct commSpecs_s));
	specs.queueLock = &queueLock;  
	specs.msgQueue = &messageQueue; 
	specs.condLock = &waitLock;
	
	specs.pluginHandle = mypluginHandle;
	specs.pluginFlag = 0;

#endif
#ifdef KERNEL
	myFunctions = malloc(sizeof(struct anaBrickFunctions));
        myFunctions->wq = wq;				/* the work queue */
        myFunctions->work = &rcv_data;			/* which function we are submitting to the work queue */
	myFunctions->recvDataQueue = &recvDataQueue;	/* the message queue where to add new messages */
	myFunctions->queueLock = &queueLock;		/*this lock protects the recvDataQueue */
	myFunctions->notification = notificationTemp;
	myFunctions->pluginFlag = 0;
#endif	
	
	if(msg==NULL || length<3){
		anaPrint(ANA_ERR, "relayAttach: Wrong attach message\n");
		return -1;
	}	
	/*<Command parsing part>*/
	
	/* The received attach message is formatted this way: [a][controlMode][controlAuxLength][controlAux][dataMode][dataAuxLength][dataAux] [desiredMode]
	
		- a is the identifier of the attach command
		- controlMode is the transfer mode (UDP,Pipes ...) the brick wishes to receive control messages on
		- controlAuxLength is the length of the controlAux field
		- controlAux is the parameter to establish the control link according to the mode (if pipes: name of the pipe, UDP : IP+port)
		- datamode is the transfer mode (UDP, Pipes..) the brick wishes to use for data
		- dataAuxLength is the length of the dataAux field
		- dataAux is the parameter to establish the data link according to the mode
		- desiredMode is the mode the brick wishes to use for sending data to minmex, if the minmex doesnt support it, it indicates the default data mode
	*/		
	
	controlMode = (int) msg[1];
	controlAuxLength = (int) msg[2]; 
	
	if( length<5+controlAuxLength){ /* counting dataMode and dataAuxLength */
		anaPrint(ANA_ERR, "relayAttach: Wrong attach message, no control gate indication\n");
		return -1;
	}
	
	controlLink= (char *) malloc((controlAuxLength+1)+sizeof(char));
	strncpy(controlLink, (char *)msg+3, controlAuxLength);
	controlLink[controlAuxLength] = 0 ;
	anaPrint(ANA_ERR, "identified  replyLink %s \n",controlLink );	

	dataMode = (int) msg[controlAuxLength+3];
	dataAuxLength = (int) msg[controlAuxLength+4]; 
	
	if( length<5+controlAuxLength+dataAuxLength){ 
		anaPrint(ANA_ERR, "relayAttach: Wrong attach message, no data gate indication \n");
		return -1;
	}
		
	dataLink = (char *) malloc((dataAuxLength+1)+sizeof(char));
	strncpy(dataLink, (char *) msg+controlAuxLength+5, dataAuxLength);
	dataLink[dataAuxLength] = 0 ;
	anaPrint(ANA_DEBUG, "identified  dataLink %s \n", dataLink);
	
	
	
	/* now we parse the name of the Brick*/
	if(length < dataAuxLength+6+controlAuxLength){
		anaPrint(ANA_ERR, "relayAttach: Wrong attach message, no bick name indication \n");
		return -1;
	}
		
	len =  (int) msg[dataAuxLength+5+controlAuxLength];
	
	
	if(length < dataAuxLength+6+controlAuxLength+len){
		anaPrint(ANA_ERR, "relayAttach: Wrong attach message, incorrect brick name length \n");
		return -1;
	}
	
	brickName = (char *) malloc((len+1));
	strncpy(brickName, (char *) msg+controlAuxLength+6+dataAuxLength, len);
	
	desiredMode=defaultDataGate->mode;
	if(length > dataAuxLength+6+controlAuxLength+len) {	/* the desiredMode feeld is optional, we dont cancel attachement if it's absent*/
		desiredMode = (int) msg[dataAuxLength+6+controlAuxLength+len];
	}
	/* <End of Command parsing part>*/
	
	entry = (struct attachedBrick_s * ) malloc(sizeof(struct attachedBrick_s));

	switch(dataMode) {
		case UDP:
		{
		#ifdef USER
			int sock ;
			#define CLIENT_DATA_DESC entry->dataDescriptor
		#endif
		#ifdef KERNEL
			struct socket *sock;
			#define CLIENT_DATA_DESC entry->dataSocket
		#endif
		
                        IP = strsep(&dataLink, ":");
                        portChar = strsep(&dataLink, ":");
			if(portChar==NULL){
				anaPrint(ANA_ERR, "relayAttach : Unable identify UDP port for brick's data Gate");
				
				free(controlLink);
				free(dataLink);
				free(brickName);
				free(entry);
				return -1;
			}
			port = atoi(portChar);
			entry->dataMode = UDP;
			
            		if(sockCreateWrapper(&sock, PF_INET, SOCK_DGRAM, IPPROTO_UDP)<0 ) {
            			anaPrint(ANA_ERR, "relayAttach : Unable to create UDP socket for brick's data Gate");
				free(brickName);
				free(controlLink);
				free(dataLink);
				free(entry);
                		return -1;
			}
			
		
			CLIENT_DATA_DESC = sock;

			tmp = (struct sockaddr_in*) malloc(sizeof(struct sockaddr_in));
			tmp->sin_addr.s_addr = inet_addr(IP);
			tmp->sin_family = AF_INET;
			tmp->sin_port = htons(port);
			
			entry->dataAux = tmp;
			
			break;
		}
#ifdef USER
		case UNIX:
		{
			int sock ;
			struct sockaddr_un *tmpu;
			char *sockPath = dataLink;
		       	entry->dataMode = UNIX;
			
            		if(sockCreateWrapper(&sock, AF_UNIX, SOCK_DGRAM, 0)<0 ) {
            			anaPrint(ANA_ERR, "relayAttach : Unable to create UNIX socket for brick's data Gate");
				free(brickName);
				free(controlLink);
				free(dataLink);
				free(entry);
                		return -1;
			}
			
		
			entry->dataDescriptor = sock;

			tmpu = (struct sockaddr_un*) malloc(sizeof(struct sockaddr_un));
			
			tmpu->sun_family = AF_UNIX;
        		strcpy(tmpu->sun_path, sockPath);
        
       		      	entry->dataAux = tmpu;
			break;
		}
                case PIPES:
		{	
			pipePath = dataLink;
			entry->dataMode = PIPES;
			entry->dataDescriptor =  open(pipePath, O_NONBLOCK|O_WRONLY);
			if(entry->dataDescriptor <0 ){
				anaPrint(ANA_ERR, "relayAttach : Unable to create named PIPE for brick's data Gate");
				free(brickName);
				free(controlLink);
				free(dataLink);
				free(entry);
				return -1;
			}
			break;
		}
#endif /*USER*/
#ifdef KERNEL
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                case GENETLINK:
                {
                        int * tmp;
                        anaPrint(ANA_DEBUG, "relayAttach : send data over genetlink\n");
                        entry->dataMode = GENETLINK;
                        tmp = malloc(sizeof(int));
                        memcpy(tmp, dataLink ,sizeof(int));
                        entry->dataDescriptor = *tmp;
                        free(tmp);
                        
                        break;
                }
#endif /*Version > 2.6.19*/
#endif /*KERNEL*/
		default:
		{	anaPrint(ANA_NOTICE, "relayAttach : Uknown data transfer mode %d \n",dataMode);
			free(brickName);
			free(controlLink);
			free(dataLink);
			free(entry);
			return -1;
		}
	}

	switch(controlMode) {

		case UDP:
		{	
		#ifdef USER
			int sock ;
			#define CLIENT_CONTROL_DESC entry->controlDescriptor
		#endif
		#ifdef KERNEL
			struct socket *sock;
			#define CLIENT_CONTROL_DESC entry->controlSocket
		#endif
		
                        IP = strsep(&controlLink, ":");
                        portChar = strsep(&controlLink, ":");
			if(portChar==NULL){
				anaPrint(ANA_ERR, "relayAttach : Unable identify UDP port for brick's control Gate");
				
				closeWrapper(&(CLIENT_DATA_DESC));
                                /* close(entry->dataDescriptor); */
				free(controlLink);
				free(dataLink);
				free(entry->dataAux);
				free(entry);free(brickName);
				return -1;
			}
				
			port =atoi(portChar);
			entry->controlMode = UDP;

                        if( sockCreateWrapper(&sock, PF_INET, SOCK_DGRAM, IPPROTO_UDP)<0 ) {
                                anaPrint(ANA_ERR, "relayAttach : Unable to create UDP socket for brick's control Gate");
				
				/* close(entry->dataDescriptor); */
				closeWrapper(&(CLIENT_DATA_DESC));
				free(controlLink);
				free(dataLink);
				free(entry->dataAux);
				free(entry);free(brickName);
                                return -1;
			}
			
			CLIENT_CONTROL_DESC = sock;
					
			tmp = (struct sockaddr_in*) malloc(sizeof(struct sockaddr_in));
			tmp->sin_addr.s_addr = inet_addr(IP);
			tmp->sin_family = AF_INET;
			tmp->sin_port = htons(port);
			
			entry->controlAux = tmp;
			break;
		}
#ifdef USER
		case UNIX:
		{			
			int sock ;
			struct sockaddr_un *tmpu;
                        char *sockPath = controlLink;
		
			entry->controlMode = UNIX;

                        if( sockCreateWrapper(&sock, AF_UNIX, SOCK_DGRAM, 0)<0 ) {
                                anaPrint(ANA_ERR, "relayAttach : Unable to create UNIX socket for brick's control Gate");
				
				/* close(entry->dataDescriptor); */
				closeWrapper(&(CLIENT_DATA_DESC));
				free(controlLink);
				free(dataLink);
				free(entry->dataAux);
				free(entry);free(brickName);
                                return -1;
			}
			
			entry->controlDescriptor = sock;
			
			tmpu = (struct sockaddr_un*) malloc(sizeof(struct sockaddr_un));
			
			tmpu->sun_family = AF_UNIX;
        		strcpy(tmpu->sun_path, sockPath);
			
			entry->controlAux = tmpu;
			break;
		}
		case PIPES:
		{	
			pipePath = controlLink;
			entry->controlMode = PIPES;
			entry->controlDescriptor =  open(pipePath, O_NONBLOCK|O_WRONLY);
			if(entry->controlDescriptor <0 ){
				anaPrint(ANA_ERR, "relayAttach : Unable to create named PIPE for brick's control Gate");
			
                                closeWrapper(&(CLIENT_DATA_DESC));
				free(controlLink);
				free(dataLink);
				free(entry->dataAux);
				free(entry);free(brickName);
				return -1;
			}
			break;
		}
#endif /*USER*/
#ifdef KERNEL
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                case GENETLINK:{
                        int pid;
                        entry->controlMode = GENETLINK;
                        pid = *(int *)controlLink;
                        entry->controlDescriptor = pid;
                        break;
                }
#endif/*Version > 2.6.19*/
#endif /*KERNEL*/
		default:
		{	anaPrint(ANA_NOTICE, "relayAttach : Uknown control transfer mode %d \n",dataMode);
			free(controlLink);
			free(dataLink);
			free(entry->dataAux);
          		closeWrapper(&(CLIENT_DATA_DESC));
			free(brickName);
			free(entry);
			return -1;
		}
	}
	free(controlLink);
	free(dataLink);


	/* we call the minmex's handleAttach */
/*	
	memset(&specs, 0,  sizeof(struct commSpecs_s));
	specs.queueLock = &queueLock;  
	specs.msgQueue = &messageQueue; 
	specs.condLock = &waitLock;
	specs.pluginHandle = mypluginHandle;
*/
#ifdef KERNEL
	attachRes = myminmex->attach(brickName, myFunctions);
#endif
#ifdef USER
	attachRes = handleAttach(brickName, &specs);
#endif
	if(attachRes==NULL){
		anaPrint(ANA_EMERG, "relayAttach : Unable to attach the remote brick to ANA minmex \n");
		/* we send an error message back */
		switch(entry->controlMode) {

		case UDP:{
	        #ifdef USER
			int sock = entry->controlDescriptor;
		#endif
		#ifdef KERNEL
			struct socket *sock = entry->controlSocket;
		#endif
		        sendSocketWrapper(&sock, AF_INET,entry->controlAux,"error", 6);
			break;
            	}
#ifdef USER
		case UNIX:{
			int sock = entry->controlDescriptor;
			sendSocketWrapper(&sock, AF_LOCAL, entry->controlAux,"error", 6);
			break;
		}
		case PIPES:{
			write(entry->controlDescriptor, "error", 6);
			break;
		}
#endif /*USER*/
#ifdef KERNEL
	  	case GENETLINK:{
			sendGenetlink(entry->controlDescriptor, "error", 6, COMMAND_CONTROL);
			break;
		}
#endif /*KERNEL*/

		}
		
		free(entry->dataAux);
		free(entry->controlAux);
		free(brickName);
		free(entry);
	}
	free(brickName);
	
	entry->handle = attachRes->handle;
	
	/* we add the entry to the attached brick's table */
	addtableEntry (&attBricksTable, entry);
	
	/* <Reply message format >
		
	['r'][handle][controlLabel][minmexDataMode][dataAuxLength][dataAux]  */


	
		
	memset(replyMsg, 0, CTRLMSG_MAXSIZE);
	replyMsg[0] ='r';
	
	handle = (anaHandle_t *)(replyMsg+1);
	*handle = htonl(attachRes->handle); /* the function htonl() should probably be changed in case the label size changes to 64 bits*/

	label = (anaLabel_t *) (replyMsg+1+sizeof(anaHandle_t));
	*label =  htonl(attachRes->nodeCompLabel); /* the function htonl() should probably be changed in case the label size changes to 64 bits*/
	

	replySize = 1 + sizeof(anaHandle_t) + sizeof(anaLabel_t);
		
	tmpGate = defaultDataGate;
	found = 0;

	while(!found && tmpGate != NULL){
		if (tmpGate->mode==desiredMode){
			anaPrint(ANA_DEBUG, "Found gate supporting data mode %d urn %s\n", desiredMode, tmpGate->urn);
			found = 1;
			replyMsg[replySize] = (unsigned char)desiredMode;
			len = strlen(tmpGate->urn);
			replyMsg[replySize+1] = (unsigned char)len ;
			strncpy((char*)replyMsg+ replySize+2, tmpGate->urn, len);
			replySize = replySize + 2 + len;
                       	break;
		}
		tmpGate = tmpGate->next;	
	}

	if(!found){
		replyMsg[replySize] = (unsigned char)defaultDataGate->mode;
		len = strlen(defaultDataGate->urn);
		replyMsg[replySize+1] = (unsigned char)len;
		strncpy((char*)replyMsg+replySize+2, defaultDataGate->urn, len);
		replySize = replySize + 2 + len;
               
	}

	len = 0;
	len = relayControl(entry->handle, replyMsg, replySize);
	
	/* <Response sending> */
	if(len<0){
		anaPrint(ANA_ERR, "relayAttach : unable to send attach reply to brick\n");
		freetableEntry(&attBricksTable, entry->handle);
		return -1;
	}
	/*<End of Response sending> */
	anaPrint(ANA_NOTICE, "Attached new brick, current number of bricks: %d\n",attBricksTable.nbAttached);	
      
      	/* we free the returned attachRes*/
	free(attachRes);
	
        return len;
}


/***************************************/
/*          kernel space               */
/***************************************/
#ifdef KERNEL
void cleanup(void);

struct wq_wrapper wq_ctr;
struct wq_wrapper wq_data;

void parseData(char *buff, int len) {
	int ackFlag;
	char ackBuf[2];
	anaLabel_t *label = (anaLabel_t *)buff;
	anaLabel_t labelVal = (anaLabel_t) ntohl(*label);
	anaHandle_t *handle = (anaHandle_t *) (buff + sizeof(anaLabel_t));
	anaHandle_t handleVal = (anaHandle_t) ntohl(*handle);

	void *data = buff + sizeof(anaLabel_t) + sizeof(anaHandle_t) + 1; /* +1 for the ack flag byte */
	int dataLength = len  - sizeof(anaHandle_t) - sizeof(anaLabel_t) -1; /* -1 for the ack flag byte */
	int ret = myminmex->callback(myHandle, labelVal, handleVal, data, dataLength); /* do we send a notification in case of failure ? */
	char * dummy = (char *) (buff + sizeof(anaLabel_t) + sizeof(anaHandle_t));
	ackFlag = dummy[0];
	if (ackFlag) { /* if the ACK flag is on*/
		if (ret<0) {
			ackBuf[0] = 'e';
			ackBuf[1] = ret;
		}
		else{
			ackBuf[0] = 'a'; /* a for ack(success)*/
			ackBuf[1] = 0;
		}
		relayControl(handleVal, ackBuf, 2);	
	}

}
void parseControl(char * buff, int len){
    	if(buff==NULL || len<0)
		return;
	switch (buff[0]) {
	case 'a':
	        relayAttach(buff, len);
       		break;
  	case 'd':
  		relayDetach(buff, len);
  		break;
  
  	case 'r':
  		relayRegister(buff, len);
  		break;
  	case 'c':
  		relayRedirect(buff, len);
  		break;
  		
  	case 'h': 
  		relayHBRequest(buff, len);
  		break; 
  	case 'u':
  		relayUnregister(buff, len);
  		break;
  
  	case 'v': 
  		relayChangeView(buff, len);
  		break;
  	case 'f':     	
  		relayForkData(buff, len);
  		break; 
  	
  	default: anaPrint(ANA_NOTICE, "Command not supported! %s\n",buff);
  	}

}

static void parseControlTemp(struct work_struct *control);
static void parseDataTemp(struct work_struct *data);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
/*the following functions and structures are used to support generic netlink */
int recvControlGN(struct sk_buff *skb_2, struct genl_info *info);
int recvDataGN(struct sk_buff *skb_2, struct genl_info *info);
/* policies for the argument for control and data messages */
static struct nla_policy controlPolicyGN[ATTR_MAX + 1] = {
        [ATTR_ID] = { .type = NLA_STRING, .len = 1024 },
        [ATTR_ID2] = { .type = NLA_UNSPEC, .len = 1024},
};
static struct nla_policy dataPolicyGN[ATTR_MAX + 1] = {
        [ATTR_ID] = { .type = NLA_STRING, .len = 1024 },
        [ATTR_ID2] = { .type = NLA_UNSPEC, .len = 1024},
};


/* family definition */
struct genl_family ANAFamGN = {
        .id = GENL_ID_GENERATE,   /* genetlink should generate an id */
        .hdrsize = 0,
       	.name = "ANA",            /* the name of this family, used by userspace application */
        .version = 1,             /* version number */  
        .maxattr = ATTR_MAX,
};
/* for control messages */
struct genl_ops controlOpsGN = {        
        .cmd = COMMAND_CONTROL,
        .flags = 0,
        .policy = controlPolicyGN,
        .doit = recvControlGN,
        .dumpit = NULL,
};
/* for data messages */
struct genl_ops dataOpsGN = {
        .cmd = COMMAND_DATA,
        .flags = 0,
        .policy = dataPolicyGN,
        .doit = recvDataGN,
        .dumpit = NULL,
};

/* callback function for the generic netlink mode, they are invoked upon 
 * receiving a message from a netlink socket
 */
int recvControlGN(struct sk_buff *skb, struct genl_info *info)
{
        struct nlattr *na;
        if (info == NULL)
                return 0;
           
        /*for each attribute there is an index in info->attrs which points to a nlattr structure
         *in this structure the data is given
         */
        if(info->attrs != NULL){
                char * mydata;
                int length;
                na = info->attrs[ATTR_ID];
                mydata = (char *)nla_data(na);
                length = nla_len(na); 
                // *(int *)(mydata+length - 5) = info->snd_pid;
                parseControl(mydata, length);
             
        }
        return 0;
}

int recvDataGN(struct sk_buff *skb, struct genl_info *info)
{
        struct nlattr *na;
        if (info == NULL)
                return 0;
       
        
        /*for each attribute there is an index in info->attrs which points to a nlattr structure
         *in this structure the data is given
         */
        if(info->attrs != NULL){
                char * mydata;
                int nbRecv;
                na = info->attrs[ATTR_ID];
                mydata = (char *)nla_data(na);
                nbRecv = nla_len(na);
                *(int *)(mydata - sizeof(int)) = info->snd_pid;
                /* anaPrint(ANA_ERR, "data: info->attrs %s\n",mydata);*/
                parseData(mydata, nbRecv);
        }
        return 0;
}
#endif

/*functions used for udp mode. The cb_* functions are invoked upon receiving a packet 
 * on the socket and are executed in interrupt context. They invoke the *Temp functions, 
 * which are executed on behalf of the work queue. They dequeue all packets from the sockets 
 * receive queue. Important: Dequeuing has to be done in the *Temp functions, since otherwise
 * we would overwrite some skb's or we would have to allocate memory in the interrupt handler.
 */
static void cb_control(struct sock *sk, int bytes){
	wq_ctr.sk = sk;
	queue_work(wq, &wq_ctr.worker);
}

void parseControlTemp(struct work_struct *control){
	struct  wq_wrapper * foo = container_of(control, struct  wq_wrapper, worker);
	int len = 0;
	while((len = skb_queue_len(&foo->sk->sk_receive_queue)) > 0){ /* as long as there are messages in the receive queue of this socket*/
		struct sk_buff *skb = NULL;
        	anaPrint(ANA_DEBUG, "packets in udp control queue: %i\n",len);
       	        skb = skb_dequeue(&foo->sk->sk_receive_queue);
		parseControl(skb->data+8, skb->len-8); //8 bytes udp header
        	kfree_skb(skb);
	}
}

static void cb_data(struct sock *sk, int bytes){
	wq_data.sk = sk;
	queue_work(wq, &wq_data.worker);
}

void parseDataTemp(struct work_struct *data){
        struct  wq_wrapper * foo = container_of(data, struct wq_wrapper, worker);
	int len = 0;
	while((len = skb_queue_len(&foo->sk->sk_receive_queue)) > 0){ /* as long as there are messages in the receive queue of this socket*/
		struct sk_buff *skb = NULL;
        	anaPrint(ANA_DEBUG, "packets in upd data queue: %i\n", len);
       	        skb = skb_dequeue(&foo->sk->sk_receive_queue);
		parseData(skb->data+8, skb->len-8); //8 bytes udp header
		kfree_skb(skb);
	}
}


int sendGenetlink(int desc, void *msg, int len, int command) {
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
	/* allocate some memory, since the size is not yet known use NLMSG_GOODSIZE*/
	struct sk_buff *skb;
	void *msg_head;
	skb = genlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);
	if (skb == NULL) {
		anaPrint(ANA_ERR, "relayControl : Unable to send control data through \
				generic Netlink socket (no memory left) \n");
		return -1;
	}
	/* create the message headers */
	/* arguments of genlmsg_put: 
	   struct sk_buff *, 
	   int (sending) pid, 
	   int sequence number, 
	   struct genl_family *,
	   int flags, 
	   u8 command index (why do we need this?)
	   */
	/* we do not implement sequence numbers */
	msg_head = genlmsg_put(skb, 0, 1, &ANAFamGN, 0, command);
	if (msg_head == NULL) {
		anaPrint(ANA_ERR, "relayControl : Unable to send control data through \
				generic Netlink socket (couldn't set header) \n");
		kfree_skb(skb);
		return -1;
	}
	/* add value to be sent */
	if (nla_put(skb, ATTR_ID, len, msg) != 0) {
		anaPrint(ANA_ERR, "relayControl : Unable to send control data through \
				generic Netlink socket (couldn't set msg) \n");
		kfree_skb(skb);
		return -1;
	}

	/* finalize the message */
	genlmsg_end(skb, msg_head);
	/* send the message back */
	if (genlmsg_unicast(skb, desc) != 0) {
		anaPrint(ANA_ERR, "relayControl : Unable to send control data through \
				generic Netlink socket (couldn't send message) \n");
		kfree_skb(skb);
		return -1;
	}
	return len;
	#else
	anaPrint("generic netlink is not supported for kernel < 2.6.20\n")
	return -1;
	#endif /*version > 2.6.19*/
}

void brick_exit() {

	anaPrint(ANA_DEBUG, "Gates Plugin EXIT\n");
	cleanup();
}

/******************************/
void cleanup(void){
        int ret, kicked = 0;
        struct commGate_s * tmpGate;
	tmpGate = defaultDataGate;
//        kickAllBricks();
        while(tmpGate != NULL){
                int mode = tmpGate->mode;
                switch(mode){
                case UDP:{
                        if( tmpGate->commGateSocket)
                                sock_release( tmpGate->commGateSocket);
                        break;
                }
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                case GENETLINK:{
                        ret = genl_unregister_ops(&ANAFamGN, &dataOpsGN);
                        if(ret != 0){
                                anaPrint(ANA_ERR, "unregister genetlink ops failed: %i\n",ret);
                                return;
                        }
                        /*unregister the family*/
                        ret = genl_unregister_family(&ANAFamGN);
                        if(ret !=0){
                                anaPrint(ANA_ERR, "unregister genetlink family failed: %i\n",ret);
                        }
			kicked = 1;  
                        break;
                }
#endif
                } //switch mode
                tmpGate = tmpGate->next;	
	}

        tmpGate = defaultControlGate;
        while(tmpGate != NULL){
                int mode = tmpGate->mode;
                switch(mode){
                case UDP:{
                        if( tmpGate->commGateSocket)
                                sock_release( tmpGate->commGateSocket);
                        break;
                }
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                case GENETLINK:{
			if(kicked == 0) {
                       		ret = genl_unregister_ops(&ANAFamGN, &controlOpsGN);
                        	if(ret != 0){
                                	anaPrint(ANA_ERR, KERN_ERR "unregister ops genetlink failed: %i\n",ret);
                                	return;
                        	}
                        	/*unregister the generic netlink family*/
                        	ret = genl_unregister_family(&ANAFamGN);
                        	if(ret !=0){
                                	anaPrint(ANA_ERR, KERN_ERR "unregister genetlink family failed%i\n",ret);
                        	}
                        	break;
			}
                }
#endif
                }
                tmpGate = tmpGate->next;	
	}
	anaPrint(ANA_DEBUG, "gates have been cleaned up\n");
}


int brick_start() {

	int errno = 0;
	int argc = 0, i = 0;
	char *IP = NULL, *portChar = NULL, *urn = NULL;
	struct commGate_s *gate = NULL; 
	char *datarep;
        struct commGate_s *datagate;

	anaPrint(ANA_DEBUG, "Gates plugin starting up !\n");
	
	while(getAuxArg(argc) != NULL) {
		anaPrint(ANA_DEBUG, "argv %i %s\n", argc, getAuxArg(argc));
		argc++;
	}

	if (initializeATTT (&attBricksTable, 10) < 0 ) {
		anaPrint(ANA_EMERG, "Unalbe to initialize the attached brick's table !");
		anaL0_stopANA(1);
	}
	anaPrint(ANA_DEBUG, "argc = %i\n", argc);	
	/* parse the arguments */
	while (i < argc) {
		char *argument = getAuxArg(i);
		char *ptr = argument;
		char *modeChar = NULL;
		int mode = -1;
		anaPrint(ANA_DEBUG, "gate %s\n", argument);
		/*check which gate it is*/
		switch(*(ptr)) {
		case 'c': {
			anaPrint(ANA_DEBUG, "control gate\n");
			/* control gate */
   			urn = malloc(strlen(ptr) + 1);
			anaPrint(ANA_DEBUG, "len = %i\n", strlen(ptr));
                	strcpy(urn, ptr);
		        modeChar = ptr + 2;
              		if (modeChar == NULL){
                        	anaPrint(ANA_EMERG, "No control mode specified");
                	}
                            
               		if(!strncmp(modeChar,"udp",3)|| !strncmp(modeChar,"UDP",3))
                        	mode = UDP;
                
                	if(!strncmp(modeChar,"genetlink",9)|| !strncmp(modeChar,"GENETLINK",9))
                        	mode = GENETLINK;
                
                	gate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
                	anaPrint(ANA_DEBUG, "mode: %i\n", mode);
			switch (mode) {
               		case UDP: {
                   	     	char *IP;
                        	char *portChar;
                        	int port;
                        	struct sockaddr_in ctrlSock;
                        	struct socket *newsocket = NULL;
                        	strsep(&ptr, "/");
                        	strsep(&ptr, "/");
                        	IP = strsep(&ptr, ":");
                        	if (IP == NULL){
                                	anaPrint(ANA_ERR, "No IP address specified");
                          		cleanup();
                                	return -EINVAL;

                                }
                        
				portChar = strsep(&ptr, ":");
                        	if (portChar == NULL){
                               		anaPrint(ANA_ERR, "No UDP port specified");
                          		cleanup();
                                	return -EINVAL;

                                }
                       
                        	port = atoi(portChar);
                        	anaPrint(ANA_ERR, "IP: %s, port %i, urn: %s \n",IP, port, urn);
                        	gate->mode = UDP;
                        	gate->urn = urn;
                        
                        	if ((errno = sock_create( PF_INET, SOCK_DGRAM, IPPROTO_UDP, &newsocket)) < 0) {
                                	anaPrint(ANA_ERR,  KERN_ERR "server: Error creating controlsocket\n" );
                                }
                        
                        	gate->commGateSocket = newsocket;
			
                        	memset(&ctrlSock, 0, sizeof(ctrlSock));
                        	ctrlSock.sin_family = AF_INET;
                        	ctrlSock.sin_addr.s_addr = htonl(INADDR_ANY);
                        	ctrlSock.sin_port = htons(port);
                        	anaPrint(ANA_DEBUG, "Binding control socket to UDP port %d\n", port);
                        
                        	/* errno = kernel_bind(gate->commGateSocket, (struct sockaddr *) &ctrlSock,  sizeof(ctrlSock));
			 	* for compatability reasons the ops->bind is used instead of the new kernel_bind function
			 	* kernel_bind is anyway only a wrapper for ops->bind
			 	*/
                        	errno = gate->commGateSocket->ops->bind(gate->commGateSocket, (struct sockaddr *) &ctrlSock,  sizeof(ctrlSock));
                        	if(errno) {
                        	        sock_release(newsocket);
                        	        gate->commGateSocket  = NULL;
                        	        anaPrint(ANA_ERR, "error binding control socket\n");
                          		cleanup();
                                	return -EINVAL;

                         	}
                        	gate->commGateSocket->sk->sk_data_ready = cb_control;
                        	gate->next = NULL;
                        	addControlGate(gate);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                        	INIT_WORK(&wq_ctr.worker, parseControlTemp);
#else
	                        INIT_WORK(&wq_ctr.worker, parseControlTemp, &wq_ctr);
#endif	
        	                break;
                	}
                	case PIPES:{
                	        break;
               		}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                	case GENETLINK:{
			//	memcpy(ANAFamGN.name, ptr + 2, strlen(ptr + 2)); /* +2 corresponds to c= */
			//	anaPrint(ANA_DEBUG, "---------------> %s %i\n", ptr + 2, strlen(ptr + 2));
				anaPrint(ANA_DEBUG, "genetlink\n");

                        	errno = genl_register_family(&ANAFamGN);
                        	if (errno != 0)
                        	        goto failurecontrol;
                        	anaPrint(ANA_ERR, "GN family ID : %i\n", ANAFamGN.id);
                        	/*register functions (commands) of the new family*/

                        	errno = genl_register_ops(&ANAFamGN, &controlOpsGN);
                        	if (errno != 0){
                                	anaPrint(ANA_ERR, "register genetlink ops failed: %i\n",errno);
                                	genl_unregister_family(&ANAFamGN);
                                	goto failurecontrol;
                        	}
                                                
                        	gate->mode = GENETLINK;
                        	gate->urn = "genetlink://ANA"; //urn;
                        	gate->descriptor = ANAFamGN.id;
                        	gate->next = NULL;
		
                        	addControlGate(gate);
               
	         		break;                              
                
			failurecontrol:
                        	anaPrint(ANA_ERR, "error on genetlink registration\n");
                        	cleanup();
                        	return errno;
                	}
#endif /*version > 2.6.19 */
              		default: {
                        	anaPrint(ANA_NOTICE, "Unknown communication Mode %s\n", modeChar); 
				cleanup();
                        	return -EINVAL;
                	}
			} /*end switch control mode */
               		break; 
		} /* end case control gate */
		
		case 'd': {
			datarep = malloc(strlen(ptr)+1);
                	strcpy(datarep, ptr);
               
			anaPrint(ANA_DEBUG, "data gate\n");
			/* control gate */
   			urn = malloc(strlen(ptr) + 1);
			anaPrint(ANA_DEBUG, "len = %i\n", strlen(ptr));
                	strcpy(urn, ptr);
		        modeChar = ptr + 2;
              		if (modeChar == NULL){
                        	anaPrint(ANA_EMERG, "No control mode specified");
				cleanup();
                                return -EINVAL;

                	}

                
                	if(!strncmp(modeChar,"udp",3)|| !strncmp(modeChar,"UDP",3))
                        	mode = UDP;
                	if(!strncmp(modeChar,"genetlink",9)|| !strncmp(modeChar,"GENETLINK",9))
                	        mode = GENETLINK;
              
			anaPrint(ANA_DEBUG, "data mode %i\n", mode); 
		        datagate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
                
                	switch (mode) {
                	case UDP: {
                        	int lenIP, lenPort, lenURN;
                        	int port;
                        	struct sockaddr_in dataSock;
                        	struct socket *newsocket;
                        
                   		IP = NULL;
                   	     	IP = strsep(&ptr, "/");
                   	     	IP = strsep(&ptr, "/");
                   	     	IP = strsep(&ptr, ":");
                        	if (IP == NULL){
                        	        anaPrint(ANA_ERR, "No IP address specified 1");
                                	cleanup();
                                	return -EINVAL;
                        	}
                        	portChar = NULL;
                        	portChar = strsep(&ptr, ":");
                        	if (portChar == NULL){
                                	anaPrint(ANA_ERR, "No UDP port specified");
                              		cleanup();
                               		return -EINVAL;
                        	}
                        
                        	port = atoi(portChar);
                        
                        	lenIP = strlen(IP);
                        	lenPort = strlen(portChar);
                        	lenURN = lenIP + lenPort + 2; /* +1 for : and ending 0*/
                        	/* the representation must be < 256 bytes since its size is coded on 1 byte in the attach message */
                        	if (lenURN > 255){
                                	anaPrint(ANA_ERR, "the data Gate representaion is too big "
                                	             "to be sent in the attach message, must be < 256\n");
                                	cleanup();
                                	return -EINVAL;
                        	}					
                        
                        	urn = (char *) malloc(lenURN * sizeof(char));
                       		memcpy(urn, IP, lenIP );
                        	urn[lenIP] = ':';
                        	memcpy(urn+lenIP+1, portChar, lenPort );
                        	urn[lenURN-1]=0;
                        	datagate->urn = urn;
                        	anaPrint(ANA_ERR, "IP: %s %i port: %s %i  representation: %s %i \n", 
                        	                  IP, lenIP, portChar, lenPort, urn, lenURN);
                    		datagate->mode = UDP;
                        
                        	if ((errno = sock_create( PF_INET, SOCK_DGRAM, IPPROTO_UDP, &newsocket)) < 0) {
                        	        anaPrint(ANA_ERR,  KERN_ERR "server: Error creating datasocket\n" );
                         	      	cleanup();
                         	      	return errno;
                        	}
                        
                        	datagate->commGateSocket = newsocket;	
                        
                        
                        	memset(&dataSock, 0, sizeof(dataSock));
                        	dataSock.sin_family = AF_INET;
                        	dataSock.sin_addr.s_addr = htonl(INADDR_ANY);
                        	dataSock.sin_port = htons(port);
                        	anaPrint(ANA_DEBUG, "Binding data socket to UDP port %d\n", port);
                        
                        	errno = datagate->commGateSocket->ops->bind(datagate->commGateSocket,  
							(struct sockaddr *)&dataSock,  sizeof(dataSock));
                       		if(errno) {
                                	sock_release(newsocket);
                                	datagate->commGateSocket  = NULL;
                                	anaPrint(ANA_ERR, "Error binding data socket\n");
                              		cleanup();
                              	 	return errno;
                        	}
                        	datagate->commGateSocket->sk->sk_data_ready = cb_data;
                        	datagate->next = NULL;
                       		addDataGate(datagate);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                        	INIT_WORK(&wq_data.worker, parseDataTemp);
#else
                        	INIT_WORK(&wq_data.worker, parseDataTemp, &wq_data);
#endif
                        	break;
                	}
                	case PIPES:{
                		break;
               		}
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
                	case GENETLINK:{
                        	anaPrint(ANA_DEBUG, "Data mode is genetlink\n");
                        
                       		errno = genl_register_ops(&ANAFamGN, &dataOpsGN);
                        	if (errno != 0){
                                	anaPrint(ANA_ERR, KERN_ERR "register genetlink ops failed: %i\n",errno);
                                	genl_unregister_family(&ANAFamGN);
                                	goto failure;
                        	}
                        	datagate->mode = GENETLINK;
                        	datagate->urn = "genetlink://ANA"; //datarep;
                        	datagate->descriptor = ANAFamGN.id;
                        	datagate->next = NULL;
                        	addDataGate(datagate);
                        	break;                              
                	failure:
                        	anaPrint(ANA_ERR, "error on registration of genetlink\n");
                       		cleanup();
                       		return errno;
				
                	}
#endif /*version > 2.6.19*/
                	default: {
                       		anaPrint(ANA_NOTICE, "Unknown communication Mode %s\n", modeChar);
                        	cleanup();
				return -EINVAL;
                	}
			} /*end switch data mode */
			break;
		} /* end case data gate */

		case 'D': /* option to set Debug mode on */
		{
			anaPrint(ANA_DEBUG, "change debug mode\n");
	//	#ifdef blub
			ptr += 2; /* +2 for the D= */	
			if(!memcmp(ptr, "ANA_NONE", 8))
				anaDebugLevel = ANA_NONE;
			if(!memcmp(ptr, "ANA_EMERG", 9))
				anaDebugLevel = ANA_EMERG;
			if(!memcmp(ptr, "ANA_ERR", 7))
				anaDebugLevel = ANA_ERR;
			if(!memcmp(ptr, "ANA_NOTICE", 10))
				anaDebugLevel = ANA_NOTICE;
			if(!memcmp(ptr, "ANA_DEBUG", 9))
				anaDebugLevel = ANA_DEBUG;
	//	#endif
			break;
		}
		
		default:
			anaPrint(ANA_DEBUG, "received unknown argument\n");
			cleanup();
			return -EINVAL;
        	} /* end switch argument */

		i++;

	}/* end argument loop*/
	anaPrint(ANA_DEBUG, "end of brick_start\n");
	return 0;
}

#endif

/*****************************************/
/*            user space                 */
/*****************************************/
#ifdef USER
void gatesDataLoop(void) {
	int ret, nbRecv,ackFlag;
	fd_set readFd;

	void *buff =  malloc(DEFAULT_UDP_MTU*sizeof(char));
	struct commGate_s *tmp;
	
	anaLabel_t *label = NULL;
	anaLabel_t labelVal;   
        anaHandle_t *handle = NULL;
        anaHandle_t handleVal;
	
	void *data=NULL;
        int dataLength;
	char *dummy = NULL;
	char ackBuf[2];
	while(attached) {

		FD_ZERO(&readFd);
		tmp = defaultDataGate;
		while(tmp!=NULL){
			FD_SET(tmp->descriptor,&readFd);
			tmp = tmp->next;
		}

		ret = select(255,&readFd,NULL,NULL,NULL);

		if(ret>0) {	
			tmp = defaultDataGate;
			while(tmp!=NULL){
				if (FD_ISSET(tmp->descriptor, &readFd)) {
						nbRecv = read(tmp->descriptor, buff, DEFAULT_UDP_MTU);
					if(nbRecv < (sizeof(anaLabel_t) + sizeof(anaHandle_t)+1)) { /* +1 for the ack flag byte*/
						anaPrint(ANA_ERR, "Data listenLoop : Wrong message Header\n");
						continue;
					}
					
					label = (anaLabel_t *)buff;
					labelVal = (anaLabel_t) ntohl(*label);
					handle = (anaHandle_t *) (buff + sizeof(anaLabel_t));
        				handleVal = (anaHandle_t) ntohl(*handle);
					
					data = buff + sizeof(anaLabel_t) + sizeof(anaHandle_t) +1; /* +1 for the ack flag byte */
       					dataLength = nbRecv  - sizeof(anaHandle_t) - sizeof(anaLabel_t) -1;/* -1 for the ack flag byte */
					ret = dispatchData(myHandle, labelVal, handleVal, data, dataLength); /* do we send a notification in case of failure ? */
					dummy = (char *) (buff+sizeof(anaLabel_t) + sizeof(anaHandle_t));
					ackFlag = dummy[0];
					if(ackFlag){ /* if the ACK flag is on*/
						if(ret<0){
							ackBuf[0] = 'e';
							ackBuf[1] = ret;
						}
						else{
							ackBuf[0] = 'a'; /* a for ack(success)*/
							ackBuf[1] = 0;
						}
						relayControl(handleVal, ackBuf, 2);	
					}
       					
										
				}
				tmp = tmp->next;
			}
		}	
	}

	free(buff);	
}

void brick_exit() {

	anaPrint(ANA_DEBUG, "Gates Plugin EXIT\n");
	/* we close open descriptors */
	while(defaultDataGate!=NULL){
		close(defaultDataGate->descriptor);
		if(defaultDataGate->mode==UNIX || defaultDataGate->mode==PIPES)
			unlink(defaultDataGate->urn);
		defaultDataGate=defaultDataGate->next;
	}
	
	while(defaultControlGate!=NULL){
		close(defaultControlGate->descriptor);
		if(defaultControlGate->mode==UNIX || defaultControlGate->mode==PIPES)
			unlink(defaultControlGate->urn);
		
		defaultControlGate=defaultControlGate->next;
	}
}

int brick_start() {

	int cflag = 0;
	int dflag = 0;

	int argc=0,i, port, ret;
	char **argv=NULL; 
	char c;
	int mode, nbRead;
	char *modeChar = NULL;
	char *IP = NULL, *portChar = NULL, *pipePath = NULL, *urn = NULL;
	struct commGate_s *gate = NULL; 
	struct commGate_s *tmpGate = NULL; 
	char buff[CTRLMSG_MAXSIZE];
	fd_set readFd;
	anaPrint(ANA_DEBUG, "Gates plugin starting up !");
	
	
	/* first we rearrange the aux args in argc,argv style in order to use getopt (makes life easier)*/
	while(getAuxArg(argc) != NULL)
		argc++;
	/* we add +1 to simulate the name of the program */	
	argc++;	
	
	argv = (char **) malloc(sizeof(void *) * argc);
	
	argv[0] = "dummy";
	 
	for(i=0; i<argc-1;i++){
		argv[i+1] = getAuxArg(i);
		anaPrint(ANA_DEBUG, "aux %s\n", argv[i+1]);
		
	}
	
	
	if (initializeATTT (&attBricksTable, 10) < 0 ) {
		anaPrint(ANA_EMERG, "Unalbe to initialize the attached brick's table !");
		anaL0_stopANA(1);
	}	
	
	
	while ((c = getopt (argc, argv, "c:d:D:")) != -1)
		switch (c) { 
			case 'c':
    			{   
				cflag=1;
				mode=-1;
				modeChar = strtok((char*)optarg, "://");
				if (modeChar == NULL){
					anaPrint(ANA_EMERG, "No control mode specified");
					anaL0_stopANA(1);
				}
				
				if(!strncmp(modeChar,"udp",3)|| !strncmp(modeChar,"UDP",3))
					mode = UDP;
				
				if(!strncmp(modeChar,"pipe",4)|| !strncmp(modeChar,"PIPE",4))
					mode = PIPES;
				
				if(!strncmp(modeChar,"unix",4)|| !strncmp(modeChar,"UNIX",4))
					mode = UNIX;
				
				gate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
				
				switch (mode) {
				case UDP: {
					struct sockaddr_in ctrlSock;
					IP = strtok(NULL, "://");
					if (IP == NULL){
						anaPrint(ANA_ERR, "No IP address specified");
						anaL0_stopANA(1);
					}
										
					portChar  = strtok(NULL, ":");
					if (portChar == NULL){
						anaPrint(ANA_ERR, "No UDP port specified");
						anaL0_stopANA(1);
					}
					
					port = atoi(portChar);
					gate->mode = UDP;
					gate->descriptor = socket(PF_INET, SOCK_DGRAM, 0);
					gate->urn= (char*)malloc(strlen(IP)+strlen(portChar)+2);
					if(gate->urn!=NULL)
						sprintf(gate->urn,"%s:%s",IP,portChar);

					memset(&ctrlSock, 0, sizeof(ctrlSock));
					ctrlSock.sin_family = AF_INET;	
					ctrlSock.sin_addr.s_addr = inet_addr(IP); /*htonl(INADDR_ANY); */
					ctrlSock.sin_port = htons(port);	
			
					anaPrint(ANA_DEBUG, "Binding control socket to UDP port %d\n", port);
			
					if(bind(gate->descriptor, (struct sockaddr *) &ctrlSock, sizeof(struct sockaddr))) {
						anaPrint(ANA_ERR, "Unable to bind control socket to UDP  port %d\n", port);
				  	    anaL0_stopANA(1);
					}
					
					addControlGate(gate);
					break;
				}
				case UNIX: {
					struct sockaddr_un ctrlSock;
					char *sockPath = optarg + 7; /* +7 for unix:// */
					gate->urn = (char *) malloc(strlen(sockPath) + 1);										
					if(gate->urn!=NULL)
						strcpy(gate->urn, sockPath);
					gate->mode = UNIX;
					gate->descriptor = socket( AF_LOCAL, SOCK_DGRAM, 0);
					memset(&ctrlSock, 0, sizeof(ctrlSock));
					
					ctrlSock.sun_family = AF_LOCAL;
        				strcpy(ctrlSock.sun_path, sockPath);
					
					unlink(sockPath); 	
					anaPrint(ANA_DEBUG, "Binding UNIX control socket to %s\n", sockPath);
					if(bind(gate->descriptor, (struct sockaddr *) &ctrlSock, SUN_LEN(&ctrlSock))) {
						anaPrint(ANA_ERR, "Unable to bind UNIX control socket \n");
				  	    anaL0_stopANA(1);
					}
					
					addControlGate(gate);
					break;
				}
				case PIPES:{
					int dummyDescriptor;
					gate->mode = PIPES;
					pipePath = optarg + 7; /* +7 for pipe:// */
					gate->urn = (char *) malloc(strlen(pipePath+1));
					if(gate->urn!=NULL)
						strcpy(gate->urn, pipePath);
					
					ret = mkfifo(pipePath, 0666);  /* the pipe is created with access rights to other users */
					if ((ret == -1) && (errno != EEXIST)) {
    					anaPrint(ANA_ERR, "Error creating control named pipe errno=%d\n",errno);
						anaL0_stopANA(1);
					}	
					gate->descriptor = open(pipePath, O_NONBLOCK|O_RDONLY);
					if(gate->descriptor <0) {
						anaPrint(ANA_ERR,  "Unable to open Control gate descriptor (PIPES) \n");
						anaL0_stopANA(1);
					}
					
					/* we open the pipe in write mode : [WARNING] we open the write mode to always have a writing descriptor pointing at the fifo
					 otherwise, on all bricks detachment, the pipes would be no more useable (even worse with select causing tight loop),
					 in which case we would have to close the pipe and reopen it => therefore we open this dummy descriptor, DONT USE IT TO WRITE */
					 
					dummyDescriptor =  open(pipePath, O_NONBLOCK|O_WRONLY);
					if(dummyDescriptor <0) {
						anaPrint(ANA_ERR,  "impossible to open dummy descriptor with write mode on PIPE \n");
						anaL0_stopANA(1);
					} 
					
					addControlGate(gate);
					break;
				}
				default: {anaPrint(ANA_NOTICE, "Unknown communication Mode %s\n", modeChar); anaL0_stopANA(1);}
				}
								
							
       			break;
			}   
			case 'd':
			{	
				dflag=1;
				mode=-1;
				urn = malloc((strlen((char*)optarg)+1)* sizeof(char));
				strcpy(urn, (char*)optarg);
				
				modeChar = strtok((char*)optarg, "://");
				if (modeChar == NULL){
					anaPrint(ANA_ERR, "No data mode specified");
					anaL0_stopANA(1);
				}
				
				if(!strncmp(modeChar,"udp",3)|| !strncmp(modeChar,"UDP",3))
					mode = UDP;
				
				if(!strncmp(modeChar,"pipe",4)|| !strncmp(modeChar,"PIPE",4))
					mode = PIPES;
				
				if(!strncmp(modeChar,"unix",4)|| !strncmp(modeChar,"UNIX",4))
					mode = UNIX;
			
				gate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
				
				switch (mode) {
				case UDP: {
					struct sockaddr_in dataSock;
					int lenIP, lenPort, lenurn;
					IP = strtok(NULL, "://");
					if (IP == NULL){
						anaPrint(ANA_ERR, "No IP address specified");
						anaL0_stopANA(1);
					}
					portChar  = strtok(NULL, ":");
					if (portChar == NULL){
						anaPrint(ANA_ERR, "No UDP port specified");
						anaL0_stopANA(1);
					}
					
					lenIP = strlen(IP);
					lenPort = strlen(portChar);
					lenurn = lenIP + lenPort + 2; /* +1 for : and ending 0*/
					/* the urn must be < 256 bytes since its size is coded on 1 byte in the attach message */
					if (lenurn > 255){
						anaPrint(ANA_ERR, "the data Gate representaion is too big to be sent in the attach message, must be < 256\n");
						anaL0_stopANA(1);
					}					
					
					urn = (char *) malloc(lenurn* sizeof(char));
					memcpy(urn, IP, lenIP );
					urn[lenIP] = ':';
					memcpy(urn+lenIP+1, portChar, lenPort );
					urn[lenurn-1]=0;
					gate->urn = urn;
										
					port = atoi(portChar);
					
					gate->mode = UDP;
					gate->descriptor = socket(PF_INET, SOCK_DGRAM, 0);
					
					

					memset(&dataSock, 0, sizeof(dataSock));
					dataSock.sin_family = AF_INET;	
					dataSock.sin_addr.s_addr = inet_addr(IP);
					dataSock.sin_port = htons(port);	
		
					anaPrint(ANA_DEBUG, "Binding data socket to UDP port %d\n", port);
			
					if(bind(gate->descriptor, (struct sockaddr *) &dataSock, sizeof(struct sockaddr))) {
						anaPrint(ANA_ERR, "Unable to bind control socket to UDP  port %d\n", port);
				  	    anaL0_stopANA(1);
					}
					
					addDataGate(gate);
					break;
				}
				
				case UNIX: {
					struct sockaddr_un dataSock;
					char *sockPath = optarg + 7; /* +7 for unix:// */
					
					gate->urn = NULL;										
					gate->mode = UNIX;
					gate->descriptor = socket( AF_LOCAL, SOCK_DGRAM, 0);
					memset(&dataSock, 0, sizeof(dataSock));
					
					gate->urn = (char *) malloc(strlen(sockPath)+1);
					strcpy(gate->urn, sockPath);
					
					
					dataSock.sun_family = AF_LOCAL;
        				strcpy(dataSock.sun_path, sockPath);
						
					anaPrint(ANA_DEBUG, "Binding UNIX data socket to %s\n", sockPath);
					
					unlink(sockPath);
					
					if(bind(gate->descriptor, (struct sockaddr *) &dataSock, SUN_LEN( &dataSock))) {
						anaPrint(ANA_ERR, "Unable to bind UNIX data socket\n");
				  	    anaL0_stopANA(1);
					}
					
					addDataGate(gate);
					break;
				}
				case PIPES:{
					int dummyDescriptor, pathLen;
					gate->mode = PIPES;
					pipePath = optarg + 7; /* +7 for pipe:// */
					pathLen = strlen(pipePath)+1;
					/* the urn must be < 256 bytes since its size is coded on 1 byte in the attach message */
					if (pathLen > 255){
						anaPrint(ANA_ERR, "the data Gate representaion is too big to be sent in the attach message, must be < 256\n");
						anaL0_stopANA(1);
					}		
					
					gate->urn = (char *) malloc(pathLen);
					strcpy(gate->urn, pipePath);
										
					ret = mkfifo(pipePath, 0666);  /* the pipe is created with access rights to other users */
					if ((ret == -1) && (errno != EEXIST)) {
    					anaPrint(ANA_ERR, "Error creating data named pipe errno=%d\n",errno);
						anaL0_stopANA(1);
					}	
					gate->descriptor = open(pipePath, O_NONBLOCK|O_RDONLY);
					if(gate->descriptor <0) {
						anaPrint(ANA_ERR,  "Unable to open data gate descriptor (PIPES) \n");
						anaL0_stopANA(1);
					}
					
					/* we open the pipe in write mode : [WARNING] we open the write mode to always have a writing descriptor pointing at the fifo
					 otherwise, on all bricks detachment, the pipes would be no more useable (even worse with select causing tight loop),
					 in which case we would have to close the pipe and reopen it => therefore we open this dummy descriptor, DONT USE IT TO WRITE */
					 
					dummyDescriptor =  open(pipePath, O_NONBLOCK|O_WRONLY);
					if(dummyDescriptor <0) {
						anaPrint(ANA_ERR,  "impossible to open dummy descriptor with write mode on PIPE \n");
						anaL0_stopANA(1);
					} 
					
					addDataGate(gate);
					break;
				}
				default: {anaPrint(ANA_NOTICE, "Unknown communication Mode %s\n", modeChar); anaL0_stopANA(1);}
				}
       				break;
			}	
			
			case 'D': /* option to set DEbug mode on */
			{	
				if(!memcmp(optarg, "ANA_NONE", 8))
					anaDebugLevel = ANA_NONE;
				if(!memcmp(optarg, "ANA_EMERG", 9))
					anaDebugLevel = ANA_EMERG;
				if(!memcmp(optarg, "ANA_ERR", 7))
					anaDebugLevel = ANA_ERR;
				if(!memcmp(optarg, "ANA_NOTICE", 10))
					anaDebugLevel = ANA_NOTICE;
				if(!memcmp(optarg, "ANA_DEBUG", 9))
					anaDebugLevel = ANA_DEBUG;
				break;
			}
			case '?':
			{	anaPrint(ANA_NOTICE, "Unknown option : -%c \n" , optopt);
				anaL0_stopANA(1);
			}
			default:
	  			abort (); 
		}
		
		
		/*if some gate is not indicated, we open a default unix sockets gate */
		if(!cflag){
			struct sockaddr_un ctrlSock;
			char ch[100];
			pid_t id = getpid();
			gate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
			
			memset(ch,0, 100);
			
		
			sprintf(ch, "/tmp/anaControl_%s_%d",mymodename, id);
			
			gate->urn = (char *) malloc(strlen(ch)+1);										
			if(gate->urn!=NULL)
				strcpy(gate->urn, ch);
			gate->mode = UNIX;
			gate->descriptor = socket( AF_UNIX, SOCK_DGRAM, 0);
			memset(&ctrlSock, 0, sizeof(ctrlSock));
			ctrlSock.sun_family = AF_LOCAL;
        		strcpy(ctrlSock.sun_path, ch);
			
			unlink(ch); 	
			anaPrint(ANA_ERR, "Binding UNIX control socket to %s\n", ch);
			if(bind(gate->descriptor, (struct sockaddr *) &ctrlSock, SUN_LEN(&ctrlSock))) {
				anaPrint(ANA_ERR, "Unable to bind UNIX control socket \n");
				anaL0_stopANA(1);
			}
				
			addControlGate(gate);
			
		}
		
		if(!dflag){
			struct sockaddr_un dataSock;
			char ch[100];
			pid_t id = getpid();
			gate = (struct commGate_s *) malloc(sizeof(struct commGate_s));
			
			memset(ch,0, 100);
	
			sprintf(ch, "/tmp/anaData_%s_%d",mymodename, id);
			
			gate->urn = (char *) malloc(strlen(ch)+1);										
			if(gate->urn!=NULL)
				strcpy(gate->urn, ch);
			gate->mode = UNIX;
			gate->descriptor = socket( AF_UNIX, SOCK_DGRAM, 0);
			memset(&dataSock, 0, sizeof(dataSock));
			dataSock.sun_family = AF_LOCAL;
        		strcpy(dataSock.sun_path, ch);
			
			unlink(ch); 	
			anaPrint(ANA_DEBUG, "Binding UNIX Data socket to %s\n", ch);
			if(bind(gate->descriptor, (struct sockaddr *) &dataSock, SUN_LEN(&dataSock))) {
				anaPrint(ANA_ERR, "Unable to bind UNIX data socket \n");
				anaL0_stopANA(1);
			}
				
			addDataGate(gate);
		}
	

	/* we free the aux arguments and argv */
	for(i=1;i<argc; i++)
		free(argv[i]);
	free(argv);
	
			
	/* First we launch the thread listening on data gates */
	
	anaThread((void *) gatesDataLoop, NULL, "gates_dataLoop");
		      	
	while(1) {
		
		memset(buff, 0, CTRLMSG_MAXSIZE);
		FD_ZERO(&readFd);
		tmpGate = defaultControlGate;
	
		while(tmpGate!=NULL){
			FD_SET(tmpGate->descriptor,&readFd);
			tmpGate = tmpGate->next;
		}
		
		ret = select(255,&readFd,NULL,NULL,NULL);
		
		if(ret>0) {
			tmpGate = defaultControlGate;
			while(tmpGate!=NULL) {
				if(FD_ISSET(tmpGate->descriptor, &readFd)) {
					nbRead = read(tmpGate->descriptor, buff, CTRLMSG_MAXSIZE);
					
					switch (buff[0]) {
       					case 'a':
					        relayAttach(buff, nbRead);
                				break;
  					case 'd':
               					relayDetach(buff, nbRead);
                				break;
       					
					case 'r':
           					relayRegister(buff, nbRead);
                				break;
        				case 'c':
                				relayRedirect(buff, nbRead);
                				break;
					case 'h': 
                				relayHBRequest(buff, nbRead);
						break; 
        				case 'u':
                				relayUnregister(buff, nbRead);
               	 				break;
        				
					case 'v': 
                				relayChangeView(buff, nbRead);
						break;
					
					case 'f':     	
						relayForkData(buff, nbRead);
						break; 
					case 'i':     	
						relayGetInfo(buff, nbRead);
						break; 
					case 's':     	
						relaySetInfo(buff, nbRead);
						break; 	
					case 'n':     	
						handleSetNotifLabel(buff, nbRead);
						break; 	
					case 'b':     	
						relayEventSubscribe(buff, nbRead);
						break; 	
	  				case 'g':     	
						relayEventUnsubscribe(buff, nbRead);
						break; 
					case 'j':     	
						relayEventNotify(buff, nbRead);
						break; 
					default: anaPrint(ANA_NOTICE, "Command not supported! a%sa\n",buff);
					}
					
					memset(buff, 0, CTRLMSG_MAXSIZE);
				}
				tmpGate = tmpGate->next;
			}
		}
	}
		
	return 0;
}
#endif
