/** \file
 * IP header file with necessary data structures, defines (documented in the report) and functions.
 */
#ifndef __IP_BRICK_
#define __IP_BRICK_

#if (defined USER) | (defined NS2)
#include <netinet/ip.h>
#else
#include <linux/ip.h>
#endif


#include "quickRepository.h"
#include "anaCommon.h"

#define MAX_INTERFACES 10

#define XRP_CMD_ROUTEREQ "req"
#define XRP_CMD_ROUTEUP "rup"
#define XRP_CMD_DISCOVER "dsc"	
#define XRP_CMD_CHANGE "chg"
#define XRP_CMD_GET "get"
#define XRP_CMD_CONFIG "cfg"

#define XRP_CLASS_DESCRIP "dsc"
#define XRP_CLASS_IPADDRESS "ipa"
#define XRP_CLASS_IPMASK "ipm"
#define XRP_CLASS_IPNHOP "ipn"
#define XRP_CLASS_METRICS "mtr"


/* flags for ip resolving */
#define IP_CMD 0x96
#define IP_DLV 0x97
#define IP_RES 0x98 /* this is for resolve request */
#define IP_RSP 0x99 /* this is for resolve reply */

/* default size of resolve request identifier */
#define RES_ID_SIZE 4

#define RESOLVE_TIMEOUT 10000

/* regular expression for IPv4 addresses: for publication in the compartment finder */
#define IPv4_REGEXP "^(([[:digit:]]|[1-9][[:digit:]]|1[[:digit:]]{2}|2[0-4][[:digit:]]|25[0-5])\\.){3}([[:digit:]]|[1-9][[:digit:]]|1[[:digit:]]{2}|2[0-4][[:digit:]]|25[0-5])$"

//! struct fullTarget
/*! A data structure to store the properties for communication targets. */
struct fullTarget {
	char *address;			/*!< IP address */
	anaLabel_t label;			/*!< Outgoing interface */
	char *destDescrip;		/*!< IP user application */
	char *source;			/*!< IP source to use */
	uint8_t  type;			/*!< CMD or next header */
	char *srcDescrip;		/*!< Source description */
};

//! struct devAddr
/*! A data structure to store IP addresses of interfaces. */
struct devAddr {
	anaLabel_t dev;			/*!< Interface */
	char *addr;			/*!< IP address */
	char *mask;			/*!< Netmask */
	struct devAddr *next;
};

//! struct resolveRequestList
/*! A data structure to store pending resolve requests. */
struct resolveRequestList{
	char resId[RES_ID_SIZE];	/*!< Resolve ID */
	anaLabel_t replyTo;		/*!< IDP to send the response to */
	void *context;			/*!< Context */
	int contextLen;			/*!< Length of the context */
	int timerId;			/*!< Timer ID */
	void *target;			/*!< Target */
	int targetLen;			/*!< Length of the target */
	char chanType;			/*!< Channel type */
	void *description;		/*!< Source description */
	int descLen;			/*!< Length of the source description */
	struct resolveRequestList *next;
};

//! A function to transform an IP address from dotted quad to hexadecimal notation.
    /*!
      \param dotted IP address in dotted quad notation
      \return IP address in hexadecimal notation
      \sa hexToDottedQuad()
    */
uint32_t dottedQuadToHex(char *dotted) {
	uint32_t hex[4];
	sscanf(dotted, "%d.%d.%d.%d", &hex[0], &hex[1], &hex[2], &hex[3]);
	return (hex[0] << 24) | (hex[1] << 16) | (hex[2] << 8) | hex[3];
}

//! A function to transform an IP address from hexadecimal to dotted quad notation.
    /*!
      \param hex IP address in hexadecimal notation
      \return IP address in dotted quad notation
      \sa dottedQuadToHex()
    */
char * hexToDottedQuad(uint32_t hex) {
	char quad1[4], quad2[4], quad3[4], quad4[4];
	char *buf;
	buf = (char *) malloc(16);
	sprintf(quad1, "%i", (hex & 0xFF000000) >> 24);
	sprintf(quad2, "%i", (hex & 0x00FF0000) >> 16);
	sprintf(quad3, "%i", (hex & 0x0000FF00) >> 8);
	sprintf(quad4, "%i", (hex & 0x000000FF));
	sprintf(buf, "%s.%s.%s.%s", quad1, quad2, quad3, quad4);
	return buf;
}

//! A function to transform a netmask to short notation (e.g. "\24").
    /*!
      \param mask Netmask in hexadecimal notation
      \return Short notation as string
    */
char * num_mask(uint32_t mask) {
	int i;
	char *num;
	uint32_t op = 1 << 31;
	num = (char *) malloc(3);
	for (i=0; i<32; i++) {
		if ((mask & op)!=op) {
			break;
		}
		else {
			op = op + (1 << (30 - i));
		}
	}
	sprintf(num, "%i", i);
	return num;
}

/* default values for the repository */
#define QREP_KEYTAB_DEFAULTMASK 10
#define QREP_ENTRYTAB_DEFAULTMASK 10
#endif
