 /* C/bricks/ip/ip_cfg.c
 *  
 * configures the ip compartment
 * depends on: ip_enc, ip_fwd, dgrmEthBrick 
 * runs in user space
 * KVR entries: none
 * 
 * use:
 *
 * ipconfig -i {ip address} -m {netmask} -e {name of dgrmEthBrick in KVR}
 * 
 * or:
 *
 * ipconfig -g {default gw ip address}
 *
 * or:
 *
 * ipconfig -a {netnumber} -a {netmask} -a {next hop} -a {name of dgrmEthBrick in KVR}
 *
 * (to manually add an arbitrary route to the forwarding table)
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch>
 * 	   Ariane Keller <arkeller@ee.ethz.ch> adapt to release 0.2 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template_ipcfg.h"
#include "anaLib2.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "ipconfig";



/*-------------------------------------------------------------------------------------*/
/* brick specific code starts here */
#include "ip.h"


anaLabel_t ethLabel = 0,  next_hop = 0, ipLabel = 0;


/* called upon leaving the ipconfig */
void brick_exit() {
	return;
}

void addRoute(void) {
	anaLabel_t fwdLabel;
	void *reply;
	char *netnumber_ch;
	int totalLength=0, arglen;
	uint32_t netnumber;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	/* get IDP of fwd brick */
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip_fwd";
	service.valueLen = strlen("ip_fwd")+1;

	fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if (fwdLabel == 0) {
		anaPrint(ANA_EMERG, "fwd brick not found!\n");
		anaL0_stopANA(0); 
	}
	else { /* write to fwd table */
		netnumber = htonl(dottedQuadToHex(getAuxArg(0)) & dottedQuadToHex(getAuxArg(1)));
		netnumber_ch = hexToDottedQuad(htonl(netnumber));
		anaPrint(ANA_NOTICE, "addRoute subnet: %s\n", netnumber_ch);
		
		msg = anaL1_allocateMessage();
		
		totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
		
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to fill command in message!\n");
			free(msg);
			anaL0_stopANA(0); 
		}  
		
		arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, netnumber_ch, strlen(netnumber_ch) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add subnet to message!\n");
			free(msg);
			anaL0_stopANA(0); 
		}
		totalLength += arglen;
		
		arglen = addXRPArg(msg, XRP_CLASS_IPMASK, getAuxArg(1), strlen(getAuxArg(1)) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add mask to message!\n");
			free(msg);
			anaL0_stopANA(0); 
		}
		
		totalLength += arglen;
		
		arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, getAuxArg(2), strlen(getAuxArg(2)) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add next hop to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		/* get IDP of eth brick */
		context.value = ".";
		context.valueLen = 2;
		service.value = getAuxArg(3);
		service.valueLen = strlen(getAuxArg(3))+1;
		
		ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (ethLabel == 0 ) {
			anaPrint(ANA_EMERG, "ethernet brick not found! %s\n", getAuxArg(3));
			anaL0_stopANA(0);
		}
		
		arglen = addXRPArg(msg, XRP_CLASS_LABEL, &ethLabel, sizeof(anaLabel_t));
		if (arglen<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add label to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
		if (anaL1_isError((xrpMsg_t)reply)) {
			anaPrint(ANA_EMERG, "ipconfig: unable to change route!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		else {
			anaPrint(ANA_NOTICE, "successfully changed route!\n");
		}
		free(msg);
	}
	anaPrint(ANA_EMERG, "########################\n");
	anaPrint(ANA_EMERG, "# ipconfig successful! #\n");
	anaPrint(ANA_EMERG, "########################\n");
	anaValidate("ipconfig successful\n");
	anaL0_stopANA(0);
}

void writeToFwd(void) {
	anaLabel_t fwdLabel;
	void *reply;
	char *netnumber_ch;
	int totalLength=0, arglen;
	uint32_t netnumber;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	/* get IDP of fwd brick */
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip_fwd";
	service.valueLen = strlen("ip_fwd")+1;
	
	anaPrint(ANA_DEBUG, "writetofwd\n");
	anaPrint(ANA_DEBUG, "gwCh %s\n", gwCh);
	
	/* get IDP of fwd brick */
	fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	anaPrint(ANA_NOTICE, "anaL2_fin 1\n");
	if (fwdLabel == 0) {
		anaPrint(ANA_NOTICE, "anaL2_fin 2\n");
		anaPrint(ANA_EMERG, "fwd brick not found!\n");
	}
	else { /* write to fwd table */
		anaPrint(ANA_NOTICE, "anaL2_fin 3\n");
		if (!strcmp(gwCh, "empty")) {	/* it is not a gateway command */	
			anaPrint(ANA_NOTICE, "anaL2_fin 4\n");
			
			anaPrint(ANA_DEBUG, "ipCh %s, mCh %s\n", ipCh, mCh);
			netnumber = htonl(dottedQuadToHex(ipCh) & dottedQuadToHex(mCh));
			netnumber_ch = hexToDottedQuad(htonl(netnumber));
			anaPrint(ANA_NOTICE, "writetofwd subnet: %s\n", netnumber_ch);
		}
		else {
			anaPrint(ANA_NOTICE, "anaL2_fin 5\n");
			
			anaPrint(ANA_DEBUG, "gwCh %s\n", gwCh); 
			netnumber_ch = "0.0.0.0"; /* we set the network number to 0, this matches all subnet -> useful for gateway */
			anaPrint(ANA_NOTICE, "writetofwd 2 subnet: %s\n", netnumber_ch);
		}
		anaPrint(ANA_NOTICE, "anaL2_fin 6\n");
			
		msg = anaL1_allocateMessage();
		
		totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
		
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to fill command in message!\n");
			free(msg);
			anaL0_stopANA(0);
		}  
		
		arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, netnumber_ch, strlen(netnumber_ch) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add subnet to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		if (!strcmp(gwCh, "empty")) {
		
			arglen = addXRPArg(msg, XRP_CLASS_IPMASK, mCh, strlen(mCh) + 1);
			if (arglen < 0) {
				anaPrint(ANA_EMERG, "ipconfig: unable to add mask to message!\n");
				free(msg);
				anaL0_stopANA(0);
			}
			totalLength += arglen;
		}
		
		if (!strcmp(gwCh, "empty")) {
		
			arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, (void *) "none", 5);
			if (arglen < 0) {
				anaPrint(ANA_EMERG, "ipconfig: unable to add next hop to message!\n");
				free(msg);
				anaL0_stopANA(0);
			}
			totalLength += arglen;
		}
		else {
			arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, gwCh, strlen(gwCh) + 1);
			if (arglen<0) {
				anaPrint(ANA_EMERG, "ipconfig: unable to add next hop to message!\n");
				free(msg);
				anaL0_stopANA(0);
			}
			totalLength += arglen;
		}
				
		arglen = addXRPArg(msg, XRP_CLASS_LABEL, &ethLabel, sizeof(anaLabel_t));
		if (arglen<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add label to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
		if (anaL1_isError((xrpMsg_t)reply)) {
			anaPrint(ANA_EMERG, "ipconfig: unable to change route!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		else {
			anaPrint(ANA_NOTICE, "successfully changed route\n");
		}
		free(msg);
	}
	anaPrint(ANA_NOTICE, "anaL2_fin 7\n");
	
	anaPrint(ANA_EMERG, "########################\n");
	anaPrint(ANA_EMERG, "# ipconfig successful! #\n");
	anaPrint(ANA_EMERG, "########################\n");
	anaValidate("ipconfig successful2\n");
	anaL0_stopANA(0);
}

void handleLookupReply(char *data, int len, anaLabel_t input, void *aux) {
	void *reply = NULL;
	int totalLength=0, arglen, j;
	xrpMsg_t msg;
	char *eth, *jch;
	anaLabel_t *ipLabelptr = NULL;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 2; 

	anaPrint(ANA_NOTICE, "received lookup reply for ip in the KVR\n");
	
	if (anaL1_isError(data)) { /* if ERROR command */
		anaPrint(ANA_EMERG, "lookup for ip failed -> cannot work without ip!\n");
		anaL0_stopANA(0);
	}
	
	if (anaL1_decResponse(data, XRP_CLASS_LABEL, 0, (void **) &ipLabelptr) < 0) {
		anaPrint(ANA_EMERG, "no ip label found -> cannot work without ip!\n");
		anaL0_stopANA(0);
	}
	ipLabel = *ipLabelptr;
	if (getAuxArg(0)!=NULL && getAuxArg(1)!=NULL && getAuxArg(2)!=NULL &&  getAuxArg(3)!=NULL) {
		anaPrint(ANA_NOTICE, "manually add new route\n");
		addRoute();
		return;
	}
	
	if (!strcmp(gwCh, "empty")) { /* configure eth brick */
		/* get IDP of eth brick */
		context.value = ".";
		context.valueLen = 2;
		service.value = ethCh;
		service.valueLen = strlen(ethCh)+1;

		ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (ethLabel == 0) {
			anaPrint(ANA_EMERG, "ethernet brick not found! %s\n", ethCh);
			anaL0_stopANA(0);
		}
		
		/* send config to ip brick */
		msg = anaL1_allocateMessage();
		
		totalLength = fillXRPCommand(msg, XRP_CMD_CONFIG);
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to fill command in message!\n");
			free(msg);
			anaL0_stopANA(0);
		}  
		
		arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, ipCh, strlen(ipCh) + 1);	
		if (arglen<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add ip address to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		arglen = addXRPArg(msg, XRP_CLASS_IPMASK, mCh, strlen(mCh) + 1);
		if (arglen<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add ip mask to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
				
		arglen = addXRPArg(msg, XRP_CLASS_LABEL, &ethLabel, sizeof(anaLabel_t));
		if (arglen<0) {
			anaPrint(ANA_EMERG, "ipconfig: unable to add label to message!\n");
			free(msg);
			anaL0_stopANA(0);
		}
		totalLength += arglen;
		
		reply = anaL2_requestReply(ipLabel, msg, totalLength, &totalLength, &timeout);
		if(reply == NULL)
			anaPrint(ANA_DEBUG, "didnt get answer\n");
		anaPrint(ANA_DEBUG, "received %s, %i\n", (char *)reply, totalLength);	
		xrpCmd_t command = (xrpCmd_t)reply;
		if (anaL1_isError((xrpMsg_t)reply) && !equalXRPCommands(command, XRP_CMD_PUBLISH)) {
			anaPrint(ANA_EMERG, "--- publish of %s in %s failed!\n", ipCh, ethCh);
			anaL0_stopANA(0);
		}
		else
			anaPrint(ANA_NOTICE, "publish of %s in %s succeeded\n", ipCh, ethCh);
		
		free(msg);
		
		/* write to fwd table */
		writeToFwd();
		
	} else { /* configure default gateway */
		
		/* find default gateway and outgoing interface */
		for (j=1; j<MAX_INTERFACES; j++) {
			eth = (char *) malloc(20);
			jch = (char *) malloc(10);
			strcat(eth, "eth0");
			sprintf(jch, "%i", j);
			strcat(eth, jch);
			context.value = ".";
			context.valueLen = 2;
			service.value = eth;
			service.valueLen = strlen(eth)+1;

			ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
			anaPrint(ANA_DEBUG, "resolved eth %s with result %i\n", eth, ethLabel);
			if (ethLabel != 0) {
				context.value = "*";
				context.valueLen = 2;
				service.value = gwCh;
				service.valueLen = strlen(gwCh)+1;


				next_hop = anaL2_resolve(ethLabel, &context, &service, 'u', &description, &timeout);
				if (next_hop != 0) {
					break;
				}
			}
			free(eth);
			free(jch);
		}
		
		if (ethLabel == 0 || next_hop == 0) {
			anaPrint(ANA_EMERG, "unable to find default gateway!\n");
			anaValidate("unable to find default gateway\n");
			anaL0_stopANA(0);
		}
		
		/* write to fwd table */
		writeToFwd();
	}
}

int brick_start() {
	int ret = 0, lenMsg, oct1 = -1, oct2 = -1, oct3 = -1, oct4 = -1, oct5 = -1, oct6 = -1, oct7 = -1, oct8 = -1;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	validate_init("ip_config"); 

	/* validate user entries */
	if (!strcmp(gwCh, "empty") && strcmp(ipCh, "empty")) {
		sscanf(ipCh, "%i.%i.%i.%i", &oct1, &oct2, &oct3, &oct4);
		sscanf(mCh, "%i.%i.%i.%i", &oct5, &oct6, &oct7, &oct8);
		if (strlen(ipCh)>15 || strlen(ipCh)<7 || oct1<0 || oct1>255 || oct2<0 || oct2>255 || oct3<0 || oct3>255 || oct4<0 || oct4>255 || 
			strlen(mCh)>15 || strlen(mCh)<7 || oct5<0 || oct5>255 || oct6<0 || oct6>255 || oct7<0 || oct7>255 || oct8<0 || oct8>255) {
			anaPrint(ANA_EMERG, "invalid ip address or netmask entered!\n");
			anaL0_stopANA(0);
		} else {
			anaPrint(ANA_NOTICE, "ip address: %s\n", ipCh);
			anaPrint(ANA_NOTICE, "subnetmask: %s\n", mCh);
			anaValidate("ip address %s with subnetmast %s\n", ipCh, mCh);
		}
	}
	else if (strcmp(gwCh, "empty")) {
		sscanf(gwCh, "%i.%i.%i.%i", &oct1, &oct2, &oct3, &oct4);
		if (strlen(gwCh)>15 || strlen(gwCh)<7 || oct1<0 || oct1>255 || oct2<0 || oct2>255 || oct3<0 || oct3>255 || oct4<0 || oct4>255) {
			anaPrint(ANA_EMERG, "invalid ip address for next hop entered!\n");
			anaL0_stopANA(0);
		} else {
			anaPrint(ANA_NOTICE, "default gateway: %s\n", gwCh);
			anaValidate("default gateway %s \n", gwCh);

		}
	}
	
	/* lookup the ip brick by name */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip";
	service.valueLen = strlen("ip")+1;
	lenMsg = anaL1_encResolve(msg, &context, &service, 'u', &description);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "unable to format lookup request for \"ip\" -> cannot work without ip!\n");
		free(msg);
		anaL0_stopANA(0);
	}
	
	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handleLookupReply, NULL, NULL, NULL, 1); 
	if (ret<0) {
		anaPrint(ANA_EMERG, "send lookup msg for \"ip\" failed -> cannot work without ip!\n");
		free(msg);
		anaL0_stopANA(0);
	}
	
	free(msg);
	
	return 1;
}
