/* C/bricks/ip/ip_fwd.c
 *  
 * contains the forwarding table
 * used by: ip_enc, routing process 
 * runs in user space
 * KVR entries: ip_fwd+forward+forwarding+fwdtable
 * 
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	   Ariane Keller <arkeller@ee.ethz.ch> Adaption for release 0.2
 *
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "ip_fwd";



/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "ip.h"

struct QREP fwdtable;
anaLabel_t fwdIDP;

static void freeValue(void *value);
struct context_s pubcontext;
struct service_s pubservice;

/* called upon leaving the ip_fwd */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	
        anaPrint(ANA_NOTICE, "IP_FWD HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	msg = anaL1_allocateMessage();
	
	/* unpublish the entries in the KVR */
	len = anaL1_encCompartmentUnpublish(msg, fwdIDP, &pubcontext, &pubservice);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0){
		anaPrint(ANA_ERR, "problem ending unpublish request for fwdLookup IDP!\n");
	}
	free(msg);
	
	freeQRepository(&fwdtable, freeValue);
	anaL0_unregisterCallback(fwdIDP, 0);
}

void handleDiscover(void *data, int len) {
	char *target, *matched = NULL, *nhop = NULL;
	int targetLen, labelLen, i, size;
	anaLabel_t *replyToptr, match = 0;
	anaLabel_t replyTo;
	xrpMsg_t replyMsg;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &fwdtable;
	uint32_t targetX, comp, compnet, test, mask, tmp;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	if (data==NULL || len<0)
		return;
	
	target =  (char *) getXRPArg((char *) data, XRP_CLASS_IPADDRESS, &targetLen, 0);
	replyToptr =  (anaLabel_t *) getXRPArg((char *) data, XRP_CLASS_LABEL, &labelLen, 0);
	
	anaPrint(ANA_NOTICE, "destination address: %s\n", target);
        
	if (target==NULL || replyToptr == NULL) {
                return;
        }
	replyTo = *replyToptr;
	targetX = dottedQuadToHex(target);
	
	/* longest prefix match */
	tmp = 0xFFFFFFFF;
	mask = 0;
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry!=NULL && strcmp((char *) entry->name, "0.0.0.0")) {
			comp = dottedQuadToHex((char *) entry->name);
			compnet = targetX & dottedQuadToHex(entry->keyWordsList->next->keyWord);
			if (comp==compnet) {
				test = targetX ^ comp;
				if (test<=tmp && htonl(dottedQuadToHex(entry->keyWordsList->next->keyWord))>mask) {
					match = *(anaLabel_t *)entry->value;
					matched = (char *) entry->name;
					nhop = entry->keyWordsList->keyWord;
					tmp = test;
				}
			}
		}
	}
	if (tmp==0xFFFFFFFF) {
		if ((entry = getQREPEntryByName(&fwdtable, "0.0.0.0", 8))!=NULL) {
			match = *(anaLabel_t *)entry->value;
			matched = (char *) entry->name;
			nhop = entry->keyWordsList->keyWord;
		}
		else {
			anaPrint(ANA_ERR, "no default gateway specified!\n");
			return;
		}
	}
	anaPrint(ANA_NOTICE, "matched subnet: %s\n", matched);
	
	/* send answer */
	replyMsg = anaL1_allocateMessage();
	context.value = nhop;
	context.valueLen = strlen(nhop) + 1;
	service.value = matched;
	service.valueLen = strlen(matched) + 1;

	size = anaL1_encData(replyMsg, &match, sizeof(anaLabel_t), &context, &service, 0);
	anaL0_send(replyTo, replyMsg, size);
}

void showFwd(void) {
	int i;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &fwdtable;
	char *tmp;
	
	anaPrint(ANA_NOTICE, "+-----------------------+-----------------------+---------------+\n");
	anaPrint(ANA_NOTICE, "| Subnet\t\t| Next Hop\t\t| IDP\t\t|\n");
	anaPrint(ANA_NOTICE, "+-----------------------+-----------------------+---------------+\n");
	
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry != NULL) {
			anaLabel_t label = *(anaLabel_t *)entry->value;
			tmp = (char *) malloc(25);
			memset(tmp, 0, 25);
			memcpy(tmp, (char *) entry->name, strlen((char *) entry->name));
			if (entry->keyWordsList->next!=NULL) {
				memcpy(tmp + strlen((char *) entry->name), "/", 1);
				memcpy(tmp + 1 + strlen((char *) entry->name), num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord)), strlen(num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord))) + 1);
			}
			if (strlen((char *) entry->name)<14) {
				if (strlen((char *) entry->keyWordsList->keyWord)>7) {
					anaPrint(ANA_NOTICE, "| %s\t\t| %s\t\t| %i\t|\n", tmp, (char *) entry->keyWordsList->keyWord, label);
					free(tmp);
				}
				else {
					anaPrint(ANA_NOTICE, "| %s\t\t| %s\t\t\t| %i\t|\n", tmp, (char *) entry->keyWordsList->keyWord, label);
					free(tmp);
				}
			}
			else {
				if (strlen((char *) entry->keyWordsList->keyWord)>7) {
					anaPrint(ANA_NOTICE, "| %s\t| %s\t\t| %i\t|\n", tmp, (char *) entry->keyWordsList->keyWord, label);
					free(tmp);
				}
				else {
					anaPrint(ANA_NOTICE, "| %s\t| %s\t\t\t| %i\t|\n", tmp, (char *) entry->keyWordsList->keyWord, label);
					free(tmp);
				}
			}
			anaPrint(ANA_NOTICE, "+-----------------------+-----------------------+---------------+\n");
		}
	}
}

void  handleChange(void *data, int len) {
	char *destination, *mask, *nhop;
	int destinationLen, IDPLen, dummy, maskLen, nhopLen, totalSize;
	anaLabel_t IDP, *IDPptr, replyTo, *replyToptr;
	xrpMsg_t replyMsg;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &fwdtable;
	
	if (data==NULL || len<0)
		return;
	
	destination =  (char *) getXRPArg((char *) data, XRP_CLASS_IPADDRESS, &destinationLen, 0);
	mask = (char *) getXRPArg((char *) data, XRP_CLASS_IPMASK, &maskLen, 0);
	nhop = (char *) getXRPArg((char *) data, XRP_CLASS_IPNHOP, &nhopLen, 0);
 	IDPptr =  (anaLabel_t *) getXRPArg((char *) data, XRP_CLASS_LABEL, &IDPLen, 0);
        IDP = *IDPptr;
	if (destination==NULL) {
                anaPrint(ANA_ERR, "handleChange: unable to identify destination!\n");
                return;
        }

        replyToptr = (anaLabel_t*) getXRPArg((xrpMsg_t)data, XRP_CLASS_LABEL, &dummy, 1);
        if (replyToptr == NULL || allocateXRPMsg(&replyMsg)<0) {
                anaPrint(ANA_ERR, "handleChange: unable to allocate memory for reply message!\n");
                return;
        }
	replyTo = *replyToptr;
	entry = getQREPEntryByName(&fwdtable, destination, destinationLen);
	if (entry!=NULL)
		freeQREPEntry(&fwdtable, entry, freeValue);
	if (strcmp(nhop, "%*&delete")) {
		if ((entry = allocateQREPEntry(destination, destinationLen, &IDP, IDPLen))!=NULL && addQREPEntry(rep, entry)>=0) {
			if (mask!=NULL)
				QREPaddKeywordToEntry(entry, mask);
			QREPaddKeywordToEntry(entry, nhop);
			anaPrint(ANA_NOTICE, "successfully changed route: %s\n", (char *) entry->name);
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_CHANGE);
			anaL0_send(replyTo, replyMsg, totalSize);
			showFwd();
		} else {
			anaPrint(ANA_ERR, "unable to change route!\n");
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
			anaL0_send(replyTo, replyMsg, totalSize);
		}
		free(replyMsg);
	}
	else {
		anaPrint(ANA_NOTICE, "successfully deleted route\n");
		totalSize = fillXRPCommand(replyMsg, XRP_CMD_CHANGE);
		anaL0_send(replyTo, replyMsg, totalSize);
		showFwd();
		free(replyMsg);
	}
}

void handleRipreq(void *data, int len) {
	int dummy, i, totalLength, arglen, send = 0;
	anaLabel_t replyTo, *replyToptr;
	xrpMsg_t replyMsg;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &fwdtable;
	
	if (data==NULL || len<0)
		return;
	
	replyToptr = (anaLabel_t *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &dummy, 0);
        if (replyToptr == NULL) {
                anaPrint(ANA_ERR, "handleRipreq: unable to find replyTo label!\n");
                return;
        }
	replyTo = *replyToptr;
	replyMsg = anaL1_allocateMessage();
	totalLength = fillXRPCommand((xrpMsg_t) replyMsg, XRP_CMD_ROUTEREQ);
	
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry!=NULL) {
			if (entry->keyWordsList->next!=NULL) {
				if (!strcmp(entry->keyWordsList->keyWord, "none")) {
					send = 1;
					arglen = addXRPArg((xrpMsg_t) replyMsg, XRP_CLASS_IPADDRESS, (char *) entry->name, strlen((char *) entry->name) + 1);
					if (arglen<0) {
						anaPrint(ANA_ERR, "handleRipreq: unable to add destination to message!\n");
						free(replyMsg);
						return;
					}
					totalLength += arglen;
					arglen = addXRPArg((xrpMsg_t) replyMsg, XRP_CLASS_IPNHOP, (char *) entry->keyWordsList->keyWord, strlen((char *) entry->keyWordsList->keyWord) + 1);
					if (arglen<0) {
						anaPrint(ANA_ERR, "handleRipreq: unable to add target to message!\n");
						free(replyMsg);
						return;
					}
					totalLength += arglen;
					arglen = addXRPArg((xrpMsg_t) replyMsg, XRP_CLASS_IPMASK, (char *) entry->keyWordsList->next->keyWord, strlen((char *) entry->keyWordsList->next->keyWord) + 1);
					if (arglen<0) {
						anaPrint(ANA_ERR, "handleRipreq: unable to add context to message!\n");
						free(replyMsg);
						return;
					}
					totalLength += arglen;
				}
			}
		}
	}
	if (send==1) {
		anaL0_send(replyTo, replyMsg, totalLength);
		anaPrint(ANA_NOTICE, "successfully sent routes to rip\n");
	}
	else {
		totalLength = fillXRPCommand((xrpMsg_t) replyMsg, XRP_CMD_ERROR);
		anaL0_send(replyTo, replyMsg, totalLength);
		anaPrint(ANA_ERR, "no routes for rip detected!\n");
	}
	free(replyMsg);
}

/* receive fwd requests */
void fwd(char *data, int len, anaLabel_t input, void *aux) {
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* discover received */
	if (equalXRPCommands(command, XRP_CMD_DISCOVER)) {
		anaPrint(ANA_NOTICE, "--> received discover request\n");
		handleDiscover(data, len);
	}
	
	/* change received */
	if (equalXRPCommands(command, XRP_CMD_CHANGE)) {
		anaPrint(ANA_NOTICE, "--> received change request\n");
		handleChange(data, len);
	}
	
	/* rip request */
	if (equalXRPCommands(command, XRP_CMD_ROUTEREQ)) {
		anaPrint(ANA_NOTICE, "--> received rip request\n");
		handleRipreq(data, len);
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_DESCRIP, 0, (void**) &name);
	
	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else
		if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
		
	/* create the fwd repository */
	initQRepository(&fwdtable, QREP_ENTRYTAB_DEFAULTMASK, QREP_KEYTAB_DEFAULTMASK);
	
	/* register fwd callback function */
	fwdIDP = anaL0_registerCallback((anaCallback_t) &fwd, NULL, NULL, NULL, 0, IDP_PERM, NOTHREAD);
	if (fwdIDP < 0) {
		anaPrint(ANA_ERR,  "fwdIDP: unable to register callback IDP to ANA minmex!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "registered fwd IDP\n");
	
	/* publish "ip_fwd" in the KVR */
	msg = anaL1_allocateMessage();
	memset(&pubservice, 0, sizeof(struct service_s));
	memset(&pubcontext, 0, sizeof(struct context_s));
	pubcontext.value = ".";
	pubcontext.valueLen = 2;
	pubservice.value = "ip_fwd+forward+forwarding+fwdtable";
	pubservice.valueLen = strlen("ip_fwd+forward+forwarding+fwdtable")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, fwdIDP, &pubcontext, &pubservice, 1);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "problem encoding publish request for fwdLookup IDP in KVR!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) &handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "problem sending publish request for fwdLookup IDP in KVR!\n");
		return -1;
	}
	
	free(msg);
	return 1;
}

/* free function which has to be provided for the QREP
 * to free all the entries
 */
static void freeValue(void *value) {
	free(value);
}


