/* C/bricks/ip/ip_sum.c
 *  
 * offers checksum computation
 * used by: ip_enc 
 * runs in user space
 * KVR entries: ip_sum+sum+checksum+chksum
 * 
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	   Ariane Keller <arkeller@ee.ethz.ch> Adaption for release 0.2
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

char mymodename[] = "ip_sum";

/*-------------------------------------------------------------------------------------*/

#include "ip.h"

/* brick specific code starts here */

anaLabel_t recvDataIDP;
struct context_s context;
struct service_s service;

/* called upon leaving the ip_sum */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	
        anaPrint(ANA_NOTICE, "IP_SUM HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	msg = anaL1_allocateMessage();
	/* unpublish the the entries in the KVR */
	len = anaL1_encCompartmentUnpublish(msg, recvDataIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0){
		anaPrint(ANA_ERR, "problem ending unpublish request for recvData IDP!\n");
	}
	free(msg);
	
	anaL0_unregisterCallback(recvDataIDP, 0);
}

/* receive Data function from ip in order to compute the checksum */
void recvData_sum(char *data, int len, anaLabel_t input, void *aux) {
	int count = 10, labelLen, bufferLen, size;
	unsigned long sum = 0;
	anaLabel_t replyTo, *replyToptr;
	xrpMsg_t replyMsg;
	unsigned short *buffer;
	
       	if (data==NULL || len<0)
       		return;
	
	anaPrint(ANA_NOTICE, "--> received checksum request\n");
	
	buffer = (unsigned short *) getXRPArg(data, XRP_CLASS_MESSAGE, &bufferLen, 0);
	replyToptr = (anaLabel_t*) getXRPArg(data, XRP_CLASS_LABEL, &labelLen, 0);
	replyTo = *replyToptr;
	while (count--) {
		sum += htons(*buffer++);
		if (sum & 0xFFFF0000) {
			sum &= 0xFFFF;
			sum++;
		}
	}
	sum = 0xFFFF - sum;
	replyMsg = anaL1_allocateMessage();
	size = anaL1_encData(replyMsg, &sum, 2,  NULL, NULL, 0);
	anaL0_send(replyTo, replyMsg, size);
	anaPrint(ANA_NOTICE, "checksum: 0x%X\n", (unsigned int) sum);
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_DESCRIP, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else
		if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	/* register data receive callback function */
	recvDataIDP = anaL0_registerCallback((anaCallback_t) &recvData_sum, NULL, NULL, NULL, 0, IDP_PERM, NOTHREAD);
	if (recvDataIDP < 0) {
		anaPrint(ANA_ERR,  "recvDataIDP: unable to register callback IDP to ANA minmex!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "registered recvData IDP\n");
	
	/* publish "ip_sum" in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip_sum+sum+checksum+chksum";
	service.valueLen = strlen("ip_sum+sum+checksum+chksum")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvDataIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "problem encoding publish request for recvData IDP in KVR!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) &handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "problem sending publish request for recvData IDP in KVR!\n");
		return -1;
	}
	
	free(msg);
	return 1;
}
