/* C/bricks/ip/rip.c
 *  
 * implements rip routing in ip compartment
 * depends on: ip_enc, ip_fwd
 * runs in user space
 * KVR entries: none
 * 
 * IMPORTANT: use minmex with -b option to extend default message size
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	   Ariane Keller <arkeller@ee.ethz.ch> Adaption for release 0.2
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "rip";

/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "ip.h"

struct QREP riptable;
anaLabel_t ripIDP;
anaLabel_t fwdLabel;
anaLabel_t ipLabel;


static void freeValue(void *value);
void showRip(void);
void hello_rip(void);


#define RIP_UP_TIME 3	/* time between two periodic rip hello messages (in seconds) */
#define RIP_DEL_TIME 3	/* after how many missing hello messages a route will be deleted */

/* called upon leaving the rip */
void brick_exit() {
        anaPrint(ANA_NOTICE, "RIP HALT\n");
/*	
	anaL2_unpublish(ipLabel, ripIDP, "rip+routing", strlen("rip+routing") + 1);
	
	anatimer_exit();

	freeQRepository(&riptable, freeValue);
	anaL0_unregisterCallback(ripIDP, NULL);
	freeLabel(ripIDP);
	freeLabel(ipLabel);
	freeLabel(fwdLabel);
	anaPrint(ANA_NOTICE, "RIP HALTED\n");
*/
}

void handleRip(void *data, int len, char *saddr) {
	int i, desLen, conLen, tarLen, scrLen, j, changed = 0, new_cost, old_cost, ttl = RIP_DEL_TIME, totalLength, arglen, k;
	xrpMsg_t msg;
	anaLabel_t next_hop, ethLabel;
	void *reply;
	char *subnet, *mask, *nhop, *cost, *eth, *jch;
	struct QREPEntry *entry = NULL;
	struct QREPEntry *new_entry = NULL;
	struct QREP *rep = (struct QREP *) &riptable;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 

	i = 0;
	while ((subnet = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_IPADDRESS, &desLen, i))!=NULL) {
		mask = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_IPMASK, &conLen, i);
		nhop = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_IPNHOP, &tarLen, i);
		cost = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_METRICS, &scrLen, i);
		sscanf(cost, "%i", &old_cost);
		
		for (j=0; j<rep->sizeEntryTab; j++) {
			entry = rep->entryTab[j];
			if (entry!=NULL) {
				if (!strcmp(entry->name, subnet) && !strcmp(entry->keyWordsList->next->keyWord, mask)) {
					if (old_cost + 1<*entry->keyWordsList->keyWord && old_cost<=15) {
						/* found a better route */
						freeQREPEntry(&riptable, entry, freeValue);
						if ((new_entry = allocateQREPEntry(subnet, desLen, saddr, strlen(saddr) + 1))!=NULL && addQREPEntry(rep, new_entry)>=0) {
							QREPaddKeywordToEntry(new_entry, (void *) &ttl);
							QREPaddKeywordToEntry(new_entry, (char *) mask);
							new_cost = old_cost + 1;
							QREPaddKeywordToEntry(new_entry, (void *) &new_cost);
							anaPrint(ANA_NOTICE, "route %s/%s changed\n", subnet, num_mask(dottedQuadToHex(mask)));
							changed = 1;
							/* write to fwd table */
							msg = anaL1_allocateMessage();
							totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
							if (totalLength<0) {
								anaPrint(ANA_EMERG, "handleRip: unable to fill command in message!\n");
								free(msg);
								break;
							}  
							arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, subnet, strlen(subnet) + 1);
							if (arglen < 0) {
								anaPrint(ANA_EMERG, "handleRip: unable to add subnet to message!\n");
								free(msg);
								break;
							}
							totalLength += arglen;
							arglen = addXRPArg(msg, XRP_CLASS_IPMASK, mask, strlen(mask) + 1);
							if (arglen < 0) {
								anaPrint(ANA_EMERG, "handleRip: unable to add mask to message!\n");
								free(msg);
								break;
							}
							totalLength += arglen;
							arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, saddr, strlen(saddr) + 1);
							if (arglen < 0) {
								anaPrint(ANA_EMERG, "handleRip: unable to add next hop to message!\n");
								free(msg);
								break;
							}
							totalLength += arglen;
							/* get eth label */
							for (k=1; k<MAX_INTERFACES; k++) {
								eth = (char *) malloc(20);
								jch = (char *) malloc(10);
								strcat(eth, "eth0");
								sprintf(jch, "%i", k);
								strcat(eth, jch);
								context.value = ".";
								context.valueLen = 2;
								service.value = eth;
								service.valueLen = strlen(eth)+1;
								ethLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
								if (ethLabel != 0) {
									context.value = "*";
									context.valueLen = 2;
									service.value = saddr;
									service.valueLen = strlen(saddr)+1;

									next_hop = anaL2_resolve(ethLabel, &context, &service, 'u', &description, &timeout);
									if (next_hop != 0) {
										break;
									}
								}
								free(eth);
								free(jch);
							}
							arglen = addXRPArg(msg, XRP_CLASS_LABEL, &ethLabel, sizeof(anaLabel_t));
							if (arglen<0) {
								anaPrint(ANA_EMERG, "handleRip: unable to add label to message!\n");
								free(msg);
								break;
							}
							totalLength += arglen;
							reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
							if (anaL1_isError((xrpMsg_t)reply)) {
								anaPrint(ANA_EMERG, "handleRip: unable to change route in fwd table!\n");
								free(msg);
								break;
							}
							else {
								anaPrint(ANA_NOTICE, "successfully changed route in fwd table\n");
							}
							free(msg);
							break;
						} else {
							anaPrint(ANA_ERR, "unable to change route!\n");
							break;
						}
					}
					else if (old_cost + 1==*entry->keyWordsList->keyWord && !strcmp(entry->value, saddr)) {
						*entry->keyWordsList->next->next->keyWord = ttl;
						anaPrint(ANA_NOTICE, "updated ttl of route %s/%s\n", subnet, num_mask(dottedQuadToHex(mask)));
						break;
					}
					else {
						/* route is not interesting */
						anaPrint(ANA_NOTICE, "route %s/%s is not interesting\n", subnet, num_mask(dottedQuadToHex(mask)));
						break;
					}
				}
			}
		}
		if (j==rep->sizeEntryTab) {
			/* new route */
			if ((new_entry = allocateQREPEntry(subnet, desLen, saddr, strlen(saddr) + 1))!=NULL && addQREPEntry(rep, new_entry)>=0 && old_cost<=15) {
				QREPaddKeywordToEntry(new_entry, (void *) &ttl);
				QREPaddKeywordToEntry(new_entry, (char *) mask);
				new_cost = old_cost + 1;
				QREPaddKeywordToEntry(new_entry, (void *) &new_cost);
				anaPrint(ANA_NOTICE, "route %s/%s added\n", subnet, num_mask(dottedQuadToHex(mask)));
				changed = 1;
				/* write to fwd table */
				msg = anaL1_allocateMessage();
				totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
				if (totalLength<0) {
					anaPrint(ANA_EMERG, "handleRip: unable to fill command in message!\n");
					free(msg);
					break;
				}  
				arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, subnet, strlen(subnet) + 1);
				if (arglen < 0) {
					anaPrint(ANA_EMERG, "handleRip: unable to add subnet to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				arglen = addXRPArg(msg, XRP_CLASS_IPMASK, mask, strlen(mask) + 1);
				if (arglen < 0) {
					anaPrint(ANA_EMERG, "handleRip: unable to add mask to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, saddr, strlen(saddr) + 1);
				if (arglen < 0) {
					anaPrint(ANA_EMERG, "handleRip: unable to add next hop to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				/* get eth label */
				for (k=1; k<MAX_INTERFACES; k++) {
					eth = (char *) malloc(20);
					jch = (char *) malloc(10);
					strcat(eth, "eth0");
					sprintf(jch, "%i", k);
					strcat(eth, jch);
					context.value = ".";
					context.valueLen = 2;
					service.value = eth;
					service.valueLen = strlen(eth)+1;
					ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
					if (ethLabel != 0) {
						context.value = "*";
						context.valueLen = 2;
						service.value = saddr;
						service.valueLen = strlen(saddr)+1;
						next_hop = anaL2_resolve(ethLabel, &context, &service, 'u', &description, &timeout);
						if (next_hop != 0) {
							break;
						}
					}
					free(eth);
					free(jch);
				}
				arglen = addXRPArg(msg, XRP_CLASS_LABEL, &ethLabel, sizeof(anaLabel_t));
				if (arglen<0) {
					anaPrint(ANA_EMERG, "handleRip: unable to add label to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
				if (anaL1_isError((xrpMsg_t)reply)) {
					anaPrint(ANA_EMERG, "handleRip: unable to change route in fwd table!\n");
					free(msg);
					break;
				}
				else {
					anaPrint(ANA_NOTICE, "successfully changed route in fwd table\n");
				}
				free(msg);
			} else {
				anaPrint(ANA_ERR, "unable to add route!\n");
			}
		}
		i++;
	}
	
	if (changed==1) {
		showRip();
		hello_rip();
	}
}

/* receive rip messages */
void rip(struct anaL2_message *data) {
	int length;
	void *saddr, *message;
	xrpCmd_t command; 	
	anaPrint(ANA_DEBUG, "----------->rip update received<----------------\n");
	showRip();
	
	length = data->dataLen;
	saddr = (char *) data->senderContext;
	message = (char *) data->data;
	
	/* parse the received xrp message */
	command = message;
	
	if (equalXRPCommands(command, XRP_CMD_ROUTEUP)) {
		anaPrint(ANA_NOTICE, "--> received rip hello message or update\n");
		handleRip(message, length, saddr);
	}
}

void showRip() {
	int i;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &riptable;
	char *tmp;
	
	anaPrint(ANA_NOTICE, "+--------------------------+-----------------------+-------+\n");
	anaPrint(ANA_NOTICE, "| Subnet\t\t\t| Next Hop\t\t| Cost\t|\n");
	anaPrint(ANA_NOTICE, "+--------------------------+-----------------------+-------+\n");
	
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry!=NULL) {
			tmp = (char *) malloc(25);
			memset(tmp, 0, 25);
			memcpy(tmp, (char *) entry->name, strlen((char *) entry->name));
			memcpy(tmp + strlen((char *) entry->name), "/", 1);
			memcpy(tmp + 1 + strlen((char *) entry->name), num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord)), strlen(num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord))) + 1);
			if (strlen((char *) entry->name)<14) {
				if (strlen((char *) entry->value)>7) {
					anaPrint(ANA_NOTICE, "| %s\t\t| %s\t\t| %i\t|\n", tmp, (char *) entry->value, (int) *entry->keyWordsList->keyWord);
					free(tmp);
				}
				else {
					anaPrint(ANA_NOTICE, "| %s\t\t| %s\t\t\t| %i\t|\n", tmp, (char *) entry->value, (int) *entry->keyWordsList->keyWord);
					free(tmp);
				}
			}
			else {
				if (strlen((char *) entry->value)>7) {
					anaPrint(ANA_NOTICE, "| %s\t| %s\t\t| %i\t|\n", tmp, (char *) entry->value, (int) *entry->keyWordsList->keyWord);
					free(tmp);
				}
				else {
					anaPrint(ANA_NOTICE, "| %s\t| %s\t\t\t| %i\t|\n", tmp, (char *) entry->value, (int) *entry->keyWordsList->keyWord);
					free(tmp);
				}
			}
			anaPrint(ANA_NOTICE, "+--------------------------+-----------------------+-------+\n");
		}
	}
}

void manage_ttl(void) {
	int i, ttl, new_ttl, totalLength, arglen;
	xrpMsg_t msg;
	void *reply;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &riptable;
	struct timespec timeout;
	timeout.tv_sec = 1; 

	anaPrint(ANA_DEBUG, "--> manage ttl\n");
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry!=NULL && strcmp(entry->value, "none")) {
			ttl = (int) *entry->keyWordsList->next->next->keyWord;
			new_ttl = ttl - 1;
			anaPrint(ANA_NOTICE, "ttl of %s/%s reduced from %i to %i\n", (char *) entry->name, num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord)), ttl, new_ttl);
			if (new_ttl<=0) {
				/* write to fwd table */
				msg = anaL1_allocateMessage();
				totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
				if (totalLength<0) {
					anaPrint(ANA_EMERG, "manageTTL: unable to fill command in message!\n");
					free(msg);
					break;
				}  
				arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, entry->name, strlen(entry->name) + 1);
				if (arglen < 0) {
					anaPrint(ANA_EMERG, "manageTTL: unable to add subnet to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				arglen = addXRPArg(msg, XRP_CLASS_LABEL, NULL, 0);
				if (arglen<0) {
					anaPrint(ANA_EMERG, "manageTTL: unable to add label to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				arglen = addXRPArg(msg, XRP_CLASS_IPNHOP, "%*&delete", 10);
				if (arglen < 0) {
					anaPrint(ANA_EMERG, "manageTTL: unable to add delete command to message!\n");
					free(msg);
					break;
				}
				totalLength += arglen;
				reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
				if (anaL1_isError((xrpMsg_t) reply)) {
					anaPrint(ANA_EMERG, "manageTTL: unable to delete route in fwd table!\n");
					free(msg);
					break;
				}
				else {
					anaPrint(ANA_NOTICE, "successfully deleted route in fwd table\n");
				}
				free(msg);
				anaPrint(ANA_NOTICE, "route %s/%s died\n", (char *) entry->name, num_mask(dottedQuadToHex((char *) entry->keyWordsList->next->keyWord)));
				freeQREPEntry(&riptable, entry, freeValue);
				showRip();
				hello_rip();
			}
			else {
				*entry->keyWordsList->next->next->keyWord = new_ttl;
			}
		}
	}
}

void hello_rip(void) {
	int i, totalLength, arglen, send = 0;
	xrpMsg_t msg;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &riptable;
	anaLabel_t receivers;
	char *cost;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 

	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand((xrpMsg_t) msg, XRP_CMD_ROUTEUP);
	
	for (i=0; i<rep->sizeEntryTab; i++) {
		entry = rep->entryTab[i];
		if (entry!=NULL) {
			send = 1;
			arglen = addXRPArg((xrpMsg_t) msg, XRP_CLASS_IPADDRESS, (char *) entry->name, strlen((char *) entry->name) + 1);
			if (arglen<0) {
				anaPrint(ANA_ERR, "handleRipreq: unable to add subnet to message!\n");
				free(msg);
				return;
			}
			totalLength += arglen;
			arglen = addXRPArg((xrpMsg_t) msg, XRP_CLASS_IPNHOP, (char *) entry->value, strlen((char *) entry->value) + 1);
			if (arglen<0) {
				anaPrint(ANA_ERR, "handleRipreq: unable to add next hop to message!\n");
				free(msg);
				return;
			}
			totalLength += arglen;
			cost = (char *) malloc(5);
			sprintf(cost, "%i", (int) *entry->keyWordsList->keyWord);
			arglen = addXRPArg((xrpMsg_t) msg, XRP_CLASS_METRICS, (char *) cost, strlen((char *) cost) + 1);
			if (arglen<0) {
				anaPrint(ANA_ERR, "handleRipreq: unable to add cost to message!\n");
				free(msg);
				return;
			}
			free(cost);
			totalLength += arglen;
			arglen = addXRPArg((xrpMsg_t) msg, XRP_CLASS_IPMASK, (char *) entry->keyWordsList->next->keyWord, strlen((char *) entry->keyWordsList->next->keyWord) + 1);
			if (arglen<0) {
				anaPrint(ANA_ERR, "handleRipreq: unable to add mask to message!\n");
				free(msg);
				return;
			}
			totalLength += arglen;
		}
	}
	if (send==1) {
		context.value = "*";
		context.valueLen = 2;
		service.value = "rip";
		service.valueLen = strlen("rip")+1;
		receivers = anaL2_resolve(ipLabel, &context, &service, 'b', &service, &timeout);
		if (receivers != 0) {
			anaL0_send(receivers, msg, totalLength);
			anaPrint(ANA_NOTICE, "--> sent rip hello message or update\n");
		}
		else
			anaPrint(ANA_ERR, "no rip nodes found!\n");
	}
	else {
		anaPrint(ANA_ERR, "--> rip table is empty, can't send hello message!\n");
	}
	free(msg);
}

int brick_start() {
	int totalLength, i, targetLength, cost = 0, desLen, contextLength, ttl = RIP_DEL_TIME;
	xrpMsg_t msgFwd;
	char *destination, *target, *contextstr;
	void *reply;
	struct QREPEntry *entry = NULL;
	struct QREP *rep = (struct QREP *) &riptable;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 

	/* create the rip repository */
	initQRepository(&riptable, QREP_ENTRYTAB_DEFAULTMASK, QREP_KEYTAB_DEFAULTMASK);
	
	/* publish "rip" in ip brick */
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip";
	service.valueLen = strlen("ip")+1;
	ipLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if (ipLabel == 0) {
		anaPrint(ANA_ERR, "discover of ip brick failed!\n");	
		return -1;
	}
	context.value = "*";
	context.valueLen = 2;
	service.value = "rip+routing";
	service.valueLen = strlen("rip+routing")+1;
	ripIDP = anaL2_publish(ipLabel, &context, &service, rip, THREAD, &timeout);
	if (ripIDP == 0) {
		anaPrint(ANA_ERR, "publish failed!\n");
		return -1;
	}
	
	/* get IDP of fwd brick */
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip_fwd";
	service.valueLen = strlen("ip_fwd")+1;

	fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if (fwdLabel == 0) {
		anaPrint(ANA_EMERG, "fwd brick not found!\n");
		return -1;
	}
	
	msgFwd = anaL1_allocateMessage();
	totalLength = fillXRPCommand((xrpMsg_t) msgFwd, XRP_CMD_ROUTEREQ);
	if (totalLength<0) {
		anaPrint(ANA_ERR, "unable to fill command in message!\n");
		free(msgFwd);
		return -1;
	}
	reply = anaL2_requestReply(fwdLabel, msgFwd, totalLength, &totalLength, &timeout);
	
	if (!equalXRPCommands((xrpCmd_t) reply, XRP_CMD_ERROR)) {
		i = 0;
		while ((destination = (char *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_IPADDRESS, &desLen, i))!=NULL) {
			target = (char *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_IPNHOP, &targetLength, i);
			contextstr = (char *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_IPMASK, &contextLength, i);
			if ((entry = allocateQREPEntry(destination, strlen(destination) + 1, target, targetLength))!=NULL && addQREPEntry(rep, entry)>=0) {
				QREPaddKeywordToEntry(entry, (void *) &ttl);
				QREPaddKeywordToEntry(entry, (char *) contextstr);
				QREPaddKeywordToEntry(entry, (void *) &cost);
			} else {
				anaPrint(ANA_ERR, "unable to add entry!\n");
			}
			i++;
		}
		showRip();
	}
	else
		return -1;
	
	anatimer_init();
	anatimer_add((1000 * RIP_UP_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)hello_rip, NULL, 0);
	anatimer_add((1000 * RIP_UP_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)manage_ttl, NULL, 0);
	
	return 1;
}

/* free function which has to be provided for the QREP
 * to free all the entries
 */
static void freeValue(void *value) {
	free(value);
}

