/* C/bricks/ip2ana/common/ip2ana_common.c
 *  
 * ANA brick for ip2ana adaptation layer 
 * 
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "brick_template.h"
#include "ip2ana.h"

#define MODULE_NAME ip2ana
char mymodename[]="ip2ana";

struct ip2ana_entry *pub_list = NULL; /* list of published ip addresses */
struct ip2ana_entry *res_list = NULL; /* list of resolved ip addresses */
struct udp_ports uports; /* communication with ip2ana netfilter module */

void brick_exit(void) {

	anaPrint(ANA_NOTICE, "brick_exit called\n");
	
#ifdef USER
	user_close_handlers();
#endif
	free_ip2ana_list(&pub_list);
	free_ip2ana_list(&res_list);
}

int brick_start(void) {

	anaPrint(ANA_NOTICE, "brick_start called\n");
	
	if (get_args() == -1)
	{
		anaPrint(ANA_DEBUG, "brick_start --> using default port numbers\n");
		
		uports.brick_port = htons(BRICK_PORT);
		uports.ipt_port = htons(IPT_PORT);
		uports.i2a_port = htons(I2A_PORT);
		uports.a2i_port = htons(A2I_PORT);	
	}
		
#ifdef USER
	if (user_socket_init(&uports) == -1)
	{
		anaPrint(ANA_ERR, "Cannot initialize ip2ana sockets ... exiting\n");
		anaL0_stopANA(-1);
		return -1;
	}
	
	/* we assume the brick is started from the bin directory */
	
	if (user_file_init("../C/bricks/ip2ana/anahosts.txt") == -1)
	{
		anaPrint(ANA_ERR, "Cannot open anahosts.txt ... exiting\n");
		anaPrint(ANA_ERR, "(need to create it by copying anahosts.txt.template?)\n");
		anaL0_stopANA(-1);
		return -1;
	}
#endif /* USER */
	
	anaL2_initDefault();
	
	publish_listen();
	
#ifdef USER
	user_loop(); /* main receive loop */
#endif
	
	return 1;
}

int get_args(void)
{
	char *arg;
	int tmp, i;
	uint16_t ports[4];
	
	for (i=0; i<4; i++)
	{
		if ((arg = getAuxArg(i)) == NULL)
		{
			anaPrint(ANA_ERR, "get_args --> cannot extract #%d port number!\n", i);
			return -1;
		}
		tmp = atoi(arg);
		if (tmp < 1024 || tmp > 65536)
		{
			anaPrint(ANA_ERR, "get_args --> port #%d out of range (%d)!\n", i, tmp);
			return -1;
		}
		ports[i] = tmp;
	}
	
	uports.brick_port = htons(ports[0]);
	uports.ipt_port = htons(ports[1]);
	uports.i2a_port = htons(ports[2]);
	uports.a2i_port = htons(ports[3]);
	
	return 1;
}

void dnsptr2ip4(char *cp, struct in_addr *addr)
{
	int i;
	uint32_t ip;

	// extract v4 addr from DNS PTR format
	for (i = 0, ip = 0; i < 4; i++) {
		int v;
		for (v = 0; *cp; cp++) {
			if (*cp == '.') {
				cp++;
				break;
			}
			v = 10*v + *cp - '0';
		}
		ip = ((v & 0xff) << 24) | (ip >> 8);
	}
	
	addr->s_addr = htonl(ip);
}


void dnsptr2ip6(char *ptr, struct in6_addr *addr)
{
	int i;

	// v6 PTR are in DNS-format
	// --> extracts v6 addr from DNS PTR format

	// ptr++;	// skip len indicator

	for (i = 15; i >= 0; i--) {
		unsigned char c;

		if ((addr->s6_addr[i] = *ptr - '0') > 10)
			addr->s6_addr[i] -= 39;

		ptr += 2;	// next value

		if ((c = *ptr - '0') > 10)
			c -= 39;

		addr->s6_addr[i] += c * 0x10;

		ptr += 2;	// next value
	}
}

int add_ip2ana_entry(struct ip2ana_entry **list, char family, void *ip, anaLabel_t s_idp, 
		     anaLabel_t c_idp, struct context_s *ctx, struct service_s *srv)
{
	struct ip2ana_entry *cur, *last = NULL, *new;
	char ipstr[INET6_ADDRSTRLEN];

	if (!s_idp || !c_idp || !ctx || !srv)
		return -1;

	if (!ctx->value || !ctx->valueLen || !srv->value || !srv->valueLen)
		return -1;

	if (*list)
	{
		/* check if entry already exists */
	
		cur = *list;
		
		while (cur)
		{
			if (family != cur->inet_family)
				goto next_entry;
			
			if (cur->inet_family == AF_INET)
			{
				if (cur->ip.ip4.s_addr == ((struct in_addr *)ip)->s_addr)
				{
					anaPrint(ANA_DEBUG, "add_ip2ana_entry --> entry already exists!\n");
					return -1;
				}
			}
			
			if (cur->inet_family == AF_INET6)
			{
				if (!memcmp(&cur->ip.ip6, ip, sizeof(struct in6_addr)))
				{
					anaPrint(ANA_DEBUG, "add_ip2ana_entry --> entry already exists!\n");
					return -1;
				}
			}
				
next_entry:
			last = cur;
			cur = cur->next;
		}
	}

	new = malloc(sizeof(struct ip2ana_entry));
	
	if (!new)
	{
		anaPrint(ANA_ERR, "add_ip2ana_entry --> cannot allocate memory\n");
		return -1;
	}
	
	if (last)
		last->next = new;
	else
		*list = new; /* first element in list */
		
	new->inet_family = family;
	
	if (family == AF_INET)
	{
		new->ip.ip4.s_addr = ((struct in_addr *)ip)->s_addr;
		inet_ntop(AF_INET, ip, ipstr, INET_ADDRSTRLEN);
	}
	else
	{
		memcpy(&new->ip.ip6, ip, sizeof(struct in6_addr));
		inet_ntop(AF_INET6, ip, ipstr, INET6_ADDRSTRLEN);
	}
		
	new->s_idp = s_idp;
	new->c_idp = c_idp;
	
	new->ctx.value = malloc(ctx->valueLen);
	if (!new->ctx.value)
	{
		anaPrint(ANA_ERR, "add_ip2ana_entry --> cannot allocate memory\n");
		return -1;
	}
	memcpy(new->ctx.value, ctx->value, ctx->valueLen);
	
	new->srv.value = malloc(srv->valueLen);
	if (!new->srv.value)
	{
		anaPrint(ANA_ERR, "add_ip2ana_entry --> cannot allocate memory\n");
		return -1;
	}
	memcpy(new->srv.value, srv->value, srv->valueLen);
	
	anaPrint(ANA_DEBUG, "add_ip2ana_entry --> added %s\n", ipstr);
	
	return 1;
}

void free_ip2ana_list(struct ip2ana_entry **list)
{
	struct ip2ana_entry *ptr = *list, *done;
	
	if (!ptr)
		return;
		
	while (ptr)
	{
		free(ptr->ctx.value);
		free(ptr->srv.value);
	
		done = ptr;
		ptr = ptr->next;
		free(done);
	}
	
	*list = NULL;
	return;
}

struct ip2ana_entry* lookup_ip2ana_entry(struct ip2ana_entry *list, char family, void *ip)
{
	struct ip2ana_entry *ptr = list;
	struct in_addr ip4;
	struct in6_addr ip6;

	if (!list || !ip)
		return NULL;
	
	if (family == AF_INET)
		memcpy(&ip4, ip, sizeof(struct in_addr));
	else if (family == AF_INET6)
		memcpy(&ip6, ip, sizeof(struct in_addr));
	
	while (ptr)
	{
		if (family == AF_INET)
		{
			if (ip4.s_addr == ptr->ip.ip4.s_addr)
				return ptr;
		}
		else if (family == AF_INET6)
		{
			if (!memcmp(&ip6, &ptr->ip.ip6, sizeof(struct in6_addr)))
				return ptr;
		}
		
		ptr = ptr->next;
	}
	
	return NULL;
}

void fwd_to_ana(char *buf, int len)
{
	struct iphdr *iph4 = (struct iphdr *)buf;
	struct ip6_hdr *iph6 = (struct ip6_hdr *)buf;
	struct ip2ana_entry *ptr;
	char ipstr[INET6_ADDRSTRLEN];

	if (!buf || len < 20)
		return;

	if (!res_list) /* list of resolved peers is empty */
	{
		/* we try to resolve the IP address into an IDP */
	
		if (iph4->version == 4)
		{
			if (inet_ntop(AF_INET, &iph4->daddr, ipstr, INET_ADDRSTRLEN))
				check_anahosts(ipstr, "A", NULL);
			else
				return;
		}
		else
		{
			if (inet_ntop(AF_INET6, &iph6->ip6_dst, ipstr, INET6_ADDRSTRLEN))
				check_anahosts(ipstr, "AAAA", NULL);
			else
				return;
		}
	}

	if (iph4->version == 4)
		ptr = lookup_ip2ana_entry(res_list, AF_INET, &iph4->daddr);
	else
		ptr = lookup_ip2ana_entry(res_list, AF_INET6, &iph6->ip6_dst);

	if (ptr)
	{
		anaL0_send(ptr->s_idp, buf, len);
		anaPrint(ANA_DEBUG, "fwd_to_ana --> sending %d bytes to %s %s\n",
			 len, (char *)ptr->ctx.value, (char *)ptr->srv.value);
	}
	else
		anaPrint(ANA_DEBUG, "fwd_to_ana --> don't know how to forward data\n");
}
