/* C/bricks/ip2ana/userspave/ip2ana.c
 *  
 * ANA brick for ip2ana adaptation layer 
 * 
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

#include <regex.h>
#include "../common/ip2ana.h"

#define MAX_PAYLOAD 2048 /* for socket recv */
#define MAX_NAMELEN 128  /* for DNS name */
#define REGEXP_ERRBUF 256
#define MAX_LINESIZE 128 /* in anahosts.txt */

#define DNS_STR "^(([[:alnum:]](\\-|_)?)+\\.)+[a-zA-Z]{2,4}$"

#define next_token() do { \
	token = strtok(NULL, " \t"); \
	if (!token) \
		continue; \
} while (0)

struct dnshdr {

	/* DNS header */

	uint16_t trans_id;
	uint16_t flags;
	uint16_t qdcount;
	uint16_t ancount;
	uint16_t nscount;
	uint16_t arcount;
};

void print_usage(char *cmd)
{
	anaPrint(ANA_DEBUG, "\nUsage: %s src_port dst_port i2a_port a2i_port\n\n", cmd);
	anaPrint(ANA_DEBUG, "where: src_port is the UDP userspace port number\n");
	anaPrint(ANA_DEBUG, "  and: dst_port is the UDP kernel port number\n");
	anaPrint(ANA_DEBUG, "  and: i2a_port is the UDP port for data packets flowing from IP to ANA\n");
	anaPrint(ANA_DEBUG, "  and: a2i_port is the UDP port for data packets flowing from ANA to IP\n\n");
}

FILE *anahosts = NULL;
regex_t *dnsregex = NULL;

int dnsreq_recv(void);
int data_recv(void);
int check_anahosts(char *query, char *type, char *result);

void publish_listen(void);
void l3deliver(struct anaL2_message *msg);

struct sockaddr_in dnslocal, dnsdest, datalocal, datadest;
int dnssock, datasock;

void user_close_handlers(void)
{
	close(dnssock);
	close(datasock);
	
	if (anahosts)
		fclose(anahosts);
	
	if (dnsregex)
		free(dnsregex);
}

int user_socket_init(struct udp_ports *ports)
{
	/* initialisation for socket for sending and receiving DNS stuff */
	
	dnssock = socket(AF_INET, SOCK_DGRAM, 0);
	if (dnssock == -1)
	{
		anaPrint(ANA_ERR, "socket_init --> problem when opening dnssock socket");
		return -1;
	}
	memset(&dnslocal, 0, sizeof(struct sockaddr_in));
	dnslocal.sin_family = AF_INET;
	inet_aton("127.0.0.1", &dnslocal.sin_addr);
	dnslocal.sin_port = ports->brick_port;
	if ( (bind(dnssock, (struct sockaddr *)&dnslocal, sizeof(struct sockaddr_in))) < 0 )
	{
		anaPrint(ANA_ERR, "socket_init --> bind anasock failed\n");
		return -1;
	}
	memset(&dnsdest, 0, sizeof(struct sockaddr_in));
	dnsdest.sin_family = AF_INET;
	inet_aton("127.0.0.1", &dnsdest.sin_addr);
	dnsdest.sin_port = ports->ipt_port;
	
	/* initialisation for socket for sending and receiving DATA */
	
	datasock = socket(AF_INET, SOCK_DGRAM, 0);
	if (datasock == -1)
	{
		anaPrint(ANA_ERR, "socket_init --> problem when opening datasock socket");
		return -1;
	}
	memset(&datalocal, 0, sizeof(struct sockaddr_in));
	datalocal.sin_family = AF_INET;
	inet_aton("127.0.0.1", &datalocal.sin_addr);
	datalocal.sin_port = ports->i2a_port;
	if ( (bind(datasock, (struct sockaddr *)&datalocal, sizeof(struct sockaddr_in))) < 0 )
	{
		anaPrint(ANA_ERR, "socket_init --> bind datasock failed\n");
		return -1;
	}
	memset(&datadest, 0, sizeof(struct sockaddr_in));
	datadest.sin_family = AF_INET;
	inet_aton("127.0.0.1", &datadest.sin_addr);
	datadest.sin_port = ports->a2i_port;

	return 1;
}

int user_file_init(char *path)
{
	int err;
	char errbuf[REGEXP_ERRBUF];

	if (!path)
		return -1;

	anahosts = fopen(path, "r");
	
	if (anahosts)
	{
		dnsregex = (regex_t *)malloc(sizeof(regex_t));
		err = regcomp(dnsregex, DNS_STR, REG_EXTENDED);

		if (err)
		{
			regerror(err, dnsregex, errbuf, REGEXP_ERRBUF);
			anaPrint(ANA_DEBUG, "file_init --> cannot initialize regexp for DNS entries - %s\n", errbuf);
			free(dnsregex);
			dnsregex = NULL;
		}
		
		anaPrint(ANA_DEBUG, "file_init --> reading entries from %s\n", path);
		return 1;
	}
	else
	{
		anaPrint(ANA_ERR, "file_init --> cannot open anahosts.txt\n");
		anahosts = NULL;
		return -1;
	}
}

int user_loop(void)
{
	int ret;
	fd_set readfd;
	
	anaPrint(ANA_DEBUG, "user_loop --> starting with udp ports: %u %u %u %u\n", ntohs(dnslocal.sin_port), 
		 ntohs(dnsdest.sin_port), ntohs(datalocal.sin_port), ntohs(datadest.sin_port));
		 
	while (1)
	{
		FD_ZERO(&readfd);
		FD_SET(dnssock, &readfd);
		FD_SET(datasock, &readfd);	
		
		ret = select(255, &readfd, NULL, NULL, NULL);							
		
		if (ret < 0)
		{
			anaPrint(ANA_ERR, "user_loop --> problem with select");
			continue;
		}
		
		if (FD_ISSET(dnssock, &readfd))
			dnsreq_recv();
			
		if (FD_ISSET(datasock, &readfd))
			data_recv();
	}

	return 0;
}

int data_recv(void)
{
	int bytes;
	struct sockaddr_in from;
	char buf[MAX_PAYLOAD];
	socklen_t slen = sizeof(struct sockaddr_in);
	
	bytes = recvfrom(datasock, buf, MAX_PAYLOAD, 0, (struct sockaddr *)&from, &slen);

	if (bytes == -1)
	{
		anaPrint(ANA_DEBUG, "data_recv --> problem when receiving data");
		return -1;
	}
		
	if (from.sin_port != datadest.sin_port)
	{
		anaPrint(ANA_DEBUG, "data_recv --> received data from alien UDP port (%d)\n", ntohs(from.sin_port));
		return -1;
	}
	
	fwd_to_ana(buf, bytes);
	
	return 1;
}

int dnsreq_recv(void)
{
	int bytes, len, err;
	char tmp_name[MAX_NAMELEN];
	char ipstr[INET6_ADDRSTRLEN];
	char buf[MAX_PAYLOAD];
	socklen_t slen = sizeof(struct sockaddr_in);
	struct dnshdr *dnsh = (struct dnshdr *)buf;
	uint16_t qtype;
	unsigned char *cp = (unsigned char *) (dnsh + 1);
	char result[MAX_LINESIZE];
	struct sockaddr_in from;

	bytes = recvfrom(dnssock, buf, MAX_PAYLOAD, 0, (struct sockaddr *)&from, &slen);

	if (bytes == -1)
	{
		perror("dnsreq_recv --> problem when receiving dns query");
		return -1;
	}
		
	if (from.sin_port != dnsdest.sin_port)
	{
		anaPrint(ANA_DEBUG, "dnsreq_recv --> received DNS query from alien UDP port (%d)\n",
			 ntohs(from.sin_port));
		return -1;
	}
		
	if (bytes < sizeof(struct dnshdr))
	{
		anaPrint(ANA_DEBUG, "dnsreq_recv --> DNS query received is too short! (%d bytes)\n", bytes);
		return -1;
	}
	
	/* contruct human-readable DNA name from DNS-encoded pattern */
	len = 0;
	while (*cp && ((char *)cp-buf)<MAX_PAYLOAD ) {
		int i = *cp;
		cp++;

		while (i--) {
			if (len < MAX_NAMELEN)
				tmp_name[len++] = *cp;
			cp++;
		}

		if (len < MAX_NAMELEN && *cp)
			tmp_name[len++] = '.';
	}
		
	if (((char *)cp-buf) >= MAX_PAYLOAD)
	{
		anaPrint(ANA_DEBUG, "dnsreq_recv --> DNS request too long!\n");
		return -1;
	}
	
	if (len >= MAX_NAMELEN)
	{
		anaPrint(ANA_DEBUG, "dnsreq_recv --> DNS name too long!\n");
		return -1;
	}
	
	tmp_name[len] = '\0';
	cp++;
	qtype = ntohs(* (uint16_t *) cp);
		
	switch (qtype) {

	case DNS_REQ_PTR:
	
		if (len <= 64) {
				
			struct in_addr addr;
			dnsptr2ip4(tmp_name, &addr);
				
			if (inet_ntop(AF_INET, &addr, ipstr, INET_ADDRSTRLEN))
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> got PTR request for %s \n", ipstr);
				err = check_anahosts(ipstr, "A", result+2);
			}
			else
				err = -1;
		}
		else /* an IPv6 PTR is always greater than 64 bytes */
		{
			struct in6_addr addr;
			dnsptr2ip6(tmp_name, &addr);
		
			if (inet_ntop(AF_INET6, &addr, ipstr, INET6_ADDRSTRLEN))
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> got PTR request for %s\n", ipstr);
				err = check_anahosts(ipstr, "AAAA", result+2);
			}
			else
				err = -1;
		}
			
		memcpy(result, &(dnsh->trans_id), 2);
			
		if (err == 1)
		{
			anaPrint(ANA_DEBUG, "dnsreq_recv --> match: %s\n", result+2);
			bytes = sendto(dnssock, result, strlen(result+2)+3, 0, (struct sockaddr *)&dnsdest, slen);
		}
		else
		{
			anaPrint(ANA_DEBUG, "dnsreq_recv --> no match: sending null reply.\n");
			qtype = htons(8583); /* DNS "no such name" reply */
			memcpy(result+2, &qtype, 2);
			bytes = sendto(dnssock, result, 4, 0, (struct sockaddr *)&dnsdest, slen);
		}
		
		if (bytes == -1)
		{
			perror("Cannot send result");
			return -1;
		}
		else
		{
			anaPrint(ANA_DEBUG, "Result was sent back to kernel ip2ana\n\n");
			return 1;
		}
			
		break;
	
	case DNS_REQ_AAAA:
	case DNS_REQ_A:
	
		/* Check if A or AAAA name belongs to .ana domain */
		
		if (strcmp(".ana", tmp_name + len - strlen(".ana")))
		{
			/* we should never receive requests not for .ana */
		
			anaPrint(ANA_ERR, "FATAL: not a request for .ana domain!\n");
			return -1;
		}
		else
		{
			anaPrint(ANA_DEBUG, "dnsreq_recv --> got request for %s %s\n", 
				 tmp_name, (qtype==DNS_REQ_A) ? "A" : "AAAA");
			err = check_anahosts(tmp_name, (qtype==DNS_REQ_A) ? "A" : "AAAA", result+2);
			memcpy(result, &(dnsh->trans_id), 2);
				
			if (err == 1)
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> match: %s\n", result+2);
				bytes = sendto(dnssock, result, strlen(result+2)+3, 0, (struct sockaddr *)&dnsdest, slen);
			}
			else
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> no match: sending null reply.\n");
				qtype = htons(8583); /* DNS "no such name" reply */
				memcpy(result+2, &qtype, 2);
				bytes = sendto(dnssock, result, 4, 0, (struct sockaddr *)&dnsdest, slen);
			}
					
			if (bytes == -1)
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> cannot send result");
				return -1;
			}
			else
			{
				anaPrint(ANA_DEBUG, "dnsreq_recv --> result was sent back to kernel ip2ana\n\n");
				return 1;
			}
		}
		break;
	default:
		return -1;
		break;
	}
	
	return -1;
}

void publish_listen(void)
{
	char line[MAX_LINESIZE];
	char *token;
	int errno;
	
	char name[64];
	char type[8];
	char inet_family;
	char addr[64];
	char cpmt[64];
	char ctxt[64];
	char csrv[64];
	char flag[8];
	
	anaLabel_t c_idp; /* compartment idp */
	anaLabel_t s_idp; /* idp of published service */
	struct context_s ctx;
	struct service_s srv;
	struct in_addr ip4;
	struct in6_addr ip6;
	void *ip;
		
	rewind(anahosts); /* make sure we read from the start */
	
	while (!feof(anahosts))
	{
		if (!fgets(line, MAX_LINESIZE, anahosts))
			return;
			
		if (line[0] == '#') /* comment */
			continue;
		
		token = strtok(line, " \t");
		if (!token)
			continue;
	
		if (dnsregex)
		{
			errno = regexec(dnsregex, token, 0, 0, 0);
		
			if (errno)
				continue; /* badly formatted DNS name */
		}
		strncpy(name, token, 64);
		
		next_token();
		if (!strcmp("A", token) || !strcmp("AAAA", token))
		{
			strncpy(type, token, 8);
			next_token();
			
			if (!strcmp("A",type))
			{
				errno = inet_pton(AF_INET, token, &ip4);
				inet_family = AF_INET;
				ip = (void *)&ip4;
			}
			else if (!strcmp("AAAA",type))
			{
				errno = inet_pton(AF_INET6, token, &ip6);
				inet_family = AF_INET6;
				ip = (void *)&ip6;
			}
			
			if (errno > 0)
				strncpy(addr, token, 64);
			else
				continue; /* format of IP address is not correct */
		}
		else
			continue; /* no "A" or "AAAA" keyword found */
		
		next_token();
		strncpy(cpmt, token, 64);
		
		next_token();
		strncpy(ctxt, token, 64);
		
		next_token();
		strncpy(csrv, token, 64);
		
		next_token();
		strncpy(flag, token, 8);
		
		if (strncmp(flag, "LISTEN", 6)) /* not LISTEN */
			continue;
		
		anaPrint(ANA_DEBUG, "publish_listen --> trying to publish %s %s %s\n", cpmt, ctxt, csrv);
		
		if (strcmp(cpmt, "NodeC"))
		{
			/* try to find compartment in KVR */
		
			ctx.value = ".";
			ctx.valueLen = strlen(".")+1;
			srv.value = cpmt;
			srv.valueLen = strlen(cpmt)+1;
		
			c_idp = anaL2_resolve(NODE_LABEL, &ctx, &srv, 'u', NULL, NULL);
			if (!c_idp)
			{
				anaPrint(ANA_DEBUG, "publish_listen --> cannot find compartment %s\n", cpmt);
				continue;
			}
		}
		else
			c_idp = NODE_LABEL; /* publish in node compartment */
		
		/* try to publish service in compartment */
		
		ctx.value = ctxt;
		ctx.valueLen = strlen(ctxt)+1;
		srv.value = csrv;
		srv.valueLen = strlen(csrv)+1;
		
		s_idp = anaL2_publish(c_idp, &ctx, &srv, l3deliver, NOTHREAD, NULL);
		if (!s_idp)
		{
			anaPrint(ANA_DEBUG, "publish_listen --> cannot publish %s %s\n", ctxt, csrv);
			continue;
		}
		
		/* register ip2ana_entry in pub_list */
		add_ip2ana_entry(&pub_list, inet_family, ip, s_idp, c_idp, &ctx, &srv);
	}
}

/* this is the callback function: prototype is fixed by API level 2 */
void l3deliver(struct anaL2_message *msg)
{
	int bytes;
	int slen = sizeof(struct sockaddr_in);
	
	if (!msg || !msg->data)
		return;
	
	/* forward data to the kernel part of ip2ana */ 
	/* for delivery of data to the IP stack      */
	
	bytes = sendto(datasock, msg->data, msg->dataLen, 0, &datadest, slen);
	
	if (bytes == -1)
	{
		anaPrint(ANA_DEBUG, "l3deliver --> cannot send data!\n");
		return;
	}

	anaPrint(ANA_DEBUG, "l3deliver --> sent %d bytes of data to ip2ana kernel\n", bytes);
}

int check_anahosts(char *query, char *qtype, char *result)
{
	char line[MAX_LINESIZE];
	char *token;
	int errno;
	
	char name[64];
	char type[8];
	char inet_family;
	char addr[64];
	char cpmt[64];
	char ctxt[64];
	char csrv[64];
	char flag[8];
	
	anaLabel_t c_idp; /* compartment idp */
	anaLabel_t s_idp; /* idp of published service */
	struct context_s ctx;
	struct service_s srv;
	struct in_addr ip4;
	struct in6_addr ip6;
	void *ip;
	
	if (!query || !qtype)
		return -1;
	
	rewind(anahosts);
		
	while (!feof(anahosts))
	{
		if (!fgets(line, MAX_LINESIZE, anahosts))
			return -1;
			
		if (line[0] == '#') /* comment */
			continue;
			
		token = strtok(line, " \t");
		if (!token)
			continue;
			
		if (dnsregex)
		{
			errno = regexec(dnsregex, token, 0, 0, 0);
		
			if (errno)
				continue; /* badly formatted DNS name */
		}
		
		strcpy(name, token);
		
		next_token();
		
		if (!strcmp("A", token) || !strcmp("AAAA", token))
		{
			strcpy(type, token);
			
			next_token();
			
			/* Obsolete: for previous anahosts.txt file format */
			/* token[strlen(token)-1] = '\0'; replaces \n character */
			
			if (!strcmp("A",type))
			{
				errno = inet_pton(AF_INET, token, &ip4);
				inet_family = AF_INET;
				ip = (void *)&ip4;
			}
			else if (!strcmp("AAAA",type))
			{
				errno = inet_pton(AF_INET6, token, &ip6);
				inet_family = AF_INET;
				ip = (void *)&ip6;
			}
			
			if (errno > 0)
				strcpy(addr, token);
			else
				continue;
		}
		else
			continue; /* no "A" or "AAAA" keyword found */
		
		if (strcmp(qtype,type))
			continue; /* not the right family */
			
		if (!strcmp(query,addr) || !strcmp(query,name))
		{
			struct ip2ana_entry *ptr;
		
			/* we have a match */
			/* 2 cases: entry exists, simply reply with DNS answer */
			/* or entry does not exist: we try to resolve it */
		
			if (inet_family == AF_INET)
				ptr = lookup_ip2ana_entry(res_list, AF_INET, &ip4);
			else
				ptr = lookup_ip2ana_entry(res_list, AF_INET6, &ip6);
		
			if (ptr) /* entry already exists */
			{
				if (result)
					sprintf(result, "%s %s %s", name, type, addr);
					
				return 1; /* exit */
			}
		
			next_token();
			strncpy(cpmt, token, 64);
		
			next_token();
			strncpy(ctxt, token, 64);
		
			next_token();
			strncpy(csrv, token, 64);
		
			next_token();
			strncpy(flag, token, 8);
		
			if (strncmp(flag, "REMOTE", 6)) /* not REMOTE */
				continue;
			
			anaPrint(ANA_DEBUG, "check_anahosts --> trying to resolve %s %s %s\n", cpmt, ctxt, csrv);
		
			if (strcmp(cpmt, "NodeC"))
			{
				/* try to find compartment in KVR */
		
				ctx.value = ".";
				ctx.valueLen = strlen(".")+1;
				srv.value = cpmt;
				srv.valueLen = strlen(cpmt)+1;
		
				c_idp = anaL2_resolve(NODE_LABEL, &ctx, &srv, 'u', NULL, NULL);
				if (!c_idp)
				{
					anaPrint(ANA_DEBUG, "check_anahosts --> cannot find compartment %s\n", cpmt);
					continue;
				}
			}
			else
				c_idp = NODE_LABEL; /* publish in node compartment */
		
			/* try to resolve remote peer via compartment */
		
			ctx.value = ctxt;
			ctx.valueLen = strlen(ctxt)+1;
			srv.value = csrv;
			srv.valueLen = strlen(csrv)+1;
		
			s_idp = anaL2_resolve(c_idp, &ctx, &srv, 'u', NULL, NULL);
			if (!s_idp)
			{
				anaPrint(ANA_DEBUG, "check_anahosts --> cannot resolve %s %s\n", ctxt, csrv);
				continue;
			}
		
			/* register ip2ana_entry in res_list */
			add_ip2ana_entry(&res_list, inet_family, ip, s_idp, c_idp, &ctx, &srv);
			
			/* "send back" DNS resolution reply */
			if (result)
				sprintf(result, "%s %s %s", name, type, addr);
			
			return 1;		
		}
	}

	return -1; /* not found */
}


