/* C/bricks/templates/brickAL0_template.c
 * 
 * template file which shows how to use API Level 0
 * For more information please refer to the doxygen docu or
 * to the ana-core documentation.
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Christophe Jelger <christophe.jelger@unibas.ch> 
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*----------------------- Default brick header code -----------------------*/
/* brick_template.h contains the functions that initialize the bricks */
#include "brick_template.h"

/* Update the fields below with the name of your brick.
 * This name in combination with the process ID 
 * will be used for the log files */
char mymodename[]="brick0";

/* Standard ANA code uses the ISC licence.
 * To be able to load our brick to the kernel we use
 * the compatible "Dual BSD/GPL" license
 */
#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif
/*----------------------- End default brick header code -------------------*/

/*----------------------- Brick specific code starts here------------------*/
/* Your code MUST always provide both 
 * the brick_start and brick_exit functions
 */

static anaLabel_t firstIDP;
static anaLabel_t secondIDP;


void callBackFct(char *data, int len, anaLabel_t input,  void *aux) {
	anaPrint(ANA_DEBUG, "received Data (%s) with len (%i) on label (%u) "\
			    "with aux (%s) \n", data, len, input, (char *)aux);

	/* this is for validation only*/
	if (input == firstIDP)
		anaValidate("received message on firstIDP\n");
	else if (input == secondIDP)
		anaValidate("received message on secondIDP\n");
	else
		anaValidate("received message on unknown IDP\n");
}


void * copyData(void * data){
	char * new_data = (char *)malloc(strlen((const char *)data)+1);
	memcpy(new_data, (char *)data, strlen((const char *)data));
	new_data[strlen((const char *)data)]='\0';
	return (char *)new_data;
}

void freeData(void * data){
	free(data);
}

/* The brick_exit function is called before a brick is unloaded
 * It should do all kind of clean up
 */
void brick_exit() {

	anaL0_unregisterCallback(firstIDP, VIEWER_PUBLIC);
	anaL0_unregisterCallback(secondIDP, VIEWER_PUBLIC);
	
	anaPrint(ANA_DEBUG, "Brick EXIT\n");
}


/* The brick_start function is the main function of each brick */
int brick_start() {

	char * msg = NULL;
	int msgLen = 0;
	int ret = 0;

	validate_init(mymodename);

	/* register a callback function with the minmex */
	firstIDP = anaL0_registerCallback((anaCallback_t)&callBackFct, /* the callback function */
				     (void *) "IDP 1", 		  /* the aux argument */
				     copyData, 			  /* function that copies the aux */
				     freeData, 			  /* function that frees the aux */
				     VIEWER_PUBLIC, 		  /* public IDP */
				     IDP_PERM, 			  /* permanent IDP */
				     NOTHREAD);			  /* callback executed not as thread */
	if (firstIDP == 0) {
		anaPerror(ANA_ERR, "error registering callback function");
	        return -1;
	}
	else
		anaPrint(ANA_DEBUG, "From brick : successfull register label = %u \n", firstIDP);
	
	/* send a message to this callback function */
	msg = "***first message***";
	msgLen = strlen(msg) + 1;
	if (anaL0_send(firstIDP, msg, msgLen) < 0)
		anaPerror(ANA_ERR, "Send Failed");
	else	
		anaPrint(ANA_DEBUG, "Send Succeeded \n");
		
	
	/* register a second callback function */
	secondIDP = anaL0_registerCallback((anaCallback_t)&callBackFct, (void *) "IDP 2", 
				     copyData, freeData, VIEWER_PUBLIC, IDP_PERM, NOTHREAD);

	/* redirect every msg received on the first IDP to the second IDP */	
	ret = anaL0_redirect(firstIDP, MY_HANDLE, secondIDP, VIEWER_PUBLIC); 
	if (ret < 0) {
		anaPerror(ANA_ERR, "redirect failed");
	}
	else
		anaPrint(ANA_DEBUG, "redirect succeded \n");

	/* send a message to the first IDP, will be received on the second IDP */
	msg = "***second message***";
	msgLen = strlen("***second message***") + 1;
	anaL0_send(firstIDP, msg, msgLen);
	
	/* on success we return a positive value, 
         * the brick_start function has finished and we are waiting for messages 
         * on the callback functions as long as the brick_exit function is not called
         */
	return 1;
}

