#ifndef NS2
#include "quickRepository.h"

/***************************************************************

Hash functions by Bob Jenkins, December 1996, Public Domain.

2005-11-09: tested by CJ with 128-bit keys to 10 bits values 
            --> hash function has a uniform distribution

***************************************************************/
ub4 QREPhash( k, length, initval, maskLength)
register ub1 *k;        /* the key */
register ub4  length;   /* the length of the key in bytes */
register ub4  initval;    /* the previous hash, or an arbitrary value */
register ub4  maskLength;    /* the previous hash, or an arbitrary value */
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */

	return (c & hashmask(maskLength));	 
}
#endif

int qrep_name2hash(struct QREP *rep, char *name, int nameLen){
	int h;
	if(rep==NULL || name==NULL) {
		anaPrint(ANA_ERR, "name2hash: NULL arguments \n");
		return -1;
	}
	h = QREPhash((ub1*)name, nameLen, 0, rep->hashMaskEntry);
	return h;
}

int  qrep_key2hash(struct QREP *rep, char *key) {
	int h;
	if(rep==NULL || key==NULL) {
		anaPrint(ANA_ERR, "key2hash: NULL arguments \n");
		return -1;
	}
	h = QREPhash((ub1*)key, strlen(key), 0, rep->hashMaskKey);
	return h;
}

void *  qrep_cloneValue(void *value, int valueSize){
	void *res=NULL;
	res = (void *) malloc(valueSize);
	memcpy(res, value, valueSize);
	return res;
}


int  initQRepository(struct QREP *rep, int hashMaskEntryTab, int hashMaskKeyTab) {
	if(rep == NULL) {
		anaPrint(ANA_ERR, "initializeQREP : NULL argument\n");
		return -1;
	}
	rep->hashMaskEntry = hashMaskEntryTab;
	rep->hashMaskKey = hashMaskKeyTab;
	rep->sizeEntryTab = pow2(hashMaskEntryTab);
	rep->sizeKeyTab = pow2(hashMaskKeyTab);
	
	/* we allocate the tables with the appropriate size*/
	rep->entryTab = (struct QREPEntry **) malloc(rep->sizeEntryTab * sizeof(void *));
	if(rep->entryTab == NULL){
		anaPrint(ANA_ERR, "initializeQREP : unable to allocate entry Tab\n");
		return -1;
	}
	memset(rep->entryTab,0, rep->sizeEntryTab * sizeof(void *));
	
	rep->keyTab = (struct QREPKey **) malloc(rep->sizeKeyTab * sizeof(void *));
	if(rep->keyTab == NULL){
		anaPrint(ANA_ERR, "initializeQREP : unable to allocate key Tab\n");
		free(rep->entryTab);
		return -1;
	}
	
	memset(rep->keyTab,0, rep->sizeKeyTab * sizeof(void *));
	
	anaPrint(ANA_DEBUG, "initializeQREP : QREP initialized with size %d for entry tab and %d for key tab \n", rep->sizeEntryTab, rep->sizeKeyTab);
	return 0;
}


struct QREPEntry *  allocateQREPEntry(char *name, int nameLen, void *value, int valueSize){
	struct QREPEntry *entry=NULL;
	if(value == NULL || name == NULL) {
		anaPrint(ANA_ERR, "allocateQREPEntry: NULL arguments\n");
		return NULL;
	}
	entry = (struct QREPEntry *) malloc(sizeof(struct QREPEntry));
	memset(entry,0,sizeof(struct QREPEntry));
	entry->name = (char *) malloc(nameLen);
	memcpy(entry->name, name, nameLen);
	entry->nameLen = nameLen;
	entry->value = qrep_cloneValue(value, valueSize);
	return entry;
}

int  QREPaddKeywordToEntry(struct QREPEntry *entry, char *keyWord) {
	struct QREPKeyword *keyW;
	if(entry == NULL || keyWord == NULL) {
		anaPrint(ANA_ERR, "addKeyToEntry: NULL arguments\n");
		return -1;
	}
	keyW = (struct QREPKeyword *) malloc(sizeof(struct QREPKeyword));
	keyW->keyWord = (char *) malloc((strlen(keyWord)+1)*sizeof(char));
	strcpy(keyW->keyWord, keyWord);
	keyW->next = entry->keyWordsList;
	entry->keyWordsList = keyW;
	return 0;
}


int qrep_updateKey(struct QREP *rep, char *keyword, struct QREPEntry *entry){
	int h, found;
	struct QREPKey *key;
	struct QREPKeyObjectList *keyEntriesList;
	
	if(rep == NULL || entry == NULL || keyword == NULL) {
		anaPrint(ANA_ERR, "updateKey: NULL arguments\n");
		return -1;
	}
	h = qrep_key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "updateKey : index %d after hash is out of bounds\n",h);
		return -1;
	}
	key = rep->keyTab[h];
	found = 0;
	while(key!=NULL && !found){
		if(!strcmp(key->keyWord, keyword))
			found = 1;
		else
			key = key->next;
	}
	if(!found){
		anaPrint(ANA_DEBUG, "updateKey : no key matching keyword %s => adding a new key\n",keyword);
		key = (struct QREPKey *) malloc(sizeof(struct QREPKey));
		key->keyWord = (char *) malloc((strlen(keyword)+1)*sizeof(char));
		strcpy(key->keyWord, keyword);
		key->next = rep->keyTab[h];
		rep->keyTab[h] = key;
	}
	
	/* we add the entry to the key's entries list*/
	keyEntriesList = (struct QREPKeyObjectList *) malloc(sizeof(struct QREPKeyObjectList));
	keyEntriesList->entry = entry;
	keyEntriesList->next = key->entriesList;
	key->entriesList = keyEntriesList;
	
	return 0;	
}

int  qrep_freeKey(struct QREP *rep, struct QREPKey *key, int h){
	struct QREPKey *tmp;
	struct QREPKey *predecessor;
	struct QREPKeyObjectList *tmpList;
	struct QREPKeyObjectList *tmpNextList;
	
	if(key==NULL || rep==NULL){
		anaPrint(ANA_ERR, "freeKey: NULL arguments\n");
		return -1;
	}
	tmp = rep->keyTab[h];
	predecessor = NULL;
	while(tmp!=NULL){
		if(key == tmp){
			if(predecessor == NULL)
				rep->keyTab[h] = tmp->next;
			else
				predecessor->next = tmp->next;
			
			free(tmp->keyWord);
			/*we free the key's object list */
			tmpList = tmp->entriesList;
			while(tmpList != NULL) {
				tmpNextList = tmpList->next;
				free(tmpList);
				tmpList = tmpNextList;
			}
			
			free(tmp);
			return 0;
		}
		else{
			predecessor = tmp;
			tmp = tmp->next;
		}
	}
	return -1;	
}


int  qrep_removeFromKeysList(struct QREP *rep, struct QREPEntry *toRemove, char *keyword){
	struct QREPKeyObjectList *objList;
	struct QREPKeyObjectList *predecessor;
	int h, foundKey;
	struct QREPKey *key;
	
	if(keyword == NULL || toRemove == NULL || rep == NULL){
		anaPrint(ANA_ERR, "removeFromKeysList: NULL arguments\n");
		return -1;
	}
	
	/* first we find the key*/
	h =  qrep_key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "removeFromKeysList : index %d after hash is out of bounds\n",h);
		return -1;
	}
	foundKey = 0;
	key = rep->keyTab[h];
	while(key!=NULL && !foundKey){
		if(!strcmp(keyword, key->keyWord))
			foundKey=1;
		else
			key = key->next;
	}
	
	if(!foundKey){
		anaPrint(ANA_ERR, "removeFromKeysList : unable to find key matching keyword %s\n",keyword);
		return -1;
	}
	
	objList = key->entriesList;
	predecessor = NULL;
	while(objList!=NULL){
	       	if(objList->entry == toRemove){
	       		if(predecessor!=NULL)
				predecessor->next = objList->next; 
			else 
				if(objList->next !=NULL)
					key->entriesList =  objList->next;
				else
					qrep_freeKey(rep, key, h); /*in case the entriesList becomes empty we free the Key */
			return 0;
		}
	       	else {
	       		predecessor = objList;
			objList = objList->next;
		}	
	}
	return -1;
}


int  addQREPEntry(struct QREP *rep, struct QREPEntry *entry){
	int hN;
	struct QREPEntry *existEntry = NULL;
	struct QREPKeyword *tmpKeyword =NULL;
	if(entry == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "addQREPEntry: NULL arguments\n");
		return -1;
	}
	/* first we add the entry to the entries Table*/	
	hN =  qrep_name2hash(rep, (char *)entry->name, entry->nameLen);
	if(hN<0 || hN>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "addQREPEntry : index %d after hash is out of bounds\n",hN);
		return -1;
	}
	existEntry = rep->entryTab[hN];
	if(existEntry==NULL){
		entry->next = NULL;
		rep->entryTab[hN] = entry;	
	}
	else{ /*hash collision*/
		/* we must check if an entry with same name exists*/
		while(existEntry!=NULL){
			if(!strcmp((const char *)existEntry->name,(const char *)entry->name)){
				anaPrint(ANA_ERR, "addQREPEntry: entry with same name already exists\n");
				return -1;
			}
			existEntry = existEntry->next;
		}	
	

		entry->next = rep->entryTab[hN];
		rep->entryTab[hN] = entry;
	}
		
	/* Then we update the keys Table*/
	tmpKeyword = entry->keyWordsList;
	/*for each keyWord*/
	while(tmpKeyword!=NULL){
		qrep_updateKey(rep, tmpKeyword->keyWord, entry);
		tmpKeyword = tmpKeyword->next;
	}
	return 0;
}

struct QREPEntry *  getQREPEntryByName(struct QREP *rep, char *name, int nameLen){
	struct QREPEntry *entry=NULL;
	int h;
	if(name == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "getQREPEntryByName: NULL arguments\n");
		return NULL;
	}
	h =  qrep_name2hash(rep, name, nameLen);
	if(h<0 || h>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "getQREPEntryByName : index %d after hash is out of bounds\n",h);
		return NULL;
	}
	entry = rep->entryTab[h];
	while(entry!=NULL){
		if(!memcmp(name, entry->name, nameLen)){
			return entry;
		}	
		entry = entry->next;
	}
	return NULL;
}

/* Warning this function is not meant to be used directly but rather with an indirection
 through findByQuery. The reason is that this function returns a pointer to QREPKeyObjectList
 that is stored in the QREP while findByQuery returns a copy of this list.
 Therefore it is safer to use findByQuery.
 */

struct QREPKeyObjectList *  QREPfindByKey(struct QREP *rep, char *keyword) {
	struct QREPKey *key;
	int h;
	if(keyword == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "findByKey: NULL arguments\n");
		return NULL;
	}
	h =  qrep_key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "findByKey : index %d after hash is out of bounds\n",h);
		return NULL;
	}
	
	key = rep->keyTab[h];
	while(key!=NULL){
		if(!strcmp(keyword, key->keyWord)){
			return key->entriesList;
		}	
		key = key->next;
	}
	return NULL;
}

/* this function submits the query to the parser */
int  QREPfindByQuery(struct QREP *rep, char *query, struct QREPKeyObjectList **res){	
	int ret;
	ret = qrep_evalQuery(rep, query, res); 
	return ret;
}	


/* this function frees a list returned as a result of a query. Important, this function does 
not free any QREP entries, it just frees the structures holding pointers to them */
void  QREPfreeQueryRes(struct QREPKeyObjectList *list){
	struct QREPKeyObjectList *next=NULL;
	while(list!=NULL){
		next = list->next;
		free(list);
		list = next;
	}	
}

/* this function frees the entry and removes it from key indexes in the repository*/
int  freeQREPEntry(struct QREP *rep, struct QREPEntry *entry, freeValFunction_t freeValueFunction){
	struct QREPKeyword *keyWordsList;
	struct QREPKeyword *nextKeyWord;
	struct QREPEntry * temp, *pred;
	int h;
	if(rep==NULL || entry ==NULL)
		return -1;
	h =  qrep_name2hash(rep, (char *)entry->name, entry->nameLen);
	if(h<0 || h>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "freeQREPEntry : index %d after hash is out of bounds\n",h);
		return -1;
	}
	/*remove the entry from the Table*/
	temp = rep->entryTab[h];
	pred = temp;
	if(temp==NULL){
		anaPrint(ANA_ERR, "freeQREPEntry : No entry to free\n");
		return -1;
	}
	
	if(!memcmp((const char *)entry->name, (const char *)temp->name, entry->nameLen)){ /*its the first entry*/
		if(temp->next == NULL) /*its the only entry*/
			rep->entryTab[h] = NULL;
		else
			rep->entryTab[h] = temp->next;
	}

	else{
		temp = temp->next; /*its not the first entry, so there must be at least a second one*/
		while(temp!=NULL){
			if(!strcmp((const char *)entry->name, (const char *)temp->name)){
				pred->next = temp->next; /*remove temp from the list */	
				break;
			}	
			temp = temp->next;
			pred = pred->next;
		}
	}
	keyWordsList = entry->keyWordsList;
	while(keyWordsList!=NULL){
		qrep_removeFromKeysList(rep, entry, keyWordsList->keyWord);
		nextKeyWord = keyWordsList->next;
		free(keyWordsList->keyWord);
		free(keyWordsList);
		keyWordsList = nextKeyWord;
	}
	free(entry->name);
	freeValueFunction(entry->value);
	free(entry);
	return 0;
}

/* this function does not free the repository i.e is does not deallocate it's table but rather frees all the
entries */
void  purgeQRepository(struct QREP *rep, freeValFunction_t freeValueFunction){
	struct QREPEntry *entry=NULL;
	struct QREPEntry *next=NULL;
	struct QREPKey *key=NULL; 
	struct QREPKey *nextkey=NULL;
	struct QREPKeyword *keyword=NULL;
	struct QREPKeyword *nextkeyword=NULL;
	struct QREPKeyObjectList *entrieslist=NULL;
	struct QREPKeyObjectList *nextentrieslist=NULL;
	int i;
	
	/* first we purge the entry tab */
	for(i=0;i<rep->sizeEntryTab;i++){
		entry = rep->entryTab[i];
		rep->entryTab[i] = NULL;
		while(entry!=NULL){
			free(entry->name);
			freeValueFunction(entry->value);
			/* we free the keyWords list */
			keyword=entry->keyWordsList;
			while(keyword!=NULL){
				nextkeyword = keyword->next;
				free(keyword->keyWord);
				free(keyword);
				keyword=nextkeyword;
			}
			
			next=entry->next;
			free(entry);
			entry=next;
		}
	}

	/* now we purge the keyTab */
	for(i=0;i<rep->sizeKeyTab;i++){
		key = rep->keyTab[i];
		rep->keyTab[i] = NULL;
		while(key!=NULL){
			nextkey = key->next;
			free(key->keyWord);
			entrieslist = key->entriesList;
			while(entrieslist!=NULL){
				nextentrieslist=entrieslist->next;
				free(entrieslist);
				entrieslist=nextentrieslist;
			}
			free(key);
			key=nextkey;
		}
	}
	return;
	
}

/* this function remove the table from memory, equivalent to free() i.e, the repository pointer should not be
accessed afterwards*/
void  freeQRepository(struct QREP *rep, freeValFunction_t fct){
	purgeQRepository(rep, fct);
	free(rep->entryTab);
	free(rep->keyTab);
//	free(rep);
	return;
}

/*this function applies the function passed as arguement to all the values in the repository*/
void  applyToAllQREPValues(struct QREP *rep, valueFunction_t valueFunction, void *aux ){
	struct QREPEntry *entry=NULL;
	int i;	
		
	for(i=0;i<rep->sizeEntryTab;i++){
		entry = rep->entryTab[i];
		while(entry!=NULL){
			valueFunction(entry->value, aux);
			entry=entry->next;
		}
	}
	return;
}
/*this function applies the function passed as arguement to all the entries in the repository*/
void  applyToAllQREPEntries(struct QREP *rep, valueFunction_t valueFunction, void *aux ){
	struct QREPEntry *entry=NULL, *temp = NULL;
	int i;	
		
	for(i=0;i<rep->sizeEntryTab;i++){
		entry = rep->entryTab[i];
		while(entry!=NULL){
			temp = entry->next;
			valueFunction(entry, aux);
			entry = temp;
		}
	}
	return;
}

















