#ifndef _ANAVLINK_H
#define _ANAVLINK_H

#ifdef USER
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/if.h>
#include <linux/if_packet.h>
#include <sys/ioctl.h>
#endif /* USER */

/* COMMON stuff */

#include "anaCommon.h"
#include "anaLib1.h"

extern unsigned int if_nametoindex(const char *if_name);

#define VL_NAME_SIZE 16
#define VL_MACLEN 6
#define ETH_P_ANAVLINK 0xACDC
#define UDP_VLCONFIG 11100
#define UDP_VLINKTUN 11200
#define MAX_BIND_TRIES 10
#define TCP_QUEUE_LEN 1

#define MX_NO_DETACH 0
#define MX_DETACH 1

/* arbitrary default MTU for vlink */
#define VLINK_MTU 1514

/* pattern to indicate "more data" when sending to vlconfig */
#define MAGIC_COOKIE "!#$&?"

/* error codes */
#define VL_NOERR 1
#define VL_ERR_NULL 0
#define VL_ERR_MISC -1
#define VL_ERR_NOT_FOUND -2
#define VL_ERR_NOMEM -3

/* macros for printf-style manipulation of MAC addresses */
#define ETH_HEX(addr) \
	((unsigned char *)addr)[0], \
	((unsigned char *)addr)[1], \
	((unsigned char *)addr)[2], \
	((unsigned char *)addr)[3], \
	((unsigned char *)addr)[4], \
	((unsigned char *)addr)[5]

#define ETH_FMT "%02x:%02x:%02x:%02x:%02x:%02x"
#define NIPQUAD_FMT "%u.%u.%u.%u"

struct ifs_ptr {

	/* structure used to point to real interfaces */
	/* points to netdevice in kernel mode and     */
	/* to ifaddrs struct in userspace mode        */

	void *ptr;
	int if_index;		/* for userspace to avoid calling if_nametoindex */
	struct ifs_ptr *next;
};

struct udp_tun {

	/* structure used to point to remote UDP peers */

	struct sockaddr_in udpdst;
	struct udp_tun *next;
};

struct vlink_s {

	char vl_name[VL_NAME_SIZE];		/* vlink object name */
	unsigned char vl_idx;			/* index, e.g. 0 for vlink0 */
	anaLabel_t sendto;			/* IDP used by brick to send data to vlink */
	anaLabel_t binding;			/* IDP used for bind/unbind operations */
	anaLabel_t upcall;			/* After bind, IDP towards brick to send incoming data to */
	anaHandle_t caller;			/* After bind, handle of brick bound to vlink */
	uint32_t vl_id;				/* vlink ID in network order */
	uint32_t vl_flags;			/* vlink flags, similar to interface flags */
	uint16_t vl_mtu;			/* lowest MTU for underlying interfaces */
	unsigned char vl_mac[VL_MACLEN];	/* vlink "virtual" MAC address */
	struct ifs_ptr *vl_iflist;		/* pointers to real interfaces */	
	int vl_ifcnt;				/* number of real interfaces it points to */
	struct udp_tun *vl_udplist;		/* pointers to remote UDP peers */
	int vl_udpcnt;				/* number of udp peers it points to */
	struct vlink_s *vl_next;		/* pointer to next */
};

#ifdef USER
#include <ifaddrs.h>

pthread_t configThread; /* added by Ghazi to be able to stop the cobfig thread at brick_exit() call*/

int user_vlink_init(void);
void user_vlink_exit(void);
int user_vlink_check_ifs(void);
void * user_vlink_loop(void *dummy);
int user_nl_recv(void);
int user_vl_recv(int fd);
int user_vlink_send(struct vlink_s*, void *data, unsigned int len);
#endif

#ifdef KERNEL
int kernel_vlink_init(void);
void kernel_vlink_exit(void);
int kernel_vlink_send(struct vlink_s*, void *data, unsigned int len);
int kernel_vlink_l2recv(struct sk_buff *skb, 
			struct net_device *indev, 
			struct packet_type *pt,
			struct net_device *outdev);
void kernel_vlink_cli(struct sock *sk, int len);
void kernel_vlink_l4recv(struct sock *sk, int len);

void kernel_vlink_cli_data(struct work_struct *ctrl);
struct wq_wrapperVlink {
        struct work_struct worker;
	struct sk_buffList *skbList;
};

struct sk_buffList {
	struct sk_buff *skb;
	struct sk_buffList *next;
};
#endif

/* Common functions */

int vlink_cli_command(char *recvbuf, int len, struct sockaddr_in *replyto);
char* vlink_create(uint32_t vl_id, char *keywords);	/* create vlink */
void vlink_bind(void *data, int len, anaLabel_t input, void *aux);

int vlink_del(char *vl_name);	/* delete vlink */
void vlink_clean(struct vlink_s *vlink); /* free vlink */
void vlink_clean_all(void);	/* delete all vlinks */

int vlink_up(char *vl_name);	/* set vlink up */
int vlink_down(char *vl_name);	/* set vlink down */

int vlink_publish(struct vlink_s *vlink, char *description);
int vlink_unpublish(struct vlink_s *vlink);

int vlink_demux(char *data, int len, int iif, struct sockaddr_in *peer);
int vlink_upcall(struct vlink_s *vlink, char *data, int len, int iif, struct sockaddr_in *peer);

int vlink_add_if(char *vl_name, char *if_name); /* add (real) interface to vlink */
int vlink_del_if(char *vl_name, char *if_name); /* remove (real) interface from vlink */

void vlink_check_mac(struct vlink_s *vlink, unsigned char *newmac); /* derive vl_mac from newmac if not set yet */

int vlink_add_udp(char *vl_name, struct sockaddr_in *peer); /* add remote UDP peer to vlink */
int vlink_del_udp(char *vl_name, struct sockaddr_in *peer); /* remove remote UDP peer from vlink */

int vlink_add_pf_if(char *if_name); /* add real interface to PF_PACKET interface list */

void vlink_send(void *data, int len, anaLabel_t input, void *aux);

struct vlink_s* vlink_lookup(char *vl_name); /* find vlink from name */

/* "global" data objects */

/* sockets */
#ifdef USER
extern int vl_clisock;
extern int vl_rawsock;
#endif
#ifdef KERNEL
extern struct socket *vl_clisock;
#endif

/* socket for UDP tunnels */
#ifdef USER
extern int vl_udptun;
#endif
#ifdef KERNEL
extern struct socket *vl_udptun;
#endif

extern struct vlink_s *vlink_head;	/* head of vlink list */
extern unsigned char vl_index_next;	/* vl_index next index to use */
extern unsigned char vl_cliport;	/* used to create unique VMAC addresses */
extern struct ifaddrs *pf_list;		/* list of PF_PACKET interfaces */

#endif /* _ANAVLINK_H */
