/* C/bricks/vlink/kernel/kernel_vlink.c
 *  
 * kernel space vlink implementation
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */


// #include "anaLib0.h"
#include "../ana_vlink.h"

#include <linux/netdevice.h>
#include <linux/udp.h>
#include <linux/ip.h>

static int vlink_netdev_event(struct notifier_block *this, unsigned long event, void *ptr);

static struct notifier_block vlink_notifier = {notifier_call: vlink_netdev_event,}; 
static void kernel_vlink_demux_queue(struct work_struct *msg);

struct packet_type anavlink_packet;
struct socket *vl_clisock = NULL;  /* socket for recv CLI commands */
struct socket *vl_udptun = NULL;   /* socket for UDP tunnels */

/* lock for list of received skb to pass to vlink_demux */
spinlock_t vl_lock = SPIN_LOCK_UNLOCKED;

struct wq_wrapper wq_cli;
struct wq_wrapperVlink wq_recv;
struct workqueue_struct *vl_wq;


void kernel_vlink_exit(void)
{
	if (vl_wq) {
		flush_workqueue(vl_wq);
		destroy_workqueue(vl_wq);
		vl_wq = NULL;
        }
	unregister_netdevice_notifier(&vlink_notifier);
	dev_remove_pack(&anavlink_packet);
	
	if (vl_clisock)
	{
		anaPrint(ANA_DEBUG, "Closing vl_clisock socket\n");
		sock_release(vl_clisock);
		vl_clisock = NULL;
	}
	
	if (vl_udptun)
	{
		anaPrint(ANA_DEBUG, "Closing vl_udptun socket\n");
		sock_release(vl_udptun);
		vl_udptun = NULL;
	}
	
	anaPrint(ANA_DEBUG, "called kernel_vlink_exit\n");
}

int kernel_vlink_init(void)
{
	struct sockaddr_in ctrl;    /* for the CLI */
	struct sockaddr_in udptun;  /* for UDP tunneling */
	uint16_t port;
	int ret;
	
	/* to receive if events (up, down, etc ...) */
	register_netdevice_notifier(&vlink_notifier);
	
	/* register ANA_VLINK packet type */
	memset(&anavlink_packet, 0, sizeof(struct packet_type));
	anavlink_packet.type = htons(ETH_P_ANAVLINK);	
	anavlink_packet.dev = NULL;
	anavlink_packet.func = kernel_vlink_l2recv;
	dev_add_pack(&anavlink_packet);
	
	/* create UDP socket for CLI sessions */
	if ((sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &vl_clisock)) < 0)
	{
		anaPrint(ANA_ERR, "cannot create UDP CLI socket");
		return -1;
	}
	
	memset(&ctrl, 0, sizeof(struct sockaddr_in));
	ctrl.sin_family = AF_INET;
	ctrl.sin_addr.s_addr = in_aton("127.0.0.1");
	port = UDP_VLCONFIG;
	
	while (1)
	{
		ctrl.sin_port = htons(port);
		ret = vl_clisock->ops->bind(vl_clisock, (struct sockaddr *) &ctrl, sizeof(ctrl));
		
		if (ret >= 0)
		{
			anaPrint(ANA_NOTICE, "Using CLI udp port %d\n", port);
			vl_cliport = port & 0x00FF;
			break;
		}
		else
			port++;
		
		if ((port - UDP_VLCONFIG) > MAX_BIND_TRIES)
		{
			anaPrint(ANA_ERR, "bind vl_clisock failed (tried ports %d to %d)\n", 
				 UDP_VLCONFIG, UDP_VLCONFIG+MAX_BIND_TRIES);
			
			sock_release(vl_clisock);
			vl_clisock = NULL;
			return -1;	
		}
	}
	/* callback to handle incoming UDP messages for CLI */
	vl_clisock->sk->sk_data_ready = kernel_vlink_cli;
		
	/* create UDP socket for UDP tunnels */
	if ((sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &vl_udptun)) < 0)
	{
		anaPrint(ANA_ERR, "cannot create UDP tunnels socket");
		return -1;
	}
	
	memset(&udptun, 0, sizeof(struct sockaddr_in));
	udptun.sin_family = AF_INET;
	udptun.sin_addr.s_addr = htonl(INADDR_ANY);
	port = UDP_VLINKTUN;
	
	while (1)
	{
		udptun.sin_port = htons(port);
		ret = vl_udptun->ops->bind(vl_udptun, (struct sockaddr *) &udptun, sizeof(udptun));
		
		if (ret >= 0)
		{
			anaPrint(ANA_NOTICE, "Using TUNNEL udp port %d\n", port);
			break;
		}
		else
			port++;
		
		if ((port - UDP_VLINKTUN) > MAX_BIND_TRIES)
		{
			anaPrint(ANA_ERR, "bind vl_udptun failed (tried ports %d to %d)\n", 
				 UDP_VLINKTUN, UDP_VLINKTUN+MAX_BIND_TRIES);
			
			sock_release(vl_udptun);
			vl_udptun = NULL;
			sock_release(vl_clisock);
			vl_clisock = NULL;
			return -1;	
		}
	}
	/* callback to handle incoming UDP packets */
	vl_udptun->sk->sk_data_ready = kernel_vlink_l4recv;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	INIT_WORK(&wq_cli.worker, kernel_vlink_cli_data, &wq_cli);
	INIT_WORK(&wq_recv.worker, kernel_vlink_demux_queue, &wq_recv);
#else 	
	INIT_WORK(&wq_cli.worker, kernel_vlink_cli_data);
	INIT_WORK(&wq_recv.worker, kernel_vlink_demux_queue);
#endif
	/*initialize the work queue*/
        vl_wq = create_singlethread_workqueue("vlink_wq"); 
        if (!vl_wq)
	{
               anaPrint(ANA_ERR, "cannot create work queue!\n");
	       return -ENOMEM;
        }
	return 1;

}

void kernel_vlink_cli(struct sock *sk, int len)
{
	/* this simply schedules kernel_vlink_cli_data */
	/* don't dequeue skb here! */
		
	wq_cli.sk = sk;
	queue_work(vl_wq, &wq_cli.worker);
}

void kernel_vlink_cli_data(struct work_struct *ctrl)
{
	struct wq_wrapper *foo = container_of(ctrl, struct wq_wrapper, worker);
	struct sockaddr_in replyto;
	struct udphdr *udph;
	int udplen = sizeof(struct udphdr);
	int len = 0;
	
	if (!foo)
	{
		anaPrint(ANA_ERR, "kernel_vlink_cli_data --> work queue data NULL!\n");
		return;
	}
	
	while ((len = skb_queue_len(&foo->sk->sk_receive_queue)) > 0) { 
	
		/* len holds the number of packets in the receive queue of this socket */
	
		struct sk_buff *skb = NULL;
 		skb = skb_dequeue(&foo->sk->sk_receive_queue);
		udph = (struct udphdr*) skb->data;
		
		memset(&replyto, 0, sizeof(struct sockaddr_in));
		replyto.sin_family = AF_INET;
		replyto.sin_addr.s_addr = in_aton("127.0.0.1");
		replyto.sin_port = udph->source;
	
		/* vlink_cli_command is shared (userspace + kernel code) */
		/* so we don't call it directly in the work queue */
	
		vlink_cli_command(skb->data + udplen, skb->len - udplen, &replyto);
		kfree_skb(skb);
	}
}

int vlink_netdev_event(struct notifier_block *this, unsigned long event, void *ptr)
{
	struct net_device *dev = (struct net_device *) ptr;
	
	if (!dev)
		return NOTIFY_DONE;
		
	switch (event) {
	
	case NETDEV_REGISTER:
	case NETDEV_UP:
		anaPrint(ANA_NOTICE, "Interface %s (%d) going up\n", dev->name, dev->ifindex);
		
		return NOTIFY_DONE;
		break;
		
	case NETDEV_UNREGISTER:
	case NETDEV_DOWN:
	
		anaPrint(ANA_NOTICE, "Interface %s (%d) going down\n", dev->name, dev->ifindex);
		
		return NOTIFY_DONE;
		break;

	default:
		break;
	}
	
	return NOTIFY_DONE;
}

int kernel_vlink_send(struct vlink_s *vlink, void *data, unsigned int len)
{
	struct net_device *dev = NULL;
	struct sk_buff *skb;
	struct ifs_ptr *cur_if = vlink->vl_iflist;
	struct udp_tun *cur_tun = vlink->vl_udplist;
	struct sockaddr bcast = {AF_UNSPEC, {0xff,0xff,0xff,0xff,0xff,0xff}};
	
	while (cur_if)
	{
		/* we call dev_get_by_index instead of using cur->ptr */
		/* to make sure netdevice still exists */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		dev = dev_get_by_index(cur_if->if_index);
#else
		dev = dev_get_by_index(&init_net, cur_if->if_index);
#endif	
		if (!dev)
			anaPrint(ANA_DEBUG, "kernel_vlink_send --> dev (%d) NULL!\n", cur_if->if_index);
		else
		{
			skb = alloc_skb(dev->hard_header_len + len, GFP_ATOMIC);
			
			if (!skb)
			{
				anaPrint(ANA_ERR, "kernel_vlink_send --> cannot allocate skb!\n");
				dev_put(dev);
				return -1;
			}
			
			skb->dev = dev;
			skb->protocol = htons(ETH_P_ANAVLINK);
			skb_reserve(skb, dev->hard_header_len); /* to insert ETH header */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
			skb->nh.raw = skb->data;
#else
			skb_reset_network_header(skb);
#endif			
			memcpy(skb->data, data, len);
			skb->len = len;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
			if (dev->hard_header)
			{
				dev->hard_header(skb, dev, ETH_P_ANAVLINK, bcast.sa_data, 0, skb->len);
#else
			if (dev_hard_header(skb, dev, ETH_P_ANAVLINK, bcast.sa_data, 0, skb->len)) {
#endif
				dev_queue_xmit(skb);
			}
			else
				anaPrint(ANA_ERR, "kernel_vlink_send --> no hard header for %s\n", dev->name);
							
			anaPrint(ANA_DEBUG, "kernel_vlink_send --> sending %i bytes to %s\n", skb->len, dev->name);
			dev_put(dev);
		}
		
		cur_if = cur_if->next;
	}
	
	if (!cur_tun)
		return 1;
	
	/* send to UDP peer(s) */
	while (cur_tun)
	{
		anaPrint(ANA_DEBUG, "kernel_vlink_send --> sending to %u.%u.%u.%u:%u\n",
			 NIPQUAD(cur_tun->udpdst.sin_addr), ntohs(cur_tun->udpdst.sin_port) );
		sendSocketWrapper(&vl_udptun, AF_INET, &cur_tun->udpdst, data, len);
		cur_tun = cur_tun->next;
	}
	
	return 1;
}

static void kernel_vlink_demux_queue(struct work_struct *msg) {
	struct wq_wrapperVlink *foo = container_of(msg, struct wq_wrapperVlink, worker);
	struct sk_buffList *skbList = foo->skbList;
	unsigned long flags;
	while ((skbList = foo->skbList) != NULL) {
		
		struct sk_buff * skb = skbList->skb;
		char *data = skb->data;
		int len = skb->len;
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
		if (skb->data == skb->nh.raw)
#else
		if (skb->data == skb_network_header(skb))
#endif
			vlink_demux(data, len, skb->dev->ifindex, NULL);
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		else if (skb->data == skb->h.raw){
			struct iphdr *ip = (struct iphdr*)skb->nh.raw;
			struct udphdr *udph = (struct udphdr*)skb->h.raw;

#else
		else if (skb->data == skb_transport_header(skb)){
			struct iphdr *ip = (struct iphdr *)skb_network_header(skb);
			struct udphdr *udph = (struct udphdr *)skb_transport_header(skb);
#endif
		
			struct sockaddr_in peer;
			int udplen = sizeof(struct udphdr);
		
			memset(&peer, 0, sizeof(struct sockaddr_in));
			peer.sin_family = AF_INET;
			peer.sin_addr.s_addr = ip->saddr;
			peer.sin_port = udph->source;
			
			vlink_demux(skb->data + udplen, skb->len - udplen, -1, &peer);		
		}
		
		kfree_skb(skb);
		spin_lock_irqsave(&vl_lock, flags);
		foo->skbList = skbList->next;
		spin_unlock_irqrestore(&vl_lock, flags);
		free(skbList);
	}
}

void kernel_vlink_l4recv(struct sock *sk, int len)
{
	/* this schedules kernel_vlink_demux_queue */
	
	int i, qlen = 0;
	struct sk_buffList *skbList;
	struct sk_buffList *temp;
	struct sk_buff *skb = NULL;
	struct iphdr *ip;
	struct udphdr *udph;
	unsigned long flags;
	
	while ((qlen = skb_queue_len(&sk->sk_receive_queue)) > 0) { 
	
		/* qlen holds the number of packets in the receive queue of this socket */
	
		skb = skb_dequeue(&sk->sk_receive_queue);
		
		if (skb->len < (sizeof(struct udphdr) + 4) ) /* udp header + vl_id */
		{
			anaPrint(ANA_ERR, "kernel_vlink_l4recv --> data too short to extract vl_id!\n");
			kfree_skb(skb);
			continue;
		}
		
 		temp = wq_recv.skbList;
		skbList = kmalloc(GFP_ATOMIC, sizeof(struct sk_buffList));
		skbList->skb = skb;
		skbList->next = NULL;
		
		spin_lock_irqsave(&vl_lock, flags);
		if (temp == NULL)
			wq_recv.skbList = skbList;
		else {
			while (temp->next != NULL)
				temp = temp->next;
			temp->next = skbList;
		}
		spin_unlock_irqrestore(&vl_lock, flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
		ip = (struct iphdr *)skb->nh.raw;
		udph = (struct udphdr *)skb->h.raw;
#else
		ip = (struct iphdr *) skb_network_header(skb);
		udph = (struct udphdr *) skb_transport_header(skb);
#endif
		if (ip && udph)
		{
			anaPrint(ANA_DEBUG, "kernel_vlink_l4recv --> recv %d bytes from %u.%u.%u.%u:%u\n", 
		 	(skb->len - sizeof(struct udphdr)), NIPQUAD(ip->saddr), ntohs(udph->source) );
		}
		else
		{
			anaPrint(ANA_ERR, "kernel_vlink_l4recv --> recv %d bytes but cannot extract peer info\n", skb->len);
			kfree_skb(skb);
			continue;
		}
			
		if (anaDebugLevel >= ANA_DEBUG)
		{
			printk("\n");
        		for (i = sizeof(struct udphdr); i < skb->len; i++) {
               			printk("%02x ",skb->data[i]);
        		}
			printk("\n\n");
		}
	}
			
	/* call vlink_demux in a work queue/task queue to avoid sleeping in interrupt context */	
	queue_work(vl_wq, &wq_recv.worker);
}

int kernel_vlink_l2recv(struct sk_buff *skb, 
		 struct net_device *indev, 
		 struct packet_type *pt,
		 struct net_device *outdev)
{
	int i = 0;
	struct sk_buffList *skbList;
	struct sk_buffList *temp = wq_recv.skbList;
	unsigned long flags;
	
	if (skb->len < 4)
	{
		anaPrint(ANA_ERR, "kernel_vlink_l2recv --> data too short to extract vl_id!\n");
		kfree_skb(skb);
		return -1;
	}
	
	skbList = kmalloc(GFP_ATOMIC, sizeof(struct sk_buffList));
	skbList->skb = skb;
	skbList->next = NULL;
	
	spin_lock_irqsave(&vl_lock, flags);
	if (temp == NULL)
		wq_recv.skbList = skbList;
	else {
		while (temp->next != NULL)
			temp = temp->next;
		temp->next = skbList;
	}
	spin_unlock_irqrestore(&vl_lock, flags);
	anaPrint(ANA_DEBUG, "kernel_vlink_l2recv --> recv %d bytes from %s\n", 
		 skb->len, skb->dev ? skb->dev->name : "skd->dev NULL");
	
	if (anaDebugLevel >= ANA_DEBUG)
	{
		printk("\n");
        	for (i = 0; i < skb->len; i++) {
               		printk("%02x ",skb->data[i]);
        	}
		printk("\n\n");
	}
	
	/*call vlink_demux in a work queue/task queue to avoid sleeping in interrupt context */	
	queue_work(vl_wq, &wq_recv.worker);
	return 1;
}


