/* C/bricks/vlink/userspace/user_vlink.c
 * 
 * user space vlink implementation
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *  
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include <linux/types.h>
#include <string.h>
#include <unistd.h>
#include <ifaddrs.h>
#include <sys/socket.h>
#include <linux/if_arp.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/if.h>
#include <pthread.h>

#include "../ana_vlink.h"

#define NETLINK_BUF_SIZE 1024
#define RECV_BUF_SIZE 2048

extern unsigned int if_nametoindex (__const char *__ifname) __THROW;
extern unsigned long int anaThread(void * (*start_routine)(void *), void *arg, char *name);

int nlsock = -1;      /* netlink socket */
int vl_rawsock = -1;  /* raw socket for sending data */
int vl_clisock = -1;  /* socket for recv CLI commands */
int vl_udptun = -1;   /* for UDP tunneling */

struct ifaddrs *ifs_list = NULL;
struct ifaddrs *pf_list = NULL;

int user_vlink_init(void)
{
	struct sockaddr_nl netlink; /* netlink specific stuff */
	struct sockaddr_in ctrl;    /* for the CLI */
	struct sockaddr_in udptun;  /* for UDP tunneling */
	uint16_t port;
	int ret, on = 1;
	
	/* get initial list of interfaces */
	user_vlink_check_ifs();
	
	/* setup netlink socket to receive notifications of link events */
	memset (&netlink, 0, sizeof(struct sockaddr_nl));
	netlink.nl_family = AF_NETLINK;
	netlink.nl_pid = getpid();
	netlink.nl_groups = RTMGRP_LINK;

	if( (nlsock = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) == -1 )
	{
		anaPrint(ANA_ERR, "netlink socket create failed %i\n", nlsock);
		return VL_ERR_MISC;
	}
	
	if ( (ret = bind(nlsock, (struct sockaddr*)&netlink, sizeof(struct sockaddr_nl))) < 0 )
	{
		anaPrint(ANA_ERR, "netlink bind failed: %i\n", ret);
		return VL_ERR_MISC;
	}

	/* socket for sending and receiving data on real interfaces */
	if ((vl_rawsock = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ANAVLINK)) ) < 0)
	{
		anaPrint(ANA_ERR, "create vl_rawsock failed %i\n", vl_rawsock);
		return VL_ERR_MISC;
	}

	/* UDP socket for receiving CLI commands */
	if ((vl_clisock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		anaPrint(ANA_ERR, "create vl_ctrlsock failed %i\n", vl_clisock);
		return VL_ERR_MISC;
	}	

	memset(&ctrl, 0, sizeof(struct sockaddr_in));
	ctrl.sin_family = AF_INET;
	inet_aton("127.0.0.1", &ctrl.sin_addr);
	port = UDP_VLCONFIG;
	
	while (1)
	{
		ctrl.sin_port = htons(port);
		ret = bind(vl_clisock, (struct sockaddr *)&ctrl, sizeof(struct sockaddr_in));
		
		if (ret == 0)
		{
			anaPrint(ANA_NOTICE, "Using CLI udp port %d\n", port);
			vl_cliport = port & 0x00FF;
			break;
		}
		else if (ret == -1)
			port++;
		
		if ((port - UDP_VLCONFIG) > MAX_BIND_TRIES)
		{
			anaPrint(ANA_ERR, "bind vl_clisock failed (tried ports %d to %d)\n", 
				 UDP_VLCONFIG, UDP_VLCONFIG+MAX_BIND_TRIES);
			return VL_ERR_MISC;	
		}
	}
	
	if ( (ret = setsockopt(vl_clisock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int))) < 0 )
	{
		anaPrint(ANA_ERR, "setsockopt vl_clisock failed %i\n", ret);
		return VL_ERR_MISC;
	}

	/* socket for sending and receiving data via UDP "tunnels" */
	if ((vl_udptun = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		anaPrint(ANA_ERR, "create vl_udptun failed %i\n", vl_udptun);
		return VL_ERR_MISC;
	}	

	memset(&udptun, 0, sizeof(struct sockaddr_in));
	udptun.sin_family = AF_INET;
	udptun.sin_addr.s_addr = htonl(INADDR_ANY);
	port = UDP_VLINKTUN;
	
	while (1)
	{
		udptun.sin_port = htons(port);
		ret = bind(vl_udptun, (struct sockaddr *)&udptun, sizeof(struct sockaddr_in));
		
		if (ret == 0)
		{
			anaPrint(ANA_NOTICE, "Using TUNNEL udp port %d\n", port);
			break;
		}
		else if (ret == -1)
			port++;
		
		if ((port - UDP_VLINKTUN) > MAX_BIND_TRIES)
		{
			anaPrint(ANA_ERR, "bind vl_udptun failed (tried ports %d to %d)\n", 
				 UDP_VLINKTUN, UDP_VLINKTUN+MAX_BIND_TRIES);
			return VL_ERR_MISC;	
		}
	}
	
	if ( (ret = setsockopt(vl_udptun, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int))) < 0 )
	{
		anaPrint(ANA_ERR, "setsockopt vl_udptun failed %i\n", ret);
		return VL_ERR_MISC;
	}

	/* run listening thread */
	
	anaThread((void *)user_vlink_loop, NULL, "user_vlink_loop");
	
	/* old stuff, still needed?
	pthread_create(&configThread, NULL, (void *) user_vlink_loop, NULL);
	pthread_detach(configThread);*/
	
	return VL_NOERR;
}

void user_vlink_exit(void)
{
	if (vl_clisock >= 0)
	{
		close(vl_clisock);
		vl_clisock = -1;
	}
	
	if (vl_udptun >= 0)
	{
		close(vl_udptun);
		vl_udptun = -1;
	}
	
	if (vl_rawsock >= 0)
	{
		close(vl_rawsock);
		vl_rawsock = -1;
	}
	
	if (nlsock >= 0)
	{
		close(nlsock);
		nlsock = -1;
	}
	
	if (ifs_list)
	{
		freeifaddrs(ifs_list);
		ifs_list = NULL;
	}
}

int user_vlink_send(struct vlink_s *vlink, void *data, unsigned int len)
{
	struct sockaddr_ll ll;
	unsigned int socklen = sizeof(struct sockaddr_ll);
	char ip_str[INET_ADDRSTRLEN];
	
	struct ifs_ptr *cur_if = vlink->vl_iflist;
	struct udp_tun *cur_tun = vlink->vl_udplist;
	struct ifaddrs *if_data;
		
	ll.sll_family = AF_PACKET;
	ll.sll_protocol = htons(ETH_P_ANAVLINK);
	ll.sll_hatype = 0;
	ll.sll_pkttype = 0;
	memset(ll.sll_addr, 0xFF, 8);
	ll.sll_halen = 6;
	
	while (cur_if)
	{
		ll.sll_ifindex = cur_if->if_index;
		if_data = (struct ifaddrs *) cur_if->ptr;
		
		if (sendto(vl_rawsock, data, len, 0, (struct sockaddr*)&ll, socklen) == -1)
		{
			anaPrint(ANA_ERR, "user_vlink_send(%s) --> sendto %s(%d) failed!\n", 
				 vlink->vl_name, if_data->ifa_name, cur_if->if_index);
			perror("sendto");
		}
		else
			anaPrint(ANA_DEBUG, "user_vlink_send(%s) --> sending to %s\n", 
				 vlink->vl_name, if_data->ifa_name);
			
		cur_if = cur_if->next;		
	}
	
	socklen = sizeof(struct sockaddr_in);
	while (cur_tun)
	{
		if (anaDebugLevel >= ANA_ERR)
			inet_ntop(AF_INET, &cur_tun->udpdst.sin_addr, ip_str, INET_ADDRSTRLEN);
	
		if (sendto(vl_udptun, data, len, 0, (struct sockaddr*)&cur_tun->udpdst, socklen) == -1)
		{
			anaPrint(ANA_ERR, "user_vlink_send(%s) --> sendto %s:%u failed!\n", 
				 vlink->vl_name, ip_str, ntohs(cur_tun->udpdst.sin_port) );
			perror("sendto");
		}
		else
			anaPrint(ANA_DEBUG, "user_vlink_send(%s) --> sending to %s:%u\n", 
				 vlink->vl_name, ip_str, ntohs(cur_tun->udpdst.sin_port) );
				 
		cur_tun = cur_tun->next;
	}
	
	return 1;
}

int user_vlink_check_ifs(void)
{
	struct ifaddrs *cur, *prev, *ifs;
	int ret;
	if (ifs_list) /* free previous before re-calling */
	{
		freeifaddrs(ifs_list);
		ifs_list = NULL;
	}
	
	if ( (ret = getifaddrs(&ifs_list)) == -1 )
	{
		anaPrint(ANA_ERR, "getifaddrs failed %i\n", ret);
		return -1;
	}
	
	ifs = ifs_list;
	
	do
	{			
		switch (ifs->ifa_addr->sa_family) {
			
		/* getifaddrs returns interfaces multiple times */
		/* we only care about the AF_PACKET instance */
		/* and store them in pf_list */
			
		case AF_PACKET:
				
			if (!pf_list) /* list is NULL */
			{
				pf_list = malloc(sizeof(struct ifaddrs));
				
				if (!pf_list)
				{
					anaPrint(ANA_ERR, "user_vlink_check_ifs --> malloc failed!\n");
					return -1;
				}
				
				anaPrint(ANA_DEBUG, "user_vlink_check_ifs --> adding %s to pf_list\n",
							ifs->ifa_name);
				memcpy(pf_list, ifs, sizeof(struct ifaddrs));
				pf_list->ifa_next = NULL;
			}
			else
			{
				cur = pf_list;
				prev = NULL;
				
				while (cur)
				{
					if (!strcmp(ifs->ifa_name, cur->ifa_name))
						break;
						
					prev = cur;
					cur = cur->ifa_next;
				}
				
				if (!cur) /* interface not in pf_list */		
				{
					cur = malloc(sizeof(struct ifaddrs));
					
					if (!cur)
					{
						anaPrint(ANA_ERR, "user_vlink_check_ifs --> malloc failed!\n");
						return -1;
					}
					
					anaPrint(ANA_DEBUG, "user_vlink_check_ifs --> adding %s to pf_list\n",
							ifs->ifa_name);
					memcpy(cur, ifs, sizeof(struct ifaddrs));
					cur->ifa_next = NULL;
					prev->ifa_next = cur;
				}
			}
			break;
		
		default:
			break;
		}
	}
	while ( (ifs = ifs->ifa_next) );
		
	/* check that all interfaces in pf_list still exist */
	prev = NULL;
	cur = pf_list;
	
	while (cur)
	{
		ifs = ifs_list;
		
		while (ifs)
		{
			if (!strcmp(ifs->ifa_name, cur->ifa_name))
				break;
			
			ifs = ifs->ifa_next;
		}
		
		if (!ifs) /* interface in pf_list does no longer exist */
		{
			if (prev)
				prev->ifa_next = cur->ifa_next;
			else
				pf_list = cur->ifa_next;
			
			anaPrint(ANA_DEBUG, "user_vlink_check_ifs --> removing %s from pf_list\n", cur->ifa_name);
			free(cur);
			
			if (prev)
				cur = prev->ifa_next;
			else
				cur = pf_list;
		}
		else
		{
			prev = cur;
			cur = cur->ifa_next;
		}
	}
	
	return 1;
}

void * user_vlink_loop(void *dummy)
{
	fd_set readfd;
	int ret = 0;
	char recvbuf[128];
	int len;
	struct sockaddr_in replyto;
	unsigned int fromlen = sizeof(struct sockaddr_in);
	
	while(attached)
	{
		FD_ZERO(&readfd);
		
		FD_SET(vl_rawsock, &readfd);
		FD_SET(nlsock, &readfd);
		FD_SET(vl_clisock, &readfd);
		FD_SET(vl_udptun, &readfd);
		
		ret = select(255, &readfd, NULL, NULL, NULL);
		
		if (ret > 0)
		{
			if (FD_ISSET(nlsock, &readfd))
				user_nl_recv();
				
			if (FD_ISSET(vl_rawsock, &readfd))
				user_vl_recv(vl_rawsock);
			
			if (FD_ISSET(vl_udptun, &readfd))
				user_vl_recv(vl_udptun);
			
			if (FD_ISSET(vl_clisock, &readfd))
			{
				/* we read here because vlink_cli_command */
				/* is a userspace + kernel function */
					
				len = recvfrom(vl_clisock, recvbuf, 128, 0,
					       (struct sockaddr*)&replyto, &fromlen);
					
				if (len == -1)
					anaPrint(ANA_ERR, "user_vlink_loop: read(vl_clisock) failed!\n");
				else
					vlink_cli_command(recvbuf, len, &replyto);
			}		
		}
	}
	
	return NULL; /* to make compiler happy */
}

int user_vl_recv(int fd)
{
	int bytes;
	struct sockaddr_ll ll;
	struct sockaddr_in in;
	unsigned int socklen;
	char recvbuf[RECV_BUF_SIZE];
	
	if (fd == vl_rawsock)
	{
		socklen = sizeof(struct sockaddr_ll);
		bytes = recvfrom(vl_rawsock, recvbuf, RECV_BUF_SIZE, 0, (struct sockaddr*)&ll, &socklen);
	}
	else if (fd == vl_udptun)
	{
		socklen = sizeof(struct sockaddr_in);
		bytes = recvfrom(vl_udptun, recvbuf, RECV_BUF_SIZE, 0, (struct sockaddr*)&in, &socklen);
	}
	else
	{
		anaPrint(ANA_ERR, "user_vl_recv --> unknown file descriptor!\n");
		return -1;
	}
	
	if (bytes == -1)
	{
		anaPrint(ANA_ERR, "user_vl_recv --> recvfrom failed!\n");
		return -1;
	}
	
	if (bytes < 4)
	{
		anaPrint(ANA_ERR, "user_vl_recv --> data too short to extract vl_id!\n");
		return -1;
	}
	
	if (fd == vl_rawsock)
	{
		anaPrint(ANA_DEBUG, "user_vl_recv --> recv %d bytes from ifindex %d\n", bytes, ll.sll_ifindex);
		vlink_demux(recvbuf, bytes, ll.sll_ifindex, NULL);
		return 1;
	}
	else
	{
		char ip_str[INET_ADDRSTRLEN];
		
		if (anaDebugLevel >= ANA_DEBUG) /* avoids doing inet_ntop if not necessary */
			inet_ntop(AF_INET, &in.sin_addr, ip_str, INET_ADDRSTRLEN);
			
		anaPrint(ANA_DEBUG, "user_vl_recv --> recv %d bytes from %s:%u\n", 
			 bytes, ip_str, ntohs(in.sin_port) );
		
		vlink_demux(recvbuf, bytes, -1, &in);
		return 1;
	}
}

int user_nl_recv(void)
{	
	int bytes, len;
	char nlbuf[NETLINK_BUF_SIZE];
	struct nlmsghdr *nh = (struct nlmsghdr *) nlbuf;
	
	/* rtnetlink specific */
	struct ifinfomsg *ifmsg;
	struct rtattr *rta;
	char *c = NULL;
	int if_mtu;
	
	memset(nlbuf, 0, NETLINK_BUF_SIZE);
	nh = (struct nlmsghdr *) nlbuf;
	
	bytes = recv(nlsock, nlbuf, NETLINK_BUF_SIZE, 0); 
				
	for (; NLMSG_OK(nh, bytes); nh = NLMSG_NEXT(nh, bytes))
	{
    		/* The end of multipart message. */
    		if (nh->nlmsg_type == NLMSG_DONE)
        		continue;
	
		if (nh->nlmsg_type == NLMSG_ERROR)
        		continue; /* Do some error handling. */
    	
		ifmsg = (struct ifinfomsg *) NLMSG_DATA(nh);
		rta = (struct rtattr*)((char*)NLMSG_DATA(nh) + NLMSG_ALIGN(sizeof(struct ifinfomsg)));
		
		len = nh->nlmsg_len - NLMSG_LENGTH(sizeof(struct ifinfomsg));
		for(; RTA_OK(rta, len); rta = RTA_NEXT(rta, len))
		{
			switch(rta->rta_type) {
			
			case IFLA_IFNAME:
				c = (char *)RTA_DATA(rta);
				break;
			
			case IFLA_MTU:
				if_mtu = *(int*)RTA_DATA(rta);
				break;
			
			default:
				break; 
			}
		}
			
		/* TODO: call functions to handle ifs ups and downs */
			
		if ( ifmsg->ifi_flags & IFF_UP )
		{	
			anaPrint(ANA_ERR, "Interface %s (%d) going up\n", c, ifmsg->ifi_index);
		}
		else
		{
			anaPrint(ANA_ERR, "Interface %s (%d) going down\n", c, ifmsg->ifi_index);
		}
			
	}
		
	memset(nlbuf, 0, NETLINK_BUF_SIZE);
	nh = (struct nlmsghdr *) nlbuf;
	
	return 1;
}
