/* C/configtools/mxconfig.c
 * 
 * Tool for configuring the ANA minmex
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Christophe Jelger <Christophe.Jelger@unibas.ch>
 *
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>

#define FAIL -1
#define NOERR 0
#define UDP_MINMEX 10100
#define BUFLEN 1024

#define MAGIC_COOKIE "!#$&?"
#define TIMEOUT 3

int send_cmd(char *cmd, int len, uint16_t port);
void print_usage(void);

char buf[BUFLEN];

int main(int argc, char **argv) {

	int cmd_index = 0, arg_index = 0, cmd_arg, port = 0;
	char *cmds[] = {"help", "show", "load", "unload", NULL}; 
	char *show_args[] = {"bricks", "gates", "IDT", "KVR", NULL};
	char *load_args[] = {"brick", NULL};
	int i;
	
	if (argc == 1)
	{
		print_usage();
		return FAIL;
	}
	
	/* argv[1] is either a command or --port_number option */
	
	if (!strncmp(argv[1], "--", 2))
	{
		port = atoi(argv[1]+2);
		cmd_arg = 2; /* argv[2] holds command */
	}
	else
		cmd_arg = 1; /* argv[1] holds command */
	
	/* check that command exists */
	
	while ( cmds[cmd_index] != NULL )
	{
		if (!strcmp(cmds[cmd_index], argv[cmd_arg]))
			break;
		cmd_index++;
	}
	
	if (cmds[cmd_index] == NULL)
	{
		printf("Command unknown!\n");
		return FAIL;
	}
		
	/* check argument if necessary then send command */
	
	if (!strncmp(argv[cmd_arg], "help", 4))
	{
		print_usage();
		return NOERR;
	}
	
	else if (!strncmp(argv[cmd_arg], "show", 4))
	{
		if (argc < 3)
		{
			printf("Not enough arguments!\n");
			return FAIL;
		}
		
		while ( show_args[arg_index] != NULL )
		{
			if (!strcmp(show_args[arg_index], argv[cmd_arg+1]))
				break;
			arg_index++;
		}
	
		if (show_args[arg_index] == NULL)
		{
			printf("Argument unknown!\n");
			return FAIL;
		}
	
		sprintf(buf, "%s_%s", cmds[cmd_index], show_args[arg_index]);
	}
	else if (!strncmp(argv[cmd_arg], "load", 4) || !strncmp(argv[cmd_arg], "unload", 6))
	{
		if (argc < 4)
		{
			printf("Not enough arguments!\n");
			return FAIL;
		}
		while ( load_args[arg_index] != NULL )
		{
			if (!strcmp(load_args[arg_index], argv[cmd_arg+1]))
				break;
			arg_index++;
		}
	
		if (load_args[arg_index] == NULL)
		{
			printf("Argument unknown!\n");
			return FAIL;
		}
		
		sprintf(buf, "%s_%s %s", argv[cmd_arg], load_args[arg_index], argv[cmd_arg+2]);
		
		i=cmd_arg+3;
		while(strlen(buf)<BUFLEN && argv[i]!=NULL){
			strcat(buf, " ");
			strcat(buf, argv[i]);
			i++;
		}
	}
	
	return send_cmd(buf, strlen(buf)+1, port);
}
	
int send_cmd(char *cmd, int len, uint16_t port)
{
	int mxsock;
	struct sockaddr_in minmex;
	unsigned int slen = sizeof(struct sockaddr_in);
	ssize_t nbRecv = -1;
	int ret, end = 0;
	int mcLen = strlen(MAGIC_COOKIE);
	fd_set readfd;
	struct timeval timeout;
	
	if (!cmd)
	{
		printf("send_cmd called with NULL pointer!\n");
		return FAIL;
	}
	
	if ( (mxsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror("socket creation failed");
		return FAIL;
	} 
	
	memset(&minmex, 0, sizeof(struct sockaddr_in));
	minmex.sin_family = AF_INET;
	inet_aton("127.0.0.1", &minmex.sin_addr);
	
	if (!port)
		port = UDP_MINMEX;
	
	printf("[sending command to udp port %d]\n", port);
	
	minmex.sin_port = htons(port);
	
	if( (sendto(mxsock, cmd, len, 0, (struct sockaddr*)&minmex, slen)) < 0)
	{
		perror("send_cmd failed");
		return FAIL;	
	}
	printf("\n");
	
	while (!end) {
	
		memset(buf, 0, BUFLEN);
		FD_ZERO(&readfd);
		FD_SET(mxsock, &readfd);
		timeout.tv_sec = TIMEOUT;
		timeout.tv_usec = 0;		
		
		ret = select(mxsock+1, &readfd, 0, 0, &timeout);
		
		if (!ret)
		{
			printf("No reply received from MINMEX (timeout=%ds, using UDP port %d)\n\n",
				TIMEOUT, port);
			return FAIL;
		}
		
		if (ret == -1)
		{
			perror("select failed");
			return FAIL;		
		}
		
		/* else, read message */
		nbRecv = recvfrom(mxsock, buf, BUFLEN, 0, (struct sockaddr*)&minmex, &slen);
				
		if ( nbRecv < 0)
		{
			perror("read_cmd failed");
			return FAIL;
		}
		
		/* we check for magic cookie */
		if (nbRecv<mcLen || strncmp(buf+nbRecv-mcLen, MAGIC_COOKIE, mcLen)) /* this is the final message*/
			end = 1;
		
		printf("%s", buf); 
	}
	
	printf("\n");
	close(mxsock);
	return NOERR;
}

void print_usage(void)
{
	printf("\nUsage: mxconfig [--port] { COMMAND } { ARGUMENT }\n\n");
	printf("  where: --port := udp port of minmex status socket\n\n");
	printf("  where: COMMAND  := { show }\n");
	printf("    and: ARGUMENT := { gates | bricks | IDT | KVR }\n\n");
	printf("  where: COMMAND  := { load | unload }\n");
	printf("    and: ARGUMENT := { brick }\n\n");
}
