/* C/configtools/vlconfig.c
 * 
 * Tool for configuring the ANA vlink brick
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 *         Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 *
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>

#define FAIL -1
#define NOERR 0
#define UDP_VLCONFIG 11100
#define BUFLEN 1024

#define TIMEOUT 3
#define MAGIC_COOKIE "!#$&?"

int send_cmd(char *cmd, int len, uint16_t port);
void print_usage(void);

char buf[BUFLEN];

int main(int argc, char **argv) {

	int cmd_index = 0, port = 0, cmd_arg;
	char *cmds[] = {"help","list","create","delete","up","down",
			"add_if","rem_if","add_udp","rem_udp", "send", NULL}; 
	char *vl_cmds[] = {"help","list","vlink_create","vlink_delete","vlink_up",
			   "vlink_down","vlink_add_if","vlink_rem_if",
			   "vlink_add_udp","vlink_rem_udp","send_dummy",};
	
	/* the "send" command is for debugging only */
	
	if (argc == 1)
	{
		print_usage();
		return FAIL;
	}
	
	/* argv[1] is either a command or --port_number option */
	
	if (!strncmp(argv[1], "--", 2))
	{
		port = atoi(argv[1]+2);
		cmd_arg = 2; /* argv[2] holds command */
		argc--; /* # of remaining args */
	}
	else
		cmd_arg = 1; /* argv[1] holds command */
	
	/* check that command exists */
	
	while ( cmds[cmd_index] != NULL )
	{
		if (!strcmp(cmds[cmd_index], argv[cmd_arg]))
			break;
		cmd_index++;
	}
	
	if (cmds[cmd_index] == NULL)
	{
		printf("Command unknown!\n");
		return FAIL;
	}
		
	/* check argument if necessary then send command */
	
	if (!strncmp(argv[cmd_arg], "help", 4))
	{
		print_usage();
		printf("       vlconfig list - displays all vlinks\n\n");
		printf("       vlconfig create vlink_id [keywords] - create vlink with given id and keywords\n");
		printf("       vlconfig delete vlink_name - delete vlink with given name\n");
		printf("       vlconfig up vlink_name - activate vlink with given name\n");
		printf("       vlconfig down vlink_name - de-activate vlink with given name\n");
		printf("       vlconfig add_if vlink_name if_name - add interface to vlink\n\n");
		printf("       vlconfig rem_if vlink_name if_name - remove interface from vlink\n");
		printf("       vlconfig add_udp vlink_name ip:udp - add udp peer to vlink\n");
		printf("       vlconfig rem_udp vlink_name ip:udp - remove udp peer from vlink\n\n");
		return NOERR;
	}
	else if (!strncmp(argv[cmd_arg], "list", 4))
		sprintf(buf, "vlink_list");
	
	else if (!strncmp(argv[cmd_arg], "delete", 6) ||
		 !strncmp(argv[cmd_arg], "up", 2) ||
		 !strncmp(argv[cmd_arg], "down", 4) ||
		 !strncmp(argv[cmd_arg], "send", 4) )
	{
		if (argc < 3)
		{
			printf("Not enough arguments!\n");
			return FAIL;
		}
		
		sprintf(buf, "%s %s", vl_cmds[cmd_index], argv[cmd_arg+1]);
	}
	
	else if (!strncmp(argv[cmd_arg], "add_if", 6) ||
		 !strncmp(argv[cmd_arg], "rem_if", 6) ||
		 !strncmp(argv[cmd_arg], "add_udp", 7) ||
		 !strncmp(argv[cmd_arg], "rem_udp", 7) )
	{
		if (argc < 4)
		{
			printf("Not enough arguments!\n");
			return FAIL;
		}
		
		sprintf(buf, "%s %s %s", vl_cmds[cmd_index], argv[cmd_arg+1], argv[cmd_arg+2]);
	}
	
	else if (!strncmp(argv[cmd_arg], "create", 6))
	{
		if (argc < 3)
		{
			printf("Not enough arguments!\n");
			return FAIL;
		}
	
		if (argc == 3)
			sprintf(buf, "%s %s", vl_cmds[cmd_index], argv[cmd_arg+1]);
		
		else if (argc == 4)
			sprintf(buf, "%s %s %s", vl_cmds[cmd_index], argv[cmd_arg+1], argv[cmd_arg+2]);
	}
	
	return send_cmd(buf, strlen(buf)+1, port);
}
	
int send_cmd(char *cmd, int len, uint16_t port)
{
	int ret, vlsock;
	struct sockaddr_in vlink;
	unsigned int slen = sizeof(struct sockaddr_in);
	fd_set readfd;
	struct timeval timeout;
	int more = 1;
	int mcLen = strlen(MAGIC_COOKIE);
	ssize_t nbRecv = -1;
	
	if (!cmd)
	{
		printf("send_cmd called with NULL pointer!\n");
		return FAIL;
	}
	
	if ( (vlsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror("socket creation failed");
		return FAIL;
	} 
	
	if (!port)
		port = UDP_VLCONFIG;
	
	memset(&vlink, 0, sizeof(struct sockaddr_in));
	vlink.sin_family = AF_INET;
	inet_aton("127.0.0.1", &vlink.sin_addr);
	vlink.sin_port = htons(port);
	
	printf("[sending command to udp port %d]\n\n", port);
	
	if( (sendto(vlsock, cmd, len, 0, (struct sockaddr*)&vlink, slen)) < 0)
	{
		perror("send_cmd failed");
		return FAIL;	
	}
	
	while (more)
	{
		FD_ZERO(&readfd);
		FD_SET(vlsock, &readfd);
		timeout.tv_sec = TIMEOUT;
		timeout.tv_usec = 0;
	
		ret = select(vlsock+1, &readfd, 0, 0, &timeout);
		
		if (!ret)
		{
			printf("\nNo reply received from VLINK (timeout=%ds, using UDP port %d)\n\n",
				TIMEOUT, UDP_VLCONFIG);
			return FAIL;
		}
		
		if (ret == -1)
		{
			perror("select failed");
			return FAIL;		
		}
	
		nbRecv = recvfrom(vlsock, buf, BUFLEN, 0, (struct sockaddr*)&vlink, &slen);
	
		if (nbRecv < 0)
		{
			perror("read_cmd failed");
			return FAIL;
		}
		
		/* we check for magic cookie */
		
		if (nbRecv > mcLen) /* must have received at least mcLen+1 characters */
		{
			if (strncmp(buf+nbRecv-mcLen, MAGIC_COOKIE, mcLen))
			{
				/* no cookie = final message */
				
				buf[nbRecv] = '\0';
				more = 0;
			}
			else
				buf[nbRecv-mcLen] = '\0'; /* remove magic cookie */
		}
		else
		{
			/* vlink list returns a final '\n' */
		
			buf[nbRecv] = '\0';
			more = 0;
		}	
		
		printf("%s", buf);
	}
	
	close(vlsock);
	return NOERR;
}

void print_usage(void)
{
	printf("\nUsage: vlconfig [--port] { COMMAND } [ argument(s) ]\n\n");
	printf("  where: --port := udp port of vlink status socket\n\n");
	printf("where: COMMAND := { help | list | create | delete | up | down |\n");
	printf("                    add_if | rem_if | add_udp | rem_udp }\n\n");
}
