#ifndef ANAERROR_H_
#define ANAERROR_H_

#ifdef USER
#include <string.h>
#include <errno.h>
#endif

/* for get_ana_errno() */
//#include "anaThread.h"

/* to set and get the ana_errno value for the calling thread */
int get_ana_errno(void);      /* returns ana_errno_value, or ??? if thread not found ? */


extern int anaDebugLevel;
extern char mymodename[];

/** \ingroup anaCommon
 * \brief ANA printing loglevels
 */
enum {
	ANA_NONE,       /**< set the Debug parameter to this value to show no messages */
	ANA_EMERG,      /**< emergency, something really bad happend -> brick/MINMEX may crash, be unusable */
	ANA_ERR,        /**< error condition, system drops a packet, a lookup fails */
	ANA_NOTICE,     /**< normal situation, worthy to note */
	ANA_DEBUG,      /**< debuging information */
	ANA_TIMER,	/**< timer debug messages (can generate a lot of messages) */
	ANA_VALIDATE	/**< set the Debug parameter to this value to activate validation output */
};

#ifdef KERNEL
/** \ingroup anaCommon
 * \def anaPrint(level, fmt, args...)
 * \brief ANA printing function *** 
 * replaces the printf and printk function in kernel space everything is printed to the system log in user space to the file log/brickname_x
 * \hideinitializer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
/* print function with some addon, this function should be used */
#define anaPrint(level, fmt, args...) do {                \
                if ((level) <= anaDebugLevel)                 \
        printk("%s: ", THIS_MODULE->name); printk(fmt,## args);    \
        } while (0)

/* only to be used if you need to print a sequence of small messages (e.g. all received bytes of a message) */
#define printWrapper(fmt, args...) printk(fmt, ## args)
#else
#define anaPrint(level, fmt, args...) do {                \
                if ((level) <= anaDebugLevel)                 \
        printk("%s: " fmt, THIS_MODULE->name, ## args);    \
        } while (0)
#endif /*version*/
#endif

#ifdef USER
FILE *log_fp;
/* in user space we write all the anaPrint messages in a log file */
#define anaPrint(level, fmt, args...) do {               \
	if ((level) <= anaDebugLevel){  \
		time_t curtime; \
		time(&curtime); \
		char anaPrintBuf[30]; \
		ctime_r(&curtime, anaPrintBuf); \
		memcpy(&anaPrintBuf[strlen(anaPrintBuf)-1], " ", 1); \
		fprintf(log_fp, "%s: " fmt, anaPrintBuf, ## args); \
		fflush(log_fp); \
		if(level == ANA_EMERG) \
			fprintf(stderr, fmt, ## args); \
	}\
}while (0)

#define printWrapper(fmt, args...) printf(fmt, ## args)
#endif

#define ANADEBUG(s)    do {if (anaDebugON) {s;} } while (0)

#ifdef KERNEL
static inline int log_init(void) {
	return 1;
}
static inline void log_exit(void) {
	return;
}

#else

static inline void log_exit(void) {
	fclose(log_fp);
}
static inline int log_init(void) {
	int n;
	int idx = 0;
	char logfile[256];
	pid_t pid = 0;
	/* we test whether the ana_base_dir environment variable is set */
	char *ana_base = getenv("ANA_BASE_DIR");

	if (ana_base != NULL) {
		memcpy(logfile, ana_base, strlen(ana_base) );
		idx = strlen(ana_base);
		memcpy(logfile + idx, "/log", 4 + 1);
		idx += 4;
	}
	else {
		/* on default we use /var/log/ANA as log directory */
		memcpy(logfile, "/var/log/ANA", strlen("/var/log/ANA")+1);
		idx = strlen("/var/log/ANA");
	}
	n = mkdir(logfile, 0777);
	if (n!= 0 && errno != EEXIST) {
		perror("could not create ANA log directory");
		goto err;
	}
	memcpy(logfile + idx, "/", 2);
	idx += 1;	/* we increment the index only by one, since we will overwrite the \0 */

	/* the log files have their pid at the start */
	pid = getpid();
	sprintf(logfile + idx, "%i", pid);
	idx = strlen(logfile);
	memcpy(logfile + idx, "_", 1);
	idx += 1;
	memcpy(logfile + idx, mymodename, strlen(mymodename)+1);
	printWrapper("logfile %s\n", logfile);
        
	/* if the file already exists we overwrite it silently */
	log_fp = fopen(logfile, "w");
	if (log_fp == NULL)
		goto err;	
	return 0;
err:
	printWrapper("log_init failed %i\n", n);
	return -1;
}

#endif


static char *errorTab[] = {
"No error",
"Unspecified error",
"This operation creates a data forwarding loop",
"Forking error : the original IDP, for which the fork was set, has disappeared",
"Forking error : at least one of the IDPs to which the data is forked does not exist anymore",
"Origin viewer brick unknown",
"Destination viewer brick unknown",
"Redirect error: there is no public IDP with given label => nothing to redirect",
"No Destination IDT entry with indicated viewer and label",
"Reached the max authorized number of entries",
"There already exists an IDP with originLabel and originViewer belonging to another owner => cannot crash it",
"No Origin IDT entry with indicated viewer and label",
"ChangeAux error : this IDP is currently being redirected to another, changing the aux would have no effect",
"Unauthorized operation",
"Unregister error : this IDP is being redirected towards another => unregister operation incoherent",
"Unregister error : removed original IDP entry but failed to remove the fork point IDP leading towards it",
"Received query from unknown brick handle",
"Unable to register new IDP to handle the query's response",
"Unable to encode AL1 query message",
"Timed out before query response reception",
"The query failed at the remote peer (XRP_ERROR response)",
"The query returned an empty result",
"Unable to send query",
"Brick is not attached to a MINMEX",
"Some function argument is NULL",
"Cannot register (and create new IDP) callback function",
"Wrong event mask",
"Querier already subscribed to this event",
"No matching subscription for this event",
"A brick with the requested name is already attached to the minmex",
"No brick with the requested name attached to the minmex"
};

/*! \addtogroup anaCommon
 *  documentation of the error values
 *  @{
 */

#define ANA_NO_ERROR 0		/*!<No error */
#define ANA_UNKNOWN_ERROR -1	/*!<Unspecified error */
#define ANA_LOOP_ERROR -2	/*!<This operation creates a data forwarding loop */
#define ANA_FORK_ERROR_1 -3	/*!<Forking error : the original IDP, for which the fork was set, has disappeared */
#define ANA_FORK_ERROR_2 -4	/*!<Forking error : at least one of the IDPs to which the data is forked does not exist anymore */
#define ANA_NOORIGINVIEWER_ERROR -5	/*!< Origin viewer brick unknown*/
#define ANA_NODESTVIEWER_ERROR -6	/*!< Destination viewer brick unknown*/

#define ANA_REDIR_ERROR_1 -7		/*!< Redirect error: there is no public IDP with given label => nothing to redirect*/

#define ANA_NODESTIDTENTRY_ERROR -8	/*!< No Destination IDT entry with indicated viewer and label*/
#define ANA_MAXNBENTRIES_ERROR -9	/*!< Reached the max authorized number of entries*/

#define ANA_REDIR_ERROR_2 -10		/*!< There already exists an IDP with originLabel and originViewer belonging to another owner => cannot crash it*/

#define ANA_NOORIGINIDTENTRY_ERROR -11	/*!< No Origin IDT entry with indicated viewer and label*/

#define ANA_CHAGEAUX_ERROR -12		/*!< ChangeAux error : this IDP is currently being redirected to another, changing the aux would have no effect*/

#define ANA_UNAUTHORIZED_ERROR -13	/*!< Unauthorized operation*/

#define ANA_UNREGISTER_ERROR -14	/*!< Unregister error : this IDP is being redirected towards another => unregister operation incoherent*/
#define ANA_UNREGISTER_ERROR_2 -15	/*!< Unregister error : removed original IDP entry but failed to remove the fork point IDP leading towards it*/

#define ANA_UNKNOWNHANDLE_ERROR -16	/*!< Received query from unknown brick handle*/


#define ANA_UNABLE_TOSET_RESPONSELABEL_ERROR -17	/*!< Unable to register new IDP to handle the query's response*/
#define ANA_UNABLE_TOENC_QUERY_ERROR -18		/*!< Unable to encode AL1 query message*/

#define ANA_TIMEOUT_ERROR -19		/*!< Timed out before query response reception*/

#define ANA_REMOTE_ERROR -20		/*!< The query failed at the remote peer (XRP_ERROR response)*/

#define ANA_EMPTY_RESULT_ERROR -21	/*!< The query returned an empty result*/

#define ANA_UNABLE_TOSEND_QUERY_ERROR -22		/*!< Unable to send query*/

#define ANA_NOMINMEX_ERROR -23		/*!< Brick is not attached to a MINMEX*/

#define ANA_FCT_ARG_NULL -24		/*!< Some function argument is NULL*/

#define ANA_REGISTER_CALLBACK_FAILED -25	/*!< AL0 register callback failed */

#define ANA_WRONG_EVENTMASK -26		/*!<  Wrong ANA event mask*/
#define ANA_ALREADY_SUBSCRIBED -27 	/*!< Querier already subscribed to this event*/ 
#define ANA_NOMATCH_SUB -28		/*!< No matching subscription for this event*/
#define ANA_ALREADY_ATTACHED -29	/*!< A brick with the requested name is already attached to the minmex*/
#define ANA_NOBRICK_ATTACHED -30	/*!< No brick with the requested name attached to the minmex*/


/** \brief returns a description of a given error value
 * @param i the error value
 * @return the string representation of this error value
 */
static inline char *getErrorString(int i) {
	if ( i > 0)
		return 0;
	else
		return errorTab[-1*i];
}
/** \brief ana equivalent to perror
 * @param level the level for the print message in case of error
 * @param str the string to be printed in front of the error message
 */
static inline void anaPerror(int level, char * str){
	int i = get_ana_errno();
	if(i > 0)
		anaPrint(ANA_DEBUG, "%s: errno > 0\n", str);
	else 
		anaPrint(level, "%s: %s\n", str, errorTab[-1*i]);
}

/** @} */

#endif 
