/** @defgroup anaValidate ANA Validate API
 * Definition of the ANA validation API
 * Each brick to be included in the "mainline" ANA project
 * need to provide an Validation script that allows to
 * verify its correct behavior. 
 */

#ifndef __anaValidate_h_
#define __anaValidate_h_
/* write some data in a file */
/* in kernel space we use /proc and in user space a "normal" file 
 * to be used only for validation purpose!
 */

#include "anaCommon.h"
static char * validatebuf;
#ifdef KERNEL
static char * validatefile;
#endif

//int validate_init(char * filename);
//int log_init(void);
//void log_exit(void);
//void validate_exit(void);
/** @ingroup anaValidate
 */
//int validate_init(char * filename);
/** @ingroup anaValidate
 */
//void validate_exit(void);

#ifdef USER
/* this is needed for version sort to work! */
#include <string.h>
#include <dirent.h>
static FILE *validateFile;
#endif



/** @ingroup anaValidate
 * anaValidate(fmt, args...)
 * \brief writes validation output in the validation file (specified in validate_init)
 * 
 * has the same syntax as printf, but prints values only if anaDebugLevel is set to ANA_VALIDATE
 */

#ifdef KERNEL
#include <linux/proc_fs.h>
static int validateoffset = 0;

#define anaValidate(fmt, args...) do {                \
	if(anaDebugLevel == ANA_VALIDATE) { 		\
		int res = snprintf((char *)validatebuf+validateoffset, 1024-validateoffset, fmt, ## args);    \
		validateoffset += res; \
		printk("offset: %i\n", validateoffset); \
	} \
} while (0)
#else
#define anaValidate(fmt, args...) do { \
	if(anaDebugLevel == ANA_VALIDATE) {  \
		char * buf = (char *)malloc(256); \
		unsigned int res = snprintf(buf, 256, fmt, ## args); \
		fwrite(buf, 1, res, validateFile); \
		fflush(validateFile); \
		free(buf); \
	}  \
} while (0)

#endif

#ifdef KERNEL
static struct proc_dir_entry *ana_dir, *ana_validate_file;
static int proc_read_file(char *page, char **start,
                            off_t off, int count, 
                            int *eof, void *data)
{
       	char * message = NULL;
	int offset=0;

	start=NULL;
        off=0;
        count=0;
        data=NULL;
        eof=NULL;

	message = kmalloc(validateoffset, GFP_KERNEL);
	anaPrint(ANA_DEBUG, "validateoffset %i\n", validateoffset);
	if(message == NULL){
		anaPrint(ANA_DEBUG, "message == NULL\n");
	}
	if(validatebuf == NULL){
		anaPrint(ANA_DEBUG, "validatebuf == NULL\n");
	}

	memcpy(message, validatebuf, validateoffset);
//	message[validateoffset] = 'EOF';
	memcpy(page+offset, message, validateoffset);
	offset += validateoffset;
	kfree(message);
	return offset;
}

static inline void validate_exit(void){
	if(anaDebugLevel == ANA_VALIDATE) {
		remove_proc_entry(validatefile, ana_dir);
	        remove_proc_entry(validatefile, NULL);
		free(validatebuf);
		free(validatefile);
	}
}

static inline int validate_init(char * filename){
	int rv = 0;
	if(anaDebugLevel == ANA_VALIDATE) {
		validatefile = malloc(strlen(filename)+1);
		memcpy(validatefile, filename, strlen(filename)+1);

		validatebuf = kmalloc(1024, GFP_KERNEL);
		ana_dir = proc_mkdir(filename, NULL);
		if(ana_dir == NULL) {
			rv = -ENOMEM;
			goto out;
		}
		ana_dir->owner = THIS_MODULE;

		ana_validate_file = create_proc_entry(filename, 0644, ana_dir);
		if(ana_validate_file == NULL) {
			rv = -ENOMEM;
			goto no_file;
		}
		ana_validate_file->read_proc = proc_read_file;
		ana_validate_file->owner = THIS_MODULE;
		printk("ok\n");
		/* everything OK */
		return 0;
no_file:
		printk("nok\n");
		remove_proc_entry(filename, ana_dir);
		remove_proc_entry(filename, NULL);
out:
		kfree(validatebuf);
	}
		return rv;
}
#else

/** @ingroup anaValidate
 * \brief initialization of the validation environment
 * \param filename the name of the file to which the output is written
 *        in kernel space this will be /proc/filename/filename and in user space
 *        trunk/validOut/filename
 * \return 0 on success a negative error value otherwise
 */
int validate_init(char * filename){
	anaPrint(ANA_DEBUG, "level %i %i\n", anaDebugLevel, ANA_VALIDATE );
	if(anaDebugLevel == ANA_VALIDATE) {
		unsigned int dirLen = strlen("test_ana/validOut/");
		unsigned int fileLen = strlen(filename);
		validatebuf = malloc(1024);
	
		char * validatefile = malloc(dirLen + fileLen + 1);
		strcpy(validatefile, "test_ana/validOut/");
		memcpy(validatefile + dirLen, filename, fileLen + 1);
		anaPrint(ANA_DEBUG, "filename %s\n", validatefile);
		validateFile = fopen(validatefile, "a"); 
		free(validatefile);
	}
	return 0;
}
/** @ingroup anaValidate
 * \brief clean up of validation environment
 */
void validate_exit(void){
	if(anaDebugLevel == ANA_VALIDATE) {
		free(validatebuf);
		fclose(validateFile);
	}
}

#endif


/* end of validation only */




#endif //__anaValidate_h_

