/* C/include/analock.h
 *  
 * helper functions to be used from minmex and bricks
 *
 * Author: Christophe Jelger <christophe.jelger@unibas.ch>
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 *
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/** @defgroup anaLock ANA Lock API
 * Definition of the ANA lock API 
 */

#ifndef _ANALOCK_H
#define _ANALOCK_H

#ifdef USER
#include <pthread.h>
#endif
#ifdef KERNEL
#include <linux/rwsem.h>
#endif


/* reader writer locks:
 * Attention: User space and Kernel space handle read and write locks differently
 * Kernel space: write locks have priority
 * User space: read locks have priority
 * this means: sequence: read_lock A , write_lock B, read_lock C, later unlock all locks
 * -> execution sequence Kernel: A, B, C
 * -> execution sequence User: A, C, B
 * */

/**
 * \ingroup anaLock
 * \brief reader writer lock
 */
typedef struct analock_rw_s {
	int initialized;
#ifdef USER
	pthread_rwlock_t rwlock;
#endif
#ifdef KERNEL
	struct rw_semaphore rwlock;
#endif
} analock_rw_t;


/**
 * \ingroup anaLock
 * \brief initialization of a read write lock
 * Attention: User space and Kernel space handle read and write locks differently
 * Kernel space: write locks have priority
 * User space: read locks have priority
 * this means: sequence: read_lock A , write_lock B, read_lock C, later unlock all locks
 * -> execution sequence Kernel: A, B, C
 * -> execution sequence User: A, C, B
 *
 * \param mylock the lock to be initialized
 * \return -1 on failure 0 on success
 */
static inline int analock_rw_init(analock_rw_t *mylock) {
	if (!mylock)
		return -1;
		
#ifdef USER
	if (!pthread_rwlock_init(&mylock->rwlock, NULL)){
		mylock->initialized = 1;
		return 0;
	}
	else{
		mylock->initialized = 0;
		return -1;
	}
#endif
#ifdef KERNEL
	init_rwsem(&mylock->rwlock);
	mylock->initialized = 1;
	return 0;
#endif

}

/**
 * \ingroup anaLock
 * \brief destroys a read write lock
 * \param mylock the lock to be destroyed
 * \return -1 on failure 0 on success
 */
static inline int analock_rw_destroy(analock_rw_t *mylock) {
	if (!mylock || mylock->initialized == 0)
		return -1;
		
#ifdef USER
	if (!pthread_rwlock_destroy(&mylock->rwlock)){
		mylock->initialized = 0;
		return 0;
	}
	else{
		/* mylock is currently locked, we cannot free it! */
		return -1;
	}
#endif
#ifdef KERNEL
	mylock->initialized = 0;
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief lock for multiple simultaneously (read) accesses
 * \param mylock the lock
 * \return -1 on failure 0 on success
 */
static inline int analock_read_lock(analock_rw_t *mylock) {
	if (!mylock || mylock->initialized == 0)
		return -1;
#ifdef USER
	return (pthread_rwlock_rdlock(&mylock->rwlock));

#endif
#ifdef KERNEL
	down_read(&mylock->rwlock);
	return 0;	
#endif
}

/**
 * \ingroup anaLock
 * \brief lock for exclusive accesses
 * \param mylock the lock
 * \return -1 on failure 0 on success
 */
static inline int analock_write_lock(analock_rw_t *mylock) {
	if (!mylock || mylock->initialized == 0)
		return -1;
#ifdef USER
	return (pthread_rwlock_wrlock(&mylock->rwlock));

#endif
#ifdef KERNEL
	down_write(&mylock->rwlock);
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief unlock a lock that was previously locked with analock_read_lock
 * \param mylock the lock
 * \return -1 on failure 0 on success
 */
static inline int analock_read_unlock(analock_rw_t *mylock) {
	if (!mylock || mylock->initialized == 0)
		return -1;
#ifdef USER
	return (pthread_rwlock_unlock(&mylock->rwlock));

#endif
#ifdef KERNEL
	up_read(&mylock->rwlock);
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief unlock a lock that was previously locked with analock_write_lock
 * \param mylock the lock
 * \return -1 on failure 0 on success
 */
static inline int analock_write_unlock(analock_rw_t *mylock) {
	if (!mylock || mylock->initialized == 0)
		return -1;
#ifdef USER
	return (pthread_rwlock_unlock(&mylock->rwlock));

#endif
#ifdef KERNEL
	up_write(&mylock->rwlock);
	return 0;

#endif
}

/**
 * \ingroup anaLock
 * \brief simple lock
 */
typedef struct analock_s {
	int initialized;
#ifdef USER
	pthread_mutex_t lock;
#endif
#ifdef KERNEL
//	spinlock_t lock;
//	unsigned int irqflags;
	struct semaphore lock;
#endif
} analock_t;

/**
 * \ingroup anaLock
 * \brief condition for the conditional locks 
 */
typedef struct analock_cond_s {
#ifdef USER
	pthread_cond_t cond;
#endif
#ifdef KERNEL
	int cond;	/* 1 = ok, go further, 0 = wait */
#endif
} analock_cond_t;

#ifdef KERNEL
static DECLARE_WAIT_QUEUE_HEAD(twq);
#endif



/**
 * \ingroup anaLock
 * \brief initializes the lock mylock
 * This is the most basic locking facility in ANA, 
 * either the lock is locked or unlocked.
 * \param mylock the lock to be initialized
 * \return -1 on failure 0 on success
 */
static inline int analock_init(analock_t *mylock)
{
	if (!mylock)
		return -1;
	
	memset(mylock, 0, sizeof(struct analock_s));	
#ifdef USER
	if (!pthread_mutex_init(&(mylock->lock), NULL)){
		mylock->initialized = 1;
		return 0;
	}
	else{	
		mylock->initialized = 0;
		return -1;
	}
#endif
#ifdef KERNEL
	//mylock->lock = SPIN_LOCK_UNLOCKED;
	init_MUTEX(&mylock->lock);
	mylock->initialized = 1;
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief locks the lock mylock
 * \param mylock the lock to be initialized
 * \return -1 on failure 0 on success
 */
static inline int analock_lock(analock_t *mylock)
{
	if (!mylock)
		return -1;
	if(mylock->initialized != 1)
		return -1;
#ifdef USER
	if (!pthread_mutex_lock(&mylock->lock))
		return 0;
	else
		return -1;
#endif
#ifdef KERNEL
	//spin_lock_irqsave(&mylock->lock, mylock->irqflags);
	down_interruptible(&mylock->lock);
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief unlocks the lock mylock
 * \param mylock the lock to be initialized
 * \return -1 on failure 0 on success
 */
static inline int analock_unlock(analock_t *mylock)
{
	if (!mylock)
		return -1;
	if(mylock->initialized != 1)
		return -1;

#ifdef USER
	if (!pthread_mutex_unlock(&mylock->lock))
		return 0;
	else
		return -1;
#endif
#ifdef KERNEL
	//spin_unlock_irqrestore(&mylock->lock, mylock->irqflags);
	up(&mylock->lock);
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief initialization of the conditional lock cond
 * Allow to lock only for a certain amount of time
 * \param cond the lock to be initialized
 * \return -1 on failure 0 on success
 */
static inline int analock_cond_init(analock_cond_t *cond)
{
#ifdef USER
	return(pthread_cond_init(&(cond->cond), NULL));
#else
	return 0;
#endif
}

/**
 * \ingroup anaLock
 * \brief lock the lock condlock for at most the given time period
 * \param cond the condition for which to wait
 * \param condlock the lock 
 * \param secs seconds until lock will be released
 * \param nsecs nano seconds until the lock will be released
 * \return -1 on failure 0 on success
 */
static inline int analock_cond_timedwait(analock_cond_t *cond ,analock_t *condlock, int secs, long int nsecs)
{
#ifdef USER
	int ret;
        struct timespec *timeout = (struct timespec *) malloc(sizeof(struct timespec));
        time(&(timeout->tv_sec));
        timeout->tv_sec += secs;
        timeout->tv_nsec += nsecs;
        ret = pthread_cond_timedwait(&(cond->cond), &(condlock->lock), timeout);
        free(timeout);
        return ret;

#endif
#ifdef KERNEL
	analock_unlock(condlock);
	wait_event_interruptible_timeout(twq, cond->cond != 0, secs * HZ);
	cond->cond = 0; /* reset to 0 in case same condition is used again */
	analock_lock(condlock);
	return 0; 
#endif
}

/**
 * \ingroup anaLock
 * \brief lock the lock condlock until the condition cond is ok
 * \param cond the condition for which to wait
 * \param condlock the lock 
 * \return -1 on failure 0 on success
 */
static inline int analock_cond_wait(analock_cond_t *cond ,analock_t *condlock)
{
#ifdef USER
	return pthread_cond_wait(&(cond->cond), &(condlock->lock));
#endif
#ifdef KERNEL
	analock_unlock(condlock);
	wait_event_interruptible(twq, cond->cond != 0);
	cond->cond = 0; /* reset to 0 in case same condition is used again */
	analock_lock(condlock);
	return 0; 
#endif
}

/**
 * \ingroup anaLock
 * \brief send a signal to cond in order to unlock it
 * \param cond the condition to which a signal is sent
 * \return 0 on success
 */
static inline int analock_cond_signal(analock_cond_t *cond)
{
#ifdef USER
	return pthread_cond_signal(&(cond->cond));
#endif
#ifdef KERNEL
	cond->cond = 1;
	wake_up_interruptible(&twq);
	return 0; 
#endif
}

#endif /* _ANALOCK_H */
