#ifndef _XRP_H
#define _XRP_H

#include "anaCommon.h"

#ifdef USER
#include <unistd.h> 
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <strings.h>
#include <arpa/inet.h>
#endif

#ifdef KERNEL
#include <linux/kernel.h>
#include <linux/stat.h>
#endif

/* 			The Format of an XRP message : 

		[command][nbArgs][class(or argId)][arg Length][arg content]...	
				  2bytes					 2bytes
				  (short)				 (length of arg content in bytes)

	the number of bytes coding the nbArgs and args Length can be optimized (to uint8 or uint16) in further development 
*/

typedef struct {  			/* All sizes are in bytes */
	
	int maxMsgSize;	 		/* Maximum size for an XRP message */
	int commandSize; 		/* number of bytes coding commands */
	int classSize; 			/* number of bytes coding classes (argument identifiers) */
	char *xrpSpecVersion;		/* indicates the version number and the usage of this XRPSpec (e.g "XRP_ETHCOMP_1.0")
					   allows two communicating peers to check if they are using the same Specs */
					
} xrpSpecs_t;  			

extern xrpSpecs_t defaultXrpSpecs;

typedef char *xrpCmd_t;
typedef char *xrpClass_t;
typedef char *xrpMsg_t;

/* Function prototypes  
   In previous versions, the first arg was [xrpSpecs_t *xrpSpec]
   --> this was now removed by Christophe to simplify the XRP API 
*/

int allocateXRPMsg(xrpMsg_t *msg);
int fillXRPCommand (xrpMsg_t msg, xrpCmd_t cmd);
int getXRPnbArgs(xrpMsg_t msg);
int addXRPArg(xrpMsg_t msg, xrpClass_t xrp_class, void *content, unsigned short int len);
void *getXRPArg(xrpMsg_t msg, xrpClass_t xrp_class, int *len, int occurence);
int equalXRPCommands(xrpCmd_t cmdA,  xrpCmd_t cmdB );
int equalXRPClasses(xrpClass_t classA,  xrpClass_t classB );

/*! \addtogroup anaXRP
 *  documentation of the XRP commands and arguments
 *  @{
 */

/* < default Values for the default XRP specification>*/
#define DEFAULT_XRP_MSGSIZE  2048		/*!< Default xrp message size*/
#define DEFAULT_XRP_CMDSIZE  3 			/*!< Default xrp command size*/
#define DEFAULT_XRP_CLASSSIZE 3			/*!< Default xrp class (argument type) size*/
#define DEFAULT_XRP_VERSION "XRP_DEFAULT_1.0"	/*!< xrp version */

#define XRP_CMD_DATA  "dta"			/*!< data message*/
#define XRP_CMD_RESOLVE  "res"			/*!< resolve request*/
#define XRP_CMD_LOOKUP  "lkp"			/*!< lookup request*/
#define XRP_CMD_PUBLISH  "pbl"			/*!< publish request*/
#define XRP_CMD_UNPUBLISH  "unp"		/*!< unpublish request*/
#define XRP_CMD_ERROR  "err"			/*!< error occurred while answering a request*/


#define XRP_CLASS_ANY "any"			/*!< used to get arbitrary class argument from an XRP message \see getXRPArg */ 
#define XRP_CLASS_NAME  "nam"			/*!< the name of a resolved brick*/

#define XRP_CLASS_LABEL "lbl"			/*!< argument is a label*/
#define XRP_CLASS_HANDLE "hdl"			/*!< argument is a handle*/

#define XRP_CLASS_SRC_CXT "sxt"			/*!< source context argument of the resolve/lookup/publish primitive*/
#define XRP_CLASS_DST_CXT "dxt"			/*!< dest context argument of the resolve/lookup/publish primitive*/

#define XRP_CLASS_SRC_SRV "ssv"			/*!< source service argument of the resolve/lookup/publish primitive*/
#define XRP_CLASS_DST_SRV "dsv"			/*!< dest service argument of the resolve/lookup/publish primitive*/

#define XRP_CLASS_LKP_DESCR "dcr"		/*!< used to store the complementary description that is result of lookup */

#define XRP_CLASS_CHANTYPE "ctp"		/*!< channel type of the resolve primitive*/
#define XRP_CLASS_PERMANENT "per"		/*!< whether idp is permanent or volatile*/
#define XRP_CLASS_MESSAGE "msg"			/*!< the actual data*/

/** @} */

#endif
