/* ip2ana_dns.c
 * 
 * IP to ANA migration/adaptation layer
 * ip2ana kernel module for the netfilter framework
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, University of Basel
 *  
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/netdevice.h>
#include <linux/kmod.h>
#include <linux/if.h>
#include <linux/inet.h>
#include <linux/inetdevice.h>
#include <linux/version.h>
#include <net/checksum.h>
#include <net/route.h>
#include <net/protocol.h>
#include <net/netfilter/nf_nat.h>
#include <net/netfilter/nf_nat_core.h>
#include <net/netfilter/nf_nat_rule.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include "ipt_ip2ana.h"

#define DNS_REQ_A           1
#define DNS_REQ_AAAA       28
#define DNS_REQ_PTR        12

#define PTR_STD_STYLE      0
#define PTR_OLD_STYLE      1
#define NOT_PTR            2

/* PTR queries can be very long with IPv6 */
#define MAX_QUERY_LEN     128

struct dnshdr {

	/* DNS header */

	uint16_t trans_id;
	uint16_t flags;
	uint16_t qdcount;
	uint16_t ancount;
	uint16_t nscount;
	uint16_t arcount;
};

struct udp_pseudo_hdr2 {

	/* for checksum calculation */

	uint32_t src, dst;
	unsigned char zero, proto;
	uint16_t len;
};

/* to check whether data and/or PTR queries are for ANA or not */
struct in_addr ana_ip4_subnet;
struct in_addr ana_ip4_mask;
struct in6_addr ana_ip6_subnet = ANA_IPV6_SUBNET;
int ana_ip6_pfxlen = ANA_IPV6_PFXLEN;

/* to maintain a list of pending DNS requests */
struct sk_buff *dnsreqs[MAX_DNS_REQS];
spinlock_t dnsreqs_lock = SPIN_LOCK_UNLOCKED; /* for locking the array */
int store_dnsreq(struct sk_buff *, uint16_t trans_id);
extern struct socket *udpsock;
extern struct timer_list dnsreqs_timer; /* for timing out dns requests */

extern int ip2ana_sendmsg(struct socket *sock, void *data, int datalen);
void dnsptr2ip4(char *cp, struct in_addr *addr);
void dnsptr2ip6(char *ptr, struct in6_addr *addr);
int fill_name_to_dns(char *dns, char *name);
int netbox_neigh_unmap(void *addr, int family);
char* eth2str(char *data);
int prefix_match_len(void *addr1, void *addr2);

/* Access to ARP and ND tables */
extern struct neigh_table arp_tbl; /* should be exported inside kernel */
struct neigh_table *my_nd_tbl_ptr;
static char eth_buf[32];
struct arp_s {
	union {
		struct in_addr ip4;
		struct in6_addr ip6;
	} addr;
	int family;
	struct arp_s *next;
};
struct arp_s *arp_entries = NULL;

/* ip2ana interface */
extern struct net_device *ip2ana_dev;

void timer_click(unsigned long dummy)
{
	unsigned long expires;
	unsigned long flags;
	int i;
	struct timeval tv;
	
	if (spin_is_locked(&dnsreqs_lock))
		return; /* we clean the list later */
		
	spin_lock_irqsave(&dnsreqs_lock, flags);
	
	do_gettimeofday(&tv);
	
	for (i=0; i<MAX_DNS_REQS; i++)
	{
		if (!dnsreqs[i])
			continue;
			
		if ((tv.tv_sec - dnsreqs[i]->tstamp.off_sec) > DNSREQS_EXPIRE_SEC)
		{
			/* no reply received before timeout */	
		
			send_dns_reply(dnsreqs[i], NULL); /* send nosuchname */
			kfree_skb(dnsreqs[i]);
			dnsreqs[i] = NULL;
		}
	}
	
	spin_unlock_irqrestore(&dnsreqs_lock, flags);
	
	del_timer(&dnsreqs_timer); /* does not harm */
	expires = jiffies + msecs_to_jiffies(1000*TIMER_CLICK_SEC);
	dnsreqs_timer.expires = expires;
	add_timer(&dnsreqs_timer);
}

void clean_dnsreqs(void)
{
	int i;
	unsigned long flags;
	
	spin_lock_irqsave(&dnsreqs_lock, flags);
	
	for (i=0; i<MAX_DNS_REQS; i++)
	{
		if (!dnsreqs[i])
			continue;
			
		kfree_skb(dnsreqs[i]);	
	}
	
	spin_unlock_irqrestore(&dnsreqs_lock, flags);
	printk("ip2ana: cleaned pending dns requests\n");
}

int store_dnsreq(struct sk_buff *skb, uint16_t trans_id)
{
	int i, e = -1;
	uint16_t *tr;
	int offset = sizeof(struct iphdr) + sizeof(struct udphdr);
	
	for (i=0; i<MAX_DNS_REQS; i++)
	{
		if (!dnsreqs[i])
		{
			if (e == -1)
				e = i; /* first empty slot */
			
			continue;
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		tr = (uint16_t *)((char *)dnsreqs[i]->nh.raw + offset);
#else
		tr = (uint16_t *)((char *)skb_network_header(dnsreqs[i]) + offset);
#endif	
		if (trans_id == *tr)
			return i; /* request is known */
	}

	if (e == -1)
		return -1; /* array is full ! */
	else
	{
		dnsreqs[e] = skb;
		return e;
	}
}

int dns_request_for_ana(struct sk_buff *skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
	struct iphdr *iph = (struct iphdr *)skb->nh.raw;
	struct udphdr *udph = (struct udphdr *)skb->h.raw;
#else
	struct iphdr *iph = (struct iphdr *) skb_network_header(skb);
	struct udphdr *udph = (struct udphdr *) skb_transport_header(skb);
#endif
	struct dnshdr *dnsh = (struct dnshdr *) (udph + 1);
	unsigned char *cp = (unsigned char *) (dnsh + 1);
	uint16_t qtype;
	int len, min_len;
	char tmp_name[MAX_QUERY_LEN + 1];
	struct timeval tv;
	
	if (!skb || !iph || !udph)
		return IP2ANA_LETGO;
	
	if (iph->protocol != IPPROTO_UDP)
	{
		/* TODO: check if this a TCP flow to be redirected to ANA */
		
		return IP2ANA_DATA;
	}
	
	if (udph->dest != htons(53))
	{
		/* TODO: check if this a UDP flow to be redirected to ANA */
		
		return IP2ANA_DATA;
	}
	
	min_len = sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr);
	if (!dnsh || ntohs(iph->tot_len) < min_len )
		return IP2ANA_LETGO; /* too short */
	
	/* contruct human-readable DNA name from DNS-encoded pattern */
	len = 0;
	while (*cp) {
		int i = *cp;
		cp++;

		while (i--) {
			if (len < MAX_QUERY_LEN)
				tmp_name[len++] = *cp;
			cp++;
		}

		if (len < MAX_QUERY_LEN && *cp)
			tmp_name[len++] = '.';
	}
	tmp_name[len] = '\0';
	cp++;
	qtype = ntohs(* (uint16_t *) cp);
	
	/* record the incoming time in the skbuff */
	do_gettimeofday(&tv);
	skb->tstamp.off_sec = tv.tv_sec;
	
	switch (qtype) {
	
	case DNS_REQ_PTR:
		
		if (len <= 64) {
			
			struct in_addr addr;
			
			dnsptr2ip4(tmp_name, &addr);
			if ((addr.s_addr & ana_ip4_mask.s_addr) == ana_ip4_subnet.s_addr)
			{
				printk("ip2ana: got PTR request for " NIPQUAD_FMT " (%u)\n",
					NIPQUAD(addr.s_addr), ntohs(dnsh->trans_id));
				
				/* forward to "ANA resolver" */
				return IP2ANA_DNSREQ;
			}
			else
				return IP2ANA_LETGO;
		}
		else /* an IPv6 PTR is always greater than 64 */
		{
			struct in6_addr addr;
			int len;
			
			dnsptr2ip6(tmp_name, &addr);
			len = prefix_match_len(&addr, &ana_ip6_subnet);
			
			if (len >= ana_ip6_pfxlen)
			{
				printk("ip2ana: got PTR request for " NIP6_FMT " (%u)\n", 
					NIP6(addr), ntohs(dnsh->trans_id));
					
				/* forward to "ANA resolver" */
				return IP2ANA_DNSREQ;
			}
			else
				return IP2ANA_LETGO;
		}
		break;
	
	case DNS_REQ_AAAA:
	case DNS_REQ_A:
	
		/* Check if A or AAAA name belongs to .ana domain */
		
		if (strcmp(".ana", tmp_name + len - strlen(".ana"))) {
		
			return IP2ANA_LETGO; /* not a ".ana" request */
		}
		else
		{
			printk("ip2ana: got %s request for %s (%u)\n", 
				(qtype==DNS_REQ_A) ? "A" : "AAAA", tmp_name, ntohs(dnsh->trans_id));
			
			/* forward to "ANA resolver" */
			return IP2ANA_DNSREQ;
		}
		break;
	default:
		return -1;
		break;
	}
}

void send_dns_reply(struct sk_buff *skb, char *result)
{
	/* packet headers of the query */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
	struct iphdr *q_iph = (struct iphdr *)skb->nh.raw;
	struct udphdr *q_udph = (struct udphdr *)skb->h.raw;
#else
	struct iphdr *q_iph = (struct iphdr *) skb_network_header(skb);
	struct udphdr *q_udph = (struct udphdr *) skb_transport_header(skb);	
#endif
	// struct dnshdr *q_dnsh = (struct dnshdr *) (q_udph + 1);
	
	unsigned char buf[512]; /* buffer for reply */
	
	/* packet headers for the reply */
	struct iphdr *iph = (struct iphdr *) buf;
	struct udphdr *udph = (struct udphdr *) (iph + 1);
	struct dnshdr *dnsh = (struct dnshdr *) (udph + 1);	
	
	struct udp_pseudo_hdr2 *uph;
	unsigned char *cp;
	uint16_t *sp, qtype;
	int len, len2, nosuchname = 0;
	
	char name[64];
	char type[8];
	char addr[64];
	char *token;
	
	if (result)
	{
		/* separate the fields of the answer */
		token = strsep(&result, " ");
		if (!token)
			nosuchname = 1;
		else
			strcpy(name, token);
	
		token = strsep(&result, " ");
		if (!token)
			nosuchname = 1;
		else
			strcpy(type, token);
	
		token = strsep(&result, " ");
		if (!token)
			nosuchname = 1;
		else
			strcpy(addr, token);
	}
	else
		nosuchname = 1;
	
	memset(buf, 0, 512);
	
	/* copy initial packet (and DNS query) */
	memcpy(buf, (char *)q_iph, ntohs(q_iph->tot_len));
	
	/* IPv4 header of reply */
	iph->ihl = 5;
	iph->version = 4;
	iph->ttl = 2;
	iph->protocol = IPPROTO_UDP;
	iph->check = 0;
	iph->saddr = q_iph->daddr;
	iph->daddr = q_iph->saddr;

	/* UDP header of reply  */
	udph->source = q_udph->dest; /* or htons(53); */
	udph->dest = q_udph->source;
	udph->check = 0;

	/* DNS reply message header */
	if (!nosuchname)
	{	
		dnsh->flags = htons(0x8180);	// standard reply
		dnsh->ancount = htons(1);
	}
	else
	{	
		dnsh->flags = htons(0x8583);	// "no such name"
		dnsh->ancount = 0;
	}
	
	dnsh->qdcount = htons(1);
	dnsh->nscount = 0;
	dnsh->arcount = 0;
	
	/* Find the type of the initial query */
	cp = (unsigned char *) (dnsh + 1);
	while (*cp)
		cp++;
	cp++;
	qtype = (*(uint16_t *)cp);
	
	sp = (uint16_t *) (buf+ntohs(q_iph->tot_len)); /* point after initial DNS query */
	
	if (nosuchname)
	{
		/* we have no answer: time to send */
	
		len = (unsigned char*) sp - (unsigned char*) buf;
		goto checksum;
	}
	
	/* reply starts here */
	
	*sp++ = htons(0xc00c); /* compression */
	*sp++ = qtype; /* record type */
	*sp++ = htons(1); /* Internet class */
	*(uint32_t *)sp = htonl(3600); /* TTL : we use 1 hour */
	sp += 2;
	
	/* append response part */
	if (ntohs(qtype) == DNS_REQ_A)
	{
		const char *end;
		int ret;
		char eth[6];
			
		*sp++ = htons(4); /* ipv4 address length */
		ret = in4_pton(addr, strlen(addr), (u8 *)sp, -1, &end);
		if (!ret || end == result)
			return; /* TODO: sendnosuchname */
		
		/* insert dummy ARP entry IP4toANA */
		eth[0] = 0x04;
		eth[1] = 0x2a;
		memcpy(eth+2, sp, 4);
		netbox_neigh_map(sp, AF_INET, eth);
		
		sp += 2;
		len = (unsigned char*) sp - (unsigned char*) buf;
	}
	else if (ntohs(qtype) == DNS_REQ_AAAA)
	{
		const char *end;
		int ret;
		char eth[6];
		
		*sp++ = htons(16); /* ipv6 address length */
		ret = in6_pton(addr, strlen(addr), (u8 *)sp, -1, &end);
		if (!ret || end == result)
			return; /* TODO: sendnosuchname */
		
		/* insert dummy ND entry IP6toANA */
		eth[0] = 0x06;
		eth[1] = 0x2a;
		memcpy(eth+2, sp, 4); /* to change: first 4 bytes of IPv6 address */
		netbox_neigh_map(sp, AF_INET6, eth);
		
		sp += 8;
		len = (unsigned char*) sp - (unsigned char*) buf;
	}
	else if (ntohs(qtype) == DNS_REQ_PTR)
	{
		int alen = fill_name_to_dns((char *)(sp+1), name);
	
		*sp++ = htons(alen);
		len = (unsigned char*) sp - (unsigned char*) buf + alen;
	}
	else
	{
		/* TODO: sendnosuchname */
		
		return;
	}
	
checksum:
	iph->tot_len = htons(len);
	udph->len = htons(len - sizeof(struct iphdr));

	// build the UDP pseudoheader (appended for doing the chksum)
	len2 = (len + 3) & ~0x03;	// dword align
	uph = (struct udp_pseudo_hdr2 *) (buf + len2);
	uph->src = iph->saddr;
	uph->dst = iph->daddr;
	uph->proto = iph->protocol;
	uph->len = udph->len;
	len2 += sizeof(*uph);

	udph->check = ip_fast_csum((char *) udph, (len2 - sizeof(struct iphdr)) / 4);
	if (!udph->check)
		udph->check = 0xffff;
	iph->check = ip_fast_csum(buf, iph->ihl);

	l3deliver(buf, len);
	
	/* reminder: skb is freed in ip2ana_msgrecv() */
}

int l3deliver(void *data, int len)
{
        struct sk_buff *skb;
	struct net_device_stats *stats = (struct net_device_stats *) ip2ana_dev->priv;
       	
	printk("ip2ana: l3deliver %d bytes, dev %s\n", len, ip2ana_dev->name);
	
	if (stats) {
		stats->rx_packets++;
		stats->rx_bytes += len;
	}
	
	skb = alloc_skb(len + 14, GFP_ATOMIC);
	if (!skb)
		return 0;
	
	skb->dev = ip2ana_dev;

	skb->mac.raw = skb->data;
	skb_reserve(skb, 14);
	memcpy(skb_put(skb, len), data, len);
	memcpy(skb->mac.raw, ip2ana_dev->dev_addr, ETH_ALEN);
	memset(skb->mac.raw + ETH_ALEN, 0, ETH_ALEN);

	if (*(char*)(skb->mac.raw + 2*ETH_ALEN +2) < 0x60) {
		// for IPv4, (char)version can be superior to 40
		// should be replaced with proper test using 4-bits bitfield

		*((short*)(skb->mac.raw + 2 * ETH_ALEN)) = htons(ETH_P_IP);
		skb->protocol = htons(ETH_P_IP);
	} else {
		*((short*)(skb->mac.raw + 2 * ETH_ALEN)) = htons(ETH_P_IPV6);
		skb->protocol = htons(ETH_P_IPV6);
	}
	netif_rx(skb);

	return 0;
}

int fill_name_to_dns(char *dns, char *name)
{
	char *cp = dns;
	int i;

	if (!dns || !name)
		return -1;
	
	/* reconstruct DNS-like name from human dotted name
	   e.g. www.renater.fr --> 03www07renater02fr00 */
	
	while (*name) {
		for (i = 0; name[i] && name[i] != '.'; i++);
		*cp = i;
		if (i) {
			cp++;
			strncpy(cp, name, i);
			cp += i;
		}
		name += i;
		if (*name) // skip the dot
			name++;
	}
	*cp++ = 0; /* set '\0' at end of name */

	return cp - dns;
}

void dnsptr2ip4(char *cp, struct in_addr *addr)
{
	int i;
	uint32_t ip;

	// extract v4 addr from DNS PTR format
	for (i = 0, ip = 0; i < 4; i++) {
		int v;
		for (v = 0; *cp; cp++) {
			if (*cp == '.') {
				cp++;
				break;
			}
			v = 10*v + *cp - '0';
		}
		ip = ((v & 0xff) << 24) | (ip >> 8);
	}
	
	addr->s_addr = htonl(ip);
}


void dnsptr2ip6(char *ptr, struct in6_addr *addr)
{
	int i;

	// v6 PTR are in DNS-format
	// --> extracts v6 addr from DNS PTR format

	// ptr++;	// skip len indicator

	for (i = 15; i >= 0; i--) {
		unsigned char c;

		if ((addr->s6_addr[i] = *ptr - '0') > 10)
			addr->s6_addr[i] -= 39;

		ptr += 2;	// next value

		if ((c = *ptr - '0') > 10)
			c -= 39;

		addr->s6_addr[i] += c * 0x10;

		ptr += 2;	// next value
	}
}

char* eth2str(char *data)
{
	sprintf(eth_buf, "%02x:%02x:%02x:%02x:%02x:%02x",
		 (unsigned char) data[0], (unsigned char) data[1],
		 (unsigned char) data[2], (unsigned char) data[3],
		 (unsigned char) data[4], (unsigned char) data[5]);

	return eth_buf;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#define NEIGH_UPDATE(n,p,a,b,c)   neigh_update(n,p,a,b,c)
#else
#define NEIGH_UPDATE(n,p,a,b,c)   neigh_update(n,p,a,b|c)
#endif

int netbox_neigh_map(void *addr, int family, char *eth)
{
	struct neighbour *neigh = 0;
	struct arp_s *new = arp_entries, *prev = NULL;

	/* Always use __neigh_lookup_errno and not neigh_lookup */
	/* as the former creates an entry if none exists */

	if (!addr || !eth || (family != AF_INET && family != AF_INET6) )
		return -1;
	
	/* check if entry is already in arp_entries list */
	while (new)
	{
		if (family == new->family)
		{
			if (family == AF_INET && !memcmp(&new->addr.ip4, addr, sizeof(struct in_addr)))
				return 0; /* address already mapped */
			
			else if (family == AF_INET6 && !memcmp(&new->addr.ip6, addr, sizeof(struct in6_addr)))
				return 0; /* address already mapped */
		}
	
		prev = new;
		new = new->next;
	}
	
	new = kmalloc(GFP_ATOMIC, sizeof(struct arp_s));
	if (prev)
		prev->next = new;
	else
		arp_entries = new;
	new->family = family;
	new->next = NULL;
	if (family == AF_INET)
		memcpy(&new->addr.ip4, addr, sizeof(struct in_addr));
	else
		memcpy(&new->addr.ip6, addr, sizeof(struct in6_addr));	
	
	if (family == AF_INET)
	{
		neigh = __neigh_lookup_errno(&arp_tbl, addr, ip2ana_dev);
	}
	else if (family == AF_INET6)
	{
		/* handling special case where ipv6 module is loaded after ip2ana module */
		
		if (!my_nd_tbl_ptr) {
			struct inet6_dev *in6dev = (struct inet6_dev*) ip2ana_dev->ip6_ptr;
			if (in6dev && in6dev->nd_parms)
				my_nd_tbl_ptr = in6dev->nd_parms->tbl;
		}
		
		if (my_nd_tbl_ptr)
			neigh = __neigh_lookup_errno(my_nd_tbl_ptr, addr, ip2ana_dev);
		else
			return -1;
	}
	
	if (neigh && !IS_ERR(neigh))
	{
		if (family == AF_INET)
			printk("ip2ana: ARP/ND update " NIPQUAD_FMT " to %s\n",
				NIPQUAD(((struct in_addr *)addr)->s_addr), eth2str(eth));
		else
			printk("ip2ana: ARP/ND update " NIP6_FMT " to %s\n",
				NIP6(*((struct in6_addr*)addr)), eth2str(eth));
		
		NEIGH_UPDATE(neigh, eth, NUD_PERMANENT, 1, 0);
                neigh_release(neigh);
        }
	else
		printk("ip2ana: cannot do ARP/ND update!\n");

	return 0;
}

void clean_arp_entries(void)
{
	struct arp_s *a = arp_entries, *n;
	
	while (a)
	{
		netbox_neigh_unmap(&a->addr, a->family);
		n = a->next;
		kfree(a);
		a = n;
	}
}

int netbox_neigh_unmap(void *addr, int family)
{
	struct neighbour *neigh = NULL;
	
	/* Always use __neigh_lookup_errno and not neigh_lookup */
	/* as the former creates an entry if none exists */

	if (!addr)
		return -1;
	
	if (family == AF_INET)
	{
		neigh = neigh_lookup(&arp_tbl, addr, ip2ana_dev);
	}
	else if (family == AF_INET6) {
		
		if (!my_nd_tbl_ptr) {
			struct inet6_dev *in6dev = (struct inet6_dev*) ip2ana_dev->ip6_ptr;
			if (in6dev && in6dev->nd_parms)
				my_nd_tbl_ptr = in6dev->nd_parms->tbl;
		}
		
		if (my_nd_tbl_ptr)
			neigh = neigh_lookup(my_nd_tbl_ptr, addr, ip2ana_dev);
		else
			return -1;
	}
	else
		return -1; /* wrong family */
		
        if (neigh)
	{
              	if (family == AF_INET)
			printk("ip2ana: ARP/ND removed " NIPQUAD_FMT "\n",
				NIPQUAD(((struct in_addr *)addr)->s_addr));
		else
			printk("ip2ana: ARP/ND removed " NIP6_FMT "\n",
				NIP6(*((struct in6_addr*)addr)));
		
                NEIGH_UPDATE(neigh, 0, NUD_FAILED, NEIGH_UPDATE_F_ADMIN, 0);
                neigh_release(neigh);
        }
	else
		printk("ip2ana: neigh_unmap failed!\n");

	return 0;
}

int prefix_match_len(void *addr1, void *addr2)
{
   /*
   *  find the first different bit between two addresses
   *  length of address must be a multiple of 32bits
   *
   *  Slightly modified copy of net/ipv6/ip6_fib.c from linux kernel source code
   *  --> the value returned is the length (in bits) of the common part
   */
	
	__u32 *a1 = addr1;
	__u32 *a2 = addr2;
	int i;
	
	for (i = 0; i < 4; i++) {
		__u32 xb;

		xb = a1[i] ^ a2[i];

		if (xb) {
			int j = 31;

			xb = ntohl(xb);

			while ((xb & (1 << j)) == 0)
				j--;

			return (i * 32 + 31 - j);
		}
	}
	
	return 128;	// addresses are identical !
}
