/* ipt_ip2ana.c
 * 
 * IP to ANA migration/adaptation layer
 * ip2ana kernel module for the netfilter framework
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, University of Basel
 *  
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
 
#include <linux/version.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/netdevice.h>
#include <linux/kmod.h>
#include <linux/if.h>
#include <linux/inet.h>
#include <linux/inetdevice.h>
#include <net/checksum.h>
#include <net/route.h>
#include <net/protocol.h>
#include <net/netfilter/nf_nat.h>
#include <net/netfilter/nf_nat_core.h>
#include <net/netfilter/nf_nat_rule.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include "ipt_ip2ana.h"

MODULE_AUTHOR("Christophe Jelger");
MODULE_DESCRIPTION("iptables ip2ana kernel module");
MODULE_LICENSE("GPL");

int ipt_port = 0;
int brick_port = 0;
int i2a_port = 0;
int a2i_port = 0;
char *if_name = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(ipt_port, "i");
MODULE_PARM(brick_port, "i");
MODULE_PARM(i2a_port, "i");
MODULE_PARM(a2i_port, "i");
#else
module_param(ipt_port, int, 0);
module_param(brick_port, int, 0);
module_param(i2a_port, int, 0);
module_param(a2i_port, int, 0);
#endif

MODULE_PARM_DESC(ipt_port, "UDP port for receiving DNS replies from ana brick");
MODULE_PARM_DESC(brick_port, "UDP port for sending DNS queries to ana brick");
MODULE_PARM_DESC(i2a_port, "UDP port for sending IP packets to ANA");
MODULE_PARM_DESC(a2i_port, "UDP port for receiving IP packets from ANA");

struct socket *dnssock = NULL;
struct socket *datasock = NULL;
struct udp_ports udpinfo;
struct net_device *ip2ana_dev = NULL;

void ip2ana_dnsrecv(struct sock *sk, int len);
void ip2ana_datarecv(struct sock *sk, int len);
int ip2ana_sendmsg(struct socket *sock, void *data, int datalen);
extern int prefix_match_len(void *addr1, void *addr2);

extern struct sk_buff *dnsreqs[MAX_DNS_REQS];
extern spinlock_t dnsreqs_lock; /* for locking the array */
struct timer_list dnsreqs_timer; /* for timing out dns requests */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define INIT_ETHERDEV(dev,len) init_etherdev(dev,len)
#define REGISTER_NETDEV(dev) NULL
#else
#define INIT_ETHERDEV(dev,len) alloc_netdev(len, "ip2ana", ether_setup)
#define REGISTER_NETDEV(dev) register_netdev(dev)
#endif
static int ip2ana_start_xmit(struct sk_buff *skb, struct net_device *dev);
static int ip2ana_open(struct net_device *dev);
static int ip2ana_stop(struct net_device *dev);
static struct net_device_stats *ip2ana_get_stats(struct net_device *dev);
#define IP2ANA_MTU 1024	/* very conservative */

static int
ipt_ip2ana_checkentry(const char *tablename,
		const void *e,
		const struct xt_target *target,
		void *targinfo,
		unsigned int hook_mask)
{
	/* all entries are fine */
 
	return 1;
}

void ip2ana_datarecv(struct sock *sk, int len)
{
	int qlen;
	struct sk_buff *skb = NULL;
	int udplen = sizeof(struct udphdr);
	char *ip;
	struct iphdr *iph;
	struct ipv6hdr *ip6h;
	char eth[6];
			
	while ((qlen = skb_queue_len(&sk->sk_receive_queue)) > 0)
	{
		skb = skb_dequeue(&sk->sk_receive_queue);
		
		/* we create a dummy ARP entry for the sender */
		ip = skb->data + udplen;
		iph = (struct iphdr *)ip;
		ip6h = (struct ipv6hdr *)ip;
		
		if (iph->version == 4)
		{
			eth[0] = 0x04; /* IP4toANA */
			eth[1] = 0x2a;
			memcpy(eth+2, &iph->saddr, 4);
			netbox_neigh_map(&iph->saddr, AF_INET, eth);
		}
		else
		{
			eth[0] = 0x06; /* IP6toANA*/
			eth[1] = 0x2a;
			memcpy(eth+2, &ip6h->saddr, 4);
			netbox_neigh_map(&ip6h->saddr, AF_INET6, eth);
		}
		
		/* we skip the UDP header and inject in IP stack */
		
		l3deliver(skb->data + udplen, skb->len - udplen);
		kfree_skb(skb); /* l3deliver made a new skb */
	}
}

void ip2ana_dnsrecv(struct sock *sk, int len)
{
	int qlen, i;
	struct sk_buff *skb = NULL;
	int udplen = sizeof(struct udphdr);
	unsigned long flags;
	
	while ((qlen = skb_queue_len(&sk->sk_receive_queue)) > 0)
	{ 
		uint16_t trans_id;
		int offset = sizeof(struct iphdr) + sizeof(struct udphdr);
		
		/* qlen holds the number of packets in the receive queue of this socket */
	
		skb = skb_dequeue(&sk->sk_receive_queue);
		
		if (skb->len < (udplen + 2) ) /* udp header + DNS trans_id */
		{
			printk("ip2ana: msgrecv --> data too short to extract trans_id!\n");
			kfree_skb(skb);
			continue;
		}
	
		trans_id = *((uint16_t *)(skb->data+udplen));
		
		spin_lock_irqsave(&dnsreqs_lock, flags);
			
		for (i=0; i<MAX_DNS_REQS; i++)
		{
			uint16_t *tr;
						
			if (!dnsreqs[i])
				continue;
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
			tr = (uint16_t *)((char *)dnsreqs[i]->nh.raw + offset);
#else
			tr = (uint16_t *)((char *)skb_network_header(dnsreqs[i]) + offset);
#endif
			if (trans_id == *tr)
			{
				uint16_t tmp = *((uint16_t *)(skb->data+udplen+2));
				
				if (tmp == htons(8583)) /* DNS no such name */
				{
					printk("ip2ana: received DNS nosuchname reply for query %u\n", ntohs(trans_id));
					send_dns_reply(dnsreqs[i], NULL);
				}
				else
				{
					printk("ip2ana: received DNS reply %s\n", skb->data+udplen+2);
					send_dns_reply(dnsreqs[i], skb->data+udplen+2);
				}
			
				kfree_skb(skb); /* free UDP reply */
				kfree_skb(dnsreqs[i]); /* free initial query */
				dnsreqs[i] = NULL;
				
				spin_unlock_irqrestore(&dnsreqs_lock, flags);
				return;
			}
		}
		
		spin_unlock_irqrestore(&dnsreqs_lock, flags);
		
		printk("ip2ana: no request found for reply %u!\n", ntohs(trans_id));
		kfree_skb(skb);
	}
	return;	
}

int ip2ana_sendmsg(struct socket *sock, void *data, int datalen)
{
	struct msghdr msgh;
	struct iovec iov;
	struct sockaddr_in dest;
	mm_segment_t oldfs;
	int bytes = -1;
		
	if (!sock || !data || !datalen )
		return -1;
	
	memset(&msgh, 0, sizeof(struct msghdr));
	memset(&dest, 0, sizeof(struct sockaddr_in));
	
	msgh.msg_name = &dest;
	msgh.msg_namelen = sizeof(struct sockaddr_in);
	
	dest.sin_family = AF_INET;
	dest.sin_addr.s_addr = in_aton("127.0.0.1");
	
	if (sock == dnssock)
		dest.sin_port = udpinfo.brick_port;
	else if (sock == datasock)
		dest.sin_port = udpinfo.i2a_port;
	
	iov.iov_base = data;
	iov.iov_len  = datalen;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_iov    = &iov;
	msgh.msg_iovlen = 1;
	
	oldfs = get_fs();
	set_fs( KERNEL_DS );
	bytes = sock_sendmsg(sock, &msgh, datalen);
	set_fs( oldfs );

	return bytes;
}

static unsigned int
ipt_ip2ana_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		unsigned int hooknum,
		const struct xt_target *target,
		const void *targinfo)
{
	struct sk_buff *skb = *pskb;
	struct udphdr *udph;
	struct dnshdr *dnsh;
	int tmp;

	tmp = dns_request_for_ana(skb);

	if (tmp == IP2ANA_DNSREQ)
	{
		/* this is a DNS request for .ana */
		int idx;
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
		udph = (struct udphdr *) skb->h.raw;
#else
		udph = (struct udphdr *) skb_transport_header(skb);
#endif
		dnsh = (struct dnshdr *) (udph + 1);
		
		idx = store_dnsreq(skb, *((uint16_t *)dnsh));		

		if (idx >= 0)
		{
			printk("ip2ana: storing/sending DNS request (idx=%d)\n", idx);
			ip2ana_sendmsg(dnssock, dnsh, ntohs(udph->len)-sizeof(struct udphdr));
			return NF_STOLEN;
		}
		else
			return NF_DROP; /* request is already known */
	}
	else
		return XT_CONTINUE;	
}

static struct xt_target ipt_ip2ana = {
	.name          = "ip2ana",
	.family        = AF_INET,
	.target        = ipt_ip2ana_target,
	.targetsize    = 0,
	.table         = "mangle",
	.hooks         = 1 << NF_IP_LOCAL_OUT,
	.checkentry    = ipt_ip2ana_checkentry,
	.me            = THIS_MODULE,
};

static int __init ipt_ip2ana_init(void)
{
	int i, ret;
	const char *end;
	unsigned long expires;		
	struct sockaddr_in tmp_in;
	uint16_t port;
		
	printk("ipt_ip2ana: init\n");
	
	/* check port values passed as arguments */
	
	if (!ipt_port || !brick_port || !i2a_port || !a2i_port)
	{
		printk("Some UDP port arguments are NULL!\n");
		return -1;	
	}
	
	if ( (ipt_port < 0 || ipt_port > 65535) ||
	     (brick_port < 0 || brick_port > 65535) ||
	     (i2a_port < 0 || i2a_port > 65535) ||
	     (a2i_port < 0 || a2i_port > 65535) )
	{
		printk("Some UDP port arguments are larger than 65535 or negative!\n");
		return -1;
	}
	
	port = ipt_port; udpinfo.ipt_port = htons(port);
	port = brick_port; udpinfo.brick_port = htons(port);
	port = i2a_port; udpinfo.i2a_port = htons(port);
	port = a2i_port; udpinfo.a2i_port = htons(port);
	
	/* Creation of UDP sockets for user-kernel communication */
		
	if ((sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &dnssock)) < 0)
	{
		printk("ip2ana: cannot create UDP socket");
		return -1;
	}
	memset(&tmp_in, 0, sizeof(struct sockaddr_in));
	tmp_in.sin_family = AF_INET;
	tmp_in.sin_addr.s_addr = in_aton("127.0.0.1");
	tmp_in.sin_port = udpinfo.ipt_port;
	if ((dnssock->ops->bind(dnssock, (struct sockaddr *) &tmp_in, sizeof(tmp_in))) < 0)
	{
		printk("ip2ana: cannot bind UDP socket\n");
		sock_release(dnssock);
		return -1;
	}
	/* callback to handle incoming UDP messages */
	dnssock->sk->sk_data_ready = ip2ana_dnsrecv;
	
	if ((sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &datasock)) < 0)
	{
		printk("ip2ana: cannot create UDP socket");
		sock_release(dnssock);
		return -1;
	}
	memset(&tmp_in, 0, sizeof(struct sockaddr_in));
	tmp_in.sin_family = AF_INET;
	tmp_in.sin_addr.s_addr = in_aton("127.0.0.1");
	tmp_in.sin_port = udpinfo.a2i_port;
	if ((datasock->ops->bind(datasock, (struct sockaddr *) &tmp_in, sizeof(tmp_in))) < 0)
	{
		printk("ip2ana: cannot bind UDP socket\n");
		sock_release(dnssock);
		sock_release(datasock);
		return -1;
	}
	/* callback to handle incoming UDP messages */
	datasock->sk->sk_data_ready = ip2ana_datarecv;
		
	/* initialize array of pending requests */
	for (i=0; i<MAX_DNS_REQS; i++)
		dnsreqs[i] = NULL;			
	
	ret = in4_pton(ANA_IP_SUBNET, strlen(ANA_IP_SUBNET), (u8 *)&ana_ip4_subnet, -1, &end);
	if (!ret)
	{
		printk("ip2ana: cannot initialize ANA_IP_SUBNET\n");
		sock_release(dnssock);
		sock_release(datasock);
		return -1;
	}
	
	ret = in4_pton(ANA_IP_MASK, strlen(ANA_IP_MASK), (u8 *)&ana_ip4_mask, -1, &end);
	if (!ret)
	{
		printk("ip2ana: cannot initialize ANA_IP_MASK\n");
		sock_release(dnssock);
		sock_release(datasock);
		return -1;
	}
	
	/* create ip2ana interface */
	
	ip2ana_dev = INIT_ETHERDEV(0, sizeof(struct net_device_stats));
		
	if (!ip2ana_dev)
	{
		printk("ip2ana: cannot create ip2ana interface ... exiting\n");
		sock_release(dnssock);
		sock_release(datasock);
		return -1;
	}
	
	SET_MODULE_OWNER(ip2ana_dev);	
	
	ip2ana_dev->open = &ip2ana_open;
        ip2ana_dev->hard_start_xmit = &ip2ana_start_xmit;
        ip2ana_dev->stop = &ip2ana_stop;
        ip2ana_dev->get_stats = &ip2ana_get_stats;

	ip2ana_dev->mtu = IP2ANA_MTU;
        ip2ana_dev->tx_queue_len = 10;
        *(short*)(ip2ana_dev->dev_addr) = htons(0x1042);
        get_random_bytes(ip2ana_dev->dev_addr + 2, 4);

	if (REGISTER_NETDEV(ip2ana_dev)) {
	
                printk("ip2ana: register_netdev failed ... exiting\n");	
		unregister_netdev(ip2ana_dev);
		sock_release(dnssock);
		sock_release(datasock);
		return -1;		
	}
	
	init_timer(&dnsreqs_timer);
	dnsreqs_timer.function = timer_click;
	expires = jiffies + msecs_to_jiffies(1000*TIMER_CLICK_SEC);
	dnsreqs_timer.expires = expires;
	add_timer(&dnsreqs_timer);
	
	return xt_register_target(&ipt_ip2ana);
}

static int ip2ana_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
        struct net_device_stats *stats = (struct net_device_stats *) dev->priv;
	struct ethhdr *ep = (struct ethhdr *) skb->data;
	struct iphdr *iph;
	struct ipv6hdr *ip6h;
	int len;
		
        if (stats) {
		stats->tx_packets++;
		stats->tx_bytes += (skb->len - sizeof(struct ethhdr));
	}

	switch( ntohs(ep->h_proto) )
	{
		case ETH_P_ARP:
		{
			printk("ip2ana: receiving ARP request on ip2ana interface! (should never happen)\n");
			break;
		}
		case ETH_P_IP:
		{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
			iph = (struct iphdr *) skb->nh.raw;
#else
			iph = (struct iphdr *) skb_network_header(skb);
#endif
			len = ntohs(iph->tot_len);
			printk("ip2ana: forwarding ipv4 packet (%d bytes) to ANA\n", len);
			ip2ana_sendmsg(datasock, iph, len);
			kfree_skb(skb);
			return 0;
			break;
		}
		case ETH_P_IPV6:
		{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)		
			ip6h = (struct ipv6hdr *) skb->nh.raw;
#else
			ip6h = (struct ipv6hdr *) skb_network_header(skb);
#endif		
			len = ntohs(ip6h->payload_len)+sizeof(struct ipv6hdr);
			printk("ip2ana: forwarding ipv6 packet (%d bytes) to ANA\n", len);
			ip2ana_sendmsg(datasock, ip6h, len);
			kfree_skb(skb);
			return 0;
			break;
		}
		
		default:
			break;
	}

        kfree_skb(skb);
        return 0;
}

static int ip2ana_open(struct net_device *dev)
{
	netif_start_queue(dev);
        return 0;
}

static int ip2ana_stop(struct net_device *dev)
{
	netif_stop_queue(dev);
        return 0;
}

static struct net_device_stats *ip2ana_get_stats(struct net_device *dev)
{
        return (struct net_device_stats *) dev->priv;
}


static void __exit ipt_ip2ana_fini(void)
{
	printk("ipt_ip2ana: exit\n");
	xt_unregister_target(&ipt_ip2ana);
	clean_arp_entries();
	del_timer(&dnsreqs_timer);
	clean_dnsreqs();
	
	unregister_netdev(ip2ana_dev);
	free_netdev(ip2ana_dev);
	ip2ana_dev = NULL;
	
	if (dnssock)
		sock_release(dnssock);
		
	if (datasock)
		sock_release(datasock);
}

module_init(ipt_ip2ana_init);
module_exit(ipt_ip2ana_fini);

