
#include "../include/KVR.h"

/***************************************************************

Hash functions by Bob Jenkins, December 1996, Public Domain.

2005-11-09: tested by CJ with 128-bit keys to 10 bits values 
            --> hash function has a uniform distribution

***************************************************************/

ub4 KVRhash( k, length, initval, maskLength)
register ub1 *k;        /* the key */
register ub4  length;   /* the length of the key in bytes */
register ub4  initval;    /* the previous hash, or an arbitrary value */
register ub4  maskLength;    /* the previous hash, or an arbitrary value */
{
   register ub4 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = initval;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */

	return (c & hashmask(maskLength));	 
}

int name2hash(struct KVR *rep, char *name){
	int h;
	if(rep==NULL || name==NULL) {
		anaPrint(ANA_ERR, "name2hash: NULL arguments \n");
		return -1;
	}
	h = KVRhash((ub1*)name, strlen(name), 0, rep->hashMaskEntry);
	/*anaPrint(ANA_DEBUG, "name2hash  : %s --> %d\n", name, h));*/
	return h;
}

int key2hash(struct KVR *rep, char *key) {
	int h;
	if(rep==NULL || key==NULL) {
		anaPrint(ANA_ERR, "key2hash: NULL arguments \n");
		return -1;
	}
	h = KVRhash((ub1*)key, strlen(key), 0, rep->hashMaskKey);
	/*anaPrint(ANA_DEBUG, "key2hash  : %s --> %d\n", key, h));*/
	return h;
}

int initializeKVR( struct KVR *rep, int hashMaskEntryTab, int hashMaskKeyTab, int deadEntryLimit ) {
	if(rep == NULL || hashMaskEntryTab<0 || hashMaskKeyTab<0 ) {
		anaPrint(ANA_ERR, "initializeKVR : NULL argument\n");
		return -1;
	}
	rep->hashMaskEntry = hashMaskEntryTab;
	rep->hashMaskKey = hashMaskKeyTab;
	rep->sizeEntryTab = pow2(hashMaskEntryTab);
	rep->sizeKeyTab = pow2(hashMaskKeyTab);
	rep->deadEntryLimit = deadEntryLimit;
	/* we allocate the tables with the appropriate size*/
	rep->entryTab = (struct KVREntry **) malloc(rep->sizeEntryTab * sizeof(void *));
	if(rep->entryTab == NULL){
		anaPrint(ANA_ERR, "initializeKVR : unable to allocate entry Tab\n");
		return -1;
	}
	rep->keyTab = (struct KVRKey **) malloc(rep->sizeKeyTab * sizeof(void *));
	if(rep->keyTab == NULL){
		anaPrint(ANA_ERR, "initializeKVR : unable to allocate key Tab\n");
		free(rep->entryTab);
		return -1;
	}
	
	anaPrint(ANA_DEBUG, "initializeKVR : KVR initialized with size %d for entry tab and %d for key tab \n", rep->sizeEntryTab, rep->sizeKeyTab);
	return 0;
}


struct KVREntry * allocateKVREntry( char *name, anaLabel_t label, anaHandle_t owner ){
	struct KVREntry *entry;
	if( name == NULL) {
		anaPrint(ANA_ERR, "allocateKVREntry: NULL arguments\n");
		return NULL;
	}
	entry = (struct KVREntry *) malloc(sizeof(struct KVREntry));
	if(entry==NULL){
		anaPrint(ANA_ERR, "allocateKVREntry: Unable to allocate memory\n");
		return NULL;
	}
	entry->name = (char *) malloc(sizeof(char)*(strlen(name)+1));
	if(entry->name==NULL){
		anaPrint(ANA_ERR, "allocateKVREntry: Unable to allocate memory for name\n");
		free(entry);
		return NULL;
	}
	strcpy(entry->name, name);
	entry->IDP = label;
	entry->owner = owner;
	return entry;
}

int addKeywordToEntry(struct KVREntry *entry, char *keyWord) {
	struct KVRKeyword *keyW;
	if(entry == NULL || keyWord == NULL) {
		anaPrint(ANA_ERR, "addKeywordToEntry: NULL arguments\n");
		return -1;
	}
	keyW = (struct KVRKeyword *) malloc(sizeof(struct KVRKeyword));
	if(keyW==NULL){
		anaPrint(ANA_ERR, "addKeywordToEntry: unable to allocate memory\n");
		return -1;
	}
	keyW->keyWord = (char *) malloc((strlen(keyWord)+1)*sizeof(char));
	if(keyW->keyWord==NULL){
		anaPrint(ANA_ERR, "addKeywordToEntry:unable to allocate memory\n");
		free(keyW); 
		return -1;
	}
	strcpy(keyW->keyWord, keyWord);
	keyW->next = entry->keyWordsList;
	entry->keyWordsList = keyW;
	return 0;
}


int updateKey(struct KVR *rep, char *keyword, struct KVREntry *entry){
	int h, found;
	struct KVRKey *key=NULL;
	struct KVRKeyObjectList *keyEntriesList=NULL;
	
	if(rep == NULL || entry == NULL || keyword == NULL) {
		anaPrint(ANA_ERR, "updateKey: NULL arguments\n");
		return -1;
	}
	h = key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "updateKey : index %d after hash is out of bounds\n",h);
		return -1;
	}
	key = rep->keyTab[h];
	found = 0;
	while(key!=NULL && !found){
		if(!strcmp(key->keyWord, keyword))
			found = 1;
		else
			key = key->next;
	}
	if(!found){
		key = (struct KVRKey *) malloc(sizeof(struct KVRKey));
		if(key==NULL){
			anaPrint(ANA_ERR, "KVR updateKey : unable to allocate memory for new key\n");
			return -1;
		}
		key->keyWord = (char *) malloc((strlen(keyword)+1)*sizeof(char));
		if(key==NULL){
			anaPrint(ANA_ERR, "KVR updateKey : unable to allocate memory for new keyword\n");
			free(key);
			return -1;
		}
		strcpy(key->keyWord, keyword);
		key->next = rep->keyTab[h];
		rep->keyTab[h] = key;
	}
	
	/* we add the entry to the key's entries list*/
	keyEntriesList = (struct KVRKeyObjectList *) malloc(sizeof(struct KVRKeyObjectList));
	if(keyEntriesList==NULL){
		if(!found){
			free(key->keyWord);
			free(key);
		}
		anaPrint(ANA_ERR, "KVR updateKey : unable to allocate memory for key's entries list\n");
		return -1;
	}	
	keyEntriesList->entry = entry;
	keyEntriesList->next = key->entriesList;
	key->entriesList = keyEntriesList;
	
	return 0;	
}

int freeKey(struct KVR *rep, struct KVRKey *key, int h){
	struct KVRKey *tmp=NULL;
	struct KVRKey *predecessor=NULL;
	struct KVRKeyObjectList *tmpList=NULL;
	struct KVRKeyObjectList *tmpNextList=NULL;
	
	if(key==NULL || rep==NULL || h<0){
		anaPrint(ANA_ERR, "freeKey: NULL arguments\n");
		return -1;
	}
	tmp = rep->keyTab[h];
	predecessor = NULL;
	while(tmp!=NULL){
		if(key == tmp){
			if(predecessor == NULL)
				rep->keyTab[h] = tmp->next;
			else
				predecessor->next = tmp->next;
			free(tmp->keyWord);
			/*we free the key's object list */
			tmpList = tmp->entriesList;
			while(tmpList != NULL) {
				tmpNextList = tmpList->next;
				free(tmpList);
				tmpList = tmpNextList;
			}
			
			free(tmp);
			return 0;
		}
		else{
			predecessor = tmp;
			tmp = tmp->next;
		}
	}
	return -1;	
}


int removeFromKeysList(struct KVR *rep, struct KVREntry *toRemove, char *keyword){
	struct KVRKeyObjectList *objList=NULL;
	struct KVRKeyObjectList *predecessor=NULL;
	int h, foundKey;
	struct KVRKey *key=NULL;
	
	if(keyword == NULL || toRemove == NULL || rep == NULL){
		anaPrint(ANA_ERR, "removeFromKeysList: NULL arguments\n");
		return -1;
	}
	
	/* first we find the key*/
	h =  key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "removeFromKeysList : index %d after hash is out of bounds\n",h);
		return -1;
	}
	foundKey = 0;
	key = rep->keyTab[h];
	while(key!=NULL && !foundKey){
		if(!strcmp(keyword, key->keyWord))
			foundKey=1;
		else
			key = key->next;
	}
	
	if(!foundKey){
		anaPrint(ANA_NOTICE, "removeFromKeysList : unable to find key matching keyword %s\n",keyword);
		return -1;
	}
	
	objList = key->entriesList;
	predecessor = NULL;
	while(objList!=NULL){
	       	if(objList->entry == toRemove){
	       		if(predecessor!=NULL){
				predecessor->next = objList->next;
			}
			else{
				if(objList->next !=NULL)
					key->entriesList =  objList->next;
				else
					freeKey(rep, key, h); /*in case the entriesList becomes empty we free the Key */
			}
			return 0;
		}
	       	else {
	       		predecessor = objList;
			objList = objList->next;
		}	
	}
	return -1;
}


int addKVREntry(struct KVR *rep, struct KVREntry *entry){
	int hN;
	struct KVREntry *existEntry = NULL;
	struct KVRKeyword *tmpKeyword =NULL;

	if(entry == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "addKVREntry: NULL arguments\n");
		return -1;
	}
	/* first we add the entry to the entries Table*/	
	hN =  name2hash(rep, entry->name);
	if(hN<0 || hN>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "addKVREntry : index %d after hash is out of bounds\n",hN);
		return -1;
	}
	existEntry = rep->entryTab[hN];
	if(existEntry==NULL){
		entry->next = NULL;
		rep->entryTab[hN] = entry;	
	}
	else{ /*hash collision*/
		/* we must check if an entry with same name exists*/
		while(existEntry!=NULL){
			if(!strcmp(existEntry->name,entry->name)){
				anaPrint(ANA_ERR, "addKVREntry: entry with name %s already exists\n",entry->name);
				return -1;
			}
			existEntry = existEntry->next;
		}	
	
		entry->next = rep->entryTab[hN];
		rep->entryTab[hN] = entry;
	}
		
	/* Then we update the keys Table*/
	tmpKeyword = entry->keyWordsList;
	/*for each keyWord*/
	while(tmpKeyword!=NULL){
		updateKey(rep, tmpKeyword->keyWord, entry);
		tmpKeyword = tmpKeyword->next;
	}
	return 0;
}

struct KVREntry * getKVREntryByName(struct KVR *rep, char *name){
	struct KVREntry *entry=NULL;
	int h;
	if(name == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "getKVREntryByName: NULL arguments\n");
		return NULL;
	}
	h =  name2hash(rep, name);
	if(h<0 || h>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "getKVREntryByName : index %d after hash is out of bounds\n",h);
		return NULL;
	}
	entry = rep->entryTab[h];
	while(entry!=NULL){
		if(!strcmp(name, entry->name)){
			if(entry->age != -1)
				entry->age = 0;
			return entry;
		}	
		entry = entry->next;
	}
	return NULL;
}

struct KVRKeyObjectList * findByKey(struct KVR *rep, char *keyword) {
	struct KVRKey *key=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	int h;
	if(keyword == NULL || rep == NULL) {
		anaPrint(ANA_ERR, "findByKey: NULL arguments\n");
		return NULL;
	}
	h =  key2hash(rep, keyword);
	if(h<0 || h>(rep->sizeKeyTab)-1) {
		anaPrint(ANA_ERR, "findByKey : index %d after hash is out of bounds\n",h);
		return NULL;
	}
	
	key = rep->keyTab[h];
	while(key!=NULL){
		if(!strcmp(keyword, key->keyWord)){
			/* we must update the age of these entries */
			tmp = key->entriesList;
			while(tmp!=NULL){
				if(tmp->entry->age !=-1)
					tmp->entry->age =0;
				tmp = tmp->next;
			}
			return key->entriesList;
		}	
		key = key->next;
	}
	return NULL;
}


int freeKVREntry(struct KVR *rep, struct KVREntry *entry){
	struct KVRKeyword *keyWordsList=NULL;
	struct KVRKeyword *nextKeyWord=NULL;
	if(rep==NULL || entry ==NULL)
		return -1;
	
	keyWordsList = entry->keyWordsList;
	while(keyWordsList!=NULL){
		removeFromKeysList(rep, entry, keyWordsList->keyWord);
		nextKeyWord = keyWordsList->next;
		free(keyWordsList->keyWord);
		free(keyWordsList);
		keyWordsList = nextKeyWord;
	}
	free(entry->name);
	free(entry);
	return 0;
}

int removeKVREntry(struct KVR *rep, char *name, struct KVREntry *pointer){
	int h;	
	struct KVREntry *entry=NULL;
	struct KVREntry *predecessor=NULL;
	
	if(rep==NULL||  name==NULL)
		return -1;
	
	h =  name2hash(rep, name);
	if(h<0 || h>(rep->sizeEntryTab)-1) {
		anaPrint(ANA_ERR, "removeKVREntry : index %d after hash is out of bounds\n",h);
		return -1;
	}
	entry = rep->entryTab[h];
	while(entry!=NULL){
		if( entry == pointer){ /*we found it*/
			if(predecessor==NULL)
				rep->entryTab[h]=entry->next;
			else
				predecessor->next=entry->next;
				
			freeKVREntry(rep, entry);
			return 0;
		}	
		predecessor = entry;
		entry = entry->next;
	}
	/* if we end up here, the entry doesn't exist*/
	return -1;
}

int checkKVREntries(struct KVR *rep){
	struct KVREntry *entry=NULL;
	struct KVREntry *predecessor=NULL;
	struct KVREntry *next=NULL;
	
	int i;
	
	if(rep == NULL)
		return -1;
	
	for(i=0;i< rep->sizeEntryTab;i++){
		entry = rep->entryTab[i];
		predecessor = NULL;
		while(entry!=NULL){
			/* first  we check if the IDP associated to the entry still exists */
			if(!existsIDP(entry->IDP)){
				anaPrint(ANA_DEBUG, "KVR Entry with name %s is associated to a non existing IDP => removed \n", entry->name);
				if(predecessor == NULL)
					rep->entryTab[i] = entry->next;
				else
					predecessor->next = entry->next;
				next = entry->next;
				freeKVREntry(rep, entry);
				entry = next; 
				continue;	
			}	
			if(entry->age != -1){
				entry->age++;
				if(entry->age > rep->deadEntryLimit){
					/*we remove the entry */
					/* first we must remove the entry from it's keys object list  */
					anaPrint(ANA_DEBUG, "KVR Entry with name %s is deprecated => removed \n", entry->name);
					if(predecessor == NULL)
						rep->entryTab[i] = entry->next;
					else
						predecessor->next = entry->next;
					next = entry->next;
					freeKVREntry(rep, entry);
					entry = next; 
				}		
				else {
					predecessor = entry;
					entry = entry->next;
				}
			}
			else {
				predecessor = entry;
				entry = entry->next;
			}
		}
	}
	return 0;
}

void freeHandleKVREntries(struct KVR *rep, anaHandle_t owner){
	int i;
	struct KVREntry *entry=NULL;
	struct KVREntry *predecessor=NULL;
	struct KVREntry *next=NULL;
	if(rep==NULL)
		return;	
		
	for(i=0;i<rep->sizeEntryTab;i++){
		entry=rep->entryTab[i];
		predecessor=NULL;
		while(entry!=NULL){
			next=entry->next;
			if(entry->owner == owner ){
				/*we remove the entry*/
				if(predecessor==NULL)
					rep->entryTab[i]=next;
				else
					predecessor->next=next;
			
				freeKVREntry(rep, entry);
			}
			entry=next;
		}
	}
}











