/* C/minmex/common/clientTable.c
 * 
 * functions to deal with the brick table
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include "../include/clientTable.h"

/* the allocated size of the table will be (2^bitSize),  */
int initBrickTable ( BrickTable_t *bt , int hashMaskSize) {	 
  	if(bt == NULL || hashMaskSize<0 ){
		anaPrint(ANA_ERR, "initBrickTable: NULL arguments \n");
		return -1;
	}
	
	bt->size = pow2(hashMaskSize);   /* the hashMaskSize will be also used in the hash  function */
	bt->hashMaskSize = hashMaskSize;
	bt->tab = (struct BTentry_s **) malloc(bt->size * sizeof(void *));
	if(bt->tab==NULL){
		anaPrint(ANA_ERR, "initBrickTable: unable to allocate memory for Bricks Table \n");
		return -1;
	}
	bt->nbBricks=0;
	anaPrint(ANA_DEBUG, "Bricks Table initialized with size %u\n", bt->size);
	return 0;
}


int handle2hash(BrickTable_t *bTable, anaHandle_t *handle) {
	int h ;
	h = hash((ub1 *)handle, sizeof(anaHandle_t), 0, bTable->hashMaskSize);
	return h;
}

anaHandle_t brickname2handle(BrickTable_t *bTable, char *brickname) {

	int i;
	struct BTentry_s *b;
	
	if (!bTable || !brickname)
		return 0;
	
	for (i=0; i<bTable->size; i++)
	{
		b = bTable->tab[i];
	
		if (!b)
			continue; /* no entry */
			
		while (b)
		{
			if (!strcmp(brickname, b->brickName))
				return b->handle;
		
			b = b->next;
		}
	}
	
	return 0;
}

char * handle2name(BrickTable_t *bTable, anaHandle_t handle) {
	struct BTentry_s* entry=NULL;
	
	if(bTable ==NULL || handle==0)
		return NULL;	
	
	entry= getBTEntry( bTable , handle );
	if(entry==0)
		return NULL;
	else
		return entry->brickName;
		
	return NULL;
}

struct BTentry_s* getBTEntry( BrickTable_t *bTable , anaHandle_t handle ) {
	int i ;
	struct BTentry_s *entry=NULL;
	if(bTable==NULL){
		anaPrint(ANA_ERR, "getBTEntry : NULL arguments \n");
		return NULL;
	}
	i= handle2hash( bTable, &handle);
	if (i<0 || i>(int) (bTable->size)-1) {
		anaPrint(ANA_ERR, "getBTEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	
	entry = bTable->tab[i];
	
	if (entry == NULL ) { 	/* meaning that there is no value mapped to this 
							 label in the Hash table */
		anaPrint(ANA_ERR, "getBTEntry : No BT entry with handle %d\n", handle);
		return NULL;	
	}
				
	if (handle == entry->handle) { /* if the keys are equal */
		return entry; 
	}
	else {  /* we have a hash collison, we must follow the linked chain to find 
			   the entry with the correct key */
		entry = entry->next; 
		while (entry != NULL) {
			if( handle == entry->handle){
				return entry;
			}
			else 
				entry = entry->next;
		} 
	}

	return NULL;
}


anaHandle_t generateUnusedHandle(BrickTable_t *bTable) {
	struct BTentry_s *entry = NULL;
	anaHandle_t res;
	while(1){
		res = generateHandle();
		anaPrint(ANA_DEBUG, " generateUnusedHandle: trying handle %d \n",res);
		entry = getBTEntry(bTable, res);
		if(entry==NULL)
			return res;
	}
}


int freeBrickTable(BrickTable_t *bTable) { 
	int i;
	struct BTentry_s *entry=NULL;
	struct BTentry_s *next=NULL;
	for (i=0; i< (int) bTable->size; i++) {
		entry = bTable->tab[i];
		while  (entry!=NULL) {
			next = entry->next;
			free(entry->brickName);
			free(entry);
			entry = next;
		}
	}
	bTable->nbBricks=0;
	bTable->size = 0;
	free(bTable->tab);
	return 0;
}


int freeBTEntry(BrickTable_t *bTable, anaHandle_t handle) { 
	int i;
	struct BTentry_s *entry=NULL;
	struct BTentry_s *predecessor=NULL;
	struct BTentry_s *next=NULL;
	 
	 if(bTable ==NULL){
		anaPrint(ANA_ERR, "freeBTEntry : NULL pointer to Brick Table \n");
		return -1;
	 }
	 
	i= handle2hash(bTable, &handle);
	
	if (i<0 || i>(int) (bTable->size)-1) {
		anaPrint(ANA_ERR, "freeBTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = bTable->tab[i];
	predecessor = NULL; 
	
	while(entry != NULL){
		next = entry->next;
		if(handle == entry->handle) {
			if(predecessor != NULL)
				predecessor->next = entry->next;
			else 
				bTable->tab[i] = entry->next;
			
			free(entry->brickName);
			free(entry); 
			bTable->nbBricks--;
			return 0;
		}
		else {
			predecessor = entry;
			entry = next;
		}	
	}
	anaPrint(ANA_ERR, "free BTentry, no entry found\n");
	return -1;
}


int addBTEntry (BrickTable_t *bTable, struct BTentry_s *toAdd) {
	
	int i;
	struct BTentry_s *entry=NULL;
	
	if(bTable ==NULL){
		anaPrint(ANA_ERR, "addBTEntry : NULL pointer to Brick Table \n");
		return -1;
	}
	i = handle2hash(bTable, &(toAdd->handle));
	
	if (i<0 || i>(int) (bTable->size)-1) {
		anaPrint(ANA_ERR, "addBTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = bTable->tab[i];
	if(entry == NULL ){
		toAdd->next = NULL;
		bTable->tab[i] = toAdd;
	}
	else { /* hash collison */
		toAdd->next = bTable->tab[i];
		bTable->tab[i]=toAdd;
	}
	
	bTable->nbBricks++;
	return 0;
}

