/* C/minmex/common/dispatchTable.c
 *  
 * functions to deal with the information dispatch table
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "../include/dispatchTable.h"
#include "../include/anaNode.h"


unsigned int label2hash( IDT_t *Table, anaLabel_t *l) {

	if(Table==NULL ){
		anaPrint(ANA_ERR, "label2hash : NULL IDT pointer \n");
		return -1;
	}
	return  hash((ub1 *)l, sizeof(anaLabel_t) , 0 , Table->hashMaskSize); 
}


int initializeIDT (IDT_t *IDT, int hashMaskSize, int deadEntryLimit) { 
/* the allocated size of the table will be (2^hashMaskSize), the hashMaskSize will be also used in the hash   function */
	if(IDT==NULL || hashMaskSize<0){
		anaPrint(ANA_ERR, "initializeIDT : NULL arguments \n");
		return -1;
	}
	IDT->size = pow2(hashMaskSize);
	IDT->hashMaskSize = hashMaskSize;
	IDT->deadEntryLimit = deadEntryLimit;
	
	IDT->tab = (struct IDTentry_s **) malloc(IDT->size * sizeof(void *));
	if(IDT->tab == NULL){
		anaPrint(ANA_ERR, "initializeIDT: malloc returned error\n");
		return -1;
	}
	anaPrint(ANA_DEBUG, "IDT table initialized with size %u\n", IDT->size);
	return 0;
}



/* The getIDTEntry first hashes the label, then runs through entire chained list 
 * to find an entry owned the asking brick with the wanted label
 * ( the mapping of the label in the bricks view ). 
 * While running though the chain it also locates the public entry corresponding 
 * to the wanted label. This public entry is returned if the brick
 * didn't previously request to create a private view of this label through a redirect() method call  */

struct IDTentry_s* getIDTEntry(IDT_t *IDT, anaHandle_t h,  anaLabel_t l) {
	int i; 
	struct IDTentry_s *entry=NULL;
	struct IDTentry_s *publicEntry=NULL;
	
	if(IDT==NULL){ 
		anaPrint(ANA_ERR, "getIDTEntry : NULL IDT pointer\n");
		return NULL;
	}	
	
	i = label2hash(IDT, &l);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "getIDTEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	
	entry = IDT->tab[i];
	publicEntry = NULL;
	
	while (entry != NULL) {
		if (entry->label==l){
			if(entry->viewer == 0)
				publicEntry = entry;
			
			if (h!=0 && h==entry->viewer){
				if(entry->age != -1)
					entry->age = 0;
				return entry;
			}
			else
				entry = entry->next;
		}
		else 
			entry = entry->next;
	}
	if(publicEntry!=NULL && publicEntry->age != -1)
		publicEntry->age = 0;
	return publicEntry;
}

struct IDTentry_s* getIDTEntryStrictView(IDT_t *IDT, anaHandle_t h,  anaLabel_t l) {
	int i; 
	struct IDTentry_s *entry=NULL;
	
	if(IDT==NULL){ 
		anaPrint(ANA_ERR, "getIDTEntryStrictViewer : NULL IDT pointer \n");
		return NULL;
	}	
	
	i = label2hash(IDT, &l);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "getIDTEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	
	entry = IDT->tab[i];
	
	while (entry != NULL) {
		if (l==entry->label){
			if (h==entry->viewer){
				if(entry->age != -1)
					entry->age = 0;
				return entry;
			}
		}
		entry = entry->next;
	}
	
	return NULL;
}




int freeIDTTable(IDT_t *IDT) { 
	int i;
	struct IDTentry_s *entry;
	struct IDTentry_s *next;
	
	if(IDT==NULL ){
		anaPrint(ANA_ERR, "freeIDTTable : NULL IDT pointer\n");
		return -1;
	}
	
	for (i=0; i< (int) IDT->size; i++) {
		entry = IDT->tab[i];
		while  (entry!=NULL) {
			next = entry->next;
			free(entry->aux);
			
			free(entry->destContext.value);
			free(entry->destService.value);
			
			free(entry);
			entry = next;
		}
	}
	IDT->size = 0;
	free(IDT->tab);
	free(IDT);
	return 0;
}


int freeHandleIDTEntries(IDT_t *IDT, anaHandle_t h) {  /* called at every detach(), too heavy :(*/
	int i;
	struct IDTentry_s *entry;
	struct IDTentry_s *predecessor;
	struct IDTentry_s *next;
	struct anaEventNotif_s notif;
	if(IDT==NULL){
		anaPrint(ANA_ERR, "freeHandleIDTEntries : NULL IDT pointer \n");
		return -1;
	}
	
	for(i=0; i< (int) IDT->size; i++) {
		entry = IDT->tab[i];
		predecessor = NULL;
		
		while(entry!=NULL) {
			next = entry->next;
			if(entry->owner == h) {
				if(predecessor!=NULL)
					predecessor->next = entry->next; 
				else  
					IDT->tab[i] =  entry->next;
				
				free(entry->aux);
				
				free(entry->destContext.value);
				free(entry->destService.value);
				
				/* we notify about the deletion of this IDP */
				memset(&notif, 0, sizeof(struct anaEventNotif_s));
				notif.eventID = IDP_DELETED_EVENT;
				notif.label = entry->label;
				notif.viewer = entry->viewer;
	
				minmex_notifyEvent(0, &notif );
				
				free(entry);
				entry = next;
			
				
			}
			else {
				predecessor = entry;
				entry = next;
			}
		}	
	}
	
	return 0;
}



int freeIDTEntry(IDT_t *IDT, anaHandle_t viewer, anaLabel_t l) { 
	int i ;
	struct IDTentry_s *entry;
	struct IDTentry_s *predecessor;
	
	if(IDT==NULL){
		anaPrint(ANA_ERR, "freeIDTEntry : NULL IDT pointer \n");
		return -1;
	}
	
	i= label2hash(IDT, &l);
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "freeIDTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	entry = IDT->tab[i];
	predecessor = NULL;
	while (entry != NULL) {
		if(l == entry->label && viewer == entry->viewer) {
			
			if(predecessor!=NULL)
				predecessor->next = entry->next; 
			else  
				IDT->tab[i] =  entry->next; 
				
			free(entry->aux);
			
			free(entry->destContext.value);
			free(entry->destService.value);
			
			free(entry); 
			return 0;
		}
		else{
			predecessor = entry;
			entry = entry->next;
		}
	}
	anaPrint(ANA_DEBUG,"freeIDTEntry: entry not found\n");
	return -1;
}




/* addIDTentry function : this function first looks if an entry with the same 
 * label and owned by the same brick exists, in which case the old
 * entry will be crashed by the new. The function also looks for the presence
 * of entries with same label owned by different bricks (handles), 
 * in which case it will declare the entry to add as a private one owned 
 * by the brick initiating the entry addition. 
 */
int addIDTEntry (IDT_t *IDT, struct IDTentry_s *toAdd, int permanent) {
	int i ;
	struct IDTentry_s *entry;
	struct IDTentry_s *predecessor;
	struct IDTentry_s *next;
	
	if(IDT==NULL || toAdd==NULL){
		anaPrint(ANA_ERR, "addIDTEntry : NULL arguments \n");
		return -1;
	}
	
	i = label2hash(IDT, &toAdd->label);
	
	if (i<0 || i>(int) (IDT->size)-1) {
		anaPrint(ANA_ERR, "addIDTEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}	
	entry = IDT->tab[i];
	predecessor = NULL;
	next = NULL;
	if(permanent) 
		toAdd->age = -1;
	else
		toAdd->age = 0;
		
	while(entry!=NULL) {
		next = entry->next;
		if (toAdd->label == entry->label) { /*if another entry with same label exists*/
			
			if (toAdd->viewer == entry->viewer) {  /*if an entry with same label and viewer exists*/
				
				anaPrint(ANA_ERR, "addIDTEntry: entry with same label exists!\n");
				if(toAdd->owner == entry->owner){ /*if it's owned by the asker, we crash the previous entry */
				
					anaPrint(ANA_NOTICE, "addIDTEntry: entry with same label, viewer and owner exists, Crashing it !\n");
				
					free(entry->aux);
					free(entry); 
					if(predecessor!=NULL) {
						predecessor->next = next;
						toAdd->next = IDT->tab[i];
					}
					else {
						toAdd->next = next;
					}
					IDT->tab[i] = toAdd;
					return 0;
				}
				else{	/*if an entry with same label and viewer exists and is owned by another brick => 
					impossible to add entry*/
					anaPrint(ANA_ERR, "addIDTEntry :Impossible to add Entry. An entry with same label, viewer and \
					Different owner exists.\n");
					return -1;
				}
			}
		}
		predecessor = entry;	
		entry = next;
	}
	
	toAdd->next = IDT->tab[i];
	IDT->tab[i] = toAdd;
	return 0;
}


int setPermanentEntry(IDT_t *IDT, anaHandle_t h,  anaLabel_t l) {
	struct IDTentry_s *entry = NULL;
	if(IDT==NULL)
		return -1;
	
	entry = getIDTEntry(IDT,h,l);
	if(entry==0) {
		anaPrint(ANA_ERR, "setPermanentEntry : IDT entry doesn't exist \n");
		return -1;
	}
	entry->age = -1;
	return 0;
}


int setVolatileEntry(IDT_t *IDT, anaHandle_t h,  anaLabel_t l) {
	struct IDTentry_s *entry = NULL;
	if(IDT==NULL)
		return -1;
	
	entry = getIDTEntry(IDT,h,l);
	if(entry==0) {
		anaPrint(ANA_ERR, "setVolatileEntry : IDT entry doesn't exist \n");
		return -1;
	}
	entry->age = 0;
	return 0;
}
