/* C/minmex/common/minmexFunctions.c
 * 
 * functions for basic communication with the bricks
 * 
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> (initial design, user space implementation
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> (kernel space extension)
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include "../include/anaNode.h"
//#include "anaError.h"

/* we need a place to define them...*/
char *validatebuf;
#ifdef USER
FILE *validateFile;
#endif

/**
 * Returns the floor form of binary logarithm for a 32 bit integer.
 * -1 is returned if n is 0.
 */
int floorLog2(unsigned int n) {
  int pos = 0;
  if (n >= 1<<16) { n >>= 16; pos += 16; }
  if (n >= 1<< 8) { n >>=  8; pos +=  8; }
  if (n >= 1<< 4) { n >>=  4; pos +=  4; }
  if (n >= 1<< 2) { n >>=  2; pos +=  2; }
  if (n >= 1<< 1) {           pos +=  1; }
  return ((n == 0) ? (-1) : pos);
}


int sendData( void *msg, int len, anaLabel_t input, struct IDTaux_s *idtAux ) {  
        struct BTentry_s* entry = NULL;
   	struct dataMessageQueue_s *message = NULL;
	struct dataMessageQueue_s *tmp = NULL;

#ifdef USER
	struct commSpecs_s *comSpec = NULL;
#endif
#ifdef KERNEL
	struct anaBrickFunctions * clifunc  = NULL;
#endif

	if(msg==NULL || len<0 ||  idtAux ==NULL)
		return -1;
	
	entry = getBTEntry(&BT, idtAux->callee);
	if (entry == NULL) {
		anaPrint(ANA_ERR, "sendData: no BT entry corresponding to destination handle \n");
		return -2;
	}

	
	/* we build our message */
	message = (struct dataMessageQueue_s *) malloc(sizeof(struct dataMessageQueue_s));
	message->callBackFct = idtAux->callBackFct;
	message->inputLabel = input; 
	message->data = (void*) malloc(len);
	memcpy(message->data, msg, len);
	message->dataLen = len;
	message->aux = idtAux->aux;
	message->thread = idtAux->thread;
	message->next = NULL;

#ifdef USER
	comSpec	= (struct commSpecs_s *)entry->commAux;

	/* we lock the queue of the brick */
	analock_lock(comSpec->queueLock);
	
	/* we queue our message */
	tmp = *(comSpec->msgQueue);
	if(tmp== NULL) /*meaning no messages on the queue*/
		*(comSpec->msgQueue) = message;
	else{
		while(tmp->next != NULL)
			tmp = tmp->next;
		tmp->next = message;
	}

	/* we wake up the brick thread  and unlock*/
	analock_cond_signal(comSpec->condLock);
	analock_unlock(comSpec->queueLock);
	
	return len;	
#endif

#ifdef KERNEL
	clifunc = (struct anaBrickFunctions *)entry->commAux; 

	/* we lock the queue of the brick */
	
	analock_lock(clifunc->queueLock);
	/* we queue our message */
	tmp = *(clifunc->recvDataQueue);
	if(tmp == NULL) {/*meaning no messages on the queue*/
		*(clifunc->recvDataQueue) = message;
	}
	else{
		while(tmp->next != NULL)
			tmp = tmp->next;
		tmp->next = message;
	}
	queue_work(clifunc->wq, clifunc->work);
	analock_unlock(clifunc->queueLock);

		
	return len;
#endif
}



/* function to increment the number of entries belonging to a brick */
int incrementNBEntries(anaHandle_t handle, int perm){
	struct BTentry_s *entry=NULL;
	entry = getBTEntry(&BT, handle);
	anaPrint(ANA_NOTICE, "incrementing counter of entries belonging to a brick \n");
	if(entry == NULL){
		return -1;
	
	}
	if(perm){
		if(entry->nbPermanentEntries ==  MAX_NB_PERMANENT)
			return -1;
		(entry->nbPermanentEntries)++;
		return entry->nbPermanentEntries;
	}
	if(entry->nbVolatileEntries ==  MAX_NB_VOLATILE)
		return -1;
	(entry->nbVolatileEntries)++;
	return entry->nbVolatileEntries;
}

/* function to decrement the number of entries belonging to a brick */
int decrementNBEntries(anaHandle_t handle, int perm){
	struct BTentry_s *entry=NULL;
	entry = getBTEntry(&BT, handle);
	if(entry == NULL){
		return -1;
	}
	if(perm){
		if(entry->nbPermanentEntries ==  0)
			return -1;
		(entry->nbPermanentEntries)--;
		return entry->nbPermanentEntries;
	}
	if(entry->nbVolatileEntries ==  0)
		return -1;
	(entry->nbVolatileEntries)--;
	return entry->nbVolatileEntries;
}

int kickAllBricks(void){
        struct BTentry_s *tmp=NULL;
        int i;
	anaPrint(ANA_NOTICE, "Sending notification of shutdown to all attached bricks \n");
        for(i=0;i<(int)BT.size;i++){
                tmp =BT.tab[i];
		while(tmp!=NULL){
                        struct BTentry_s *nextBrick = tmp->next;
#ifdef KERNEL
			if(tmp->pluginFlag) { /* we notify only plugin bricks */
				struct anaBrickFunctions * clifunc  = NULL;
				anaPrint(ANA_DEBUG, "pluginFlag set\n");

				/*only a temporary hack to avoid the brick crashing if the
 				 * minmex is unloaded before the brick
 				 */
				clifunc = (struct anaBrickFunctions *)tmp->commAux; 
				clifunc->notification();
			}
#endif	
                        /*no error check since minmex will disappear anyway*/
			kickBrick(tmp->handle);
                        tmp=nextBrick;
                }
        }

        return 0;
}




int existsIDP(anaLabel_t label){
	struct IDTentry_s *entry = NULL;
	
	entry = getIDTEntry(&IDT, 0, label);
	if(entry==0) 
		return 0;
	
	return 1;
}


int kickBrick(anaHandle_t handle){
	/* we first notify the brick */
	struct anaEventNotif_s notif;
	char *brickName = NULL;
	
#ifdef USER
	struct commSpecs_s *specs = NULL;
	struct BTentry_s *entry = NULL;
	void *pluginHandle = NULL;
	/*void (*plug_brick_exit)(void);*/
	int (*brick_exit)(void);

#endif	
	
	if(handle==0)
		return -1;
#ifdef USER
	entry = getBTEntry( &BT , handle);
	if(entry != NULL)
		specs = (struct commSpecs_s *)entry->commAux;
	
	if(specs)
		pluginHandle = specs->pluginHandle;
			
		
	brickName = (char *) malloc(strlen(entry->brickName)+1);
	strcpy(brickName, entry->brickName);
	

	/* we call the brick's brick_exit function */
	if(entry->pluginFlag && pluginHandle){
		dlerror();
		brick_exit = dlsym(pluginHandle, "whenKicked");
		if(brick_exit ==NULL) 
			anaPrint(ANA_DEBUG, "unable to load symbol : %s\n", dlerror());
		else
			brick_exit();
	
		anaThread_killbrick(handle); 
		dlclose(pluginHandle);	 
	}
	else{
		int (*whenBrickKicked)(anaHandle_t h);
		/* we alert the gates plugin that the brick is being kicked */
		dlerror();
		brick_exit = dlsym(pluginHandle, "whenBrickKicked");
		if(whenBrickKicked ==NULL) 
			anaPrint(ANA_DEBUG, "unable to load symbol : %s\n", dlerror());
		else
			whenBrickKicked(handle);
	} 
#endif
/* we free the brick entry in BT */
	if(freeBTEntry(&BT, handle)<0){
		anaPrint(ANA_ERR, "kickBrick: Unable to free brick %d from Brick Table\n", handle);
		return -1;
	}
	
     	/* we free the IDT entries corresponding to the brick */
	if(freeHandleIDTEntries(&IDT, handle)<0){
		anaPrint(ANA_ERR, "kickBrick: Unable to free IDPs belonging to %d from Dispatch Table\n", handle);
		return -1;
	}
	
	/* we also free the KVR entries owned by this brick*/
	freeHandleKVREntries(&kValRep, handle);
	
	
	/* we clear all the event subscriptions belonginf to this brick */
	clearBrickEvents(&NT, handle);
	
	/* we notify bricks interested in this event */
	memset(&notif, 0, sizeof(struct anaEventNotif_s));
	notif.eventID = BRICK_DETACHED_EVENT;
	notif.name = brickName; /* brickName will be freed by the notifWrapper in anaLib0_so*/ 
	
	minmex_notifyEvent(0, &notif );
	
	
	anaPrint(ANA_DEBUG, "Kicked brick, current number of bricks: %d\n",BT.nbBricks);
	return 0;		
}

int createsLoop(anaHandle_t originViewer, anaLabel_t originLabel, anaHandle_t destViewer, anaLabel_t destLabel){
	struct IDTentry_s *idtEntry=NULL;	
	
	if(originViewer == destViewer && destLabel == originLabel)
		return 1;


	idtEntry  = getIDTEntry(&IDT, destViewer, destLabel);

	/* we check for redirection loops */
	while(idtEntry->callBackFct == (minmexCallback_t)& redirectData && idtEntry->aux != NULL){
		
		/*we check if the label to redirect to is equal to origin*/
		
		struct redirectAux_s *tmpau = (struct redirectAux_s *)idtEntry->aux;
		
		if( tmpau->handle == originViewer && tmpau->label == originLabel)
			/* There is a loop */
			return 1;
				
		
		idtEntry  = getIDTEntry(&IDT, tmpau->handle, tmpau->label);
	}

	/* here the redirect chain ends*/
	if(idtEntry->callBackFct == (minmexCallback_t)& forkData && idtEntry->aux != NULL){
		int offset;
		int *nbForks = (int *)idtEntry->aux;
		int auxSize;
					
		offset = sizeof(int);
		auxSize =  sizeof(int) + *nbForks * (sizeof(struct redirectAux_s));

		/* then we call contains loops recursively for all forks */
		while(offset< auxSize){
			struct redirectAux_s *auxil = idtEntry->aux + offset;
			anaHandle_t handle = (*auxil).handle;
			anaLabel_t label = (*auxil).label;
			
			if(createsLoop(originViewer, originLabel, handle, label) )
				return 1;
			offset += sizeof(struct redirectAux_s);
		}
	}
	return 0;
}


/* this function is to be called when a brick is kicked  */
int  clearBrickEvents(notifTable_t *Table, anaHandle_t owner){ 
	int i; 
	struct notifAux_s *tmpNotif = NULL;
	struct notifAux_s *predecessor = NULL;
	struct notifEntry_s *tmpEntry = NULL;
	struct notifEntry_s *entryPredecessor = NULL;	
	struct notifEntry_s *entryNext = NULL;
	/*we first clear all the allEvents*/
	for(i=0;i<32;i++){
		tmpNotif = allEvents[i];
		while(tmpNotif!=NULL){
			if(tmpNotif->owner == owner){
				if(predecessor == NULL)
					allEvents[i] = tmpNotif->next;
				else
					predecessor->next = tmpNotif->next;
			
				free(tmpNotif->aux);
				free(tmpNotif);
				break;
			}
			predecessor = tmpNotif;
			tmpNotif = tmpNotif->next;	
		}
	}
	
	/* then we go through all the notification table, yes very heavy ! */
	for(i=0;i<NT.size;i++){
		tmpEntry = NT.tab[i];
		predecessor = NULL;
		while(tmpEntry!=NULL){
			tmpNotif = tmpEntry->notifAux;
			while(tmpNotif != NULL){
				if(tmpNotif->owner == owner){
					if(predecessor == NULL)
						tmpEntry->notifAux = tmpNotif->next;
					else
						predecessor->next = tmpNotif->next;
						
					free(tmpNotif->aux);
					free(tmpNotif);
					break;
				}
				predecessor = tmpNotif;
				tmpNotif = tmpNotif->next;	
			}
			if(tmpEntry->notifAux == NULL){
				/* nobody is subscribed to this event anymore, we free the entry */
				if(entryPredecessor == NULL)
					NT.tab[i] = tmpEntry->next;
				else
					entryPredecessor->next = tmpEntry->next;	
			
				free(tmpEntry->eventKey);
				entryNext = tmpEntry->next;
				free(tmpEntry);
				tmpEntry = entryNext;
			}
			else{
				entryPredecessor = tmpEntry;
				tmpEntry = tmpEntry->next;
			}
		}
	
	}
	return 0;
}

int  minmex_unregisterEventHandler(anaHandle_t owner, struct anaEventSub_s* sub ){

/* first we check if the querier exists */
	struct BTentry_s *querierEntry = NULL;
	struct notifEntry_s *entry = NULL;
	struct notifAux_s *predecessor = NULL;
	struct notifAux_s *tmpNotif = NULL;
	uint32_t flag;
	void *param = NULL;
	int i, paramLen = 0;
	
	if(sub==NULL)
		return -1;
	
	querierEntry = getBTEntry( &BT , owner);
	if(querierEntry == NULL)
		return ANA_NOORIGINVIEWER_ERROR;
		
	if( (IDP_DELETED_EVENT&sub->eventMask) || (IDP_REDIRECTED_EVENT&sub->eventMask) || (IDP_UNPUBLISHED_EVENT&sub->eventMask) ){
			
		anaPrint(ANA_DEBUG, "minmex_unregisterEventHandler : unsubscribe regarding IDP events\n");
		
		/* as a security measure we verify that the mask does not contains the ATThed brick events */
		if(sub->eventMask&BRICK_ATTACHED_EVENT || sub->eventMask&BRICK_DETACHED_EVENT)
			return ANA_WRONG_EVENTMASK;
	
		param = &(sub->label);
		paramLen = sizeof(anaLabel_t) + sizeof(anaHandle_t);
	}
	else{
		if(sub->eventMask&BRICK_ATTACHED_EVENT || sub->eventMask&BRICK_DETACHED_EVENT){
			anaPrint(ANA_DEBUG, "minmex_unregisterEventHandler : unsubscribe regarding BRICK events, brick name: %s \n", sub->name);
			if( sub->name ) {
				param = sub->name;
				paramLen = strlen(sub->name)+1;
			}
		}
		else
			return ANA_WRONG_EVENTMASK;
			
	}
	
	if( sub->label == 0 && sub->name == NULL){
		
		for(i=0;i<32;i++){ /* feeling lazy :)*/
			flag = 1 << i;
			if(sub->eventMask&flag){
				tmpNotif = allEvents[i];
				predecessor = NULL;
				
				while(tmpNotif!=NULL){
					if(tmpNotif->owner == owner){
						/* we free the entry */
						if(predecessor == NULL)
							allEvents[i] = tmpNotif->next;
						else
							predecessor->next = tmpNotif->next;
						
						free(tmpNotif->aux);
						free(tmpNotif);
						
						break;
					}
					predecessor = tmpNotif;
					tmpNotif = tmpNotif->next;
				}
				
			} 
		}
		return 0;
	} 
	
	entry = getNotifEntry(&NT, param,  paramLen);
		
	if( entry == NULL)
		return ANA_NOMATCH_SUB;
	
	tmpNotif = entry->notifAux;
	predecessor = NULL;
	
	while(tmpNotif!=NULL){
		if(tmpNotif->owner == owner && tmpNotif->eventID ==  sub->eventMask){
			/* we free the entry */
			if(predecessor == NULL)
				entry->notifAux = tmpNotif->next;
			else
				predecessor->next = tmpNotif->next;
						
			free(tmpNotif->aux);
			free(tmpNotif);
			break;
		}
		predecessor = tmpNotif;
		tmpNotif = tmpNotif->next;
	}	
	
	if((entry->notifAux) == NULL){/* if none is registered to this event anymore, we delete it */
		freeNotifEntry(&NT, param, paramLen);
	} 

		
	
	return 0;

}

int  minmex_registerEventHandler(anaHandle_t callee, anaHandle_t owner, struct anaEventSub_s* sub, anaCallback_t notificationWrapper, void *aux, int auxLen, int threadFlag ){
	
	/* first we ceck if the querier exists */
	struct BTentry_s *querierEntry = NULL;
	struct notifEntry_s *entry = NULL;
	struct notifAux_s *notifAux;
	struct notifAux_s *tmpNotif;
	uint32_t flag;
	void *param = NULL;
	int ret, i, paramLen=0;
	eventHandler_t *t;
	struct IDTentry_s *idtentry=NULL;
	
	if(sub==NULL || notificationWrapper==NULL || aux == NULL)
		return -1;
	
	querierEntry = getBTEntry( &BT , owner);
	if(querierEntry == NULL)
		return ANA_NOORIGINVIEWER_ERROR;
	
	/* we first Identify the events in the mask */
	
	if(sub->eventMask&IDP_DELETED_EVENT || sub->eventMask&IDP_REDIRECTED_EVENT ||  sub->eventMask&IDP_UNPUBLISHED_EVENT){
			
		anaPrint(ANA_DEBUG, "minmex_registerEventHandler : subscribe regarding IDP events\n");
		
		/* as a security measure we verify that the mask does not contains the ATThed brick events */
		if(sub->eventMask&BRICK_ATTACHED_EVENT || sub->eventMask&BRICK_DETACHED_EVENT)
			return ANA_WRONG_EVENTMASK;
		
		param = &(sub->label);
		paramLen = sizeof(anaLabel_t) + sizeof(anaHandle_t);
	
		/* we check that the IDP exists in the IDT */
		idtentry = getIDTEntryStrictView(&IDT, sub->viewer, sub->label);
		if(idtentry == NULL)
			return ANA_NODESTIDTENTRY_ERROR;
	}
	else{
		if(sub->eventMask&BRICK_ATTACHED_EVENT || sub->eventMask&BRICK_DETACHED_EVENT){
			anaPrint(ANA_DEBUG, "minmex_registerEventHandler : subscribe regarding BRICK events\n");
			if(sub->name){
				param = sub->name;
				paramLen = strlen(sub->name)+1;
			}
			
			/* if it is a brick attached Event we check that the brick is not already attached */
			if(sub->eventMask&BRICK_ATTACHED_EVENT){
				if(brickname2handle(&BT, sub->name) != 0)	
					return ANA_ALREADY_ATTACHED;
			}
			/* if it is a brick detached Event we check that the brick is attached */
			if(sub->eventMask&BRICK_DETACHED_EVENT){
				if(brickname2handle(&BT, sub->name) == 0)	
					return ANA_NOBRICK_ATTACHED;
			}
			
			
		}
		else
			return ANA_WRONG_EVENTMASK;
			
	}
		
	notifAux = (struct notifAux_s *) malloc(sizeof(struct notifAux_s));
	notifAux->callee = callee;
	notifAux->owner =  owner;
	notifAux->eventID = sub->eventMask;
	notifAux->callBackFct = notificationWrapper;
	notifAux->aux = (void *) malloc(auxLen);
	memcpy(notifAux->aux, aux, auxLen);
	
	t = (eventHandler_t *) aux; 
		
	notifAux->auxLen = auxLen;
	notifAux->thread = threadFlag;
	
		
	if(sub->label ==0 &&sub->name == NULL){ /* if params are NULL it means that the querier wants to subscribe to all events of the specified types */
		
		for(i=0;i<32;i++){ /* feeling lazy :)*/
			flag = 1 << i;
			if(sub->eventMask&flag){
				anaPrint(ANA_DEBUG, "**** registering the subscrption at index %d\n",i);
			
				tmpNotif = allEvents[i];
				/* we check that the querier was not subscribed to this event previously */
				while(tmpNotif!=NULL){
					if(tmpNotif->owner == owner){
						anaPrint(ANA_DEBUG, "minmex_registerEventHandler: querier already subscribed to this global event => abort\n");
						free(notifAux->aux);
						free(notifAux);
						return ANA_ALREADY_SUBSCRIBED;
					}
					tmpNotif = tmpNotif->next;
				}
				
				/* if not previously subscribed we add the entry */
				notifAux->next = allEvents[i];
				allEvents[i] = notifAux;	
			} 
		}

		return 0;
	} 
	
	
	/* we look for an existing entry handling the same event (same key + same eventID) */
	
	entry = getNotifEntry(&NT, param,  paramLen);
		
	if( entry == NULL) { /* meaning there was no previous subscription regarding this object (IDP or Brick) */
	
		entry = (struct notifEntry_s *) malloc(sizeof(struct notifEntry_s));
	
		entry->eventKey = (void *) malloc(paramLen);
		memcpy(entry->eventKey, param, paramLen );
		entry->keyLen = paramLen;
		
		entry->notifAux = notifAux;
		
		/* we add the entry to the notification table */
		ret = addNotifEntry(&NT, entry);
		if(ret < 0){
			anaPrint(ANA_DEBUG, "minmex_registerEventHandler: Unable to add event entry to notification table\n");
			free(entry->eventKey);
			free(entry);
			free(notifAux->aux);
			free(notifAux);
			return -1;
		} 
	}
	else{ /* a subscribtion for the same event already exists */
		struct notifAux_s *tmp = entry->notifAux;
		
		while(tmp!=NULL){
			if(tmp->owner == owner && (!(tmp->eventID&sub->eventMask))){  /* we check that the querier was not previously subscribed to one of the events in the Mask */
				anaPrint(ANA_DEBUG, "minmex_registerEventHandler: querier already subscribed to this event => abort\n");
				free(notifAux->aux);
				free(notifAux);
				return ANA_ALREADY_SUBSCRIBED;
			}
			tmp = tmp->next;
		}
		/* we append ou notif Aux to the */
		notifAux->next = entry->notifAux;
		entry->notifAux	= notifAux;
		
	}
	return 0;
}

int  minmex_notifyEvent(anaHandle_t notifier, struct anaEventNotif_s *notif ){ /*notification is done for a single event (no masks)*/
	
	struct notifEntry_s *entry = NULL;
	void *param;
	int paramLen=0;
	int found=0;
	struct IDTaux_s tmp;
	struct notifAux_s *tmpNotif;	
	struct notifAux_s *next;
	int i;
	
	if(notif==NULL)
		return -1;
	
	/* first we check authorization to notify for example the notifier must be the owner of the IDP, however for the unpublish it is difficult to check rights */
	/*
	if(eventMask&IDP_DELETED_EVENT || eventMask&IDP_REDIRECTED_EVENT ||  eventMask&IDP_DELETED_EVENT){
	
		struct IDPnotif_s *IDP = (struct IDPnotif_s *) params;
		
		struct IDTentry_s * IDTentry = getIDTEntryStrictView(&IDT, IDP->viewer, IDP->label);
		if(IDTentry==NULL)
			return ANA_NODESTIDTENTRY_ERROR;
		if(entry->owner != notifier)
			
	}	
	*/
	
	/* we first find the entry that matches */
	
	if(notif->eventID&IDP_DELETED_EVENT || notif->eventID&IDP_REDIRECTED_EVENT ||  notif->eventID&IDP_UNPUBLISHED_EVENT){
			
		anaPrint(ANA_DEBUG, "minmex_notifyEvent : notify regarding IDP %8x\n", notif->label);
		
		param = &(notif->label);
		paramLen = sizeof(anaLabel_t) + sizeof(anaHandle_t);
			
	}
	else{
		if(notif->eventID&BRICK_ATTACHED_EVENT || notif->eventID&BRICK_DETACHED_EVENT){
			anaPrint(ANA_DEBUG, "minmex_notifyEvent : notify regarding BRICK events\n");
			param = notif->name;
			paramLen = strlen(notif->name)+1;
		}
		else
			return ANA_WRONG_EVENTMASK;
			
	}
	
	
	entry = getNotifEntry(&NT, param,  paramLen);
	
	if(entry != NULL) {
		tmpNotif = entry->notifAux;
		while(tmpNotif!=NULL){
			next = tmpNotif->next;
			if(tmpNotif->eventID&notif->eventID) {
				found = 1;
				memset(&tmp, 0, sizeof(struct IDTaux_s));
				tmp.callee = tmpNotif->callee;
				tmp.callBackFct = tmpNotif->callBackFct;
				tmp.aux = (void*) malloc(tmpNotif->auxLen);
				memcpy(tmp.aux, tmpNotif->aux, tmpNotif->auxLen);
				tmp.thread = tmpNotif->thread;
				
				if(sendData( notif, sizeof(struct anaEventNotif_s), 0, &tmp )<0)
					anaPrint(ANA_DEBUG, "minmex_notifyEvent: unable to transmit notification via sendData\n");
				
			
			}
			tmpNotif = next;
		}
		
		if(notif->eventID&IDP_DELETED_EVENT || notif->eventID&BRICK_ATTACHED_EVENT || notif->eventID&BRICK_DETACHED_EVENT ){
			/* we free the entry for this object */
			freeNotifEntry(&NT, param, paramLen);
		}
	}
	
	
	/* we also alert the interested in all events of this type */
	i = floorLog2(notif->eventID);
	
	
	tmpNotif = allEvents[i];
	
	if(!found && allEvents[i]==NULL ){
		anaPrint(ANA_DEBUG, "minmex_notifyEvent: no matching subscription to this event\n");
		return ANA_NOMATCH_SUB;
	}
		
	while(tmpNotif!=NULL){
		memset(&tmp, 0, sizeof(struct IDTaux_s));
		tmp.callee = tmpNotif->callee;
		tmp.callBackFct = tmpNotif->callBackFct;
		tmp.aux = (void*) malloc(tmpNotif->auxLen);
		memcpy(tmp.aux, tmpNotif->aux, tmpNotif->auxLen);
		tmp.thread = tmpNotif->thread;
			
		if(sendData( notif, sizeof(struct anaEventNotif_s), 0, &tmp )<0)
			anaPrint(ANA_DEBUG, "minmex_notifyEvent: unable to transmit notification via sendData\n");
		
		tmpNotif = tmpNotif->next;
	}
	
	return 0;
}



struct attachRes_s *handleAttach(char *brick_name, void *commAux){

#ifdef USER
	struct commSpecs_s *auxComm = commAux; 	
	struct commSpecs_s *tmp = NULL;
#endif
#ifdef KERNEL
	struct anaBrickFunctions *auxComm = (struct anaBrickFunctions*)commAux;
#endif
	anaHandle_t handle;
	struct BTentry_s *toAdd = NULL;
	struct attachRes_s *res = NULL;
	anaLabel_t nodeCompLabel; 
	struct IDTentry_s *nodeCompEntry = NULL;
	struct anaEventNotif_s notif;
	
	if(brick_name == NULL || commAux == NULL)
		return NULL;
	
	/* we generate a handle for this brick */
	handle = generateUnusedHandle(&BT);
	
	toAdd = (struct BTentry_s *) malloc(sizeof(struct BTentry_s));
	toAdd->pluginFlag = auxComm->pluginFlag;
	
	toAdd->handle = handle;
	toAdd->age = 0;
	toAdd->nbPermanentEntries=0;
	toAdd->nbVolatileEntries=0;
	toAdd->brickName = (char *) malloc(strlen(brick_name)+1);
	strcpy(toAdd->brickName, brick_name);
	
	/* we store the communication auxiliary*/
#ifdef USER
	tmp = (struct commSpecs_s *) malloc(sizeof(struct commSpecs_s));
	
	memcpy(tmp, auxComm, sizeof(struct commSpecs_s));
	toAdd->commAux = tmp;
	toAdd->pluginFlag = tmp->pluginFlag;
	
#endif
#ifdef KERNEL
	toAdd->commAux = malloc(sizeof (struct anaBrickFunctions));
	memcpy(toAdd->commAux, commAux, sizeof(struct anaBrickFunctions));
#endif		
	/* we add the entry to the brickTable*/
	if(addBTEntry(&BT, toAdd)<0){
		free(toAdd->commAux);
		free(toAdd->brickName);
		free(toAdd);
		return NULL;
	}
	
	/* we add a node Compartment entry */
	nodeCompLabel = generateLabel(); 
	nodeCompEntry = (struct IDTentry_s *) malloc(sizeof(struct IDTentry_s));
	nodeCompEntry->label = nodeCompLabel;
	nodeCompEntry->viewer = handle;
	nodeCompEntry->owner = handle;
	nodeCompEntry->callBackFct = (minmexCallback_t)& dispatchNodeCompMsg;
	nodeCompEntry->aux = malloc(sizeof(anaHandle_t));
	*((anaHandle_t *) (nodeCompEntry->aux)) = handle;
	nodeCompEntry->auxLen = sizeof(anaHandle_t);
	
	if (addIDTEntry(&IDT, nodeCompEntry, 1)<0){
		anaPrint(ANA_ERR, "Unable to add nodeCompartment IDT entry for new brick => attach aborted \n");
		free(nodeCompEntry->aux);
		free(nodeCompEntry);
		
		/* we free the BT entry */
		freeBTEntry(&BT, handle);
		return NULL;
	}
	
	
	res = (struct attachRes_s *) malloc(sizeof(struct attachRes_s));
	res->handle = handle;
	res->nodeCompLabel = nodeCompLabel;
	
	/* we notify the interested bricks of this detach */
	memset(&notif, 0, sizeof(struct anaEventNotif_s));
	notif.eventID = BRICK_ATTACHED_EVENT;
	notif.name = (char *) malloc(strlen(brick_name)+1); /* brickName will be freed by the notifWrapper in anaLib0_so*/ 
	strcpy(notif.name, brick_name);
	
	minmex_notifyEvent(0, &notif );
	
	/* we return the handle*/
	return res;
}

void handleDetach(anaHandle_t handle, int exitVal){ 
	char *brickName;
	struct BTentry_s *entry = NULL;
	struct anaEventNotif_s notif;
#ifdef USER
	struct commSpecs_s *specs = NULL;
	void *pluginHandle = NULL;
	int pluginFlag; 

#endif		
	entry = getBTEntry(&BT, handle);
#ifdef USER
	if(entry != NULL)
		specs = (struct commSpecs_s *) entry->commAux;
	
	if(specs != NULL)
		pluginHandle = specs->pluginHandle;

	pluginFlag = entry->pluginFlag;
#endif	
	anaPrint(ANA_DEBUG, "Received Detach request from Brick \n");
	
	brickName = (char *) malloc(strlen(entry->brickName)+1);
	strcpy(brickName, entry->brickName);
	
	/* we free the brick entry in BT */
	if(freeBTEntry(&BT, handle)<0){
		anaPrint(ANA_ERR, "handleDetach: Unable to free brick %d from Brick Table\n", handle);
		return;
	}

	/* we free the IDT entries owned by the detaching brick */
	if(freeHandleIDTEntries(&IDT, handle)<0){
		anaPrint(ANA_ERR, "handleDetach: Unable to free IDPs belonging to %d from Dispatch Table\n", handle);
		return;
	}
	
	/* we also free KVR entries owned by this brick*/
	freeHandleKVREntries(&kValRep,handle);
	
	/* we clear all the event subscriptions belonginf to this brick */
	clearBrickEvents(&NT, handle);
	
#ifndef KERNEL
	if(exitVal != -255){
		/* we kill all the threads of the .so brick */
		anaThread_killbrick(handle);
	}
#endif	

#ifdef USER
	/* we unload the brkic's .so code  : */
	if(pluginFlag && pluginHandle!=NULL)
		dlclose(pluginHandle); 
#endif

	/* we notify the interested bricks of this detach */
	memset(&notif, 0, sizeof(struct anaEventNotif_s));
	notif.eventID = BRICK_DETACHED_EVENT;
	notif.name = brickName; /* brickName will be freed by the notifWrapper in anaLib0_so*/ 
	
	minmex_notifyEvent(0, &notif );
	
	
	anaPrint(ANA_NOTICE, "Successfully detached brick, current number of bricks: %d\n",BT.nbBricks);		
}



/* triggered at reception of a HeartBeat request from brick => sets the age of the brick to 0  */
void handleHBRequest(anaHandle_t handle){ 		
	
	struct BTentry_s *entry=NULL;

	entry = getBTEntry(&BT, handle);
	if(entry==NULL){
		anaPrint(ANA_DEBUG, "didn't find an entry for %8x\n", handle);
		return;
	}

	entry->age = 0;
	return;
}

int  minmex_setIDPinfo(anaHandle_t querier, struct IDPinfo_s *info){
	
	struct IDTentry_s *entry = NULL;
	int index;
	anaHandle_t callee;
	struct IDTaux_s *tmp = NULL;
	
	if(info==NULL || info->IDPtype != BRICK) /* it is forbidden to set an info on redirected and forked IDPs*/
		return -1;

	/* we need to find the entry with the right callee*/
	callee = (info->binding).brick.handle;
	
	index = label2hash(&IDT, &(info->label));
	
	entry = IDT.tab[index];
	while(entry!=NULL){
		if(entry->label == (anaLabel_t)(info->label) && entry->callBackFct == (minmexCallback_t)&sendData){
			tmp = (struct IDTaux_s *)entry->aux ;
			if(callee == tmp->callee){
				/* we need to check authorization */
				if(querier == entry->owner ){
					if(info->ICflag){
						int valueLen = (info->destContext).valueLen;
						entry->ICflag = 1;
						entry->destContext.value = (void *) malloc( valueLen );
						memcpy(entry->destContext.value, (info->destContext).value, valueLen );
						entry->destContext.valueLen = valueLen;
					
						valueLen = (info->destService).valueLen;
						entry->destService.value = (void *) malloc( valueLen );
						memcpy(entry->destService.value, (info->destService).value, valueLen );
						entry->destService.valueLen = valueLen;					
				
						entry->MTU = info->MTU;
						entry->chanType = info->chanType;
					}
					else
						entry->ICflag = 0;
					
					entry->nextIDP = (info->binding).brick.outputIDP;
					return 0;
				}
				else
					return ANA_UNAUTHORIZED_ERROR;
			}
		}
		entry = entry->next;
	}
	return ANA_NODESTIDTENTRY_ERROR;
}

int  minmex_getIDPinfo(anaHandle_t querier, anaHandle_t viewer, anaLabel_t label, struct IDPinfo_s* toBeFilled){
/*	int allowed = 1; */
	
	struct IDTentry_s *entry = NULL;
	
	if(toBeFilled==NULL)
		return -1;
	
/*	if(!allowed)
		return UNATUHORIZED */

	/* we get the IDP in strict view */
	entry =	getIDTEntryStrictView(&IDT, viewer, label);
	if(entry == NULL)
		return ANA_NODESTIDTENTRY_ERROR;
	
	toBeFilled->label = label;
	toBeFilled->viewer = viewer;
	toBeFilled->owner = entry->owner;	
	
	if(entry->age ==-1)
		toBeFilled->permFlag = 1;
	else
		toBeFilled->permFlag = 0;
	
	if(entry->ICflag) {
		int valueLen = entry->destContext.valueLen;
		toBeFilled->ICflag = 1;
		
		toBeFilled->destContext.value = (void *) malloc(valueLen);
		memcpy(toBeFilled->destContext.value, entry->destContext.value, valueLen);
		toBeFilled->destContext.valueLen = valueLen;

		valueLen = entry->destService.valueLen;
		toBeFilled->destService.value = (void *) malloc(valueLen);
		memcpy(toBeFilled->destService.value, entry->destService.value, valueLen);
		toBeFilled->destService.valueLen = valueLen;
		
		toBeFilled->MTU = entry->MTU;
		toBeFilled->chanType = entry->chanType;
	
	}
	else
		toBeFilled->ICflag = 0;
	
	
	if(entry->callBackFct == (minmexCallback_t)&sendData){
		struct IDTaux_s *tmp;
		toBeFilled->IDPtype = BRICK;
		tmp = (struct IDTaux_s *)entry->aux ;
		(toBeFilled->binding).brick.handle = tmp->callee;
		(toBeFilled->binding).brick.outputIDP = entry->nextIDP;
					
	}
	if(entry->callBackFct == (minmexCallback_t)&redirectData){
		struct redirectAux_s *tmp = (struct redirectAux_s *)(entry->aux);
		struct nextIDP_s *tmpN  = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));
		
		toBeFilled->IDPtype = REDIRECT;
		
		tmpN->label = tmp->label;
		tmpN->viewer = tmp->handle;
		tmpN->next = NULL;
		
		(toBeFilled->binding).IDPs = tmpN;
			
	}
	if(entry->callBackFct == (minmexCallback_t)&forkData){	
		int *nbForks = (int *)entry->aux;
		int i;
		struct redirectAux_s *tmps = (struct redirectAux_s *) ((entry->aux)+sizeof(int));
		struct nextIDP_s *tmpN  = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));
		(toBeFilled->binding).IDPs = tmpN;
		for(i=0;i<*nbForks;i++){
			tmpN->label = tmps->label;
			tmpN->viewer = tmps->handle;
			
			if(i!=(*nbForks-1)){
				tmpN->next = (struct nextIDP_s *) malloc(sizeof(struct nextIDP_s));
				tmpN = tmpN->next;
				tmps = (struct redirectAux_s *) ((entry->aux)+sizeof(int) + (i+1)*sizeof(struct redirectAux_s));
			}
		}
		
		toBeFilled->IDPtype = FORK;
				
	}	
		
	return 0;
}


int checkIDTAges() {
	int i;
	struct IDTentry_s *entry;
	struct IDTentry_s *next;
	struct IDTentry_s *predecessor;
	struct IDTaux_s *idtAux = NULL;
	freeAuxFunction freeAux;
	struct anaEventNotif_s notif;

	for (i=0; i< (int) IDT.size; i++) {
		entry = IDT.tab[i];
		predecessor = NULL;
		while  (entry!=NULL) {
			next = entry->next;
			if(entry->age != -1) {  /* if not a permanent entry */
				entry->age = entry->age+1;
				if(entry->age > IDT.deadEntryLimit) { /* the Entry hasn't been used for too long => remove it*/
					
					/*
					WE NEED TO NOTIFY ABOUT THIS DELETION
					*/
					if(entry->callBackFct == (minmexCallback_t)&  sendData){
		
						idtAux = entry->aux;
						freeAux = idtAux->freeAux;
						if(idtAux->aux != NULL && idtAux->freeAux != NULL)
							freeAux(idtAux->aux);
						free(idtAux);
							
					}
					else
						free(entry->aux);
					
					decrementNBEntries(entry->owner, 0);
					
					if(predecessor!=NULL)
						predecessor->next = entry->next; 
					else 
						IDT.tab[i] =  next; 	
					
					/* we notify about this deletion */
					memset(&notif, 0, sizeof(struct anaEventNotif_s));
					notif.eventID = IDP_DELETED_EVENT;
					notif.label  = entry->label; 
					notif.viewer = entry->viewer;
					
					minmex_notifyEvent(0, &notif);
						
					free(entry);
					entry = next;
				}
				else {
					predecessor = entry;
					entry = next;
				}		
			}
			else {
				predecessor = entry;
				entry = next;
			}
		}
	}
	return 0;
}

int forkData(void *msg, int len, anaLabel_t input, void *dest){ 

	anaHandle_t viewer=0;
	anaLabel_t label=0;

	struct IDTentry_s *entry = NULL;
	int offset=0;
	int numfork=0;
	int *nbForks = (int *)dest;
	int destSize;
	int ret = 0;
	struct redirectAux_s *tmps = NULL;
	if(msg==NULL || len<0 || dest==NULL){
		anaPrint(ANA_ERR, "forkData: NULL arguments\n");
		return -1;
	}
	
	
	destSize = sizeof(int) + (*nbForks) * sizeof(struct redirectAux_s);
	
	offset+=sizeof(int);
	
	while(offset< destSize){
		tmps = (struct redirectAux_s *)(dest+offset);
		viewer = tmps->handle;
		label = tmps->label;

		entry  = getIDTEntryStrictView(&IDT, viewer, label);
		if(entry==NULL){ 
			if(numfork == 0){ /* this means that the original IDP, for which the fork was set, has disappeared => this is a major error*/
				return ANA_FORK_ERROR_1; /*  that the original IDP, for which the fork was set, has disappeared*/
			}
			
			anaPrint(ANA_ERR, "forktData : No entry mapped to label %d in IDT\n",label);
			ret = ANA_FORK_ERROR_2; /* at least one of the IDPs to which the data is forked does not exist anymore */	
		}
		else{
			anaPrint(ANA_DEBUG, "Forking data from label %d to label %d\n", input, label);
			if((entry->callBackFct) != NULL)    /*now we call the callback function fct*/
			{
				if (numfork==0)
					entry->callBackFct(msg, len, input, entry->aux, entry->auxLen); /*The first IDP to fork is the origingnal, we need to call its callback function with the right input label*/
				else
					entry->callBackFct(msg, len, label, entry->aux, entry->auxLen);	
			}
			else 
				anaPrint(ANA_ERR, "forkData : callBack function mapped to IDP %d is NULL \n", label);
             	}
		offset += sizeof(struct redirectAux_s);
		numfork++;
	}
	
	return ret;
}



int redirectData(void *msg, int len, anaLabel_t input, struct redirectAux_s *dest){  /* redirects the message to the IDP  In dest */

								
	struct IDTentry_s *entry = NULL;
		
	if(msg==NULL || len<0 || dest==NULL){
		anaPrint(ANA_ERR, "redirectData: NULL arguments\n");
		return -1;
	}
	

	entry  = getIDTEntry(&IDT, dest->handle, dest->label);
	anaPrint(ANA_DEBUG, "Redirecting label %d to label %d\n", input , dest->label);
		
	if(entry==NULL) {
		anaPrint(ANA_ERR, "redirectData : No entry mapped to label %d in IDT, removing this\n", dest->label);
		return -1;
	}
	
	if((entry->callBackFct) != NULL) {   /*now we call the callback function fct*/
		
		return entry->callBackFct(msg, len, dest->label, entry->aux, entry->auxLen);
	}
	else {
		anaPrint(ANA_ERR, "redirectData : callBack function mapped to IDP %d is NULL \n", dest->label);
          	return -1;
	}

	return 0;
}




int handleRedirect(anaHandle_t querier, anaHandle_t originViewer, anaLabel_t originLabel, anaHandle_t destViewer, anaLabel_t destLabel ) { 
	
	struct IDTentry_s *toAdd=NULL;
	
	struct IDTentry_s *originIdtEntry =NULL;
	struct IDTentry_s *publicOriginIdtEntry =NULL;
		
	struct IDTentry_s *destIdtEntry =NULL;
	struct BTentry_s *ctEntry=NULL;
	
	int perm = 0;
	
	struct redirectAux_s *tmpr = NULL;
	
/*	struct anaEventNotif_s notif; */
	
	/* First we should check the credentials of the querier */
			/* to come */
	
	/* then we check if origin == dest */
	if(originLabel == destLabel && originViewer == destViewer){
		return -2;  /* -2 means : redirection creates a loop*/
	}
	
	
	/* we check if the originViewer brick exists */
	if(originViewer != 0){
		ctEntry = getBTEntry(&BT, originViewer);
		if(ctEntry ==NULL){
			anaPrint(ANA_DEBUG, "origin FB not found\n");
			return ANA_NOORIGINVIEWER_ERROR; /*  origin viewer brick unknown*/	
		}
	}
	
	
	if(destViewer != 0){
		ctEntry = getBTEntry(&BT, destViewer);
		if(ctEntry ==NULL){
			anaPrint(ANA_DEBUG, "dest FB not found\n");
			return ANA_NODESTVIEWER_ERROR; /*  dest viewer brick unknown*/	
		}
	}
	
	
	/* we check if a public origin IDP exists, if not there is nothing to Redirect (i.e another function should be used) */
	publicOriginIdtEntry = getIDTEntryStrictView(&IDT, 0, originLabel);
	if(publicOriginIdtEntry==NULL)
		return ANA_REDIR_ERROR_1; /* there is no public IDP with this label => nothing to redirect */		
	
	
	
	/* we check if the origin IDP exists */
	if(originViewer==0)
		originIdtEntry = publicOriginIdtEntry;
	else
		originIdtEntry  = getIDTEntryStrictView(&IDT, originViewer, originLabel);
	
		
	/* we check if the dest IDP exists */
	destIdtEntry  = getIDTEntryStrictView(&IDT, destViewer, destLabel);
	
	if(destIdtEntry==NULL) {
		anaPrint(ANA_DEBUG, "dest IDP not found\n");
		return ANA_NODESTIDTENTRY_ERROR; /* no dest IDT entry with indicated viewer and label*/	
	}
	
	if(originIdtEntry != NULL){
		if(originIdtEntry->age == -1)
			perm = 1 ;
	}
	else
		if(publicOriginIdtEntry->age == -1)
			perm = 1 ;
	
	/* we check for fork + redirection loops */
	if(createsLoop(originViewer, originLabel, destViewer, destLabel))
		return -2;

	/* we check if the querier did not reach the maximum of nb Entries */
	if(incrementNBEntries(querier, perm)<0)
		return ANA_MAXNBENTRIES_ERROR; /*reached the max number of entries */
	
		
	if(originIdtEntry!=NULL){
		/* we test if the querier is the owner of the previous entry */
		if (originIdtEntry->owner == querier){
			/* in this case we will crash the existing entry ( we need to remove here and add it later)*/
			if(originIdtEntry->callBackFct == (minmexCallback_t)& sendData){
				struct IDTaux_s *idtAux =  originIdtEntry->aux;
				freeAuxFunction freeAux = idtAux->freeAux;
				freeAux(idtAux->aux);
			}
			freeIDTEntry(&IDT, originViewer, originLabel);
		}
		else
			return ANA_REDIR_ERROR_2; /* there already exists an IDP with originLabel and originViewer belonging to another owner => cannot crash it */		
	}
		
	
	toAdd = (struct IDTentry_s*) malloc(sizeof(struct IDTentry_s));
	
	toAdd->label = originLabel;
	toAdd->owner = querier;
	toAdd->viewer = originViewer;

	toAdd->callBackFct = (minmexCallback_t)& redirectData;
		
	tmpr = (struct redirectAux_s *) malloc(sizeof(struct redirectAux_s));
	tmpr->handle = destViewer;
	tmpr->label  = destLabel;
	
	toAdd->aux = tmpr;
	toAdd->auxLen = sizeof(struct redirectAux_s);
	
	toAdd->next = NULL;

	if (addIDTEntry(&IDT, toAdd, perm)<0) {
		anaPrint(ANA_ERR, "handleRedirect : Unable to add entry owner=%d viewer=%d label=%d  to IDT\n", 
			querier, originViewer, originLabel);
		free(toAdd->aux);
		free(toAdd);
		
		/* we decrement the number we incremented few lines before */
		decrementNBEntries(querier, perm);
		return -1; 
	}
	

	/* we notify the interested bricks in case there are some subscribers*/
	
/*	memset(&notif, 0, sizeof(struct anaEventNotif_s));
	notif.eventID = IDP_REDIRECTED_EVENT;
	notif.label = originLabel;
	notif.viewer = originViewer;
	
	notif.destLabel = destLabel;
	notif.destViewer = destViewer;
	minmex_notifyEvent(0, &notif );

*/	
	return 0;
}




int  handleForkData(anaHandle_t querier, anaHandle_t originViewer, anaLabel_t originLabel, anaHandle_t destViewer, anaLabel_t destLabel) { 

	struct IDTentry_s *toAdd;
	struct IDTentry_s *originIdtEntry; 	
	struct IDTentry_s *destIdtEntry; 
	struct BTentry_s *ctEntry=NULL;
	/* first we check if origin == dest */
	if(originLabel  == destLabel &&  originViewer == destViewer){
		return -2; /* -2 means : this fork creates a data forwarding loop */
	}
		
	/* First we should check the credentials of the querier */
			/* to come */
	
	
	/* we check if the originViewer brick exists */
	if(originViewer != 0){
		ctEntry = getBTEntry(&BT, originViewer);
		if(ctEntry ==NULL){
			return ANA_NOORIGINVIEWER_ERROR; /*  origin viewer brick unknown*/	
		}
	}
		
	/* we check if the destViewer brick exists */
	if(destViewer != 0){
		ctEntry = getBTEntry(&BT, destViewer);
		if(ctEntry ==NULL){
			return ANA_NODESTVIEWER_ERROR; /* dest viewer brick unknown*/	
		}
	}
		
	/* we check if the origin IDP exists */
	originIdtEntry  = getIDTEntryStrictView(&IDT, originViewer, originLabel);
		
	if(originIdtEntry ==NULL)
		return ANA_NOORIGINIDTENTRY_ERROR; /*  there exists no IDT entry with originViewer and originLabel to fork */		
		
		
	/* we check if the dest IDP exists */
	destIdtEntry  = getIDTEntryStrictView(&IDT, destViewer, destLabel);
	
	if(destIdtEntry==NULL) 
		return ANA_NODESTIDTENTRY_ERROR; /*  there exists no dest IDT entry with given viewer and label */		
	
	/* we check for fork (redirection) loops */
	if(createsLoop(originViewer, originLabel, destViewer, destLabel))
		return -2;
	
	if(originIdtEntry->callBackFct==(minmexCallback_t) & forkData){
	/* in this case we extend the list of IDPs to fork to*/
		int *nbForks = (int *)originIdtEntry->aux;
		int *nbForks2;
		struct redirectAux_s *tmps = NULL;
		int offset = sizeof(int) + (*nbForks) * sizeof(struct redirectAux_s);
		
		void *aux2 = malloc(offset + sizeof(struct redirectAux_s));
		memcpy(aux2, originIdtEntry->aux, offset);
		
		/* we increment the number of forks */
		nbForks2  = (int*)aux2;
		*nbForks2 = *nbForks+1;
		anaPrint(ANA_DEBUG, "*nbForks %i\n", *nbForks2);	
		free(originIdtEntry->aux);
		
		tmps = (struct redirectAux_s *)(aux2 + offset);
		tmps->handle = destViewer;
		tmps->label = destLabel;
						
		originIdtEntry->aux = aux2;
		originIdtEntry->auxLen = offset + sizeof(struct redirectAux_s);
	}
	else { 
		/* in this case we create a new IDP with the callback function of origin
 		   and replace the one of origin with forkData	
		*/

		/* < construct IDT entry   > */
		int perm =0;
		int *intP=NULL;
		struct redirectAux_s *tmps = NULL;
		
		
		if(originIdtEntry->age == -1)
			perm = 1;
		else
			perm = 0;
		
		
		/* we check if we did not reach the limit of authorized entries for the querier*/
		if(incrementNBEntries(querier,perm)<0)
			return ANA_MAXNBENTRIES_ERROR; /*  reached max number of IDT entries*/
				
		
		toAdd = (struct IDTentry_s*) malloc(sizeof(struct IDTentry_s));
	
		toAdd->label =  generateLabel();
	
		toAdd->owner = originIdtEntry->owner;
		toAdd->viewer = originIdtEntry->viewer;

		toAdd->callBackFct = originIdtEntry->callBackFct;
		
		toAdd->aux = originIdtEntry->aux;
		toAdd->auxLen = originIdtEntry->auxLen;
		
		toAdd->next = NULL;
	
	
					
		/* < End of construct IDT entry > */
		/* < adding entry > */ 
		if (addIDTEntry(&IDT, toAdd,perm)<0) {
			anaPrint(ANA_ERR, "handleFork : Unable to add new entry \n");
			
			/* we decrement the number of entries we just incremented a few lines above */
			decrementNBEntries(querier, perm);
			return -1; 
		}
		
	
		/*now we change the callback and aux of the destIDtEntry */		
		
		originIdtEntry->owner = querier;
		originIdtEntry->callBackFct = (minmexCallback_t)& forkData;
		
		originIdtEntry->aux = (void*) malloc( sizeof(int)+ 2 * sizeof(struct redirectAux_s));
		intP = (int *)originIdtEntry->aux; /*we set the number of forks first in the aux field*/
		*intP = 2; /* 2 as we have only 2 forks for now */ 
		
		tmps = (struct redirectAux_s *)(originIdtEntry->aux + sizeof(int));
		tmps->handle = toAdd->viewer;
		tmps->label = toAdd->label;
		
		tmps = (struct redirectAux_s *)(originIdtEntry->aux + sizeof(int) + sizeof(struct redirectAux_s));
		tmps->handle = destViewer;
		tmps->label = destLabel;

	}


        return 0;
}

int handleRegister (anaHandle_t querier, anaCallback_t fct, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux, anaHandle_t viewer, int perm, int thread, anaLabel_t *result){
	struct BTentry_s *entry=NULL;
	struct IDTentry_s *toAdd=NULL;
	anaLabel_t label;
	struct IDTaux_s *idtAux = NULL;
	if(result == NULL)
		return -1;
	
		/* check the viewer */
	if(viewer!=0){
		entry = getBTEntry(&BT, viewer);
		if(entry == NULL){
			anaPrint(ANA_ERR, "handleRegister: Unknown viewer brick \n");
			return -1;
		}
	}
	/*  check querier existence  */	
	entry = getBTEntry(&BT, querier);
	if(entry == NULL){
		anaPrint(ANA_ERR, "handleRegister: Unknown querier brick %d\n", querier);
		return -1;
	}
	/* we check the number of permanent and volatile entries */
	anaPrint(ANA_DEBUG, "brick %s: perm entries: %i, vol entries %i\n", entry->brickName, entry->nbPermanentEntries, entry->nbVolatileEntries);
	if(perm)
		if(entry->nbPermanentEntries == MAX_NB_PERMANENT)
			return ANA_MAXNBENTRIES_ERROR; /* reached maximum allowed number of permanent entries */	
		else
			entry->nbPermanentEntries++;	
	else
		if(entry->nbVolatileEntries == MAX_NB_VOLATILE)
			return ANA_MAXNBENTRIES_ERROR; /* reached maximum allowed number of volatile entries */
		else
			entry->nbVolatileEntries++;
	
	label = generateLabel();

	toAdd = (struct IDTentry_s*) malloc(sizeof(struct IDTentry_s));
	if(toAdd==NULL){
		anaPrint(ANA_ERR, "handleRegister : Unable to allocate memory for new entry \n");
		return -1;
	}
	
	toAdd->label = label ;
	toAdd->owner = querier;
	toAdd->viewer = viewer;
	toAdd->callBackFct = (minmexCallback_t)& sendData;
		
	idtAux = (struct IDTaux_s*) malloc(sizeof(struct IDTaux_s));
	idtAux->callee = querier;
	idtAux->callBackFct = fct;
	if(aux != NULL){
		idtAux->aux = copyAux(aux);
		idtAux->copyAux = copyAux;
		idtAux->freeAux = freeAux;
	}
	idtAux->thread = thread;
	
	toAdd->aux = idtAux;
	
	if (addIDTEntry(&IDT, toAdd, perm)<0) {
		anaPrint(ANA_ERR, "handleRegister : Unable to add new entry to IDT\n");
		freeAux(idtAux->aux);
		free(idtAux);
		free(toAdd);
		return -1; 
	}
	
	*result = label;
	
	return 0;
}


int handleChangeLabelAux (anaHandle_t querier, anaLabel_t label, void *newAux, copyAuxFunction copyAux, freeAuxFunction freeAux) {

	struct IDTentry_s *entry = NULL;
	struct IDTaux_s *idtAux = NULL;
	freeAuxFunction freeaux;
	int allowed = 0;
	int index = 0;

	if(newAux != NULL && (copyAux == NULL || freeAux == NULL))
		return -1;
		
	index = label2hash(&IDT, (anaLabel_t *)&label);
		
	entry = IDT.tab[index];
	while(entry!=NULL){
		if(entry->label == label && entry->callBackFct == (minmexCallback_t)sendData){
			
			/* we check if the querier is the owner of the entry */
			if(!allowed)
				if(querier != entry->owner)
					return ANA_UNAUTHORIZED_ERROR; /*  Must be the owner of the IDP in order to change the auxiliary argument */
	
			idtAux = entry->aux;
			freeaux = idtAux->freeAux;
			freeaux(idtAux->aux);
	
			idtAux->copyAux = copyAux;
			idtAux->freeAux = freeAux;
			idtAux->aux = copyAux(newAux);
			return 0;
		}
		entry = entry->next;		
	}
		
	/* if we arrive here means we didn't find the entry */
	return 	ANA_NODESTIDTENTRY_ERROR;
}	 


int handleUnregister(anaHandle_t querier, anaHandle_t viewer, anaLabel_t label) {
	
	int allowed = 0;
	struct IDTentry_s *entry=NULL;
	int perm;
	struct IDTaux_s *idtAux = NULL;
	freeAuxFunction freeAux;
	struct anaEventNotif_s notif;
	
	entry = getIDTEntryStrictView(&IDT, viewer, label);
	if(entry==NULL)
		return -1;
	
	
	/* we check the credentials here */
		/* to come */
	

	if(entry->age == -1)
		perm =1;
	else
		perm =0;
	
	
		
	if(entry->callBackFct == (minmexCallback_t)& redirectData)
		return ANA_UNREGISTER_ERROR; /* This IDP is being redirected towards another  */
	
	if(entry->callBackFct == (minmexCallback_t)& forkData){
		struct redirectAux_s *tmps = (struct redirectAux_s *)entry->aux + sizeof(int);
		struct IDTentry_s *originalEntry = NULL;
		int ret=0;
		originalEntry = getIDTEntryStrictView(&IDT, tmps->handle, tmps->label);
		if(originalEntry == NULL){
			return ANA_FORK_ERROR_1; /*this IDP label corresponds to a fork point and the original IDP does not exist anymore */
		}
		
	 	ret = handleUnregister(querier, tmps->handle, tmps->label);
		
		if (ret >= 0){
			/* we removed the original entry, now we can remove this fork point entry */
			decrementNBEntries(entry->owner, perm);
			if(freeIDTEntry(&IDT, viewer, label)<0){
				incrementNBEntries(entry->owner, perm);
				return ANA_UNREGISTER_ERROR_2; /* removed original IDP entry but failed to remove the fork point IDP leading towards it */
			}
			
			/* we notify the interested bricks in case there are some subscribers*/
	
			memset(&notif, 0, sizeof(struct anaEventNotif_s));
			notif.eventID = IDP_DELETED_EVENT;
			notif.label = label;
			notif.viewer = viewer;
	
			minmex_notifyEvent(0, &notif );
			return 0;
		}
		return ret;
	}
		
	if(entry->callBackFct == (minmexCallback_t)&  sendData){
		
		if(!allowed){ /*we check if the querier is the owner of the IDP */
			if(entry->owner != querier)
				return ANA_UNAUTHORIZED_ERROR; /* unauthorized to delete this Entry */	
		}
	
		
		idtAux = entry->aux;
		freeAux = idtAux->freeAux;
		if(idtAux->aux != NULL && idtAux->freeAux != NULL)
			freeAux(idtAux->aux);	
	
		decrementNBEntries(entry->owner, perm);	
	
		if(freeIDTEntry(&IDT, viewer, label)<0){
			incrementNBEntries(entry->owner, perm);
			return -1;
		}
		
		/* we notify the interested bricks in case there are some subscribers*/
	
		memset(&notif, 0, sizeof(struct anaEventNotif_s));
		notif.eventID = IDP_DELETED_EVENT;
		notif.label = label;
		notif.viewer = viewer;
	
		minmex_notifyEvent(0, &notif );
		return 0;
	}
	
	
	
	
	return -1;
}


int handleChangeView(anaHandle_t querier, anaHandle_t oldViewer, anaLabel_t label, anaHandle_t newViewer) {	

	struct IDTentry_s *entryOld = NULL;
	struct BTentry_s *btEntry=NULL;
	
	int allowed = 0 ;
	

	/* we check the existense of the new viewer */
	if(newViewer!=0)
		btEntry = getBTEntry(&BT, newViewer);
	if(btEntry == NULL){
		return ANA_NODESTVIEWER_ERROR; /* no known brick with the new viewer*/
	}
	entryOld = getIDTEntryStrictView(&IDT, oldViewer, label);
	if(entryOld ==NULL)
		return ANA_NODESTIDTENTRY_ERROR; /* entry deos not exist */
		
		
	/* we check the credentials of the querier to do so */
		/* to come */
	
	if(!allowed){
		if(entryOld->owner != querier)
			/* we return an error code to brick */
			return ANA_UNAUTHORIZED_ERROR; /* not authorized and not the owner of the IDP */
	}

	/*now we do the change */
	entryOld->viewer = newViewer;
	
	return 0;
}

int dispatchData(anaHandle_t querier, anaLabel_t label, anaHandle_t viewer, void *data, int len){
        
	struct BTentry_s *BTentry = NULL;
	struct IDTentry_s *entry = NULL;
	if(data == NULL || label == 0)
		return -1;
	
	/* check the querier existence */
       	anaPrint(ANA_ERR, "looking for label %i\n", label); 
        BTentry = getBTEntry(&BT, querier);
        if(BTentry == NULL){
        	return ANA_UNKNOWNHANDLE_ERROR; /* unknown brick handle*/      
        }
        entry  = getIDTEntry(&IDT, viewer, label);
        if(entry==NULL) 
             return ANA_NODESTIDTENTRY_ERROR; /* no entry matching this viewer and label */
	     
	  
	if((entry->callBackFct) != 0) {
          	
		return entry->callBackFct(data, len, label, entry->aux, entry->auxLen);
        }
	
        /* else */
	anaPrint(ANA_ERR, "listenLoop: callBack function mapped to IDP %d is NULL \n", label);
	return -1;     

}


/* this function is meant to be used by the nodeCompartment */
int sendToIDP(void *msg, int len, anaLabel_t label){
	struct IDTentry_s *entry=NULL;
	
	if(msg==NULL || len<0)  
		return -1;
	
	entry = getIDTEntry(&IDT, 0, label);
	if(entry==NULL || entry->callBackFct == NULL) {
		anaPrint(ANA_ERR, "sendToIDP: entry not found\n");
		return -1;
	}
	
	return entry->callBackFct(msg, len, label, entry->aux, entry->auxLen);
}












