/* C/minmex/common/nodeComp.c
 * 
 * functions for the node compartment
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include "../include/nodeComp.h"

extern IDT_t IDT;
xrpSpecs_t defaultXrpSpecs;   /*  defines the default xrp specifics used by AL1 */
struct KVR kValRep;

int initNodeComp(int hashMaskEntryTab, int hashMaskKeyTab, int deadKVREntryLimit) {

	/* first we intialize the default XRP Specs */
	defaultXrpSpecs.maxMsgSize = DEFAULT_XRP_MSGSIZE ;	 		
	defaultXrpSpecs.commandSize = DEFAULT_XRP_CMDSIZE; 		
	defaultXrpSpecs.classSize = DEFAULT_XRP_CLASSSIZE; 
	defaultXrpSpecs.xrpSpecVersion = (char *)malloc(strlen(DEFAULT_XRP_VERSION)+1);
	strcpy(defaultXrpSpecs.xrpSpecVersion, DEFAULT_XRP_VERSION); 
	
	/* we then initialize the Key Val Repository s*/
	if(initializeKVR(&kValRep, hashMaskEntryTab, hashMaskKeyTab, deadKVREntryLimit) < 0){
		anaPrint(ANA_ERR, "initNodeComp : unable to initialize KVR\n");
		return -1;
	}		
	return 0;	
}


void sendMulticast(void *msg, int len, anaLabel_t input, void *query, int queryLen){
	/* we look in the KVR everything matching a description and we forward data to it */

	struct KVRKeyObjectList *queryRes=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	int ret = evalQuery((char *)query, &queryRes);
	
		
	anaPrint(ANA_DEBUG, " **** Multicast Function\n");
	
	if(ret!=0 || queryRes==NULL)
		return;

	tmp=queryRes;
	while(tmp!=NULL){
		anaPrint(ANA_DEBUG, " *** multicasting to IDP %d\n", tmp->entry->IDP);
		sendToIDP(msg, len, tmp->entry->IDP);
				
		tmp=tmp->next;		
	}
	
	
	/* in this case the result returned eval() is a copy => we can free it */

	while(queryRes!=NULL){
		tmp = queryRes->next;
		free(queryRes);
		queryRes=tmp;
	}
}	


void handleResolve(void *msg, int len, anaLabel_t input, void *aux, int auxLen) {
	
	int nameLen, queryLen, contextLen, labelSize, totalSize, arglen, dummy;
	char *query, *context, *chanType;
	
	anaLabel_t *replyTo=NULL;
	anaLabel_t result;
	xrpMsg_t replyMsg=NULL;
	struct KVRKeyObjectList *queryRes=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	struct KVRKeyObjectList *next=NULL;
	int ret=0;
	
	if(msg==NULL || len<0)
		return;
	
	
	replyTo = (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &labelSize,0);
	if(replyTo == NULL ) { /* if no label to reply to it's pointless to continue*/
		anaPrint(ANA_ERR, "handleResolve : No label to reply to \n");
		return;
	}
	

	context = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_CXT, &contextLen, 0);
	query = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_SRV, &queryLen, 0);
	chanType = (char *) getXRPArg((char *)msg, XRP_CLASS_CHANTYPE, &dummy,0);
	
	/* for the moment the node compartment accepts by default the contexts '.' or '*' */
	if (context)
	{
		if (context[0] != '.' && context[0] != '*')
			contextLen = 0; 	
	}
	anaPrint(ANA_DEBUG, "handleResolve :received resolve with query:%s\n", query);
	ret = evalQuery(query, &queryRes);
	if(ret!=0 || queryRes==NULL || contextLen==0){
		/* we send an error message */
		int ret;
		ret = fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
		
		if (ret < 0) {
			anaPrint(ANA_ERR, "handleResolve : Unable to fill ERROR command in message\n");
			return;
		}	
			
		sendToIDP(msg, len, *replyTo);
		return;
	}
	
	if(allocateXRPMsg(&replyMsg)<0) {
		anaPrint(ANA_ERR, "handleResolve : Unable to allocate memory for reply message\n");
		
		/* we send an error message */
		fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
		sendToIDP(msg, len, *replyTo);
		/* we free the query result */
		tmp = queryRes;
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp);
			tmp=next;
		}
		return ;
	}
	
	totalSize = fillXRPCommand(replyMsg, XRP_CMD_RESOLVE);
	
	if (totalSize < 0) {
		anaPrint(ANA_ERR, "handleResolve : Unable to fill command in message\n");
		/* we send an error message */
		fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
		sendToIDP(msg, len, *replyTo);
		/* we free the query result */
		tmp = queryRes;
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp);
			tmp=next;
		}
		
		free(replyMsg);
		return;
	}
	
	if(chanType[0]=='u' || chanType[0]=='a') /*if anycast or unicast, the first answer is enough */{
		if(queryRes->entry->name != NULL) {			
			
			nameLen = strlen(queryRes->entry->name)+1;
			arglen = addXRPArg(replyMsg, XRP_CLASS_NAME, queryRes->entry->name, nameLen);	
			
			if (arglen < 0) {
				anaPrint(ANA_ERR, "handleResolve : Unable to add Name to message\n");
				
				/* we send an error message */
				fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
				sendToIDP(msg, len, *replyTo);
				
				/* we free the query result */
				tmp = queryRes;
				while(tmp!=NULL){
					next = tmp->next;
					free(tmp);
					tmp=next;
				}
				
				free(replyMsg);
				return ;
			}
			
			totalSize += arglen;
			
			result = queryRes->entry->IDP;
			arglen = addXRPArg(replyMsg, XRP_CLASS_LABEL, &result, sizeof(anaLabel_t));
					
			if (arglen < 0) {
				anaPrint(ANA_ERR, "handleResolve : Unable to add label to message\n");
				
				/* we send an error message */
				fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
				sendToIDP(msg, len, *replyTo);
				/* we free the query result */
				tmp = queryRes;
				while(tmp!=NULL){
					next = tmp->next;
					free(tmp);
					tmp=next;
				}
				
				free(replyMsg);
				return ;
			}
			
			totalSize += arglen;
			
					
		}
		else{ /*we send error back*/
			int ret = fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); /* should crush the previous cmd*/
		
			if (ret < 0) {
				anaPrint(ANA_ERR, "handleResolve : Unable to fill ERROR command in message\n");
				/* we free the query result */
				tmp = queryRes;
				while(tmp!=NULL){
					next = tmp->next;
					free(tmp);
					tmp=next;
				}
				
				free(replyMsg);
				return;
			}	
						
			sendToIDP(msg, len, *replyTo);
			free(replyMsg);	
			return;
		}
	}
/*	else { *//*experimental, creating a multicast IDP here*//*
		anaLabel_t tmpLabel;
		if(createIDP(& sendMulticast, &tmpLabel, query, queryLen, (anaHandle_t)aux, (anaHandle_t)aux, 0)<0){  *//*the IDP is in the view of the querier only*//*
			anaPrint(ANA_ERR, "handleResolve : Unable to create Multicast IDP\n");
			*//* we send an error message *//*
			fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
			sendToIDP(msg, len, replyTo, NULL);
			*//* we free the query result *//*
			tmp = queryRes;
			while(tmp!=NULL){
				next = tmp->next;
				free(tmp);
				tmp=next;
			}
				
			free(replyMsg);
			return ;
		}
		

		
		arglen = addXRPArg(replyMsg, XRP_CLASS_NAME, query, queryLen);	
		if (arglen < 0) {
			anaPrint(ANA_ERR, "handleResolve : Unable to add query to reply message\n");
			*//* we send an error message *//*
			fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
			sendToIDP(msg, len, replyTo, NULL);
			*//* we free the query result *//*
			tmp = queryRes;
			while(tmp!=NULL){
				next = tmp->next;
				free(tmp);
				tmp=next;
			}

			free(replyMsg);
			free(tmpLabel);
			return ;
		}
		
		totalSize+=arglen;
		
		arglen = addXRPArg(replyMsg, XRP_CLASS_LABEL, tmpLabel, labelSize);	
		if (arglen < 0) {
			anaPrint(ANA_ERR, "handleResolve : Unable to add query to reply message\n");
			*//* we send an error message *//*
			fillXRPCommand ((xrpMsg_t)msg, XRP_CMD_ERROR); 
			sendToIDP(msg, len, replyTo, NULL);
			*//* we free the query result *//*
			tmp = queryRes;
			while(tmp!=NULL){
				next = tmp->next;
				free(tmp);
				tmp=next;
			}
				
			free(replyMsg);
			free(tmpLabel);
			return ;
		}
		totalSize+=arglen;
		
		free(tmpLabel);
	}
*/		
	/* we send the message containing names and IDPs */
	sendToIDP(replyMsg, totalSize, *replyTo);
	
	/* in this case the result returned eval() is a copy => we can free it */
	tmp = queryRes;
	while(tmp!=NULL){
		next = tmp->next;
		free(tmp);
		tmp=next;
	}
	
	free(replyMsg);
	anaPrint(ANA_DEBUG, "Resolve succeeded\n");
	return;
}

void handleLookup(void *msg, int len, anaLabel_t input, void *aux, int auxLen) {
	
	int nameLen, queryLen, contextLen, labelSize, totalSize, arglen;
	char *query, *context;
	anaLabel_t *replyTo=NULL;

	xrpMsg_t replyMsg=NULL;
	struct KVRKeyObjectList *queryRes=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	struct KVRKeyObjectList *next=NULL;
	int ret=0;
	
	if(msg==NULL || len<0)
		return;
	
	replyTo = (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &labelSize,0);

	if(replyTo == NULL ) { /* if no label to reply to it's pointless to continue*/
		anaPrint(ANA_ERR, "handleLookup : No label to reply to \n");
		return;
	}
	
	
	context = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_CXT, &contextLen, 0);
	query = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_DST_SRV, &queryLen, 0);
	
	/* for the moment the node compartment accepts by default the contexts '.' or '*' */
	
	if (context)
	{
		if (context[0] != '.' && context[0] != '*')
			contextLen = 0; 	
	}
	
	anaPrint(ANA_DEBUG, "Received Lookup with query: %s \n", query);	
	
	ret = evalQuery(query, &queryRes);
	if(ret!=0 || queryRes==NULL || contextLen==0){
		/* we send an error message */
		fillXRPCommand ((char *)msg, XRP_CMD_ERROR); /* should crush the previous cmd*/
		sendToIDP(msg, len, *replyTo);
		return;
	}
	
	
	if(allocateXRPMsg(&replyMsg)<0) {
		anaPrint(ANA_ERR, "handleLookup : Unable to allocate memory for reply message\n");
		/* we send an error message */
		fillXRPCommand ((char *)msg, XRP_CMD_ERROR); /* should crush the previous cmd*/
		sendToIDP(msg, len, *replyTo);	
		
		/* we also free the query result (thx Sylvain ;) ) */
		tmp = queryRes;
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp);
			tmp=next;
		}
		return ;
	}
	
	totalSize = fillXRPCommand(replyMsg, XRP_CMD_RESOLVE);
	
	if (totalSize < 0) {
		anaPrint(ANA_ERR, "handleLookup : Unable to fill command in message\n");
		/* we send an error message */
		fillXRPCommand ((char *)msg, XRP_CMD_ERROR); /* should crush the previous cmd*/
		sendToIDP(msg, len, *replyTo);
		free(replyMsg);
		/* we also free the query result (thx Sylvain ;) ) */
		tmp = queryRes;
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp);
			tmp=next;
		}
		
		return;
	}
		
	tmp=queryRes;
	while(tmp!=NULL){
		if(tmp->entry->name != NULL ) {			
			
			nameLen = strlen(tmp->entry->name)+1;
			arglen = addXRPArg(replyMsg, XRP_CLASS_LKP_DESCR, tmp->entry->name, nameLen);	
			
			if (arglen < 0) {
				anaPrint(ANA_ERR, "handleLookup : Unable to add Name to message\n");
				/* we send an error message */
				fillXRPCommand ((char *)msg, XRP_CMD_ERROR); 
				sendToIDP(msg, len, *replyTo);
				free(replyMsg);
				
				/* we also free the query result (thx Sylvain ;) ) */
				tmp = queryRes;
				while(tmp!=NULL){
					next = tmp->next;
					free(tmp);
					tmp=next;
				}
				return ;
			}
	
			totalSize += arglen;
			arglen = addXRPArg(replyMsg, XRP_CLASS_LABEL, &(tmp->entry->IDP), sizeof(anaLabel_t));
					
			if (arglen < 0) {
				anaPrint(ANA_ERR, "handleLookup : Unable to add label to message\n");
								
				/* we send an error message */
				fillXRPCommand ((char *)msg, XRP_CMD_ERROR); /* should crush the previous cmd*/
				
				sendToIDP(msg, len, *replyTo);
				free(replyMsg);
				
				/* we also free the query result */
				tmp = queryRes;
				while(tmp!=NULL){
					next = tmp->next;
					free(tmp);
					tmp=next;
				}
				return ;
			}

			totalSize += arglen;
			tmp=tmp->next;		
		}
	}
	
	/* we send the message containing names and IDPs */
	sendToIDP(replyMsg, totalSize, *replyTo);
	
	/* in this case the result returned eval() is a copy => we can free it */
	tmp = queryRes;
	while(tmp!=NULL){
		next = tmp->next;
		free(tmp);
		tmp=next;
	}
	
	free(replyMsg);
	anaPrint(ANA_DEBUG, "Lookup succeeded\n");
	return;
}


/* the publish query contains the name of the entry to publish, a label, and key words   */
void handlePublish(void *msg, int len, anaLabel_t input, void *aux, int auxLen) {
	anaLabel_t *IDP = NULL;
	
	anaLabel_t *replyTo=NULL;
	anaHandle_t owner;
	int nameLen,labelLen, descLen, dummy;
	char *name=NULL;
	char *description=NULL;
	int permanent;
	char *permVal = NULL;
	struct KVREntry *entry=NULL;
	struct IDTentry_s *idtEntry=NULL;
	char *keyWord=NULL;

	permanent  = 0;
	if(msg==NULL || len< 0)
		return ;
	
	anaPrint(ANA_DEBUG, "received publish\n");			
	
	IDP =  (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &labelLen, 0);

	owner = *((anaHandle_t *) aux);
	
	description = (char*) getXRPArg((char *)msg, XRP_CLASS_SRC_SRV, &descLen, 0);
		
	if( description==NULL || IDP==NULL ){
		anaPrint(ANA_ERR, "handlePublish: unable to identify name or IDP\n");
		return;
	}

	/* the replyTo label is optional, we look for an argument with class label
	at the second occurence => the first label is the one associated to the entry
	*/
	replyTo = (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &dummy,1);
	
	permVal = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_PERMANENT, &dummy,0);
	
	if(permVal != NULL ) {		
		permanent = 1;
	}

	/*first we check if there is an Entry in IDT with the IDP to publish */
	idtEntry = getIDTEntry(&IDT, 0, *IDP);
	if(idtEntry==NULL){
		if(replyTo!=NULL){
			/* to send an error we just put the XRP_ERROR command in front of the recevied message*/
			fillXRPCommand((char *)msg, XRP_CMD_ERROR);
			sendToIDP(msg, len, *replyTo);
		}
		return;
	}
	
	name=strsep(&description,"+");		
	if(name==NULL){
		if(replyTo!=NULL){
			/* to send an error we just put the XRP_ERROR command in front of the recevied message*/
			fillXRPCommand((char *)msg, XRP_CMD_ERROR);
			sendToIDP(msg, len, *replyTo);
		}
		return;
	}
	
	nameLen = strlen(name);
	
	entry = allocateKVREntry(name, *IDP, owner);
	if(entry==NULL){ 
		/*we send an error message, if a replyTo label was specified*/
		if(replyTo!=NULL){
			/* to send an error we just put the XRP_ERROR command in front of the recevied message*/
			fillXRPCommand((char *)msg, XRP_CMD_ERROR);
			sendToIDP(msg, len, *replyTo);
		}
		return;
	}
	if(permanent)
		entry->age = -1;
	else 
		entry->age = 0;
	
	/* now we add the keys to the entry */
/*	nbKeys =  getXRPnbArgs(msg) - 2;  */ /*nbArguments - ( name argument + returnTo label) */
	
	keyWord=strsep(&description,"+");
	
	while(keyWord!=NULL){
		addKeywordToEntry(entry, keyWord);
		keyWord=strsep(&description,"+");
	}
	
	/* we add the name to the keys */
	addKeywordToEntry(entry, name);	
	
	if(addKVREntry(&kValRep, entry)<0){
		struct KVRKeyword *tmp=NULL;
		struct KVRKeyword *next = NULL;
		if(replyTo!=NULL){
			/* to send an error we just put the XRP_ERROR command in front of the recevied message*/
			fillXRPCommand((char *)msg, XRP_CMD_ERROR);
			sendToIDP(msg, len, *replyTo);
		}
		/* we need to free the allocated entry in this case */
		free(entry->name);
		tmp = entry->keyWordsList;
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp->keyWord);
			free(tmp);
			tmp = next;
		}
		free(entry);
	}
	else{ /*everything went well*/
		if(replyTo!=NULL){ /*as a confirmation we send thereceived message back*/
			sendToIDP(msg, len, *replyTo);
		}
		anaPrint(ANA_DEBUG, "Publish succeeded\n");
	}	

}

void handleUnpublish(void *msg, int len, anaLabel_t input, void *aux, int auxLen) {

	int descLen ,labelSize, found, dummy;
	char *descrip = NULL;
	anaLabel_t *replyTo=NULL;	
	anaLabel_t *toUnpublish=NULL;
	
	anaHandle_t owner;
	
	struct KVRKeyObjectList *queryRes=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	
	toUnpublish = (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &labelSize, 0);
		
	replyTo = (anaLabel_t *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_LABEL, &dummy, 1);
		
	descrip = (char *) getXRPArg((xrpMsg_t)msg, XRP_CLASS_SRC_SRV, &descLen, 0);	 /* we get the name to lookup*/
		
	owner = *((anaHandle_t *)aux);

	if(toUnpublish==NULL || descrip == NULL ) { /* if no owner handle handle specified, NO UNPUBLISH possible */
		anaPrint(ANA_ERR, "handleUnpublish : No owner handle or no label to unpublish or description NULL\n");
		fillXRPCommand ((char *)msg, XRP_CMD_ERROR);
		sendToIDP(msg, len, *replyTo);
		return;
	}
	
	anaPrint(ANA_DEBUG, "handleUnpublish : description to unpublish = %s  IDP=%d \n",descrip, *toUnpublish);

	
	evalQuery(descrip, &queryRes);	
	
	found = 0;
	while(queryRes!=NULL && !found){
		if(queryRes->entry->IDP == *toUnpublish && queryRes->entry->owner == owner) {			
			removeKVREntry(&kValRep, queryRes->entry->name, queryRes->entry);
			found=1;	
		}
		
		tmp = queryRes;
		queryRes = queryRes->next;
		free(tmp);
	}
	
	/* we send as confirmation message, the same received message */
	if(replyTo!=NULL){
		if(found){
			sendToIDP(msg, len, *replyTo);
			anaPrint(ANA_DEBUG, "Unpublish succeeded\n");
		}else{
			anaPrint(ANA_ERR, "handleUnpublish : No entry (owned by  querier) fitting description and label \n");
			fillXRPCommand ((char *)msg, XRP_CMD_ERROR);
			sendToIDP(msg, len, *replyTo);
		}
		return;
	}
}



void dispatchNodeCompMsg(void *msg, int len, anaLabel_t input, void *aux, int auxLen) {
	
	xrpCmd_t command = (xrpCmd_t)msg;
	
	if(equalXRPCommands(command, XRP_CMD_RESOLVE )) /*if lookup command*/
		handleResolve(msg, len, input, aux, auxLen);
	if(equalXRPCommands(command, XRP_CMD_LOOKUP )) /*handleREsolve function does the same job, last arg 0: measn its resolve, 1 lookup*/
		handleLookup(msg, len, input, aux, auxLen);
	if(equalXRPCommands(command, XRP_CMD_PUBLISH )) /*if publish command*/
		handlePublish(msg, len, input, aux, auxLen);
	if(equalXRPCommands(command, XRP_CMD_UNPUBLISH )) /*if Unpublish command*/
		handleUnpublish(msg, len, input, aux, auxLen);
}

void heartBeatNCCheck(void){  /*this function is triggered at each heartBeat of the ana Minmex*/
	if(checkKVREntries(&kValRep)<0)
		anaPrint(ANA_ERR, "heartBeatNCCheck: check KVR Entries age failed \n");
}


