/* C/minmex/common/notifTable.c
 *  
 * functions to deal with the notification dispatch table
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "../include/anaNode.h"
#include "hashFunction.h"

unsigned int notifkey2hash( notifTable_t *Table, void *key, int keyLen) {

	if(Table==NULL || key == NULL){
		anaPrint(ANA_ERR, "notifkey2hash : NULL pointers \n");
		return -1;
	}
	return  hash((ub1 *)key, keyLen , 0 , Table->hashMaskSize); 
}


int initializeNotifTable(notifTable_t *Table, int hashMaskSize) { 
/* the allocated size of the table will be (2^hashMaskSize), the hashMaskSize will be also used in the hash   function */
	if(Table==NULL || hashMaskSize<0){
		anaPrint(ANA_ERR, "initializeNotifTable : NULL arguments \n");
		return -1;
	}
	Table->size = pow2(hashMaskSize);
	Table->hashMaskSize = hashMaskSize;
		
	Table->tab = (struct notifEntry_s **) malloc(Table->size * sizeof(void *));
	if(Table->tab == NULL){
		anaPrint(ANA_ERR, "initializeNotifTable: malloc returned error\n");
		return -1;
	}
	anaPrint(ANA_DEBUG, "Notif table initialized with size %u\n", Table->size);
	return 0;
}



/* The getIDTEntry first hashes the label, then runs through entire chained list 
 * to find an entry owned the asking brick with the wanted label
 * ( the mapping of the label in the bricks view ). 
 * While running though the chain it also locates the public entry corresponding 
 * to the wanted label. This public entry is returned if the brick
 * didn't previously request to create a private view of this label through a redirect() method call  */

struct notifEntry_s* getNotifEntry(notifTable_t *Table, void *key,  int keyLen) {
	int i; 
	struct notifEntry_s *entry=NULL;
	
	if(Table==NULL){ 
		anaPrint(ANA_ERR, "getNotifEntry : NULL Table pointer\n");
		return NULL;
	}	
	
	i = notifkey2hash(Table, key, keyLen);
	if (i<0 || i>(int) (Table->size)-1) {
		anaPrint(ANA_ERR, "getNotifEntry : index %d after hashing is out of bounds \n",i);
		return NULL;
	}
	
	entry = Table->tab[i];

	while (entry != NULL) {
		if (keyLen==entry->keyLen && (!memcmp(entry->eventKey, key, keyLen)))
			return entry;
		else 
			entry = entry->next;
	}
	
	return NULL;
}


int freeNotifTable(notifTable_t *Table) { 
	int i;
	struct notifEntry_s *entry;
	struct notifEntry_s *next;
	
	struct notifAux_s *notifAux;
	struct notifAux_s *notifAuxNext;
	
	if(Table==NULL ){
		anaPrint(ANA_ERR, "freeNotifTable : NULL Table pointer\n");
		return -1;
	}
	
	for (i=0; i< (int) Table->size; i++) {
		entry = Table->tab[i];
		while  (entry!=NULL) {
			next = entry->next;
			/* we free the aux(s) of the entry */
			notifAux = entry->notifAux;
			
			while(notifAux!=NULL){
				notifAuxNext = notifAux->next;
				free(notifAux->aux);
				free(notifAux);
				notifAux = notifAuxNext;
			}
			
			free(entry->eventKey);
						
			free(entry);
			entry = next;
		}
	}

	free(Table->tab);
	free(Table);
	return 0;
}




int freeNotifEntry(notifTable_t *Table, void *key, int keyLen) { 
	
	int i; 
	struct notifEntry_s *entry=NULL;
	struct notifEntry_s *predecessor=NULL;
	
	struct notifAux_s *notifAux;
	struct notifAux_s *notifAuxNext;
	
	if(Table==NULL){ 
		anaPrint(ANA_ERR, "getNotifEntry : NULL Table pointer\n");
		return -1;
	}	
	
	i = notifkey2hash(Table, key, keyLen);
	if (i<0 || i>(int) (Table->size)-1) {
		anaPrint(ANA_ERR, "freeNotifEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	entry = Table->tab[i];

	while (entry != NULL) {
		if ( keyLen == entry->keyLen && (!memcmp(entry->eventKey, key, keyLen))){
			
			if(predecessor!=NULL)
				predecessor->next = entry->next;
			else
				Table->tab[i] = entry->next;
			notifAux = entry->notifAux;
			
			while(notifAux!=NULL){
				notifAuxNext = notifAux->next;
				free(notifAux->aux);
				free(notifAux);
				notifAux = notifAuxNext;
			}
			
			free(entry->eventKey);
			free(entry);
			
			return 0;	
			
		}
		else {
			predecessor = entry;
			entry = entry->next;
		}
	}
	
	return -1;
}


int addNotifEntry (notifTable_t *Table, struct notifEntry_s *toAdd) {
	int i ;
	
	if(Table==NULL || toAdd==NULL){
		anaPrint(ANA_ERR, "addIDTEntry : NULL arguments \n");
		return -1;
	}
	
	i = notifkey2hash(Table, toAdd->eventKey, toAdd->keyLen);
	if (i<0 || i>(int) (Table->size)-1) {
		anaPrint(ANA_ERR, "addNotifEntry : index %d after hashing is out of bounds \n",i);
		return -1;
	}
	
	if(Table->tab[i] == NULL)
		Table->tab[i] = toAdd;
	else{
		toAdd->next = Table->tab[i];
		Table->tab[i] = toAdd;
	}	
	
	return 0;
}


