/* C/minmex/kernel/kernel_minmex.c
 * 
 * kernel specific minmex functions
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 *	
 * Copyright (c) 2008, Ariane Keller, Swiss federal institute of technology (ETH Zurich) 
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */



#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/random.h>
#include <linux/net.h>
#include <linux/socket.h>
#include <linux/time.h>
#include <linux/times.h>
#include <linux/wait.h>
#include <linux/rwsem.h>
#include <linux/mm.h>
#include <linux/in.h>
#include <linux/workqueue.h>

#include "../include/anaNode.h"
#include "anaValidate.h"

/* contains  all data gates open by this minmex*/
struct commGate_s *defaultDataGate;
/* contains all control gates open by this minmex */
struct commGate_s *defaultControlGate;
/* The information dispatch table */
IDT_t IDT;
/* table storing the  */
BrickTable_t BT;

notifTable_t NT; /* notification Table*/

struct  notifAux_s* allEvents[32];

/* sensitivity of anaPrint */
int anaDebugLevel = ANA_DEBUG;
//int validateoffset = 0;


/* the MINMEX in kernel space supports UDP sockets, kernel intern function call and 
 * for Linux Kernel Versions >= 2.6.20 generic netlink sockets for the communication 
 * between MINMEX and Bricks.
 */

/*module parameter, may be specified upon insmod */
static int t = 6;       /* option for the dispatch table size */
static int b = 10;      /* option for the brick table size */
static char * D  = "ANA_DEBUG";        /* indicates debug level*/

/* 
 * module_param(foo, int, 0000)
 * The first param is the parameters name
 * The second param is it's data type
 * The final argument is the permissions bits, 
 * for exposing parameters in sysfs (if non-zero) at a later stage.
 */
module_param(t, int, 0);
MODULE_PARM_DESC(t, "option for the dispatch table size");
module_param(b, int, 0);
MODULE_PARM_DESC(b, "option for the brick table size ");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
module_param(D, charp, 0);
#else
MODULE_PARM(D, "s");
#endif
MODULE_PARM_DESC(debug, "debug Level");

MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Ariane Keller");

/* work queue initialization, a wrapper function is needed
 * since the workqueue function prototype does not hava a 
 * data argument (for kernel 2.6.20 and newer)
 */
struct wq_wrapper wq_ctr;
struct wq_wrapper wq_data;
struct workqueue_struct *wq;

void heartBeatCheck(struct work_struct *unused);
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
DECLARE_DELAYED_WORK( work_hbc, heartBeatCheck); 
#else
DECLARE_WORK(work_hbc, heartBeatCheck, NULL);
#endif

/* kernel<->kernel: this function is called from a brick to let the MINMEX know
 * its data-receiving-function
 * */
void handleBrickDataChannel(struct anaBrickFunctions * brickDataFunctions, anaHandle_t handle){
        struct BTentry_s* entry = getBTEntry(&BT, handle);
        entry->commAux = brickDataFunctions;
}
/* struct which holds all MINMEX function which are accesible for a brick */
static struct anaMinmexFunctions minmex1Functions = {
        .attach            = handleAttach,
        .detach            = handleDetach,
        .regist            = handleRegister,
        .redirect          = handleRedirect,
	.forkData	   = handleForkData,
        .unregist          = handleUnregister,
        .heartBeat         = handleHBRequest,
        .notification      = NULL, //handleNotificationDecl,
        .callback          = dispatchData,
        .brickDataChannel  = handleBrickDataChannel,
	.changeView        = handleChangeView,
	.changeLabelAux	   = handleChangeLabelAux,
  	.setIDPinfo 	   = minmex_setIDPinfo,
  	.getIDPinfo 	   = minmex_getIDPinfo,
	.registerEventHandler 	= minmex_registerEventHandler,
	.unregisterEventHandler 	= minmex_unregisterEventHandler,
	.notifyEvent	   = minmex_notifyEvent,
};


/* heart beat check function, increases the age of of each Brick every 
 * HEART_BEAT_INTERVAL seconds
 */
void heartBeatCheck(struct work_struct *unused){
	/* we loop on bricks incrementing their age*/
	struct BTentry_s *tmp;
	int i;
        for(i=0;i<BT.size;i++){
                tmp =BT.tab[i];
                while(tmp!=NULL){
                        struct BTentry_s *nextBrick = tmp->next;
			tmp->age = tmp->age +1;
        		anaPrint(ANA_DEBUG, "\n+++++++++++++ brick: %s age: %i\n\n", tmp->brickName, tmp->age);
	                if(tmp->age > DEAD_CLIENT_THRESHOLD) { /* no life sign ==> we detach the brick*/
                                anaPrint(ANA_NOTICE, "Brick %s has given no life sign (age %d)==> detach\n", tmp->brickName, tmp->age);
                                if(kickBrick(tmp->handle)<0)
                                        anaPrint(ANA_ERR, "Unable to kick out brick %s\n", tmp->brickName);
                                
                                
                        }
                        tmp=nextBrick;
                }
        }
	
	/* then we check the IDT to remove unused entries*/
	if( checkIDTAges() < 0)
		anaPrint(ANA_NOTICE, "heartBeatCheck: IDT entries age check failed\n");
	/* we trigger the heartBeat check for the node Compartment */
	heartBeatNCCheck();	
	/* put us again in the queue */
	queue_delayed_work(wq, &work_hbc, HZ * HEART_BEAT_INTERVAL);
	return;
}

void cleanup(void){
        kickAllBricks();
	if (unregisterAnaMinmex(THIS_MODULE->name) < 0) {
		anaPrint(ANA_ERR, "could not unregister kern mode\n");
	}
	minmex1Functions.attach = NULL;
	minmex1Functions.detach = NULL;
	minmex1Functions.regist = NULL;
	minmex1Functions.redirect = NULL;
	minmex1Functions.unregist = NULL;
	minmex1Functions.heartBeat = NULL;
	minmex1Functions.notification = NULL;
	minmex1Functions.callback = NULL;
	minmex1Functions.brickDataChannel = NULL;
	minmex1Functions.registerEventHandler = NULL;
	minmex1Functions.unregisterEventHandler = NULL;
	minmex1Functions.notifyEvent = NULL;
}

static int __init anaMinmex_init(void) {

	int hashMaskSize, brickHashMaskSize;
	int ret = 0;
       
	/*default: print all messages */
	anaDebugLevel = ANA_DEBUG ;
	if(!memcmp(D, "ANA_NONE", 8))
		anaDebugLevel = ANA_NONE;
	if(!memcmp(D, "ANA_EMERG", 9))
		anaDebugLevel = ANA_EMERG;
	if(!memcmp(D, "ANA_ERR", 7))
		anaDebugLevel = ANA_ERR;
	if(!memcmp(D, "ANA_NOTICE", 10))
		anaDebugLevel = ANA_NOTICE;
	if(!memcmp(D, "ANA_DEBUG", 9))
		anaDebugLevel = ANA_DEBUG;
	if(!memcmp(D, "ANA_VALIDATE", 12))
		anaDebugLevel = ANA_VALIDATE; 

	if (anaDebugLevel == ANA_VALIDATE) {
		if (validate_init(THIS_MODULE->name) < 0)
			return -1;
	}


	hashMaskSize = t;
	brickHashMaskSize = b;

        /* Parsing the command Line */
	anaPrint(ANA_DEBUG, "INIT ANA NODE\n");
   	anaPrint(ANA_DEBUG, "hash Mask size and size of brick table: %d\n", brickHashMaskSize);
	anaPrint(ANA_DEBUG, "hash Mask size and size of the IDT: %d\n", hashMaskSize);

             
   	/* the 0 was the gate identifier, we keep it for now, but it is unused */
        if (registerAnaMinmex(&minmex1Functions, 0, THIS_MODULE->name) < 0) {
		anaPrint(ANA_ERR, "registration of kern mode MINMEX failed\n");
                cleanup();
                return -EEXIST;
	}
              
	
       		
	/* we initialize the IDT */
	ret = initializeIDT(&IDT, hashMaskSize, DEAD_IDTENTRY_THRESHOLD);
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for IDT \n");
		anaValidate("Unable to allocate memory for IDT \n");
                cleanup();
		return -ENOMEM;
	}
	
	/* we initialize the BT */
	ret = initBrickTable (&BT, brickHashMaskSize);
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for BT \n");
		anaValidate("Unable to allocate memory for BT \n");
                cleanup();
		return -ENOMEM;
	}
        
	/* we initialize the Notification table */
	ret =  initializeNotifTable(&NT, DEFAULT_NT_HASHMASK);
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for NT \n");
		anaValidate("Unable to allocate memory for NT \n");  
		cleanup();
		return -ENOMEM;
	}
	
	
	/* we initialize the node Compartment (that will initialize the KVR)*/
	initNodeComp(KVR_ENTRYTAB_DEFAULTMASK, KVR_KEYTAB_DEFAULTMASK, DEAD_KVRENTRY_THRESHOLD);

	/* initialize the work queue: The name has to be less than 10 characters for 
	 * compatibility reasons with older 2.6. kernels
	 */
        wq = create_singlethread_workqueue( "nodeWq" ); 
        if (!wq){
                cleanup();
                return -ENOMEM;
        }
        
        /*start the heard beat check*/
	if( !queue_delayed_work( wq, &work_hbc, HZ ) ) {                 	
                anaPrint(ANA_NOTICE,  "queue_work for heartbearcheck failed...\n");
		anaValidate("queue_work for heartbearcheck failed... \n");
        }

	ret = initStatusInterface();
	anaValidate("initialized status interface %i\n", ret); 
	anaValidate("minmex correctly initialized\n");

	return 0;
}

	
static void __exit anaMinmex_exit( void )
{
        kernel_status_exit();
        cleanup();
	if( wq ) {
                cancel_delayed_work(&work_hbc);
                flush_workqueue(wq);
                destroy_workqueue( wq );
        }
	if (anaDebugLevel == ANA_VALIDATE) {
		validate_exit();
	}

        anaPrint(ANA_DEBUG, "EXIT ANA NODE\n");
}

module_init(anaMinmex_init);
module_exit(anaMinmex_exit);

