/* C/minmex/userspace/user_minmex.c
 * 
 * userspace specific functions for the minmex
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "anaNode.h"
#include "anaValidate.h"

int anaDebugLevel = ANA_DEBUG;
char mymodename[]="anaMinmex";


IDT_t IDT;   /* The information dispatch table */

BrickTable_t BT;   /* table storing the bricks */

notifTable_t NT; /* notification Table*/

int status_clisock;  /* socket for recv CLI commands */
struct  notifAux_s* allEvents[32];

/* triggered at timeout elapse. Increments bricks ages and detaches bricks who havent sent HeartBeat above a certain threshold */
void heartBeatCheck(int sig){
	struct BTentry_s *nextBrick;
	struct BTentry_s *tmp;
	int i;
	/* we loop on bricks incrementing their age*/
	for(i=0;i<BT.size;i++){
		tmp =BT.tab[i];
		while(tmp!=NULL){
			tmp->age = tmp->age +1;
			nextBrick = tmp->next;
			if(tmp->age > DEAD_CLIENT_THRESHOLD) { /* no life sign ==> we detach the brick*/
				anaPrint(ANA_NOTICE, "Brick %s has given no life sign  (age %d)==> detach\n", tmp->brickName, tmp->age);
				if(kickBrick(tmp->handle)<0)
					anaPrint(ANA_ERR, "Unable to kick out Brick with handle %s\n", tmp->brickName);
			}
			tmp=nextBrick;
		}
	}
	
	/* then we check the IDT to remove unused entries*/
	if( checkIDTAges() < 0)
		anaPrint(ANA_NOTICE, "heartBeatCheck: IDT entries age check failed\n");
	
	/* we trigger the heartBeat check for the node Compartment */
	heartBeatNCCheck();  
		

}

static void anaMinmex_exit( int sig )
{
	user_status_exit();
        kickAllBricks();
		
        anaPrint(ANA_DEBUG, "EXIT ANA NODE\n");
	validate_exit();
	exit(0);
}


void user_status_exit(void)
{
      	if (status_clisock >= 0) 
	{
		anaPrint(ANA_DEBUG, "Closing status_clisock socket\n");
                close(status_clisock);
		status_clisock = -1;
        }
}


int user_status_init(void)
{
	struct sockaddr_in ctrl;    /* for the CLI */
	int ret;
	int i;
	
	status_clisock = socket(AF_INET, SOCK_DGRAM, 0);
	/* UDP socket for receiving CLI commands */
	if (status_clisock < 0)
	{
		anaPrint(ANA_ERR, "create status_clisock failed %i \n", status_clisock);
		return -1;
	}	

	memset(&ctrl, 0, sizeof(struct sockaddr_in));
	ctrl.sin_family = AF_INET;
	inet_aton("127.0.0.1", &ctrl.sin_addr);
	ctrl.sin_port = htons(UDP_ANASTATUS);
	
	/* since one can run multiple minmexes on a device, each minmex */
	/* must use a different UDP port for the status socket */
	
	for (i=0; i<MAX_UDP_BIND_TRIES; i++)
	{
		ctrl.sin_port = htons(UDP_ANASTATUS+i);
		ret = bind(status_clisock, (struct sockaddr *)&ctrl, sizeof(struct sockaddr_in));
		
		if (!ret)
			break;
	}

	if (ret!=0)
	{	
		anaPrint(ANA_ERR, "cannot bind UDP status_clisock socket!\n");
		return -1;
	}
	else /* we use ANA_EMERG because it also prints out on the console */
		anaPrint(ANA_EMERG, "Using CLI udp port %d\n", ntohs(ctrl.sin_port));
	
	return 0;
}


int main(int argc, char **argv) {
	int hashMaskSize, brickHashMaskSize;
	int ret = 0;
	char recvbuf[BUFSIZE];
	int len, c;
	struct sockaddr_in replyto;
	unsigned int fromlen = sizeof(struct sockaddr_in);

	if (log_init() < 0)
		return -1;


	anaDebugLevel = ANA_DEBUG;
	hashMaskSize = DEFAULT_IDT_HASHMASK;
	brickHashMaskSize = DEFAULT_BT_HASHMASK;

	opterr = 0;
	while ((c = getopt (argc, argv, "t:b:D:")) != -1)
		switch (c) { 
			case 't': /* option for the dispatch table size */
			{	
				hashMaskSize = atoi((char*)optarg); /* the size of the table is determined according to the hashMaskSize */
				if(hashMaskSize>25){
					anaPrint(ANA_ERR, "The IDT table size is too big, may slowdown or even crash the minmex, must be < 25\n");
					return 1;
				}	
				break;
			}
			case 'b': /* option for the brick table size */
			{	
				brickHashMaskSize = atoi((char*)optarg); /* the size of the table is determined according to the hashMaskSize */
				if(brickHashMaskSize>25){
					anaPrint(ANA_ERR, "The Bricks table size is too big, may slowdown or even crash the minmex, must be < 25 \n");
					return 1;
				}
				break;
			}
			case 'D': /* option to set DEbug mode on */
			{	
				if(!memcmp(optarg, "ANA_NONE", 8))
					anaDebugLevel = ANA_NONE;
				if(!memcmp(optarg, "ANA_EMERG", 9))
					anaDebugLevel = ANA_EMERG;
				if(!memcmp(optarg, "ANA_ERR", 7))
					anaDebugLevel = ANA_ERR;
				if(!memcmp(optarg, "ANA_NOTICE", 10))
					anaDebugLevel = ANA_NOTICE;
				if(!memcmp(optarg, "ANA_DEBUG", 9))
					anaDebugLevel = ANA_DEBUG;
				if(!memcmp(optarg, "ANA_VALIDATE", 12))
					anaDebugLevel = ANA_VALIDATE;

				break;
			}
			case '?':
			{	anaPrint(ANA_NOTICE, "Unknown option : -%c \n" , optopt);
				return 1;
			}
			default:
	  			abort (); 
		}
	
	if (validate_init(mymodename) < 0)
		return -1;

	
	/* we initialize the random generator */
	srandom(time(0));
			
	/* we initialize the IDT */
	ret = initializeIDT(&IDT, hashMaskSize, DEAD_IDTENTRY_THRESHOLD); 
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for IDT \n");
		anaValidate("Unable to allocate memory for IDT \n");
		exit(1);
	}
	
	/* we initialize the BT */ 
	ret = initBrickTable (&BT, brickHashMaskSize);
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for BT \n");
		anaValidate("Unable to allocate memory for BT \n");
		exit(1);
	}
	
	/* we initialize the Notification table */
	ret =  initializeNotifTable(&NT, DEFAULT_NT_HASHMASK);
	if (ret<0) {
		anaPrint(ANA_ERR,  "Unable to allocate memory for NT \n");
		anaValidate("Unable to allocate memory for NT \n");
		exit(1);
	}
		 
 	/* we arm the heartBeat timer*/	
	anatimer_init();
	if ( anatimer_add(HEART_BEAT_INTERVAL*1000, 1, 1, (fct_t)heartBeatCheck, NULL, 0) == -1 ){
		anaPrint(ANA_ERR,  "Unable to set HeartBeat timer \n");
		anaValidate("Unable to set HeartBeat timer \n");
		exit(1);
	}
	
	
	/*we set the exit function as the handler of Ctrl-C*/
	signal(SIGINT, anaMinmex_exit);
	
	/* we initialize the node Compartment ( that will initialize the KVR)*/
	initNodeComp(KVR_ENTRYTAB_DEFAULTMASK, KVR_KEYTAB_DEFAULTMASK, DEAD_KVRENTRY_THRESHOLD); 
		 

	/* First we launch the thread listening for communications encapsulated with IDPs only*/
	ret = user_status_init();
	if(ret < 0){
		anaPrint(ANA_ERR,  "Unable to initialize CLI status socket \n");
		anaValidate("Unable to initialize CLI status socket \n");
		exit(1);
	}
	anaValidate("initialized status interface %i\n", ret); 
	anaValidate("minmex correctly initialized\n");
	
       	/* now we loop indefinetly listening on the status socket */
	while(1)
	{	
		memset(recvbuf,0, BUFSIZE);
		len = recvfrom(status_clisock, recvbuf, BUFSIZE , 0,(struct sockaddr*)&replyto, &fromlen);
		if (len != -1){
			status_cli_command(recvbuf, len, &replyto);
			anaPrint(ANA_DEBUG,  "Returned from command handler\n");
		}
	}

	return 1;	
}
