/* C/shared/anaCommon.c
 * 
 * helper functions to be used from minmex and bricks
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> 
 * 	   Christophe Jelger <christophe.jelger@unibas.ch>
 *
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include "anaCommon.h"

unsigned int pow2(unsigned int x){
	return ((unsigned int)1)<<x;
}


anaLabel_t generateLabel(){
	/* for now we just generate random uint32 */
	return random();
}


anaHandle_t generateHandle(){
	return random();
}

int sendSocketWrapper(void *sock, int domain, void *aux, void *msg, int len){
	int nbSent = -1;
#ifdef USER
	int descriptor;
	int size;
	if (sock==NULL || aux==NULL || msg==NULL || len<=0)
		return -1;
	
	descriptor = *(int*)sock;

	if (domain == AF_LOCAL)
		size = sizeof(struct sockaddr_un);
	else
		size = sizeof(struct sockaddr);	 
	
	nbSent = sendto(descriptor, msg, len, 0,(struct sockaddr *) aux,  size );	 
#endif
#ifdef KERNEL
	struct msghdr msgh;
	struct iovec iov;
	mm_segment_t oldfs;
	struct socket *s = NULL;
	
	if (sock==NULL || aux==NULL || msg==NULL || len<=0 )
		return -1;
	
	memset(&msgh, 0, sizeof(msgh));
	msgh.msg_name = aux;
	if (aux)
		msgh.msg_namelen = sizeof(struct sockaddr_in);
	else
		msgh.msg_namelen = 0;

	iov.iov_base = msg;
	iov.iov_len  = len;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_iov    = &iov;
	msgh.msg_iovlen = 1;
	
	s = *((struct socket **)sock);

	if (s == NULL) {
        	anaPrint(ANA_ERR,  "send socket Wrapper: socket == NULL\n");
        	return -1;
    	}
	
	oldfs = get_fs();
	set_fs( KERNEL_DS );
	nbSent = sock_sendmsg(s, &msgh, len );
	set_fs( oldfs );
#endif
	return nbSent;
}


/* Wrapper functions to "port" userspace functions for the Linux kernel */
#ifdef KERNEL
/** \ingroup anaCommon
 * \brief platform independent random function.
 * returns a random integer 
 * \return random integer, in kernel space generated with get_random_bytes
 */
int random(void){
	unsigned int r;
	get_random_bytes(&r, sizeof(unsigned int));
	return r;
}
/** \ingroup anaCommon
 * \brief platform independent atoi function.
 * converts an ascii string to an integer 
 * \return the converted integer (0 if nptr was NULL)
 */
int atoi(char *nptr){
	if (nptr == NULL)
		return 0; /* we do not want a segfault...*/
	else
		return simple_strtol(nptr,NULL,10);
}
/** \ingroup anaCommon
 * \brief platform independent inet_addr function.
 * converts ip address from number and dot representation to binary
 * \param cp pointer to a human readable ip address
 * \return the binary ip address
 */
int inet_addr(const char *cp){
	return in_aton(cp);
}

/** \ingroup anaCommon
 * \brief  platform independent sleep function.
 * sleeps for the given amount of time
 * \param seconds number of seconds to sleep
 * \return failure: -1 success 0 (or a positive value if we have been woken up to early)
 */
int sleep(int seconds){
	int ret;
	
	if (seconds < 0)
		return -1;
		
	set_current_state(TASK_INTERRUPTIBLE);
	ret = schedule_timeout(seconds*HZ);
	return ret/HZ;
}

/** \ingroup anaCommon
 * \brief  platform independent gettimeofday function.
 * The timezone argument is officially deprecated therefore we do not support it
 * \param tv will hold the result
 * \param tz ignored
 * \return failure: -1 success 0
 */
int gettimeofday(struct timeval *tv, struct timezone *tz){
	if (!tv)
		return -1;
	do_gettimeofday(tv);
	return 0;
}
#endif


/* Wrapper functions for standard system calls: configured in Makefiles variable GWRAP */
/* e.g. using 'malloc' will call __wrap_malloc (refer to man ld) */

#ifdef USER
extern void __real_free(void *ptr);
void __wrap_free(void *pointer){
	if (pointer != NULL)
		__real_free(pointer);
	pointer = NULL;
}
#endif
#ifdef KERNEL
/** \ingroup anaCommon
 * \brief  platform independent free function.
 * first sets freed pointer to NULL
 * \param pointer the pointer to the memory to be freed
 */
void free(void *pointer){
	if (pointer != NULL)
		kfree(pointer);
	pointer = NULL;
}
#endif

#ifdef USER
void * __wrap_malloc(size_t size){
	void *pointer=NULL;
	if (size<=0)
		return NULL;
	pointer = __real_malloc(size);

	if (pointer != NULL)
		memset(pointer, 0, size);
	return pointer;
}
#endif

#ifdef KERNEL
/** \ingroup anaCommon
 * \brief  platform independent malloc function.
 * initializes the memory with 0
 * \param size number of bytes to allocate
 * \return pointer to allocated memory
 */
void * malloc(int size){
        void *pointer=NULL;
	if (size<=0)
		return NULL;
        pointer = kmalloc(size, GFP_KERNEL);
	if (pointer != NULL)
		memset(pointer, 0, size);
	return pointer;
}

	
#endif

/** \ingroup anaL0
 * \brief frees correctly the memory allocated inside an IDPinfo_s structure (we don't free the structure itself) 
 * \param result pointer to the structure to be freed  
 * \return 0 on success, negative error value otherwise
 */
int anaL0_clearIDPinfo(struct IDPinfo_s *result){
	struct nextIDP_s *tmp  = NULL;
	struct nextIDP_s *next = NULL;
	if (result == NULL)
		return -1;
	
	if(result->ICflag){
		free(result->destContext.value);
		free(result->destService.value);
	}
	if(result->IDPtype == FORK || result->IDPtype == REDIRECT ){
		tmp = (struct nextIDP_s *)((result->binding).IDPs); 
		while(tmp!=NULL){
			next = tmp->next;
			free(tmp);
			tmp = next;		
		}
	}
	
	memset(result, 0, sizeof(struct IDPinfo_s));
	return 0;
}
