/* C/shared/anaThread.c
 * 
 * ana thread library for user space and kernel space
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 *         Christophe Jelger <Christophe.Jelger@unibas.ch>
 *
 * Copyright (c) 2008, Ariane Keller, Swiss federal institute of technology (ETH Zurich) 
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/** @defgroup anaThread ANA Thread API
 * Definition of the ANA Thread API 
 */

#include "../minmex/include/anaNode.h"
#include "anaCommon.h"
#include "analock.h"

#ifdef USER
#include <pthread.h>
#include <signal.h>
//#include <dlfcn.h>
#include "../minmex/include/clientTable.h"
#endif

#ifdef KERNEL
#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/completion.h>
#include <linux/kthread.h>

static DECLARE_COMPLETION(onExit);
#endif /* KERNEL */

struct anaThread_s {
	void * (*function)(void *); /* the function to be executed */
	void *arg;		    /* argument for the function */
	anaHandle_t brickhandle;    /* the brick handle, 0 for minmex */
	char *threadname;	    /* the name of the thread */
	char *brickname;	    /* the name of the brick, or NULL (if minmex thread) */
	int ana_errno;		    /* the ana_errno for this anathread */
#ifdef USER
	pthread_t pid;
#else
	analock_t startLock;	/* needed to prevent race condition when starting a thread */
	pid_t pid;		/* the pid of the thread */
	int terminated;		/* this is set by anaThread exit: only if it is set */
				/* the anaKernelThread has to call complete_and_exit */
	struct task_struct *task; /* describes the thread */
#endif
	struct anaThread_s *next;
};

/* a list with all active threads, used upon rmmod to kill all threads */
static struct anaThread_s *thread_list = NULL;
analock_t threadLock;

int set_ana_errno(int value)
{
	struct anaThread_s *t = thread_list;
#ifdef USER
	pthread_t id = pthread_self();
#endif
#ifdef KERNEL 
	pid_t id = current->pid;
#endif

	while (t)
	{
		if (t->pid == id)
		{
			t->ana_errno = value;
			return 0;
		}
	
		t = t->next;
	}

	anaPrint(ANA_ERR, "set_ana_errno --> cannot find thread %lu!\n", (long unsigned) id);
	return -1; /* thread id not found! */
}

int get_ana_errno(void)
{
	struct anaThread_s *t = thread_list;
#ifdef USER
	pthread_t id = pthread_self();
#endif
#ifdef KERNEL 
	pid_t id = current->pid;
#endif

	while (t)
	{
		if (t->pid == id)
			return t->ana_errno;
			
		t = t->next;
	}

	anaPrint(ANA_ERR, "get_ana_errno --> cannot find thread %lu!\n", (long unsigned) id);
	return 0; /* thread id not found --> what shall we return? */
}

#ifdef USER /* to catch segfaults inside each thread and avoid crashing the entire process*/
struct sigaction action;

void anaThread_signal(int sig)
{	
	struct anaThread_s *t = thread_list;
	pthread_t id = pthread_self();

	#ifdef PLUGIN_MODE
	struct BTentry_s *b;
	void (*so_log_signal)(pthread_t pid, int sig, char *bname, char *threadname);
	#endif

	while (t)
	{
		if (id == t->pid)
		{
			if (sig == SIGSEGV)
			{
				anaPrint(ANA_EMERG, "Thread %lu (%s %s) doing segmentation fault!\n", 
					 id, t->brickname ? t->brickname : "minmex", t->threadname);
			}
			else if (sig == SIGTERM)
			{
				anaPrint(ANA_EMERG, "Thread %lu (%s %s) receiving SIGTERM!\n", 
					 id, t->brickname ? t->brickname : "minmex", t->threadname);
			}
			else
			{
				anaPrint(ANA_ERR, "Thread %lu (%s %s) receiving unknown signal\n", 
					 id, t->brickname ? t->brickname : "minmex", t->threadname);
					 
				return;
			}

			/* if brick is a .so plugin, the log above goes in the minmex log */
			/* --> so we also log the seg fault in the brick's log file */
#ifdef PLUGIN_MODE
			if (t->brickhandle)
			{
				b = getBTEntry(&BT, t->brickhandle);
			
				if (b)
				{
					struct commSpecs_s *specs = (struct commSpecs_s *)b->commAux;
					char *error = NULL;
				
					so_log_signal = dlsym(specs->pluginHandle, "so_log_signal");
				
					if ((error = dlerror()))
					{
						anaPrint(ANA_ERR, "Error resolving so_log_signal symbol for brick %s: %s\n", 
						 	 t->brickname, error);
					}
					else
						so_log_signal(t->pid, sig, t->brickname, t->threadname);
				}	
			}
#endif /* PLUGIN_MODE */
			free(t->brickname);
			free(t->threadname);
			t->pid = 0; /* free slot */
			pthread_exit(0);
		}
		
		t = t->next;
	}

	if (sig == SIGSEGV)
		anaPrint(ANA_EMERG, "Thread %lu (unknown) doing segmentation fault!\n", id);
	else if (sig == SIGTERM)
		anaPrint(ANA_EMERG, "Thread %lu (unknown) receiving SIGTERM!\n", id);
	else
	{
		anaPrint(ANA_ERR, "Thread %lu (unknown) receiving unknown signal\n", id);
		return;
	}
	if(t!=NULL){
		free(t->threadname);
		free(t->brickname);	
	}
	pthread_exit(0);
}

#endif /* USER */

/** \ingroup anaThread
 * \brief Indicates if the given thread should stop.
 *
 * This function is only important in kernel space, since one cannot 
 * explicitly kill a kernel thread, but the thread has to terminate itself.
 * (e.g. call exit(0));
 * 
 * \return 1 if the thread should stop 0 otherwise
 */
int anaThreadShouldStop(void){
#ifdef USER
	return 0;
#endif
#ifdef KERNEL
	return kthread_should_stop();
#endif

}
#ifdef KERNEL
/* function which launches the kernel thread.
 */
static int anaKernelThread(void * arg){
	struct anaThread_s * mythread = (struct anaThread_s *)arg;
	
	/* we need to wait until the _anaThread function has done the bookkeeping*/
	analock_lock(&mythread->startLock);
	anaPrint(ANA_TIMER, "starting thread %s (pid = %lu) of brick %s\n",
		 mythread->threadname, (long unsigned)mythread->pid, mythread->brickname);
	
	mythread->function(mythread->arg);
	
	anaPrint(ANA_TIMER, "Thread %s (pid = %lu) of brick %s has terminated\n",
		 mythread->threadname, (long unsigned)mythread->pid, mythread->brickname);	
	mythread->pid = 0; /* slot can be reused */
	analock_unlock(&mythread->startLock);
	return 0;
}
#endif /*KERNEL*/

#ifdef USER
void *anaUserThread(void * arg){
	struct anaThread_s * mythread = (struct anaThread_s *)arg;
	
	/* make sure seg faults do not crash entire process */
	sigaction(SIGSEGV, &action, NULL);
	sigaction(SIGTERM, &action, NULL);
	
	/* execute the function */
/*	anaPrint(ANA_DEBUG, "New anaThread: %s %s (thread id %lu)\n", 
		 mythread->brickname ? mythread->brickname : "minmex", mythread->threadname, mythread->pid); */
	
	mythread->function(mythread->arg);
	
/*	anaPrint(ANA_DEBUG, "anaThread %s %s (thread id %lu) has terminated\n",
		 mythread->brickname ? mythread->brickname : "minmex", mythread->threadname, mythread->pid); */
	mythread->pid = 0; /* slot can be reused */
	return NULL;
}
#endif /* USER */

/* \ingroup anaThread 
 * \brief Initialize ANA Thread Infrastructure
 */
void anaThreadInit(void){

#ifdef USER
	action.sa_handler = anaThread_signal; /* to catch seg faults and sig terms */
#endif
	analock_init(&threadLock);
}

/* \ingroup anaThread
 * \brief cleans up ANA Thread Infrastructure
 */
void anaThreadExit(void){

	struct anaThread_s *done, *ptr = thread_list;
#ifdef USER
	pthread_t me = pthread_self();
	struct anaThread_s *m = NULL;
#endif
	analock_lock(&threadLock); /* we lock the entire list */
	
	while (ptr)
	{
#ifdef USER	
		if (ptr->pid == me)
		{
			/* if the brick runs as a standalone process, the thread */
			/* that called anaThreadExit() must kill itself last ! */
			
			m = ptr;
			ptr = ptr->next;
			continue;		
		}
#endif	
		if (ptr->pid)
		{
#ifdef USER
			pthread_kill(ptr->pid, SIGTERM);
#else			
			anaPrint(ANA_TIMER, "stopping thread %s (pid = %i) of brick %s\n", ptr->threadname, ptr->pid, ptr->brickname);	
			kthread_stop(ptr->task);

#endif /* KERNEL */
		}
		
		done = ptr;
		ptr = ptr->next;
		free(done);
	}
	thread_list = NULL;
#ifdef USER
	if (m)
	{
		pthread_kill(m->pid, SIGTERM); /* to get proper log */
		free(m); /* not sure this is ever executed */	
	}
#endif
	
	analock_unlock(&threadLock);

	anaPrint(ANA_DEBUG, "All threads have been killed\n");
}

int anaThread_killbrick(anaHandle_t brickhandle)
{
	struct anaThread_s *t = thread_list;
	int i = 0;
	
	if (!brickhandle)
		return -1;
		
	analock_lock(&threadLock);
	
	while (t)
	{
		if (t->pid && (t->brickhandle == brickhandle))
		{
#ifdef USER
			anaPrint(ANA_TIMER, "Stopping thread %s %s (thread id %lu)\n", 
				 t->brickname, t->threadname, t->pid);
			pthread_kill(t->pid, SIGTERM);
#else			
			anaPrint(ANA_TIMER, "stopping thread %s of brick %s\n", t->threadname, t->brickname);
			kthread_stop(t->task);
			t->pid = 0;
#endif /* KERNEL */
			/* for .so plugins, t->pid is set to 0 by anaThread_signal */
			i++;
			t = t->next;
		}
		else
			t = t->next;
	}
	analock_unlock(&threadLock);
	return i; /* number of killed threads */
}

//extern char mymodulename[]; 
unsigned long int _anaThread(anaHandle_t brickhandle, void * (*start_routine)(void *), void *arg, char *threadname){
	
	struct anaThread_s *last = NULL, *cur = thread_list;
#ifdef USER
	int a;
#endif
	analock_lock(&threadLock);
	
	while (cur)
	{
		if (!cur->pid)
			break; /* found an unused slot */
			
		last = cur;
		cur = cur->next;
	}
	
	if (!cur) /* allocate new space */
	{
		cur = malloc(sizeof(struct anaThread_s));
		
		if (!cur)
		{
			anaPrint(ANA_ERR, "Cannot allocate new anaThread_s!\n");
			analock_unlock(&threadLock);
			return 0;
		}
		
		if (last)
			last->next = cur;
		else
			thread_list = cur;
			
		cur->next = NULL; /* cur->next is already set for existing slots */
	}

	cur->brickhandle = brickhandle;
	cur->function = start_routine;
	cur->arg = arg;
#ifdef KERNEL
	analock_init(&cur->startLock);
#endif 
	
	if (brickhandle) {
#ifdef USER
#ifdef PLUGIN_MODE
		char *tmp = handle2name(&BT, brickhandle);
		if (tmp) {
			cur->brickname = (char *)malloc(strlen(tmp)+1);
			strcpy(cur->brickname, tmp);
		}
#else
		cur->brickname = mymodename;
#endif
#else
		cur->brickname = mymodename;
#endif
	}
	else
		cur->brickname = NULL;
	
	if (threadname) {
		cur->threadname = (char *) malloc(strlen(threadname)+1);
		strcpy(cur->threadname, threadname);
	}
	else {
		cur->threadname = (char *)malloc(strlen("anaThread") + 1);
		strcpy(cur->threadname, "anaThread");
	}	
#ifdef KERNEL
	cur->terminated = 0;

	/* we assume that all slots with pid != 0 are running threads.
	 * if a threads stops it sets its pid to 0. If the thread
	 * terminates *before* we assign its pid to cur->pid the 0 will
	 * be overwritten and we wrongly assume the thread is still running...
	 * -> we do the locking here
	 */
	analock_lock(&cur->startLock);
	cur->task = kthread_run(anaKernelThread, cur, cur->threadname);
	cur->pid = cur->task->pid;
	analock_unlock(&cur->startLock);

	analock_unlock(&threadLock);
	return cur->pid;
#endif /* KERNEL */

#ifdef USER
	pthread_attr_t tattr;
	pthread_attr_init(&tattr);
	pthread_attr_setdetachstate(&tattr,PTHREAD_CREATE_DETACHED);
	a = pthread_create(&cur->pid, &tattr, anaUserThread, cur);
	
	if (a != 0)
	{
		anaPrint(ANA_ERR, "Cannot create thread: %i\n", a);
		pthread_attr_destroy(&tattr);
		cur->pid = 0; /* slot can be reused */
		analock_unlock(&threadLock);
		return 0;
	}
	
	pthread_attr_destroy(&tattr);
	analock_unlock(&threadLock);
	return cur->pid;
#endif /* USER */
}

