/* C/shared/anatimer.c
 * 
 * ana timers for user space, kernel space and ns2
 *
 * Author: Christophe Jelger <christophe.jelger@unibas.ch> (initial design)
 *  	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> (testing and feedback)
 * 	  
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/** @defgroup anaTimer ANA Timer API
 * Definition of the ANA Timer API 
 */


#include "anatimer.h"
#include "analock.h"
#include "anaThread.h"

inline long timeval_diff(struct timeval *t1, struct timeval *t2);

int timerthread(void *data);

static int anatimer_resched(struct timeval *now);
static int anatimer_periodic(struct anatimer_s *cur, long delay);
int anatimer_insert(struct anatimer_s *new_timer, long msec);

#ifdef USER
static void anatimer_handler(int sig, siginfo_t *siginfo, void *context);
#endif

#ifdef KERNEL
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void anatimer_handler(void *ctrl);
#else
static void anatimer_handler(struct work_struct *ctrl);
#endif /* VERSION */
#endif /* KERNEL */

unsigned int anatimer_nextid = 0;
struct anatimer_s *anatimer_list;
analock_t list_lock;

/* id of thread which started the anatimers */
/* --> in .so mode, it prevents bricks from stopping timers */
/* not needed in kernel mode: 
 * we call anatimer_exit from the module_exit function.
 * rmmod has its own pid and could therefor not stop any timer */
#ifdef USER
pthread_t anatimer_owner;
#endif

inline long timeval_diff(struct timeval *t1, struct timeval *t2)
{
    long msec;

    if (!t1 || !t2)
	return -1;
    else 
    {
	msec = (t1->tv_sec - t2->tv_sec) * 1000;
	msec += (t1->tv_usec / 1000);
	msec -= (t2->tv_usec / 1000);
	
	return msec;
    }
}

#ifdef USER
static struct sigaction oldact;
#endif

#ifdef KERNEL
static struct timer_list myanatimer;
struct work_struct wp_anatimer;

struct workqueue_struct *anatimerwq;

void anatimer_fired(unsigned long dummy)
{
	queue_work(anatimerwq, &wp_anatimer);
}
#endif /* KERNEL */

/** \ingroup anaTimer
 * \brief adds a new timer
 *
 * \param msec time in mili seconds until when a 
 * \param period
 * \param fix_clock
 * \param action the function to be executed when the timer expires
 * \param fct_data data passed to the function. The data is /b copied.
 * \param datalen lenght of the data
 * \return returns the id of this timer
 */

int anatimer_add(long msec, unsigned char period, unsigned char fix_clock, 
		 fct_t action, void *fct_data, int datalen)
{
	struct anatimer_s *new_timer;
		
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "Call anatimer_init before adding timers!\n");
		return -1;
	}
	
	if ( msec <= 0 || !action )
		return -1;

	
	new_timer = (struct anatimer_s *)malloc(sizeof(struct anatimer_s));
	if (!new_timer)
		return -1;
	
	/* setup timer parameters */
	new_timer->id = anatimer_nextid++;
	new_timer->action = action;

	if (datalen > 0 && fct_data)
	{
		new_timer->data = malloc(datalen);
		
		if (!new_timer->data)
			return -1;
			
		memcpy(new_timer->data, fct_data, datalen);
	}
	else
		new_timer->data = NULL;


	new_timer->fix_clock = fix_clock;
	new_timer->running = 0;		
	/* compute timeout value */
	gettimeofday(&new_timer->timeout, NULL);
	new_timer->timeout.tv_usec += msec * 1000;
	new_timer->timeout.tv_sec += new_timer->timeout.tv_usec / 1000000;
	new_timer->timeout.tv_usec = new_timer->timeout.tv_usec % 1000000;

	/* if periodic timer, setup value */
	if (period)
	{
		new_timer->period.tv_sec = msec / 1000;
		new_timer->period.tv_usec = (msec % 1000) * 1000;
	}

	analock_lock(&list_lock);
	anatimer_insert(new_timer, msec);
	analock_unlock(&list_lock);

	return new_timer->id;
}


/** \ingroup anaTimer
 * \brief deletes the timer with the given ID
 * \param id the id to be deleted
 * \return 1 on success
 */

int anatimer_del(unsigned int id)
{
	struct anatimer_s *cur, *prev;
	
	if (!id)
		return -1; /* id 0 is used for garbage collection */
		
	analock_lock(&list_lock);
	
	cur = anatimer_list;
	prev = cur;
	if (!cur)
	{
		analock_unlock(&list_lock);
		return 0; /* list is NULL */
	}

	while (cur)
	{
		if (cur->id == id)
			break;
		prev = cur;	
		cur = cur->next;
	}
	
	if (!cur)
	{
		analock_unlock(&list_lock);
		return -1; /* id not found */
	}
	
	/* if the cur->action is currently executed we set the id to 0
 	 * this frees the data and the anatimer_s when the action has returned
 	 * if cur->action is not running we free the data and the anatimer_s
 	 */
	if(cur == anatimer_list)
		anatimer_list = cur->next;
	else
		prev->next = cur->next;
	cur->next = NULL;

	if(cur->running)
		cur->id = 0;
	else{
		free(cur->data);
		free(cur);
	}

	analock_unlock(&list_lock);
	
	anaPrint(ANA_TIMER, "Removing Timer_%d\n", id);
	return 1;

}

/* \ingroup anaTimer
 * \brief initializes anatimer environment
 */

void anatimer_init(void)
{
#ifdef USER	
	struct sigaction sigact;
	sigset_t sigset;

	if (anatimer_nextid != 0) /* anatimer_init already called */
		return;
				
	/* block SIGALRM */
	sigemptyset (&sigset);
	sigaddset (&sigset, SIGALRM);
	sigprocmask (SIG_BLOCK, &sigset, NULL);
		
	/* set up signal handler */
	memset(&sigact, 0, sizeof(struct sigaction));
	sigact.sa_sigaction = anatimer_handler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = SA_SIGINFO;
	sigaction(SIGALRM, &sigact, &oldact);
		
	/* unblock SIGALRM */
	sigprocmask(SIG_UNBLOCK, &sigset, NULL);
#endif

#ifdef KERNEL
	if (anatimer_nextid != 0) /* anatimer_init already called */
		return;

	init_timer(&myanatimer);
	myanatimer.function = anatimer_fired;
		
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)	
	INIT_WORK(&wp_anatimer, anatimer_handler);
#else	
	INIT_WORK(&wp_anatimer, anatimer_handler, &wp_anatimer);
#endif /*version*/

	/*initialize the work queue*/
	anatimerwq = create_singlethread_workqueue("anatimerwq"); 
	if (!anatimerwq)
	       anaPrint(ANA_ERR, "Cannot create anatimer work queue!\n");
#endif /*kernel*/

	anaThreadInit();

#ifdef USER
	anatimer_owner = pthread_self();
#endif

	analock_init(&list_lock);
	anatimer_nextid = 1;
	anatimer_list = NULL;
}	

/* \ingroup anaTimer
 * \brief cleanup the anatimer environment
 */

void anatimer_exit(void)
{
	struct anatimer_s *next, *cur = anatimer_list;
#ifdef USER
	struct itimerval utimer;
#endif

#ifdef PLUGIN_MODE
#ifdef USER
	pthread_t id = pthread_self();
	if (id != anatimer_owner) {
		return;
	}
#endif
#endif	
	if (anatimer_nextid == 0) /* timers are already off */
		return;

#ifdef USER
	sigaction(SIGALRM, &oldact, NULL); /* restore SIGALRM catching */
		
	utimer.it_interval.tv_sec = 0;
	utimer.it_interval.tv_usec = 0;
	utimer.it_value.tv_sec = 0;
	utimer.it_value.tv_usec = 0;
		
	if ( setitimer(ITIMER_REAL, &utimer, NULL) < 0 )
		anaPrint(ANA_ERR, "anatimer_exit --> cannot stop timer?\n");
	else
		anaPrint(ANA_DEBUG, "anatimer_exit --> timer stopped\n");
#endif

#ifdef KERNEL
	del_timer_sync(&myanatimer);

	if (anatimerwq)
	{
		flush_workqueue(anatimerwq);
		destroy_workqueue(anatimerwq);
	}
#endif

	anatimer_nextid = 0;
		
	if (!anatimer_list)
		return;

	/* freeing the list
 	 * if cur->action is currently executed we set the id to 0
 	 * this frees the data and the anatimer_s when the action has returned
 	 * if cur->action is not running we free the data and the anatimer_s
 	 */

	while (cur)
	{
		next = cur->next;
		if(cur->running)
			cur->id = 0;
		else {
			if (cur->data)
				free(cur->data);
			free(cur);
		}
		cur = next;
	}
	
	anatimer_list = NULL;
}

/** \ingroup anaTimer
 * \brief restarts a timer befor it is expired
 * \param id the of the timer
 * \param msec mili seconds until the timer should fire
 * \return -1 on failure, id of the timer otherwise
 */
int anatimer_restart(unsigned int id, long msec)
{
	struct anatimer_s *prev = NULL, *cur;
	
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_restart --> timers are off!\n");
		return -1;
	}

	if ( msec <= 0 || !id )
	{
		anaPrint(ANA_ERR, "anatimer_restart --> msec or id NULL!\n");
		return -1;
	}
	
	analock_lock(&list_lock);
	
	cur = anatimer_list;
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "anatimer_restart --> list NULL!\n");
		analock_unlock(&list_lock);
		return -1;
	}
	
	while (cur)
	{
		if (cur->id == id)
			break;
			
		prev = cur;
		cur = cur->next;
	}

	if (!cur)
	{
		anaPrint(ANA_ERR, "anatimer_restart --> Timer_%d not found!\n", id);
		analock_unlock(&list_lock);
		return -1;
	}

	cur->id = 0; /* temporarily de-activate */

	if (!prev)
		anatimer_list = cur->next;
	else
		prev = cur->next;
		
	cur->next = NULL;
	cur->id = id;
	/* re-compute timeout value */
	gettimeofday(&cur->timeout, NULL);
	cur->timeout.tv_usec += msec * 1000;
	cur->timeout.tv_sec += cur->timeout.tv_usec / 1000000;
	cur->timeout.tv_usec = cur->timeout.tv_usec % 1000000;
	
	anaPrint(ANA_TIMER, "anatimer_restart --> restarting Timer_%d in %ld msec\n", id, msec);
	anatimer_insert(cur, msec);
	analock_unlock(&list_lock);
	return id;
}

int timerthread(void *data){
	struct anatimer_s * timer = (struct anatimer_s *)data;
	timer->action(timer->data);
	if(timer->id == 0){  /*we have been deleted*/
		free(timer->data);
		free(timer);
	}
	else
		timer->running = 0;
	return 0;
}


#ifdef USER
static void anatimer_handler(int sig, siginfo_t *siginfo, void *context)
#endif
#ifdef KERNEL
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void anatimer_handler(void *ctrl)
#else
static void anatimer_handler(struct work_struct *ctrl)

#endif /* VERSION */
#endif /* KERNEL */

{
	struct anatimer_s *cur, *old, *prev = NULL;
	struct timeval now;
	long delay;
	
	anaPrint(ANA_TIMER, "anatimer_handler called\n");	
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_handler --> anatimers are off!\n");
		return;
	}
			
#ifdef USER
	if (sig != SIGALRM)
		return;
#endif
	analock_lock(&list_lock);

	if (!anatimer_list)
	{
		anaPrint(ANA_ERR, "anatimer_handler --> anatimer_list is NULL!\n");
		goto exit_handler;
	}
	
	cur = anatimer_list;	
	gettimeofday(&now, NULL);

	while (cur)
	{
		if (!cur->id) /* timer is de-activated */
		{
			prev = cur;
			cur = cur->next;
			continue;
		}
				
		anaPrint(ANA_TIMER, "Now %ld.%ld Timer_%d %ld.%ld\n", now.tv_sec, now.tv_usec,
			 cur->id, cur->timeout.tv_sec, cur->timeout.tv_usec); 
		delay = timeval_diff(&cur->timeout, &now);
			
		if ( delay <= 0 )
		{
			/* cur has expired */
			
			old = cur;
			cur = cur->next;
		
			if (!old->period.tv_sec && !old->period.tv_usec)
			{
				/* timer is not periodic: 
 				 * set id = 0 -> timerthread will delete this timer
 				 * remove timer from list
 				 */
				anaPrint(ANA_TIMER, "Removing Timer_%d\n", old->id);
				old->id = 0;
				if (prev)
					prev->next = old->next;
				else
					anatimer_list = old->next;
				old->next = NULL; /* avoid list loops! */

			}
			else
			{	/* remove from list before re-inserting */
				
				if (prev)
					prev->next = old->next;
				else
					anatimer_list = old->next;
			
				old->next = NULL; /* avoid list loops! */
			
				/* restart periodic timer */
				/* delay is a negative number */
				/* we here pass the absolute value */
			
				anatimer_periodic(old, -delay);
			}
			
			if (old->action)
			{	
				/* launch action in a separate thread 
 				 * set running to 1 indicates to del_timer 
 				 * and exit_timer to not free the memory for this timer*/
				old->running = 1;
				_anaThread(0, (void *)timerthread, old, "anatimer action");
			}
		}
		else
		{	/* cur not expired : reschedule timer */
			
			anatimer_resched(&now);
			goto exit_handler;
		}
	}
	
	anatimer_resched(&now); /* reached end of list, recheck in case something went wrong */
exit_handler:
	analock_unlock(&list_lock);
}

static int anatimer_periodic(struct anatimer_s *cur, long delay)
{

	/* restart periodic timer */
	/* lock is already acquired: do not recall analock_lock */

	long msec;
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_restart --> anatimers are off!\n");
		return -1;
	}
	
	if (!cur)
		return -1;
		
	if (!cur->id) /* trying to reschedule de-activated timer? */
		return -1;

	/* previous time + period */
	
	cur->timeout.tv_sec += cur->period.tv_sec;
	cur->timeout.tv_usec += cur->period.tv_usec;
	
	cur->timeout.tv_sec += cur->timeout.tv_usec / 1000000;
	cur->timeout.tv_usec = cur->timeout.tv_usec % 1000000;
	
	/* delay until next timer (= period) */
	
	msec = cur->period.tv_sec*1000 + cur->period.tv_usec/1000;
	
	/* ABSOLUTE flag means: re-schedule according to delay */
	/* That is: if timer was late, re-schedule earlier to  */
	/* compensate delay */
	
	if (cur->fix_clock == ANATIMER_ABSOLUTE)
	{
		msec = msec - delay;
		cur->timeout.tv_sec -= delay / 1000;
		
		delay = (delay % 1000) * 1000; /* usec */
				
		if (cur->timeout.tv_usec < delay )
		{
			cur->timeout.tv_sec -= 1;
			cur->timeout.tv_usec += 1000000;
			cur->timeout.tv_usec -= delay;
		}
		else
			cur->timeout.tv_usec -= delay;
	}
	
	anaPrint(ANA_TIMER, "Periodic Timer_%d restarts in %ld msec\n", cur->id, msec);
	anatimer_insert(cur, msec);
	
	return 1;
}


int anatimer_insert(struct anatimer_s *new_timer, long msec)
{

	struct anatimer_s *cur, *prev;
	long delay;
	int idx = 0;

	/* lock is already acquired: do not recall analock_lock */
	
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_insert --> anatimers are off!\n");
		return -1;
	}
	
	if (!new_timer || msec <= 0)
		return -1;
	
	if (!new_timer->id) /* trying to insert de-activated timer? */
		return -1;
	
	if (!anatimer_list) /* list is NULL */
	{
		anatimer_list = new_timer;
		new_timer->next = NULL;
		anaPrint(ANA_TIMER, "Timer_%d inserted in anatimer_list[0]\n", new_timer->id);
		anatimer_resched(NULL);

		return 1;
	}	
	else
	{
		cur = anatimer_list;
		prev = NULL;
	}

	while (cur) /* insert timer in anatimer_list (chronological order) */
	{
		if (!cur->id) /* timer is de-activated (waiting to be garbage cleaned) */
		{
			prev = cur;
			cur = cur->next;
			continue;
		}
	
		delay = timeval_diff(&new_timer->timeout, &cur->timeout);
	
		if ( delay < 0 )
		{
			/* insert between prev and cur */
			new_timer->next = cur;
			
			if (prev)
				prev->next = new_timer;
			else
			{	/* start of list: change list head */
				anatimer_list = new_timer;
			}
			
			anaPrint(ANA_TIMER, "Timer_%d inserted in anatimer_list[%d]\n", new_timer->id, idx);
			break;	
		}
		
		prev = cur;
		cur = cur->next;
		idx++; /* index in active timer list, for debugging purposes */
	}
	
	if (!cur) /* reached end of list (prev is always not NULL) */
	{
		new_timer->next = NULL;
		prev->next = new_timer;
		anaPrint(ANA_TIMER, "Timer_%d inserted in end of anatimer_list\n", new_timer->id);
	}
	
	/* check real timer */
	anatimer_resched(NULL);
	return 1;
}

int anatimer_change(unsigned int id, long msec)
{
	struct anatimer_s *prev = NULL, *cur;
	
	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_change --> timers are off!\n");
		return -1;
	}

	if ( msec <= 0 || !id )
	{
		anaPrint(ANA_ERR, "anatimer_change --> msec or id NULL!\n");
		return -1;
	}
	
	analock_lock(&list_lock);
	
	cur = anatimer_list;
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "anatimer_change --> list NULL!\n");
		analock_unlock(&list_lock);
		return -1;
	}
	
	while (cur)
	{
		if (cur->id == id)
			break;
			
		prev = cur;
		cur = cur->next;
	}

	if (!cur)
	{
		anaPrint(ANA_ERR, "anatimer_change --> Timer_%d not found!\n", id);
		analock_unlock(&list_lock);
		return -1;
	}
	/* the only thing we do is setting the new period. 
 	 * this does not affect the current timeout,
 	 * but takes effect, the next time a new timeout is set
 	 * (e.g. after having fired once again)
 	 */	
	cur->period.tv_sec = msec / 1000;
	cur->period.tv_usec = (msec % 1000) * 1000;

	analock_unlock(&list_lock);
	return id;
}

static int anatimer_resched(struct timeval *now)
{

	struct anatimer_s *head = anatimer_list;
	struct timeval lnow;
	long msec;
#ifdef USER
	struct itimerval next;
	int ret;
#endif
	/* lock is already acquired: do not recall analock_lock */

	anaPrint(ANA_TIMER, "anatimer_resched\n");
	if (!now)
	{
		/* anatimer_insert does not set "now" */
		/* but anatimer_handler sets "now" (to "freeze" time while rescheduling) */
		gettimeofday(&lnow, NULL);
		now = &lnow;
	}

	if ( !anatimer_nextid )
	{
		anaPrint(ANA_ERR, "anatimer_resched --> timers are off!\n");
		return -1;
	}

	if ( !anatimer_list )
	{
		anaPrint(ANA_ERR, "anatimer_resched --> anatimer_list is NULL!\n");
		return -1;
	}

	while (head)
	{ 	
		/* skip de-activated timers, if any */
		if (!head->id)
			head = head->next;
		else
			break;
	}
	
	if (!head)
	{
		anaPrint(ANA_NOTICE, "anatimer_resched --> nothing to re-schedule!\n");
		return -1;
	}
	
	/* get delay until next timer fires */
	msec = timeval_diff(&head->timeout, now);
	
	if (msec <= 0)
	{
		anaPrint(ANA_TIMER, "anatimer_resched --> delay until next Timer_%d is < 0! (firing in 10ms)\n", head->id);
		msec = 10; /* this will call anatimer_handler in 10 ms and hopefully clean the list */
	}
	
#ifdef USER
	next.it_interval.tv_sec = 0;
	next.it_interval.tv_usec = 0;
	next.it_value.tv_sec = msec / 1000;
	next.it_value.tv_usec = (msec % 1000) * 1000;
	
	if ( (ret = setitimer(ITIMER_REAL, &next, NULL)) < 0 )
	{
		anaPrint(ANA_ERR, "setitimer failed %i\n", ret);
		return -1;
	}
#endif

#ifdef KERNEL
	/* kernel seems to fire timer always too early by less than 10msec */
	/* so we add 10msec to the expected firing time */
	
	del_timer(&myanatimer);
	myanatimer.expires = jiffies + msecs_to_jiffies((unsigned int)msec+10);
	add_timer(&myanatimer);
#endif 

	anaPrint(ANA_TIMER, "Next timer in %ld.%03ld seconds\n", msec/1000, msec%1000);
	return 1;
}
