/* C/shared/xrp.c
 * 
 * functions to handle xrp messages
 * 
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 *
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */



#include "xrp.h"

/** @defgroup anaXRP XRP 
 * ANA uses XRP as the standard message format for the communication
 * between Bricks
 */
/*
 *			The Format of an XRP message : 
 *
 *		[command][nbArgs][xrpClass(or argId)][arg Length][arg content]...	
 *	        variable  2bytes  		      2bytes
 *		size	 (short)		      (length of arg content)
 */

/** \ingroup anaXRP
 * \brief allocates memory space to store the XRP message, the size is indicated in the xrpSpec structure
 * \param msg pointer to the address of the message to allocate
 * \return failure: -1, success: 0
 */
int allocateXRPMsg(xrpMsg_t *msg){
	
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if (xrpSpec == NULL || msg == NULL)
		return -1;
		
	*msg = (xrpMsg_t) malloc(xrpSpec->maxMsgSize * sizeof(char));
	
	if (*msg == NULL) {
		anaPrint(ANA_ERR, "unable to allocate memory for XRP message\n");
		return -1;
	}
	return 0;
}

/** \ingroup anaXRP
 * \brief  Stores the XRP command cmd in the XRP message
 * \param msg message in which the command will be written
 * \param cmd XRP Command to be written to the message
 * \return failure: -1, success: size of the newly added element
 */
int fillXRPCommand (xrpMsg_t msg, xrpCmd_t cmd){
	
	uint16_t *tmp;
	int size;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if (xrpSpec == NULL || msg == NULL || cmd == NULL)
		return -1;
			
	memcpy(msg, cmd, xrpSpec->commandSize);

	/* we set the number of arguments to zero */
	tmp = (uint16_t *)msg + xrpSpec->commandSize;
	*tmp = htons(0);
	
	size = defaultXrpSpecs.commandSize + sizeof(uint16_t);
	return size; /* now returns total size of argument */
}

/** \ingroup anaXRP
 * \brief returns the number of arguments of the XRP command by reading the nbArgs field of the message
 * \param msg the message from which we want to know the number of arguments
 * \return failure: -1, success: number of xrp arguments in msg
 */
int getXRPnbArgs(xrpMsg_t msg){  
	
	uint16_t *nbArg;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if (xrpSpec == NULL || msg == NULL )
		return -1;	
	
	nbArg =  (uint16_t *)(msg + xrpSpec->commandSize);
	return ntohs(*nbArg);
}

/** \ingroup anaXRP
 * \brief add a new XRP argument to the message
 * \param msg the message
 * \param xrpClass the argument type to be added
 * \param content the argument itself
 * \param len the length of the argument 
 * \return failure: -1 e.g. if the argument is too large, success: length of the newly added argument
 */
int addXRPArg(xrpMsg_t msg, xrpClass_t xrpClass, void *content, unsigned short int len){  
	
	int nbArguments, i, size, argLen; 
	char *pointer;
	uint16_t *nbArg;
	uint16_t *lenArg;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if (xrpSpec == NULL || msg == NULL || xrpClass == NULL)
		return -1;	

	/* we must find the pointer to the end of the last argument*/
	nbArg =  (uint16_t *)(msg + xrpSpec->commandSize);
	nbArguments = ntohs(*nbArg);
	pointer = msg + xrpSpec->commandSize + sizeof(uint16_t);
	size = xrpSpec->commandSize + sizeof(uint16_t);
	
	for (i=0;i<nbArguments;i++){
		pointer = pointer + xrpSpec->classSize;
		lenArg = (uint16_t *)pointer;
		argLen = ntohs(*lenArg);
		pointer = pointer + argLen + sizeof(uint16_t);
		size = size + xrpSpec->classSize +sizeof(uint16_t) + argLen;
	}
	
	/* <check fitting possibility>*/
	if ((int)(size + xrpSpec->classSize +sizeof(uint16_t) + len) >=  xrpSpec->maxMsgSize){
		anaPrint(ANA_ERR, "addXRPArg : impossible to add argument, reaching total message size limit\n");
		return -1;
	}
	/* <nbArgs incrementation> */
	*nbArg = htons((uint16_t)nbArguments + (uint16_t)1);
	
	/*<adding the argument>*/
	
	memcpy(pointer, xrpClass, xrpSpec->classSize);
			
	pointer = pointer + xrpSpec->classSize; 
	lenArg = (uint16_t *)pointer;
	*lenArg = htons(len);
	if (content!=NULL){
		pointer = pointer + sizeof(uint16_t);
		memcpy(pointer, content, len);
	}
	
	size = xrpSpec->commandSize + sizeof(uint16_t) + len;
	return size; /* now returns total size of argument */
}

/** \ingroup anaXRP
 * \brief parses an xrp message and returns the argument that was searched for.
 * \param msg the message
 * \param xrpClass the argument type to looked for.
 * 	If XRP_CLASS_ANY is specified it is searched for any argument and xrpClass is set to the
 * 	actual Class of the argument returned (attention: the caller has to allocate memory for xrpClass)
 * \param len will contain the length of the XRP argument
 * \param  occurrence which argument of class xrpClass should be returned. E.g. the first -> occurrence = 0 
 * \return failure: NULL, success: pointer to the argument within the XRP message (no memory is allocated!)
 */
void * getXRPArg(xrpMsg_t msg, xrpClass_t xrpClass, int *len, int occurrence) {
	
	uint16_t *nbArgs;
	int nbArguments, lenArg;
	char *ch;
	char *pointer;
	uint16_t *argLen;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
		
	if (xrpSpec == NULL || msg == NULL || xrpClass == NULL || len == NULL )
	{
		if (len)
			*len = 0;
		
		return NULL;
	}
		
	nbArgs = (uint16_t *)(msg + xrpSpec->commandSize);
	nbArguments = ntohs(*nbArgs);
		
	pointer = msg+sizeof(uint16_t)+xrpSpec->commandSize;
	
	while (nbArguments>0) {
		ch = (char *)pointer;
		if(equalXRPClasses((xrpClass_t)pointer, xrpClass)||equalXRPClasses(xrpClass, XRP_CLASS_ANY)){
			if(occurrence==0) {
				uint16_t *argLen = (uint16_t *)(pointer + xrpSpec->classSize);
				lenArg  = ntohs(*argLen);
				*len = lenArg;
				if(equalXRPClasses(xrpClass, XRP_CLASS_ANY))
					memcpy(xrpClass, pointer, xrpSpec->classSize);
				return (uint16_t *)(pointer+sizeof(uint16_t) + xrpSpec->classSize);
			}
			else {
				occurrence--;
			}	
		}
		argLen = (uint16_t *)(pointer + xrpSpec->classSize);
		pointer = pointer + xrpSpec->classSize + sizeof(uint16_t) + ntohs(*argLen);
		nbArguments--;
	}

	*len = 0;
	return NULL;
}

/** \ingroup anaXRP
 * \brief Compares two XRP command Fields
 * \param cmdA first command
 * \param cmdB second command.
 * \return 1 if the commands are equal, 0 otherwise
 */
int equalXRPCommands(xrpCmd_t cmdA,  xrpCmd_t cmdB ) {  
	
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if(cmdA==NULL && cmdB==NULL)
		return 1;
		
	if(cmdA==NULL || cmdB==NULL)
		return 0;
	
	if(!memcmp(cmdA,cmdB, xrpSpec->commandSize ))
		return 1;
	else
		return 0;
}

/** \ingroup anaXRP
 * \brief Compares two XRP Class Fields
 * \param xrpClassA first argument type
 * \param xrpClassB second argument type
 * \return 1 if the argument types are equal, 0 otherwise
 */
int equalXRPClasses(xrpClass_t xrpClassA,  xrpClass_t xrpClassB ) { 
	
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	
	if(xrpClassA == NULL && xrpClassB ==NULL)
		return 1;
	if(xrpClassA == NULL || xrpClassB ==NULL)
		return 0;
	
	if(!memcmp(xrpClassA, xrpClassB, xrpSpec->classSize ))
		return 1;
	else
		return 0;
}




























